package alejandro;

import deb.BaseFeedDynamic;
import deb.Parameters;
import deb.Population;
import deb.Simulator;
import deb.control.Controller;
import deb.fishtank.RotiferPopulation;

public class MyController implements Controller {

	private double ref, base;
	private double Kp, Ki, Kd;
	private double int_e, ek1, sat_e;
	private boolean first;
	private int gain;
	private double dz;
	private double total;
	
	private double referenceLastRun;
	private double referenceInterval;
	private double growth_ref;
	private boolean r_first_up;
	private boolean r_first_down;
	
	public MyController(){
		ref = 0.01100;
//		ref = 0.01300;
//		ref = 0.02400;

		Kp = 1.75;
		Ki = 0.15;
		Kd = 0.06;
		int_e = 0;
		ek1 = 0;
//		sat_e = ref / 10;
		sat_e = ref / 8;
		first = true;
		gain = 1;
		dz = 0.000001;
		total = 0;
		base = 0.000625;
		
		referenceLastRun = 0;
		referenceInterval = 1;
		growth_ref = 0.31;
		r_first_up = true;
		r_first_down = true;
	}
	
	public void setParams(double p, double i, double d){
		this.Kp = p;
		this.Ki = i;
		this.Kd = d;
	}
	
	public void setRef(double r){
		this.ref = r;
		sat_e = this.ref / 8;
	}
	
	public double getCallInterval() {
		// TODO Auto-generated method stub
		return 0.000625; //0.000625 of a day = 0.9 minutes
	}

	public void makeControlRun(double t, Simulator sim, Population pop, Parameters envPar) {
		// TODO Auto-generated method stub
		BaseFeedDynamic bfd = pop.getBaseFeedDynamic();
		double feed, feed_c, err, dif_e, u;
		
		double rot_den = pop.getHerbivoreCount() / envPar.get("V_w0", 0) / 1000;

		RotiferPopulation rotPop = (RotiferPopulation)pop;
		
//		OUTER LOOP
		if(t-referenceLastRun >= referenceInterval){
			
			if(rot_den >= 300){
				if(r_first_up){
					growth_ref = 0.14;
					r_first_up = false;
				}
				else if(rot_den >= 500){
					growth_ref = 0.14;
					r_first_down = false;
				}
			}
			else if(rot_den < 500 && !r_first_down){
				growth_ref = 0.165;
			}
			
			double gr_e = growth_ref - (0.9*rotPop.getEggRate() - 0.16);
			System.out.println("Growth rate error = "+gr_e);
			this.setRef(ref + 0.02*gr_e);
			referenceLastRun = t;
		}
	
		System.out.println("Density: "+rot_den+" rot/ml");
		
		//INNER LOOP
		
		if (rot_den < 100.0 && gain!=1){
			this.setParams(1.75, 0.15, 0.06);
			this.base = 0.000625;
			gain = 1;
		}
		
		if (rot_den > 100.0 && rot_den < 225.0 && gain!=2){
			this.setParams(1.75, 0.15, 0.06);
			this.base = 0.00100;
			gain = 2;
		}
		
		if (rot_den > 225.0 && rot_den < 300.0 && gain!=3){
			this.setParams(2.0, 0.18, 0.06);
			this.base = 0.00150;
			gain = 3;
		}
		
		if (rot_den > 300.0 && rot_den < 3500.0 && gain!=4){
			this.setParams(2.25, 0.15, 0.06);
			this.base = 0.00150;
//			this.setRef(0.0110);
//			envPar.put("T_w", 22);
			gain = 4;
		}

		System.out.println("Ref: "+ref);
	
		if(first){
			u = ref;
			first = false;
		}
		else{
		
			feed = ref*base;
		
			feed_c = pop.getBaseFeedConcentration();
			System.out.println("Concentration = "+feed_c);
		
			err = ref - feed_c;
			dif_e = err - ek1;
			ek1 = err;
			int_e = int_e + err;
			if (int_e < -sat_e) int_e = -sat_e;
			else if (int_e > sat_e) int_e = sat_e;
			System.out.println("Accumulated error: "+int_e);
	
			u = feed + (err*Kp) + (int_e*Ki) + (dif_e*Kd);
			
		}
		if (u >= dz){
			bfd.increaseLevel(u);
			total = total + u;
			System.out.println(u+" g/l added");
			System.out.println("Total feed added: "+total+" g/l");
		}
		else if (u < 0){
			System.out.println("Let the rotifers it the excess");
		}
	}
}
