/*
 * Created on 29.sep.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.util.ArrayList;

/**
 * @author alver
 *
 * This class implements {@link SimulationMonitor}, and stores states and derived outputs, as
 * chosen by the user, in ArrayList objects. These can be retrieved and, if desired, stored in
 * tab-separated text file format using the appropriate method in {@link Functions}.
 * To perform extra operations once the simulation is finished, create a subclass that overrides
 * simulationFinished() with a method that initially calls super.simulationFinished(t, pop), followed 
 * by the operations you want to perform afterwards.
 */
public class ArrayListSimulationMonitor implements SimulationMonitor {

	protected ArrayList herbivoreStates, meanHerbivoreStates, stdevHerbivoreStates, predatorStates, meanPredatorStates, stdevPredatorStates;
	protected boolean recordAllHerbivoreStates = true,
		recordMeanHerbivoreStates = true,
		recordStdevHerbivoreStates = true,
		recordAllPredatorStates = true,
		recordMeanPredatorStates = true,
		recordStdevPredatorStates = true;
	
	
	/**
	 * 
	 */
	public ArrayListSimulationMonitor(boolean recordAllHerbivoreStates, boolean recordMeanHerbivoreStates, boolean recordStdevHerbivoreStates, 
				boolean recordAllPredatorStates, boolean recordMeanPredatorStates, boolean recordStdevPredatorStates) {
		super();
		this.recordAllHerbivoreStates = recordAllHerbivoreStates;
		this.recordMeanHerbivoreStates = recordMeanHerbivoreStates;
		this.recordStdevHerbivoreStates = recordStdevHerbivoreStates;
		this.recordAllPredatorStates = recordAllPredatorStates;
		this.recordMeanPredatorStates = recordMeanPredatorStates;
		this.recordStdevPredatorStates = recordStdevPredatorStates;
		
		if (recordAllHerbivoreStates)
			herbivoreStates = new ArrayList();
		if (recordMeanHerbivoreStates)
			meanHerbivoreStates = new ArrayList();
		if (recordStdevHerbivoreStates)
			stdevHerbivoreStates = new ArrayList();
		if (recordAllPredatorStates)
			predatorStates = new ArrayList();
		if (recordMeanPredatorStates)
			meanPredatorStates = new ArrayList();
		if (recordStdevPredatorStates)
			stdevPredatorStates = new ArrayList();
	}

        
        
	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationStarted(double, deb.Population)
	 */
	public void simulationStarted(double t, Population pop) {
		record(t, pop);
	}

	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationUpdate(double, deb.Population)
	 */
	public void simulationUpdate(double t, Population pop) {
		record(t, pop);

	}

	public void simulationInterval(double t, Population pop) {
		simulationFinished(t, pop);
	}
	
	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationFinished(double, deb.Population)
	 */
	public void simulationFinished(double t, Population pop) {
		record(t, pop);
	}

	
	
	/**
	 * This method does the actual storage of data.
	 */
	protected void record(double t, Population pop) {
		if (recordAllHerbivoreStates) {
			herbivoreStates.add(pop.getHerbivoreBioStates());
		}
		if (recordMeanHerbivoreStates) {
			meanHerbivoreStates.add(pop.getMeanHerbivoreStates());
		}
		if (recordStdevHerbivoreStates) {
			double[][] val = pop.getMeanAndStdevHerbivoreStates();
			stdevHerbivoreStates.add(val[1]);
		}
		if (recordAllPredatorStates) {
			predatorStates.add(pop.getPredatorBioStates());
		}
		if (recordMeanPredatorStates) {
			meanPredatorStates.add(Functions.mean(pop.getPredatorBioStates()));
		}
		if (recordStdevPredatorStates) {
			stdevHerbivoreStates.add(Functions.stdev(pop.getPredatorBioStates()));
		}
	}

        public SimulationMonitor copy(Parameters par, String suffix) {
            ArrayListSimulationMonitor newMon = new ArrayListSimulationMonitor(recordAllHerbivoreStates, recordMeanHerbivoreStates, recordStdevHerbivoreStates,
                    recordAllPredatorStates, recordMeanPredatorStates, recordStdevPredatorStates);
            if (herbivoreStates != null) newMon.herbivoreStates = (ArrayList)herbivoreStates.clone();
            if (meanHerbivoreStates != null) newMon.meanHerbivoreStates = (ArrayList)meanHerbivoreStates.clone();
            if (stdevHerbivoreStates != null) newMon.stdevHerbivoreStates = (ArrayList)stdevHerbivoreStates.clone();
            if (predatorStates != null) newMon.predatorStates = (ArrayList)predatorStates.clone();
            if (meanPredatorStates != null) newMon.meanPredatorStates = (ArrayList)meanPredatorStates.clone();
            if (stdevPredatorStates != null) newMon.stdevPredatorStates = (ArrayList)stdevPredatorStates.clone();
            return newMon;
        }
   
}
