/*
 * Created on 15.sep.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import deb.control.InternalController;
import java.util.*;

/**
 * @author alver
 *
 * Abstract convenience class for Individual, which implements some methods of the Individual interface.
 */
public abstract class BasicIndividual implements Individual {

	protected double[] states = null;
	protected boolean died = false, eaten = false;
	
	protected Parameters par;
	protected InitialValues initVal;
	protected double n_0;
	protected Strain type;
	protected List regulators = new ArrayList();
    protected List<Double> regulatorLastCallTime = new ArrayList<Double>();

    /**
	 * This variable is reported as the number of individuals represented by this BasicIndividual,
	 * and can be manipulated to implement a dynamic value for the number of subindividuals. Alternately,
	 * getN() can be overridden to ignore n.
	 */
	protected double n;
	
	public BasicIndividual(Parameters par, Strain type, InitialValues initVal, double n_0) {
		this.par = par;
		this.type = type;
		this.initVal = initVal;
		this.n_0 = n_0;
		reset();
	}

	public BasicIndividual(Parameters par, Strain type, InitialValues initVal) {
		this.par = par;
		this.type = type;
		this.initVal = initVal;
		this.n_0 = 1;
		reset();
	}
	
        public BasicIndividual(BasicIndividual individual) {
            this.par = individual.par;
            states = (double[])individual.states;
            for (Iterator i=individual.regulators.iterator(); i.hasNext();) {
                regulators.add(((InternalController)i.next()).copy(this.par));
                regulatorLastCallTime.add(0.0);
            }
        }
        
        
        
	public void reset() {
		states = initVal.getValues();	
		n = n_0;
	}

	public boolean isAlive() {
		return !died;
	}
	
	/* (non-Javadoc)
	 * @see deb.Individual#kill()
	 */
	public void kill() {
            died = true;
	}
	
	public void eat() {
		died = true;
		eaten = true;
	}

	public boolean isEaten() {
		return eaten;
	}
	
	public void killFraction(double fraction) {
		n = (1-fraction)*n;
		if (n < 1) {
			n = 0;
                        kill();
                }
	}
	
	
	/* (non-Javadoc)
	 * @see deb.Individual#getBioStates()
	 */
	public double[] getBioStates() {
		return states;
	}

	/* (non-Javadoc)
	 * @see deb.Individual#getStateCount()
	 */
	public int getStateCount() {
		return states.length;
	}

	/**
	 * Returns the number of individuals represented by this individual.
	 */
	public double getN() {
		return n;
	}

	/* (non-Javadoc)
	 * @see deb.Individual#getInitialValuesFromCurrentState()
	 */
	public InitialValues getInitialValuesFromCurrentState() {
		double[] stateVal = (double[])states.clone();
		return new InitialValues(stateVal);
	}

	/**
	 * @param n The n to set.
	 */
	public void setN(double n) {
		this.n = n;
	}
	
	/* (non-Javadoc)
	 * @see deb.Individual#getType()
	 */
	public Strain getType() {
		return type;
	}
	
	public void addRegulator(InternalController reg) {
		assert reg != null;
		regulators.add(reg);
        regulatorLastCallTime.add(0.0);
    }
	
	public int getRegulatorCount()  {
		return regulators.size();
	}
	
	public InternalController getRegulator(int number) {
		assert number < regulators.size();
		return (InternalController)regulators.get(number);
	}
	
	/**
	 * This implementation of update() takes care of updating all Regulator instances attached to this
	 * Individual. Extending classes should call super.update() in their overriding update() method.
	 */
	public IndividualReturnValues update(double t, double dt, Population population,
			Parameters environmentalParameters) {

        int piv=0;
        for (Iterator i=regulators.iterator(); i.hasNext();) {
			InternalController reg = (InternalController)i.next();
            double interval = reg.getCallInterval();
            if ((interval < 0) ||
                (t - regulatorLastCallTime.get(piv) >= interval)) {
                reg.update(this, t, dt, environmentalParameters);
                regulatorLastCallTime.set(piv, t);
            }

        piv++;
        }

		return null;
	}
        
        
        /*public BasicIndividual copy(boolean cloneParameters) {
            try {
                BasicIndividual clone = (BasicIndividual)super.copy();
                    //new BasicIndividual(cloneParameters ? (Parameters)par.clone() : par,
                    //this.getType(), initVal);
                return clone;
            } catch (UnsupportedOperationException ex) {
                ex.printStackTrace();
            }
		return null;            
        }*/
        
        public Individual copy(Parameters par) throws UnsupportedOperationException {
            throw new UnsupportedOperationException("Copy not supported by class "+getClass());
        }

     
    public boolean vulnerableToDilution() {
        return true;
    }

    public Parameters getParameters() {
        return par;
    }
}
