/*
 * Created on 29.okt.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public abstract class CombinationTools {

	/**
	 * This method takes a List, assumed to contain compatible Individual objects, and creates a List of the most similar
	 * pairs - either all pairs closer than the give threshold (with a lower compareTo() value), or the maxComb most similar
	 * ones, whichever is fewest.
	 * @param individuals
	 * @param threshold
	 * @param nLimit A limit signifying a low n value for a super individual - meaning that individuals with lower n
	 * should have a very minor effect on the end result of the simulation. If we are comparing two individuals with an
	 * n value this low, we should suggest combining them even if they are not very similar.
	 * @param maxComb
	 * 
	 * @return
	 * @throws ClassCastException
	 */
	public static List findClosestComparisons(List individuals, double threshold, double nLimit, int maxComb) throws ClassCastException {
		
		TreeSet cand = new TreeSet();
		int count = 0;
		double total = 0;
		Object[] ind = individuals.toArray();
		
		// Outer loop - iterate over all individuals except the last one:
		for (int i=0; i<ind.length-1; i++) {
			Individual one = (Individual)ind[i];
			
			// Inner loop - iterate over all individuals AFTER the i-th one:
			for (int j=i+1; j<ind.length; j++) {
				Individual two = (Individual)ind[j];
				
				// See if they are below the threshold. If so, add the pair.
				double distance = Double.MAX_VALUE;
				if ((one.getClass() == two.getClass()) && (one.getType() == two.getType())) {
					distance = Math.min(one.compareTo(two), two.compareTo(one));
					total += distance;
					count += 1;
				
                                    if (distance <= threshold) {
					cand.add(new CombinedPair(one, two, distance));
                                    } else if ((one.getN()<nLimit) && (two.getN()<nLimit))
					cand.add(new CombinedPair(one, two, distance));
                                }
			}
		}
		total /= count;
		//System.out.println("Mean distance: "+total);
		// Remove duplicates:
		HashSet used = new HashSet();
		for (Iterator i=cand.iterator(); i.hasNext();) {
			CombinedPair cp = (CombinedPair)i.next();
			if (used.contains(cp.getOne()) || used.contains(cp.getTwo()))
				i.remove();
			else {
				used.add(cp.getOne());
				used.add(cp.getTwo());
			}
		}
		
		// If we now don't have too many pairs, just return the whole Set, converted to List:
		if (cand.size() <= maxComb)
			return Arrays.asList(cand.toArray());
		
		// We have too many, so build a new List using the best ones:
		else {
			List res = new ArrayList(maxComb);
			int piv = 0;
			Iterator i = cand.iterator();
			while (piv < maxComb) {
				res.add(i.next());
				piv++;
			}
			return res;
		}
		
	}
}
