/*
 * Created on 13.sep.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public abstract class Functions {

	public static double su(double max, double J_1, double J_2, double n_1, double n_2) {
		double inv = 1/max + n_1/J_1 + n_2/J_2 - 1/(J_1/n_1 + J_2/n_2);
		return Math.max(0, 1/inv);
	}
	
        /**
         * Extended to three substrates. Kooijman (2000) p. 48.
         */
	public static double su(double max, double J_1, double J_2, double J_3, double n_1, double n_2, double n_3) {
            double x1 = J_1/n_1,
                    x2 = J_2/n_2,
                    x3 = J_3/n_3;
		double inv = 1/max + 1/x1 + 1/x2 + 1/x3 - (1/(x1+x2) + 1/(x1+x3)) + 1/(x1+x2+x3);
		return Math.max(0, 1/inv);
	}
	
        /**
         * Determine production depending on three substrates using Liebig's law of the minimum.
         */
	public static double non_su(double max, double J_1, double J_2, double J_3, double n_1, double n_2, double n_3) {
            double cand_1 = J_1/n_1,
                   cand_2 = J_2/n_2,
                   cand_3 = J_3/n_3;
            return Math.min(cand_1, Math.min(cand_2, cand_3));
		
	}
	
    /**
     * Determine production depending on three substrates using Liebig's law of the minimum. A certain percentage
     * overhead is added to the cost, and this overhead is drawn from nonlimiting substrates, in effect providing
     * some plasticity in the stoichiometry of the production.
     */
	public static double[] hybrid_su(double max, double J_1, double J_2, double J_3, double n_1, double n_2, double n_3, double overhead) {
            double[] result = new double[4]; // 0: growth, 1: use flux 1, 2: use flux 2, 3: use flux 3
            double cand_1 = J_1/n_1,
                   cand_2 = J_2/n_2,
                   cand_3 = J_3/n_3;
            double res0 = Math.min(cand_1, Math.min(cand_2, cand_3)); // Nutrient limited growth before overhead
            double tot = res0*(n_1+n_2+n_3), // Total energy before overhead
                    overh = overhead*tot, // Overhead
                    rest_1 = J_1 - res0*n_1, // Nutrient fluxes before overhead
                    rest_2 = J_2 - res0*n_2,
                    rest_3 = J_3 - res0*n_3,
                    rest_sum = rest_1+rest_2+rest_3;
            if (rest_sum >= overh) {
                // There are enough nonlimiting fluxes to cover overhead.
                result[0] = res0;
                result[1] = res0*n_1 + overh*rest_1/rest_sum;
                result[2] = res0*n_2 + overh*rest_2/rest_sum;
                result[3] = res0*n_3 + overh*rest_3/rest_sum;
                
            } else {
                // The in fluxes are so well balanced that there are not large enough rejected fluxes to cover
                // overhead. The growth must be reduced exactly enough to allow the rejected fluxes to cover overhead.
                double sum_n = (n_1 + n_2 + n_3);
                res0 -= (overh - rest_sum)/sum_n/(1+overhead);
                //tot = res0*(n_1+n_2+n_3);
                //overh = overhead*tot;
                //rest_1 = J_1 - res0*n_1;
                //rest_2 = J_2 - res0*n_2;
                //rest_3 = J_3 - res0*n_3;
                //rest_sum = rest_1+rest_2+rest_3;
                result[0] = res0;
                result[1] = J_1;//res0*n_1 + overh*rest_1/rest_sum;
                result[2] = J_2;//res0*n_2 + overh*rest_2/rest_sum;
                result[3] = J_3;//res0*n_3 + overh*rest_3/rest_sum;
                //System.out.println("Nopa: "+result[2]+"\t "+J_2+"\t "+result[3]+"\t "+J_3);
            }
		
            return result;
	}

    /**
     * Similar to the hybrod_su above, except for 4 reserves.
     */
    public static double[] hybrid_su(double max, double J_1, double J_2, double J_3, double J_4,
                                 double n_1, double n_2, double n_3, double n_4, double overhead) {
            double[] result = new double[5]; // 0: growth, 1: use flux 1, 2: use flux 2, 3: use flux 3
            double cand_1 = J_1/n_1,
                   cand_2 = J_2/n_2,
                   cand_3 = J_3/n_3,
                   cand_4 = J_4/n_4;
            double res0 = Math.min(cand_1, Math.min(cand_2, Math.min(cand_3, cand_4))); // Nutrient limited growth before overhead
            double tot = res0*(n_1+n_2+n_3+n_4), // Total energy before overhead
                    overh = overhead*tot, // Overhead
                    rest_1 = J_1 - res0*n_1, // Nutrient fluxes before overhead
                    rest_2 = J_2 - res0*n_2,
                    rest_3 = J_3 - res0*n_3,
                    rest_4 = J_4 - res0*n_4,
                    rest_sum = rest_1+rest_2+rest_3+rest_4;
            if (rest_sum >= overh) {
                // There are enough nonlimiting fluxes to cover overhead.
                result[0] = res0;
                result[1] = res0*n_1 + overh*rest_1/rest_sum;
                result[2] = res0*n_2 + overh*rest_2/rest_sum;
                result[3] = res0*n_3 + overh*rest_3/rest_sum;
                result[4] = res0*n_4 + overh*rest_4/rest_sum;

            } else {
                // The in fluxes are so well balanced that there are not large enough rejected fluxes to cover
                // overhead. The growth must be reduced exactly enough to allow the rejected fluxes to cover overhead.
                double sum_n = (n_1 + n_2 + n_3 + n_4);
                res0 -= (overh - rest_sum)/sum_n/(1+overhead);

                result[0] = res0;
                result[1] = J_1;
                result[2] = J_2;
                result[3] = J_3;
                result[4] = J_4;

            }

            return result;
	}

    /**
     * Similar to the hybrod_su above, except only for two reserves.
     */
	public static double[] hybrid_su(double max, double J_1, double J_2, double n_1, double n_2, double overhead) {
            double[] result = new double[3]; // 0: growth, 1: use flux 1, 2: use flux 2
            double cand_1 = J_1/n_1,
                   cand_2 = J_2/n_2;
            double res0 = Math.min(cand_1, cand_2); // Nutrient limited growth before overhead
            double tot = res0*(n_1+n_2), // Total energy before overhead
                    overh = overhead*tot, // Overhead
                    rest_1 = J_1 - res0*n_1, // Nutrient fluxes before overhead
                    rest_2 = J_2 - res0*n_2,
                    rest_sum = rest_1+rest_2;
            if (rest_sum >= overh) {
                // There are enough nonlimiting fluxes to cover overhead.
                result[0] = res0;
                result[1] = res0*n_1 + overh*rest_1/rest_sum;
                result[2] = res0*n_2 + overh*rest_2/rest_sum;
            } else {
                // The in fluxes are so well balanced that there are not large enough rejected fluxes to cover
                // overhead. The growth must be reduced exactly enough to allow the rejected fluxes to cover overhead.
                double sum_n = (n_1 + n_2);
                res0 -= (overh - rest_sum)/sum_n/(1+overhead);
                  result[0] = res0;
                result[1] = J_1;//res0*n_1 + overh*rest_1/rest_sum;
                result[2] = J_2;//res0*n_2 + overh*rest_2/rest_sum;
            }
		
            return result;
	}
        
        public static double TC(double T, double T_1, double[] Tpars) {
		T += 273;
		T_1 += 273;
		double 
			T_A = Tpars[0], // Arrhenius temperature
			T_L = Tpars[1], // Lower temp boundary
			T_H = Tpars[2], // Upper temp boundary
			T_AL = Tpars[3], // Arrh. temp for lower boundary
			T_AH = Tpars[4]; // Arrh. temp for upper boundary

			return Math.exp(T_A/T_1 - T_A/T)*(1 + Math.exp(T_AL/T_1 - T_AL/T_L) +
					Math.exp(T_AH/T_H - T_AH/T_1))/(1 + Math.exp(T_AL/T - T_AL/T_L) + 
							Math.exp(T_AH/T_H - T_AH/T));
	}
	
	/**
	 * Calculates the mean of a one-dimensional array.
	 * @param values The array to compute the mean of (dimension n)
	 * @return The mean. If the array is empty, 0 is returned.
	 */
	public static double mean(double[] values) {
		if (values.length == 0)
			return 0;
		double res = 0;
		for (int i=0; i<values.length; i++) {
			res += values[i];
		}

		return res/((double)values.length);
	}
	
	/**
	 * Calculates the mean of a two-dimensional array, along the first dimension.
	 * @param values The array to compute the mean of (dimension m * n)
	 * @return The mean (dimension n)
	 */
	public static double[] mean(double[][] values) {
		if (values.length == 0)
			return null;
		double[] res = new double[values[0].length];
		for (int i=0; i<values.length; i++) {
			for (int j=0; j<values[i].length; j++) {
				res[j] += values[i][j];
			}
		}
		for (int j=0; j<res.length; j++) {
			res[j] /= values.length;
		}
		return res;
	}
	
	/**
	 * Calculates the standard deviation of a one-dimensional array.
	 * @param values The array to compute the st.dev. of (dimension n)
	 * @return The mean. If the array is empty, 0 is returned.
	 */
	public static double stdev(double[] values) {
		double mean = mean(values);
		double sqSum = 0;
		if (values.length == 0)
			return 0;
		for (int ind=0; ind<values.length; ind++) {
			double diff = values[ind] - mean;
			sqSum += diff*diff;
		}
		if (values.length >= 2)
			return Math.sqrt(sqSum/(values.length-1));
		else
			return 0;
	}
	
	/**
	 * Calculates the standard deviation of a two-dimensional array, along the first dimension.
	 * @param values The array to compute the st.dev. of (dimension m * n)
	 * @return The mean (dimension n)
	 */
	public static double[] stdev(double[][] values) {
		double[] means = mean(values);
		if (means == null)
			return null;
		double[] sqSum = new double[means.length];
		if (values.length == 0)
			return null;
		for (int ind=0; ind<values.length; ind++)
			for (int st=0; st<values[ind].length; st++) {
				double diff = values[ind][st] - means[st];
				sqSum[st] += diff*diff;
			}
		double[] res = new double[means.length];
		for (int i=0; i<res.length; i++) {
			if (values.length >= 2)
				res[i] = Math.sqrt(sqSum[i]/(values.length-1));
			else
				res[i] = 0;
			//System.out.println(sqSum[i]+"\t"+states.length);
		}
		return res;
	}
	
	/**
	 * Calculates the mean and standard deviation of a two-dimensional array, along the first dimension.
	 * @param values The array to compute the st.dev. of (dimension m * n)
	 * 
	 * The current implementation is not efficient - it simply calls the mean() and stdev() methods
	 * and puts the results together.
	 * 
	 * @return The mean and standard deviation (dimension 2*n)
	 */
	public static double[][] meanAndStdev(double[][] values) {
		double[]
			mean = mean(values),
			stdev = stdev(values);
		double[][] res = new double[2][mean.length];
		for (int i=0; i<mean.length; i++) {
			res[0][i] = mean[i];
			res[1][i] = stdev[i];
 		}
		return res;
	}

    /**
     * Builds a new array containing the elements from array given by the index
     * array idx.
     */
    public static double[] pick(double[] array, int[] idx) {
	double[] res = new double[idx.length];
	for (int i=0; i<res.length; i++)
	    res[i] = array[idx[i]];
	return res;

    }

    /**
     * Builds a new array containing the elements from array between
     * the given indices.
     */
    public static double[] pick(double[] array, int from, int to) {
	double[] res = new double[to-from];
	for (int i=0; i<res.length; i++)
	    res[i] = array[from+i];
	return res;

    }
	
	public static double distortValue(double oldValue, double coefficientOfVariance) {
		// TODO: Implement distortValue with real normal distribution?
		double randomValue = 1+coefficientOfVariance*(2*Math.random()-1);
		return oldValue*randomValue;
	
	}
	
	/**
	 * Stores the Double or double[] data in the given ArrayList to a tab-separated text file.
	 * @param data
	 * @param file
	 * @throws IOException
	 */
	public static void saveToFile(List data, File file) throws IOException {
		FileWriter fw = new FileWriter(file);
			
		for (int i=0; i<data.size(); i++) {
			Object o = data.get(i);
			if (o instanceof double[]) {
				double[] val = (double[])o;
				for (int j=0; j<val.length; j++) {
					fw.write(val[j]+"\t");
				}
			}
			else if (o instanceof Double) {
				fw.write(((Double)o).doubleValue()+"\t");
			}
			fw.write("\n");
		}
		fw.close();
	}
	
	/**
	 * This function compiles an array containing the derived output values of the given List of
	 * Individuals. It is assumed that they are of the same type.
	 * @param individuals The list of individuals to analyze.
	 * @param t The current time.
	 * @return An array of derived outputs.
	 */
	public static double[][] getDerivedOutputs(List individuals, double t) {
		if (individuals.size() < 1)
			return new double[0][0];
		double[][] res = new double[individuals.size()][];
		for (int i=0; i<individuals.size(); i++) {
			Individual ind = (Individual)individuals.get(i);
			res[i] = ind.getDerivedOutputs(t);
		}
		return res;
	}

        /**
	 * This function compiles an array containing the weighted means of the derived outputs of the
	 * given Individuals. It is assumed that they are of the same type.
	 * @param individuals The list of individuals to analyze.
	 * @param t The current time.
	 * @return An array of derived outputs.
	 */
	public static double[] getWeightedMeanDerivedOutputs(List individuals, double t) {
            if (individuals.size() < 1)
                return new double[0];
            double[] derivs = ((Individual)individuals.get(0)).getDerivedOutputs(t);
            double[] res = new double[derivs.length];
            double nTot = 0;
            for (int i=0; i<individuals.size(); i++) {
                Individual ind = (Individual)individuals.get(i);
                derivs = ind.getDerivedOutputs(t);
                double n = ind.getN();
                nTot += n;
                for (int j=0; j<res.length; j++) {
                    res[j] += n*derivs[j];
                }
            }
            if (nTot > 0)
                for (int j=0; j<res.length; j++) {
                    res[j] /= nTot;
                }
            return res;
        }
        
	/**
	 * This function compiles an array containing the state values of the given List of
	 * Individuals. It is assumed that they are of the same type.
	 * @param individuals The list of individuals to analyze.
	 * @param t The current time.
	 * @return An array of state values.
	 */
	public static double[][] getBioStates(List individuals) {
		if (individuals.size() < 1)
			return new double[0][0];
		double[][] res = new double[individuals.size()][];
		for (int i=0; i<individuals.size(); i++) {
			Individual ind = (Individual)individuals.get(i);
			res[i] = ind.getBioStates();
		}
		return res;
	}	
	
	/**
	 * This function computes the age structure histogram for the List of individuals given.
	 * The caller specifies the age range to consider, and the number of age categories.
	 * Individuals falling outside of the last category are either ignored, or added to the
	 * last category.
	 * The calculation takes the individuals' getN() value into consideration.
	 * 
	 * @param individuals The List of individuals to analyze.
	 * @param t The current time.
	 * @param maxAge The upper age limit of the last age category.
	 * @param boxes Number of age categories to split the individuals into.
	 * @param ignoreOverflow Specifies whether individuals falling outside of the last category should
	 * be ignored. If not, they will be added to the last category.
	 * @return
	 */
	public static double[] ageStruct(List individuals, double t, double maxAge, int boxes, boolean ignoreOverflow) {
		double[] res = new double[boxes];
		double boxWidth = maxAge/((double)boxes);
		for (Iterator i=individuals.iterator(); i.hasNext();) {
			Individual ind = (Individual)i.next();
			int box = (int)Math.floor(ind.getAge(t)/boxWidth);
			if (box >= boxes) {
				if (ignoreOverflow)
					continue;
				else
					box = boxes-1;
			}

			res[box] += ind.getN();
		}
		return res;
		
	}
	
	/**
	 * This method iterates through the given List of Individual objects, and adds together their getN() value.
	 * @param list The list of Individuals.
	 * @return The total N.
	 */
	public static double countIndividuals(List list) {
		double count = 0;
		for (Iterator i=list.iterator(); i.hasNext();) {
			Individual rot = (Individual)i.next();
			if (rot.isAlive())
				count += rot.getN();
		}
		return count;
	}

    public static void main(String[] args) {
        double[] tpar = new double[] {
                7300, 273, 308, 10000, 30000
        };
        double[] temps = new double[] {5, 10, 15, 20, 25, 30};
        for (int i = 0; i < temps.length; i++) {
            double temp = temps[i];
            System.out.println("Temp="+temp+"\t TC="+TC(temp,  25, tpar)+
            "\t TC/temp="+(TC(temp, 25, tpar)/temp));
        }
    }
}
