/*
 * Created on 27.apr.2005
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.awt.Point;
import java.io.File;
import java.io.IOException;

import deb.fishtank.RotiferModel;


/**
 * @author alver
 * 
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class GeneralDEB extends BasicIndividual implements Herbivore {

	InitialValues childInitialValue;
	boolean usesBaseFeed, usesWaste;
	/**
	 * @param par
	 * @param type
	 * @param initVal
	 * @param n_0
	 */
	public GeneralDEB(Parameters par, Strain type, InitialValues initVal, InitialValues childInitalValue,
			double n_0) {
		super(par, type, initVal, n_0);
		this.childInitialValue = childInitalValue;
		this.usesBaseFeed = par.getBool("usesBaseFeed", 0);
		this.usesWaste = par.getBool("usesWaste", 0);
	}

	/**
	 * @param par
	 * @param type
	 * @param initVal
	 */
	public GeneralDEB(Parameters par, Strain type, InitialValues initVal, InitialValues childInitalValue) {
		super(par, type, initVal);
		this.childInitialValue = childInitalValue;
	}

	public IndividualReturnValues update(double t, double dt, Population population,
			Parameters envPar) {
		super.update(t, dt, population, envPar);
		
		double[] states_bak = (double[])states.clone();
		// Step ordinary states:
		states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);
		
		if (getE(states) < 0) {
			kill();
			return new IndividualReturnValues();
		}
		
		for (int j=0; j<states.length; j++) {
			if (Double.isNaN(states[j]) || (states[j]<0)) {
				for (int i=0; i<states.length; i++)
					System.out.println("State "+i+": "+states[i]+" old: "+states_bak[i]);
				System.exit(0);
			}
		}
		
		return ingestion(t, population, envPar);
		
	}
	
	/**
	 * Calculates the ingestion rate of this DEB individual.
	 * 
	 * The current implementation calls f() to calculate the ingestion.
	 * 
	 * @param t The current time.
	 * @param population The population this individual is a member of.
	 * @param envPar The environmental parameters.
	 * @return The ingestion rate of this individual in grams per time unit.
	 */
	protected IndividualReturnValues ingestion(double t, Population population, Parameters envPar) {
		double[] f = new double[states.length];
		IndividualReturnValues rate = f(f, t, states, population, envPar);
		//System.out.println(rate);
		return rate;
	}

	protected double getHazardRate() {
		return states[4];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#checkMortality(double, double)
	 */
	public boolean checkMortality(double t, double dt) {
		// Hazard rate:
		if (!died) {
			double h = getHazardRate(), randNum = Math.random();
			died = randNum < h * dt;
		}
		return died;
	}

	/**
	 * @param states
	 * @return Returns the energy reserve total.
	 */
	protected double getE(double[] states) {
		return states[1];
	}
	

	/**
	 * @param states
	 * @return Returns the structural volume total.
	 */
	protected double getV(double[] states) {
		return states[0];
	}
	
	public double getDryWeight(double t) {
		return 0;//getE(states)/par.get("mu_E", t) + getV(states)/par.get("DW_to_VOL", t);
	}
	
	/**
	 * @param states
	 * @return The current reproductive buffer level.
	 */
	protected double getReproductiveBuffer(double[] states) {
		return states[2];
	}
	
	/**
	 * Sets the reproductive buffer level. Limits the value to 0 or higher.
	 * @param value The new value for the reproductive buffer.
	 */
	protected void setReproductiveBuffer(double value) {
		states[2] = value;
		if (states[2] < 0)
			states[2] = 0;
	}
	
	/**
	 * Calculates the catabolic rate for the given state vector and time.
	 * @param states The state vector.
	 * @param TC The temperature correction coefficient.
	 * @param t The current time.
	 * @return Catabolic rate.
	 */
	protected double calculateCatabolicRate(double[] states, double TC, double t) {
		// Catabolic energy:
		double V = getV(states),
			E_ = getE(states)/V,
			v_dot = par.get("v_dot", t)*TC,
			E_G_ = par.get("E_G_", t),
			kappa = par.get("kappa", t),
			p_Am_ = ((par.containsKey("p_Am_baseFeed") ? par.get("p_Am_baseFeed", t) : 0) 
					+ (par.containsKey("p_Am_waste") ? par.get("p_Am_waste", t) : 0))*TC,
			p_dot_M_ = par.get("p_dot_M_", t)*TC;
		
		return E_*(E_G_*v_dot*Math.pow(V, 0.66) + p_dot_M_*V)/(E_G_+E_*kappa);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#f(double[], double, double[], deb.Population,
	 *      deb.Parameters)
	 */
public IndividualReturnValues f(double[] f, double t, double[] states,
			Population population, Parameters envPars) {
		
//		 Convenience variables for states:
		double
			V = states[0],
			E = states[1],
			E_ = E/V,
			R = states[2],   // Reproductive buffer
			M_Q = states[3], // Damage-inducing compounds
			h = states[4];   // Hazard rate
		
		double
			v_dot = par.get("v_dot", t),
			k_as = par.get("k_as", t),
			p_dot_M_ = par.get("p_dot_M_", t),
			E_G_ = par.get("E_G_", t),
			kappa = par.get("kappa", t),
			p_Am_ = 0,//par.get("p_Am_", t),
			T_ref = par.get("T_ref", t),
			V_p = par.get("V_p", t);
			
		double[] Tpars = par.getArray("Tpars", t);
		
		// Environmental parameters
		double
			T_w = envPars.get("T_w", t),
			V_w = envPars.get("V_w", t),
			V23 = Math.pow(V, 0.66667),
			V13 = Math.pow(V, 0.33333);
		
		//		 Temperature corrections:
		double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature
													 // Correction factor (fnct.
													 // by Kooijman)
		// Assimilated energy:
		double energyIntake = 0, faeces = 0, baseFeedConsumption = 0, wasteConsumption = 0;
		if (usesBaseFeed) {
			double X_K = par.get("X_K_baseFeed", t),
				p_Am_bf = par.get("p_Am_baseFeed", t);
			p_Am_ += p_Am_bf;
			double X = population.getBaseFeedConcentration();
			double f_feed = X/(X + X_K);
			baseFeedConsumption = f_feed*p_Am_bf*TC*V23;;
			energyIntake += baseFeedConsumption;
			faeces += (1-k_as)*baseFeedConsumption;
			
		}
		if (usesWaste) {
			double X_K = par.get("X_K_waste", t),
				p_Am_wst = par.get("p_Am_waste", t);
			p_Am_ += p_Am_wst;
			double X = population.getWasteLevel();// getBaseFeedConcentration();
			double f_feed = X/(X + X_K);
			wasteConsumption = f_feed*p_Am_wst*TC*V23;
			energyIntake += wasteConsumption;
			//if (t > 20)
			//	System.out.println("f = "+f_feed);
			//faeces += 0.1*(1-k_as)*wasteConsumption; // The faeces must be even lower quality than the initial waste :/
		}
		
		if ((t > 70) &&(energyIntake > 0))
			System.out.println("Intake: "+energyIntake+"\tfeed: "+population.getBaseFeedConcentration()+"\twaste: "+population.getWasteLevel());
		
		double p_A = k_as * energyIntake;
		
		double mu_X = 25000; // "Reference" energy density of feed
		double pred_g = energyIntake/mu_X; // Predation measured in grams. Return this value.
		double baseFeedCons_g = baseFeedConsumption/mu_X;
		double wasteConsumption_g = wasteConsumption/mu_X;
		double faeces_g = faeces/mu_X;
		//System.out.println("Faeces: "+faeces+"\tIntake: "+energyIntake+"\tWasteIntake: "+wasteConsumption);
		// Catabolic energy:
		double p_C = calculateCatabolicRate(states, TC, t);
		if (Double.isNaN(p_C)) {
			System.out.println(getV(states)+" "+getE(states));
			System.exit(0);
		}
		// State equations:
		double d_E = p_A - p_C;
		double d_V = (kappa*p_C - p_dot_M_*TC*V)/E_G_;
		if (d_V < 0) {
			d_V = 0;
			p_C = p_dot_M_*TC*V;
			d_E = -p_C;
		}
		
		// Reproduction, only if V >= V_p:
		double d_R = V >= V_p ? Math.max(0, (1-kappa)*p_C - p_dot_M_*V_p) : 0;
		
//		 Damage-inducing compounds and hazard rate:
		double
			nu_QC = par.get("eta_QC", t),
			d_M_Q = nu_QC*p_C,
		    d_h = (M_Q/V);

		f[0] = d_V;
		f[1] = d_E;
		f[2] = d_R;
		f[3] = d_M_Q;
		f[4] = d_h;
		//System.out.println("Waste cons: "+(pred_g-baseFeedCons_g));
		return new IndividualReturnValues(baseFeedCons_g, wasteConsumption_g, 0*faeces_g); // Return predation in grams per time unit.
	}
	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#getDerivedOutputs(double)
	 */
	public double[] getDerivedOutputs(double t) {
		// TODO Auto-generated method stub
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#getDerivedOutputNames()
	 */
	public String[] getDerivedOutputNames() {
		// TODO Auto-generated method stub
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#addNutritionalContent(deb.NutrientIntake, double,
	 *      double)
	 */
	public void addNutritionalContent(NutrientIntake ni, double multiplier,
			double t) {
		// TODO Auto-generated method stub

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#getPosition()
	 */
	public Point getPosition() {
		// TODO Auto-generated method stub
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#getAge(double)
	 */
	public double getAge(double t) {
		// TODO Auto-generated method stub
		return 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#readyToSpawn(double)
	 */
	public boolean readyToSpawn(double t) {
		double E_spawn = par.get("E_spawn", t);
		return getReproductiveBuffer(states) >= E_spawn;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#storeCurrentState(java.io.File)
	 */
	public void storeCurrentState(File file) throws IOException {
		Parameters state = new Parameters();
		state.putArray("initValues", getBioStates());
		state.storeValues(file);
	}

	private static double ZERO_THRESHOLD = 1e-9;
	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#compareTo(deb.Individual)
	 */
	public double compareTo(Individual ind) throws ClassCastException {
		GeneralDEB deb = (GeneralDEB)ind;
		if (type != deb.getType())
			return 100; // Different types cannot be combined.
		double[] otherStates = deb.getBioStates();
		double stateDiff = 0, biggestDiff = -1;
		int biggestI = -1;
		// Calculate relative, weighted sum of deviances:
		for (int i=0; i<states.length; i++) {
			if (states[i] != 0) {
				// Check if both are above the zero threshold before counting the relative difference:
				if ((Math.abs(states[i]) > ZERO_THRESHOLD) || (Math.abs(otherStates[i]) > ZERO_THRESHOLD))
					stateDiff += Math.abs(states[i]-otherStates[i])/states[i];
			}
			else {
				stateDiff += (otherStates[i] != 0) ? 1 : 0;
			}
		}
		//System.out.println(stateDiff);
		return stateDiff;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Individual#combineWith(deb.Individual, double)
	 */
	public Individual combineWith(Individual ind, double t)
			throws ClassCastException {
		
		GeneralDEB oth = (GeneralDEB)ind;
		double othN = oth.getN(),
		sumN = othN + getN();
	
		double[] states = getBioStates(), 
			othSt = oth.getBioStates(),
			othChildInit = oth.childInitialValue.getValues(),
			thisChildInit = childInitialValue.getValues();
	
		// The new state vector should be a weighted mean of the old ones:
		double[] newSt = new double[states.length],
			newChildInit = new double[states.length];
		for (int i=0; i<newSt.length; i++) {
			newSt[i] = (n*states[i] + othN*othSt[i])/sumN;
			newChildInit[i] = (n*thisChildInit[i] + othN*othChildInit[i])/sumN;
		}
		GeneralDEB newDEB = new GeneralDEB(par, type, new InitialValues(newSt), new InitialValues(newChildInit), sumN);
		return newDEB;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Herbivore#spawnAndGetChild(double)
	 */
	public Herbivore spawnAndGetChild(double t) {
		setReproductiveBuffer(getReproductiveBuffer(states) - par.get("E_spawn", t));
		Herbivore newHerb = new GeneralDEB(par, type, childInitialValue,
				childInitialValue,  getN());
		return newHerb;
	}

}
