/*
 * Created on 23.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.awt.Point;
import java.io.IOException;
import java.io.File;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public interface Individual {

	/**
	 * This method shold take care of advancing all dynamic values of this Individual one time step.
	 * @param dt
	 * @param feedConcentration
	 * @return The feed ingestion rate.
	 */
	
	public IndividualReturnValues update(double t, double dt, Population population, Parameters environmentalParameters);
	
	/**
	 * This method returns a string describing the type of organism. For populations containing a mixture of different
	 * species/strains (useful e.g. for simulating competing species), this can be useful to sort types.
	 * @return
	 */
	public Strain getType();
	
	/**	
 	* Check whether this individual is still alive. If and only if mortality is simulated, this method will be called
 	* from the simulator exactly once per individual per time step.
 	* @parame t The current time.
	* @param dt The time step.
 	* @return
 	*/
	public boolean checkMortality(double t, double dt);

	/**
	 * Check whether this individual is still alive. This method can be called an arbitrary number of times
	 * for each time step.
	 * @return
	 */
	public boolean isAlive();
	
	/**
	 * Remove this individual from the simulation.
	 */
	public void kill();

	/**
	 * Remove this individual from the simulation, and signal that it has been eaten. The distinction with
	 * the kill() method is useful when dead, uneaten individuals are considered additional waste. 
	 */
	public void eat();

	/**
	 * Returns true if this individual has been eaten.
	 * @return 
	 */
	public boolean isEaten();

	
	/**
	 * Remove the given fraction of individuals represented by this individual.
	 * The resulting n should not be rounded, but should be set to 0 for all values < 1.
	 */
	public void killFraction(double fraction);
	
	/**
	 * Reset to initial value.
	 */
	public void reset();

	/**
	 * This method should calculate the time derivatives of all states in its normal state vector,
	 * and put the values in the array f. The method should return the ingestion rate of the individual.
	 * @param f The array to be filled with the states' time derivatives.
	 * @param states The current state values. These values must override the individual's own account of its states, 
	 * because these values can be from an intermediate step of the solver algorithm.
	 * @param population The population to which this individual belongs. 
	 * @param environmentalParameters Environmental parameters.
	 * @return The ingestion rate of this individual.
	 */
	public IndividualReturnValues f(double[] f, double t, double[] states, Population population, Parameters environmentalParameters);
	
	/**
	 * Return a set of biologically meaningful outputs, if any. These should correspond to
	 * the names returned from {@link getDerivedOutputNames()}.
	 * @param t The current time.
	 * @return
	 */
	public double[] getDerivedOutputs(double t);
	
	/**
	 * Returns the array of names corresponding to the output values returned by {@link getDerivedOuts()}.
	 * @return String array containing the derived state names.
	 */
	public String[] getDerivedOutputNames();
	
	/**
	 * Returns the protein content of the individual.
	 * @param t The current time
	 * @return The protein content in g
	 
	public double getProteinContent(double t);
	
	/**
	 * Returns the lipid content of the individual.
	 * @param t The current time
	 * @return The lipid content in g
	 
	public double getLipidContent(double t);*/
	
	/**
	 * Add the nutrients contained in this Individual to the {@link NutrientIntake} object, multiplied
	 * by the given number to represent the summation of several individuals.
	 * @param ni The NutrientIntake object where the values are to be added.
	 * @param multiplier The multiplier for all values added.
	 */
	public void addNutritionalContent(NutrientIntake ni, double multiplier, double t);
	
	/**
	 * Returns the current dry weight of the individual.
	 * @param t The current time
	 * @return The dry weight in grams
	 */
	public double getDryWeight(double t);
	
	public double[] getBioStates();
	
	public int getStateCount();
	
	public Point getPosition();
	
	/**
	 * Returns the time (simulated time, not real world time) since this individual was created. In the initial phase of the
	 * simulation, this age may not correspond to biological age, unless care has been taken to initialize the age properly.
	 * @param t The current time.
	 * @return
	 */
	public double getAge(double t);
	
	/**
	 * Returns the number of individuals represented by this Individual. This property can be used for super-individual models.
	 * The data type is double to allow for non-integer numbers.
	 * @return
	 */
	public double getN();
	
	/**
	 * Sets the number of individuals represented by this Individual. The value does not have to be an integer.
	 * @param n The new number of individuals.
	 * @return
	 */
	public void setN(double n);
	
	/**
	 * Called by the {@link Simulator} to discover whether this individual is ready to produce an offspring.
	 * This method is always called once per time step. If <code>true</code> is returned, the appropriate spawn
	 * method will be called to obtain the offspring.
	 * @param t The current time
	 * @return
	 */
	public boolean readyToSpawn(double t);

        
        /**
         * Find out whether this individual can be affected by dilution. For instance, a rotifer individual
         * attached to a substrate, and thus not following the water flow, may not be affected by water dilution
         * in a culture tank.
         */
        public boolean vulnerableToDilution();
        
	/**
	 * Create an InitalValues instance containing the current state values of this individual. This method can be useful for
	 * creating various initial value sets, which can be serialized and used for later simulations.
	 * 
	 * @return An InitialValues instance.
	 */
	public InitialValues getInitialValuesFromCurrentState();
	
	/**
	 * Store an XML representation of the Individual's current state - all dynamic elements included.
	 * @param file
	 * @throws IOException
	 */
	public void storeCurrentState(File file) throws IOException;
	
	
	/**
	 * Compares this Individual to another Individual. This method is only required to work with
	 * and Individual of the same class as this Individual.
	 * 
	 * The comparison should be based on how similar the individuals are in state values, for the
	 * purpose of deciding if they are eligible for combining (in order to reduce model dimension).
	 * Note that the comparison should *not* take the N of the individuals into account - it is a
	 * matter of comparing the state values of the individuals represented by the super individuals,
	 * not of comparing the super individuals.
	 *
	 * @param ind The individual to compare this Individual with. 
	 * @return A value signifying the "distance" between the two individuals, where 0 denotes that
	 * they are exactly similar. Values should range from 0 and upwards.
	 * @throws ClassCastException Thrown if the individuals are not comparable.
	 */
	public double compareTo(Individual ind) throws ClassCastException;
	
	/**
	 * Combines this Individual with another individual of the same class.
	 * 
	 * Implementers of this method should note that the combined indiual cannot legally be 
	 * the same object as either this Individual or the argument Individual.
	 * 
	 * The combined Individual should have an N corresponding to the sum of the N's of the two
	 * individuals that are combined. The physiological states should be calculated with regard to
	 * the relative N's of the two.
	 * @param ind The combined Individual
	 * @param t The current time.
	 * @throws ClassCastException Thrown if the individuals are not compatible.
	 */
	public Individual combineWith(Individual ind, double t) throws ClassCastException;
        
    /**
     * If supported, creates an independent copy of the Individual, set up with the given
     * parameter set.
     *
     * @param par The Parameter object to be used by the new individual.
     * @return A copy of the individual with the exact same state.
     * @throws UnsupportedOperationException Thrown if the Individual does not support the copying.
     */
    public Individual copy(Parameters par) throws UnsupportedOperationException;

    /**
     * Accessor for this Individual's Parameters.
     * @return The Parameters for this Individual.
     */
    public Parameters getParameters();
}
