/*
 * Created on 15.okt.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.util.Iterator;

/**
 * This class is a simple container that represents the nutritional contents of a predator's
 * feed intake. The values are the result of a mapping from number of herbivores eaten, to 
 * the actual nutritional value of those herbivores.
 */
public class NutrientIntake extends Parameters {

	/**
	 * This method works in the same way as the put() method in Parameters when the parameter has not been
	 * set before. If the parameter has already been set, the new value is added to the old one.
	 * @param key The name of the parameter.
	 * @param value The value to add.
	 */
	public void add(String key, double value) {
		if (par.containsKey(key)) {
			put(key, get(key, 0)+value);
		} else
			put(key, value);
		
	}
	
	/**
	 * This method works in the same way as the get() method in Parameters when the parameter has
	 * already been set. If the parameter has not been set, 0 is returned. No error condition can
	 * therefore happen for non-null keys.
	 * @param key The name of the parameter.
	 * @return The value of the parameter.
	 */
	public double get(String key) {
		if (par.containsKey(key)) 
			return get(key, 0);
		else
			return 0;
	}

	/**
	 * When this object represents an absolute amount of nutrients, it must be converted so it
	 * represents nutrient fluxes, to be usable by the dynamic model.
	 * @param dt
	 * @return
	 */
	public void convertToFlux(double dt) {
		for (Iterator i=par.keySet().iterator(); i.hasNext();) {
			// Assume only simple double values are set. Will fail otherwise.
			String s = (String)i.next();
			double number = get(s);
			par.put(s, new SimpleParameter(number/dt));
		}
	}
	
}
