
package deb;

import org.jfree.data.general.AbstractDataset;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.DomainOrder;
import java.util.*;
import javolution.util.FastTable;

/**
 * Dataset for recording an arbitrary number of time steps.
 */
public class OpenEndedDataset extends AbstractDataset implements XYDataset {

    private FastTable states;
    private int piv = 0, seriesCount = 0;
    private double dt, minimumEndTime;
    private boolean online = true; // This variable determines whether the dataset will notify the plot of changes.
    private int windowSize = -1, offsetSteps = 0; 
    // Allows "rolling window" plotting, where
    // time steps slide out of view after a while.

    private	String[] names;
    private String xLabel, yLabel;
    
    public OpenEndedDataset(double dt, double minimumEndTime, String[] names) {
	this(dt, minimumEndTime,"","");
	this.names = names;
    }

    public OpenEndedDataset(double dt, double minimumEndTime, String xLabel, String yLabel) {
	states = new FastTable();
	this.minimumEndTime = minimumEndTime;
	//System.out.println("dt="+dt);
	//	states = new double[2+steps][];
	this.dt = dt;
    }

    
    public OpenEndedDataset(double dt, double minimumEndTime, int windowSize, String[] names, String xLabel, String yLabel) {
	this(dt, minimumEndTime, xLabel, yLabel);
	this.windowSize = windowSize;
        this.names = names;
    }

    public void setOnline(boolean online) {
        this.online = online;
    }
    
    public void record(double t, double[] x) {
	states.add(x);
	if ((windowSize > 0) && (states.size() > windowSize)) {
	    states.remove(0);
	    offsetSteps++; // To keep the time increasing properly.
	}

	if (seriesCount == 0)
            seriesCount = x.length;
        
        if (online) 
	    fireDatasetChanged();
	
    }

    public String getXLabel() {
        return xLabel;
    }
    
    public String getYLabel() {
        return yLabel;
    }
    
    public double getEndTime() {
	return Math.max(minimumEndTime, (states.size()-1)*dt);
    }

    public boolean isEndTimeIncreasing() {
	return minimumEndTime < (states.size()-1)*dt;
    }

   public DomainOrder getDomainOrder() {
	return DomainOrder.ASCENDING;
    }    

    /**
     * Returns the number of series in the dataset.
     */
    public int getSeriesCount() {
   
	return seriesCount;
	//return states[0].length;
    }
          
    /**
     * Returns the name of the given series.
     */
    public String getSeriesName(int i) {
	if (names != null)
	    return names[i];
	else
	    return "";
    }

    /**
     * Returns the number of items in a series.
     */
    public int getItemCount(int series) {
	//	System.out.println(states.size());
	return states.size();
	//return piv;//states.length;
    }
     
    /**
     * Returns the x-value (as a double primitive) for an item within
     * a series.
     */
    public double getXValue(int series, int item) {
	return (((double)item)+offsetSteps)*dt;
    }

    /**
     * Returns the x-value for an item within a series.
     */
    public Number getX(int series, int item) {
	return new Double(getXValue(series, item));
    }

    /**
     * Returns the y-value (as a double primitive) for an item within
     * a series.
     */
    public double getYValue(int series, int item) {
	double[] x = (double[])states.get(item);
	return x[series];
    }
     
    /**
     * Returns the y-value for an item within a series.
     */
    public Number getY(int series, int item) {
	return new Double(getYValue(series, item));
    }

/*  Fra Bubble...
    public XYDataset getSubDataset(int[] toPlot) {
    
	//System.out.println(""+names);
	double[][] stateArray = new double[states.size()][getSeriesCount()];
	for (int i=0; i<stateArray.length; i++) {
	    stateArray[i] = (double[])states.get(i);
	}
	return new ArraySubDataset(stateArray, names, dt, toPlot);
    }*/

}
