/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import de.olikurt.parser.Function;
import de.olikurt.parser.Variable;
import java.util.Map;
import java.util.Iterator;
import java.util.TreeMap;
import java.io.IOException;
import java.io.File;
import java.util.Vector;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
class ParameterHandler extends DefaultHandler {

	final static String PARAMETERS = "parameters"; // Element name for parameter set.
	final static String PARAM = "parameter"; // Element name for parameter.
        final static String ITERATION = "iteration"; // Element name for iteration value set.
	final static String PARNAME = "name"; // Attribute name for parameter name.
	final static String VALUE = "value"; // Attribute name for parameter value.
	final static String TYPE = "type"; // 
	final static String ELEMENT = "element"; //
	final static String NUMBER = "number"; //
	final static String TIME = "time"; //
	final static String SCHEDULE = "schedule";
        
        final static String STRUCT = "struct"; // "struct" is a parameter set inside a single parameter.
        final static String VARSTRUCT = "varstruct"; // "varstruct" is a time-varying struct that references different
                                                     // structs at different times.
	final static String VARIABLE = "variable";
	final static String EXTEND = "extend";
        final static String FUNCTION = "function"; // Simple expression dependent on time.
	final static String PATH_NAME = "path"; // Parameter type that references a path, possibly relative to the current file.
	
	boolean inParameters = false, inArray = false, inVariable = false, inSchedule = false, inPath = false,
                inStruct = false, inVarStruct = false, inFunction = false, inIteration = false;
		
        private Parameters currentStruct = null;
	private Map par, iter;
	private TreeMap currentList;
	private String currentPar;
	private File file;
	private StringBuffer functionData = null;
	/**
	 * 
	 */
	public ParameterHandler(Map par, Map iter, File f) 
	{
		super();
		this.par = par;
                this.iter = iter;
		this.file = f;
	}

	public void startElement(String namespaceURI, String localName,
		String qName, Attributes atts)
	{
		
		Double number = null, time = null;
		Object value = null;
		String name = null, type = null;
		
		
		for (int i=0; i<atts.getLength(); i++) {
			String nm = atts.getQName(i);
			if (VALUE.equals(nm)) {
				try {
					value = new SimpleParameter(Double.parseDouble(atts.getValue(i)));
				} catch (NumberFormatException ex) {
					// Try to parse boolean:
					if ("true".equals(atts.getValue(i)))
						value = new SimpleBooleanParameter(true);
					else if ("false".equals(atts.getValue(i)))
						value = new SimpleBooleanParameter(false);
					else value = new StringParameter(atts.getValue(i)); // String
				}
			}
			else if (PARNAME.equals(nm)) {
				name = atts.getValue(i);
			}
			else if (TYPE.equals(nm)) {
				type = atts.getValue(i);
			}
			else if (NUMBER.equals(nm)) {
				number = new Double(Double.parseDouble(atts.getValue(i)));
			}
			else if (TIME.equals(nm)) {
				time = new Double(Double.parseDouble(atts.getValue(i)));
			}
		}

		if (EXTEND.equals(qName)) {
			// "Extend" means that we should read the parameter file specified in, 
			// the "name" attribute, and use all parameter values given there, except 
			// those overridden in this file.
			File f = new File(name);
			// If the filename is given relative, we resolve the file starting from the one we are currently parsing:
			if (!f.isAbsolute()) {
				f = new File(file.getParentFile(), f.getPath());
			}
			try {
				// Parse the parameters:
				// XXX: If the extended file in turn is specified to extend this file, we are in trouble.
				Parameters origPar = new Parameters(f);
				
				// Iterate through all settings in origPar, and coyp all those not already set in par.
				for (Iterator i=origPar.keySet().iterator(); i.hasNext();) {
					String parName = (String)i.next();
                                        //System.out.println(parName);
					if (!par.containsKey(parName)) {
						par.put(parName, origPar.getParameter(parName));
					}
				}
				
			} catch (IOException ex) {
				ex.printStackTrace();
			}
			return;
		}
		
		if ((inParameters) && (PARAM.equals(qName)) && (name != null)) {
			
			if (value != null) {
				if (PATH_NAME.equals(type)) {
					// Store path name:
					File f = new File(value.toString());
					// If the filename is given relative, we resolve the file starting from the one we are currently parsing:
					if (!f.isAbsolute()) {
						f = new File(file.getParentFile(), f.getPath());
					}
					par.put(name, new StringParameter(f.getPath()));
					//System.out.println("Path name: "+f.getPath());
				}
				else
					par.put(name, value);
			} else if ("array".equals(type)) {
				currentList = new TreeMap();
				currentPar = name;
				inArray = true;
			} else if (VARIABLE.equals(type)) {
				currentList = new TreeMap();
				currentPar = name;
				inVariable = true;
			} else if (SCHEDULE.equals(type)) {
				currentList = new TreeMap();
				currentPar = name;
				inSchedule = true;
			} else if (STRUCT.equals(type)) {
                            currentStruct = new Parameters();
                            currentPar = name;
                            inStruct = true;
                        } else if (VARSTRUCT.equals(type)) {
                            currentList = new TreeMap();
                            currentPar = name;
                            inVarStruct = true;
                        } else if (FUNCTION.equals(type)) {
                            functionData = new StringBuffer();
                            currentPar = name;
                            inFunction = true;
                            //System.out.println("In function");
                        }
		}
		else if (PARAMETERS.equals(qName)) {
			inParameters = true;
		}
		else if (ELEMENT.equals(qName)) {
                    if (inArray) {
			if ((number != null) && (value != null)) {
				//System.out.println("param: "+number);
				currentList.put(number, value);
			}
                    }
                    else if (inVariable) {
			if ((time != null) && (value != null)) {
				//System.out.println("param: "+number);
				currentList.put(time, value);
			}
                    }
                    else if (inSchedule) {
			if ((time != null) && (value != null)) {
				//System.out.println("time: "+time+"\tvalue: "+value);
				currentList.put(time, value);
			}
                    }
                    else if (inStruct) {
                        //System.out.println("In struct: name: "+name+"\tvalue: "+value+"\tStruct: "+currentPar);
                        try {
                            currentStruct.putParameter(name, (Parameter)value);
			} catch (Exception ex) {
                            
                            System.err.println("Error parsing struct value of '"+name+"' in struct '"+currentPar+"'");
                            ex.printStackTrace();
                            System.exit(0);
                        }
                        
                    }
                    else if (inVarStruct) {
                        if ((time != null) && (value != null)) {
				currentList.put(time, value);
			}
                    }
                    else if (inIteration) {
                        if ((number != null) && (value != null)) {
                            //System.out.println("param: "+number+"\t "+value);
                            currentList.put(number, value);
			}
                    }
                }
                else if (inParameters && ITERATION.equals(qName) && (name != null)) {
                    inIteration = true;
                    currentPar = name;
                    currentList = new TreeMap();
                }
	}

	public void endElement(String namespaceURI, String localName,
		String qName)
	{
		if (inParameters && inArray && PARAM.equals(qName)) {
			// Finished array: store result
			if (currentList.size() > 0) {
				// Decide what kind of array we should create:
				Object o = currentList.get(currentList.firstKey());
				//System.out.println("en "+o);
				if (o instanceof SimpleParameter) {
					double[] value = new double[currentList.size()];		    
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						SimpleParameter val = (SimpleParameter)currentList.get(i.next());
						value[count] = val.getValue(0);
						count++;
					}
                                        
					par.put(currentPar, new ArrayParameter(value));
				} else if (o instanceof Boolean) {					
					boolean[] value = new boolean[currentList.size()];		    
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						Boolean val = (Boolean)currentList.get(i.next());
						value[count] = val.booleanValue();
						count++;
					}
					par.put(currentPar, value);
					
				}
				//System.out.println("to");
			}			
			currentPar = null;
			currentList = null;
			inArray = false;
		}
		else if (inParameters && inVariable && PARAM.equals(qName)) {
			// Finished variable: store result
			if (currentList.size() > 0) {
				// Decide what kind of variable we should create:
				Object o = currentList.get(currentList.firstKey());
				if (o instanceof SimpleParameter) {
					double[][] value = new double[currentList.size()][2];
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						Object key = i.next();
						double val = ((SimpleParameter)currentList.get(key)).getValue(0);
						value[count][0] = ((Double)key).doubleValue();
						value[count][1] = val;
						//System.out.println("entering time: "+value[count][0]+" value: "+value[count][1]);
						count++;
					}
                                        //System.out.println(currentPar);
					par.put(currentPar, new VariableParameter(value));
				} else if (o instanceof SimpleBooleanParameter) {
					boolean[] values = new boolean[currentList.size()];
                                        double[] times = new double[currentList.size()];
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						Object key = i.next();
						SimpleBooleanParameter val = (SimpleBooleanParameter)currentList.get(key);
						times[count] = ((Double)key).doubleValue();
						values[count] = val.getBoolean(0);
						count++;
					}
					par.put(currentPar, new VariableBoolean(times, values));
				}		
			}	
			currentPar = null;
			currentList = null;
			inVariable = false;
		}

		else if (inParameters && inSchedule && PARAM.equals(qName)) {
			// Finished schedule: store result
			if (currentList.size() > 0) {
				// Decide what kind of variable we should create:
				Object o = currentList.get(currentList.firstKey());
				if (o instanceof SimpleParameter) {
					double[][] value = new double[currentList.size()][2];
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						Object key = i.next();
						double val = ((SimpleParameter)currentList.get(key)).getValue(0);
						value[count][0] = ((Double)key).doubleValue();
						value[count][1] = val;
						
						count++;
					}
                                        par.put(currentPar, new ScheduleParameter(value));
				}	
			}				
				
			currentPar = null;
			currentList = null;
			inSchedule = false;
		}
                else if (inParameters && inStruct && PARAM.equals(qName)) {
                    // Finished struct: store result
                    //System.out.println(currentPar+"\t"+currentStruct);
                    par.put(currentPar, new StructParameter(currentStruct));
                    currentStruct = null;
                    currentPar = null;
                    inStruct = false;
                }
                else if (inParameters && inVarStruct && PARAM.equals(qName)) {
                    // Finished variable: store result
                    if (currentList.size() > 0) {
                        double[] times = new double[currentList.size()];
                        String[] values = new String[currentList.size()];
                        int count = 0;
			for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
                            Object key = i.next();
                            times[count] = ((Double)key).doubleValue();
                            values[count] = ((StringParameter)currentList.get(key)).getString(0);
                            count++;
                        }
                        par.put(currentPar, new VarStructParameter(times, values, par));
                        
                    }
                    inVarStruct = false;
                    currentPar = null;
                }
                else if (inParameters && inFunction && PARAM.equals(qName)) {
                    try {
                        Function f = new Function(functionData.toString().trim());
                        //TimeFunction timeF = new TimeFunction(functionData.toString());
                        par.put(currentPar, new FunctionParameter(f));
                        
                        //System.exit(0);
                    } catch (IllegalArgumentException ex) {
                        ex.printStackTrace();
                    }
                    inFunction = false;
                    //System.out.println("Out of function.");
                }
                else if (inParameters && inIteration && ITERATION.equals(qName)) {
                    // Finished reading iteration value set: store result
			if (currentList.size() > 0) {
				// Decide what kind of array we should create:
				Object o = currentList.get(currentList.firstKey());
				//System.out.println("en "+o);
				if (o instanceof SimpleParameter) {
					double[] value = new double[currentList.size()];		    
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						SimpleParameter val = (SimpleParameter)currentList.get(i.next());
						value[count] = val.getValue(0);
						count++;
					}
                                        
					iter.put(currentPar, new ArrayParameter(value));
				} else if (o instanceof Boolean) {					
					boolean[] value = new boolean[currentList.size()];		    
					int count = 0;
					for (Iterator i=currentList.keySet().iterator(); i.hasNext();) {
						Boolean val = (Boolean)currentList.get(i.next());
						value[count] = val.booleanValue();
						count++;
					}
					iter.put(currentPar, value);
					
				}
				//System.out.println("to");
			}			
			currentPar = null;
			currentList = null;
			inIteration = false;
                }
		else if (inParameters && (PARAMETERS.equals(qName)))
			inParameters = false;
	}

    public void characters(char[] ch, int start, int length) throws org.xml.sax.SAXException {
  
        if (inFunction) {
            for (int i=start; i<start+length; i++) {
                functionData.append(ch[i]);
            }
        }
    }
        
        
}
