/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import de.olikurt.parser.Function;
import de.olikurt.parser.Variable;
import java.util.Map;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.Collections;
import javax.xml.parsers.*;
import javolution.util.FastMap;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.File;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Vector;

public class Parameters {

	protected Map<String,Parameter> par = new FastMap<String,Parameter>();
        protected Map<String,Parameter> iterationValues = new FastMap<String,Parameter>();
        
        /*
         * The iterationMode determines if we should use normal parameter values only (when iterationMode == -1),
         * or if iteration values (where available) should take precedence over normal parameter values. If iterationMode
         * is nonnegative, its value determines the iteration index to use.
         */
        protected int iterationMode = -1;
        
	public static void main(String[] args) {
		System.out.println(args[0]);
		try {
			Parameters p = new Parameters(new File(args[0]));
			//System.out.println("kappa = "+p.get("kappa"));
		} catch (IOException ex) {
			ex.printStackTrace();
		}
				
	}

	/**
	 * Instantiate an empty parameter container.
	 */
	public Parameters() {
		super();
 	}

        public void setIterationMode(int mode) {
            iterationMode = mode;
        }
        
	/**
	 * Instantiate a parameter object with an existing hashmap.
	 * @param par
	 */
	public Parameters(Map par) {
		this.par = par;
	}
	
	/**	
 	* Read parameters from an XML file. 
 	* @param f The file to parse.
 	*/
	public Parameters(File f) throws IOException {
		parseParameters(f);
 	}

	public void parseParameters(File f) throws IOException {
		ParameterHandler ph = new ParameterHandler(par, iterationValues, f);
		SAXParserFactory parserFactory = SAXParserFactory.newInstance();
		try {
			SAXParser parser = parserFactory.newSAXParser();
			parser.parse(f, ph);
		} catch (Exception ex) {
			ex.printStackTrace();
		}	
	}
	
        /**
         * Adds a set of iteration values for a given parameter.
         */
        public void putIterationValues(String key, Parameter values) {
            iterationValues.put(key, values);
        }
        
        /**
         * Returns a Set containing the names of all parameters that have registered
         * sets of iteration values.
         */
        public Set getIterationParameters() {
            return iterationValues.keySet();
        }
        
        /**
         * Returns an ArrayParameter containing the iteration values for the given parameter, or
         * null if there are no iteration values for the parameter.
         */
        public Parameter getIterationValues(String key) {
            return iterationValues.get(key);
        }
        
	public void put(String key, double value) {
		putParameter(key, new SimpleParameter(value));
	}
	
	public void putVariable(String key, double[][] value) {
		par.put(key, new VariableParameter(value));
	}
	
	public void putArray(String key, double[] value) {
		par.put(key, new ArrayParameter(value));	
	}
	/*
        public void putStruct(String key, Parameters struct) {
            par.put(key, struct);
        }
        */
        public void putParameter(String key, Parameter parameter) {
            par.put(key, parameter);
        }
	/**
	 * Check whether the named parameter is defined.
	 * @param key
	 * @return true if the parameter is defined.
	 */
	public boolean containsKey(String key) {
		return par.containsKey(key);
	}
	
	public double get(String key, double t) throws IllegalArgumentException {
            // If we are in iteration mode, check first if we have an iteration value for this parameter:
            if (iterationMode >= 0) {
                Object o = iterationValues.get(key);
                if (o != null) {
                    if (o instanceof ArrayParameter)
                        return (((ArrayParameter)o).getArray(t))[iterationMode];
                }
                    
            }
	
            Parameter p = par.get(key);
            if (p == null) {
                System.err.println("Undefined parameter: "+key);
            }
		
            return p.getValue(t);
             
	}

	public String getString(String key, double t) throws IllegalArgumentException {
		Object o = par.get(key);
		if (o == null) {
			System.err.println("Undefined parameter: "+key);
		}
		return o.toString();
	}
	
	public double[] getArray(String key, double t) {
		Parameter p = par.get(key);
		if (p == null) {
			System.err.println("Undefined parameter: "+key);
		}
		return p.getArray(t);
	}
	
	public boolean getBool(String key, double t) {
            Parameter p = par.get(key);
            if (p == null) {
                    System.err.println("Undefined parameter: "+key);
            }
		
            return p.getBoolean(t);
        }
        
	public boolean getBoolean(String key, double t) {
		return getBool(key, t);
	}
	
        /**
         * Returns the named struct.
         */
        public Parameter getStruct(String key, double t) {
            Object o = par.get(key);
            if (o == null) {
                System.err.println("Undefined parameter: "+key);
            }
            if (o instanceof Object[][]) {
                Object[][] var = (Object[][])o;
                for (int i=0; i<var.length-1; i++)
                    if (t < ((Double)var[i+1][0]).doubleValue())
			return getStruct((String)var[i][1], t);
		return getStruct((String)var[var.length-1][1], t);
            }
            return (Parameter)o;
        }
        
        /**
         * Looks up the named parameter in the named struct.
         */
        public double getFromStruct(String structKey, String paramKey, double t) {
            return getStruct(structKey, t).getFromStruct(paramKey, t);
        }
        
	/**
	 * Returns the parameter in "raw" form. Works for any parameter, but provides no interpretation of
	 * the parameter type (array, variable, boolean, double etc.).
	 * @param key The parameter name.
	 * @return The object defining the parameter.
	 */
	public Parameter getParameter(String key) {
		Parameter p = par.get(key);
		if (p == null) {
			System.err.println("Undefined parameter: "+key);
		}
		return p;
	}
	
	/**
	 * Get the Set of all parameter names.
	 * @return An unmodifiable Set containing all parameters in this instance.
	 */
	public Set keySet() {
		return Collections.unmodifiableSet(par.keySet());
	}
	
	/**
	 * Checks the schedule parameter @param key against the time @param t. If @param t is greater than or equal to
	 * one of the times in the schedule, returns the time since the last scheduled time. If not, returns -1.
	 * @param key The parameter name.
	 * @param t The current time.
	 * @return The time since the last scheduled time, or -1 if none have yet occured.
	 */
	public double getTimeSinceScheduled(String key, double t) {
		Parameter p = par.get(key);
		if (p == null) {
			System.err.println("Undefined parameter: "+key);
		}
		return p.getTimeSinceScheduled(t);
	}
	
	/**
	 * Checks the schedule parameter @param key against the time @param t. If @param t is greater than or equal to
	 * one of the times in the schedule, returns the last scheduled time value. If not, returns -1.
	 * @param key The parameter name.
	 * @param t The current time.
	 * @return The last scheduled value, or -1 if no scheduled times have yet occured.
	 */
	public double getLastScheduledValue(String key, double t) {
		Parameter p = par.get(key);
		if (p == null) {
			System.err.println("Undefined parameter: "+key);
		}
		return p.getLastScheduledValue(t);
	}
	
	public void distortConstantParameter(String key, double coefficientOfVariance) {
		Object o = par.get(key);
		if ((o == null) || !(o instanceof Double))
			return; // Error, the variable doesn't exist, or is not constant.
		// TODO: Should distortConstantParameter throw exception if parameter is not ok?
		double val = ((Double)o).doubleValue();
		val = Functions.distortValue(val, coefficientOfVariance);
	}
	
	public void checkRequiredParameters(String[] pars) throws MissingParameterException {
		for (int i=0; i<pars.length; i++) {
			if (!par.containsKey(pars[i]))
				throw new MissingParameterException(pars[i]);
		}
	}
	
	public Object clone() {
		Map newMap = new FastMap();
                newMap.putAll(par);
                return new Parameters(newMap);
	}
	

	/**
	 * Stores these parameter values in XML format parsable by this class. Only works if
         * all parameters support this operation.
	 * @param f The file to save to.
	 * @throws IOException For any IO exception occuring during the save.
	 */
	public void storeValues(File f) throws IOException {
        Writer fw = new OutputStreamWriter(new FileOutputStream(f), "UTF8");
		fw.write("<?xml version='1.0' encoding='UTF-8'?>\n<parameters>\n");
		for (Iterator i=(new TreeSet(par.keySet())).iterator(); i.hasNext();) {
			String name = (String)i.next();
			Parameter param = par.get(name);
                        param.write(name, fw);

		}
		fw.write("</parameters>\n");
		fw.close();        
    }
}
