/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import deb.fishtank.RotiferMale;
import java.util.LinkedHashSet;
//import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import javolution.util.FastTable;
import deb.operations.Operation;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class Population {

	/*protected ArrayList herbivores = new ArrayList(),
		predators = new ArrayList(),
		operations = new ArrayList();*/
	protected FastTable<Herbivore> herbivores = new FastTable<Herbivore>();
        protected FastTable<Predator> predators = new FastTable<Predator>();
        protected FastTable<Operation> operations = new FastTable<Operation>();
		
		
	protected LinkedHashSet <Herbivore>
		spawnedHerbivores = new LinkedHashSet();
        protected LinkedHashSet
		spawnedPredators = new LinkedHashSet();

	protected boolean simulateReproduction = false,
		simulateMortality = false,
		removeDeadIndividuals = true,
		simulatePredation = true;
	protected BaseFeedDynamic baseFeedDynamic, wasteDynamic;
	protected Parameters par;
    private double herbivoreCount = 0;

    /**
     *
     */
    public Population(BaseFeedDynamic bfd, BaseFeedDynamic wsd, Parameters par) {
        super();
        baseFeedDynamic = bfd;
        wasteDynamic = wsd;
        this.par = par;
    }

        /**
         * Set up this population to use iteration values corresponding to the given iteration index.
         * If the value -1 is given, no iteration values will be used.
         */
        public void setIterationMode(int iterationMode) {
            par.setIterationMode(iterationMode);
        }
        
        public BaseFeedDynamic getBaseFeedDynamic() {
            return baseFeedDynamic;
        }
        
        public BaseFeedDynamic getWasteDynamic() {
            return wasteDynamic;
        }
        
        public Population copy(boolean cloneParameters) throws UnsupportedOperationException {
            //System.out.println("Pop. before: "+getHerbivoreCount()+" / "+getHerbivoreObjectCount());
            Parameters newPar = cloneParameters ? (Parameters)par.clone() : par;
            Population newPop = new Population(baseFeedDynamic.copy(par), wasteDynamic.copy(par), par);
            copyFaunaAndOperations(newPop, cloneParameters);
            newPop.simulateReproduction = this.simulateReproduction;
            newPop.simulateMortality = this.simulateMortality;
            newPop.simulatePredation = this.simulatePredation;
            //System.out.println("Pop. after: "+newPop.getHerbivoreCount()+" / "+newPop.getHerbivoreObjectCount());
            return newPop;
        }
        
        /**
         * Copies all the Individuals and Operations from this Population into the Population given as argument.
         *
         * @param newPop The Population that should receive copies of the objects from this Population.
         */
        public void copyFaunaAndOperations(Population newPop, boolean cloneParameters) {
            if (herbivores.size() > 0) {
                Parameters newPar = cloneParameters ?
                    (Parameters)herbivores.get(0).getParameters().clone() :
                    herbivores.get(0).getParameters();

                for (Herbivore herb : herbivores) {
                    newPop.herbivores.add((Herbivore)herb.copy(newPar));
                }
            }
            if (predators.size() > 0) {
                Parameters newPar = cloneParameters ?
                    (Parameters)predators.get(0).getParameters().clone() :
                    predators.get(0).getParameters();

                for (Predator pred : predators) {
                    newPop.predators.add((Predator)pred.copy(newPar));
                }
            }

            Parameters newPar = cloneParameters ?
                (Parameters)par.clone() : par;

            for (Operation op : operations) {
                newPop.operations.add(op.copy(newPar));
            }
                    
        }
        
        
	public void addHerbivore(Herbivore herb) {
		herbivores.add(herb);
        herbivoreCount += herb.getN();
        //System.out.println("Adding herbivore. Pop count: "+herbivores.size());
	}

	public void addSpawnedHerbivore(Herbivore herb) {
		addHerbivore(herb);
        herbivoreCount += herb.getN();
    }
	
	public void addSpawnedPredator(Predator pred) {
		addPredator(pred);
	}
	
	public void addPredator(Predator pred) {
		predators.add(pred);
	}

	/**
	 * Set whether dead individuals should be removed from the population. If set to false, dead individuals will be retained but not updated.
	 * This setting has no effect if mortality not enabled. Default value is true.
	 * @param value
	 */
	public void setRemoveDeadIndividuals(boolean value) {
		removeDeadIndividuals = value;
	}
	
	/**
	 * Set whether the simulation should include reproduction in individuals. Default is false.
	 * @param value
	 */
	public void setSimulateReproduction(boolean value) {
		simulateReproduction = value;
	}

	/**
	 * Set whether the simulation should include mortality of individuals. Default is false.
	 * @param value
	 */
	public void setSimulateMortality(boolean value) {
		simulateMortality = value;
	}
	
	/**
	 * Iterate through all individuals in the population, and update their states.
	 * @param t The current time.
	 * @param dt The time step.
	 * @param envPar Environmental parameters.
	 */
	public void stepAll(double t, double dt, Parameters envPar) {
		int count = 0, mort = 0;
		double ingestionByHerbivores = 0, ingestionByPredators = 0, wasteProduction = 0;
		
		// First step herbivores:
        	for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
                        Herbivore ind = i.next();
			
			// Check if the individual is alive, if we are simulating mortality.
			if (simulateMortality && ind.isAlive())
				ind.checkMortality(t, dt);
			
			
			if (!simulateMortality || ind.isAlive()) {
							
				IndividualReturnValues retVal = ind.update(t, dt, this, envPar);
				ingestionByHerbivores += ind.getN()*retVal.getBaseFeedConsumption();
                                wasteProduction += ind.getN()*(retVal.getFaecesProduction() - retVal.getWasteConsumption());
                                
				//if (retVal.getWasteConsumption() > 0)
				//	System.out.println(ind.getType().getName()+"\tProd: "+retVal.getFaecesProduction() +"\tCons: "+retVal.getWasteConsumption());
				
				// Check spawning, if we are simulating reproduction.
				if (simulateReproduction && ind.readyToSpawn(t)) {
					//System.out.println("Spawned herbivore from no. "+count);
					Herbivore newHerb = ind.spawnAndGetChild(t);
					// Check if the new individual is viable. The egg may have a chance of not
					// surviving past hatch, and in this case the new individual produced will
					// return false from isAlive().
					if (newHerb.isAlive())
						addSpawnedHerbivore(newHerb);
				}
			} else if (simulateMortality && !ind.isAlive() && removeDeadIndividuals) {
				// If this Individual died but was not eaten, its dry weight should be added in as waste:
				if (!ind.isEaten()) {
					wasteProduction += ind.getN()*ind.getDryWeight(t)/dt;
					//System.out.println("dead: res = "+(ind.getN()*ind.getDryWeight(t)/dt));
					// Dividing by dt to convert from an absolute value to a rate (per time unit).
				}/* else
					System.out.println("eaten");*/
				mort++;
				i.remove();
			}
			count++;
		}
		
		//if (ingestionByHerbivores*dt/getBaseFeedConcentration() > 10)
		//	System.out.println("n="+herbivores.size()+"\ting="+ingestionByHerbivores+"\tbasef="+getBaseFeedConcentration());
		// Update base feed:
                if (baseFeedDynamic != null)
			baseFeedDynamic.update(t, dt, ingestionByHerbivores, envPar);
		
		//if (mort > 10)
		//	System.out.println("Mortality: "+mort);
		
		
		// Predators:

		count = 0;
		// First a questionnaire where each predator reports how much it is planning to eat.
		double[] predation = new double[predators.size()];
                for (int i=0; i<predators.size(); i++) {
		//for (Iterator i=predators.iterator(); i.hasNext();) {
			//Predator ind = (Predator)i.next();
                        Predator ind = (Predator)predators.get(i);
			if (!simulateMortality || ind.isAlive()) {
				predation[count++] = dt*ind.getN()*ind.predation(t, this, envPar);
			}
		}
		
		
		
		// Then we remove the eaten herbivores, calculate how much nutrients each of the
		// predators is going to gain from it, and put the relevant information into each
		// Predator object.
		resolvePredation(predation, t, dt);
			
		count = 0;
		for (Iterator i=predators.iterator(); i.hasNext();) {
			Predator ind = (Predator)i.next();
			// Check if the individual is alive, if we are simulating mortality.
			if (simulateMortality && ind.isAlive()) {
				ind.checkMortality(t, dt);			
				
				// Step the predator:
				IndividualReturnValues retVal = ind.update(t, dt, this, envPar);
				wasteProduction += ind.getN()*retVal.getFaecesProduction();
				
				// Check spawning, if we are simulating reproduction.
				if (simulateReproduction && ind.readyToSpawn(t)) {
					//System.out.println("Spawned herbivore from no. "+count);
					Predator newPred = ind.spawnAndGetChild(t);
					addSpawnedPredator(newPred);
				}
			} else if (simulateMortality && !ind.isAlive() && removeDeadIndividuals) {
				//System.out.println("Herbivore no. "+count+" died.");
				mort++;
				i.remove();
			}


			count++;	
		}
		
		// Perform necessary operations, if any:
		for (Iterator it=operations.iterator(); it.hasNext();) {
			Operation op = (Operation)it.next();
			if (op.performNow(t, dt))
				op.performOperation(t, dt, this, envPar);
		}
		
		// Resolve spawnings:
		for (Iterator<Herbivore> it=spawnedHerbivores.iterator(); it.hasNext();) {
			Herbivore o = it.next();
			herbivores.add(o);
			it.remove();
		}
		for (Iterator it=spawnedPredators.iterator(); it.hasNext();) {
			Object o = it.next();
			predators.add((Predator)o);
			it.remove();
		}

		// Step waste dynamic:
		wasteDynamic.update(t, dt, -wasteProduction, envPar);

        // Count the current number of herbivores
        updateHerbivoreCount();
    }

	/**
	 * Picks out the given number of herbivores, removes them from the population, and
	 * gathers information on the total nutritional contents of the prey.
	 * @param nEaten The number of herbivores eaten.
	 */
	public void resolvePredation(double[] predation, double t, double dt) {
		//NutrientIntake[] res = new NutrientIntake[predation.length]; 
		for (int i=0; i<predation.length; i++) {
			NutrientIntake ni = removeHerbivoresAtRandom(predation[i], t);
			
			// Tell the predator about the nutrients it is about to gain:
			Predator pred = getPredator(i);
			ni.convertToFlux(dt*pred.getN());
			//System.out.println(pred.getN());
			pred.setNutrientIntake(ni);
		}
	}
	
	/**
	 * This method removes n individuals from the herbivore population, by random.
	 */
	public NutrientIntake removeHerbivoresAtRandom(double amount, double t) {
                double amount_0 = amount;
            	NutrientIntake res = new NutrientIntake();
		loop: while ((amount > 0) && (herbivores.size() > 0)) {
                    	int n = (int)(Math.random()*herbivores.size());
			if (n >= herbivores.size())
				continue loop;
			Herbivore herb = getHerbivore(n);
			double ant = herb.getN();
			if (ant > amount) {
				// This Individual contains enough to satisfy the rest of amount. Reduce its numbers and finish.
				// Since the Individual has not been removed from the simulation, there is no danger of the eaten
				// part of being calculated as waste.
				herb.setN(ant-amount);
				
				herb.addNutritionalContent(res, amount, t);
				amount = 0;
			} else {
				// This Individual does not contain enough, so we kill it off completely.
				amount -= ant;
				herb.addNutritionalContent(res, ant, t);
				
				herb.eat();
			}
		}
                //System.out.println(res.get("protein")/amount_0);
                res.add("numbers", amount_0-amount);
		return res;
	}

	public void resetAll() {
		for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
			i.next().reset();
		}
		for (Iterator i=predators.iterator(); i.hasNext();) {
				((Individual)i.next()).reset();
		}
		
	}

	/**
	 * Returns a compact collection of all the bio states of all herbivores. 
	 * @return A two-dimensional array containing all states of each individual.
	 */
	public double[][] getHerbivoreBioStates() {
		double[][] res = new double[herbivores.size()][];
		int piv = 0;
		for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
			res[piv++] = i.next().getBioStates();
		}
		return res;
	}
	
	/**
	 * Returns a compact collection of all the bio states of all herbivores. 
	 * @return A two-dimensional array containing all states of each individual.
	 */
	public double[] getHerbivoreAges(double t) {
		double[] res = new double[herbivores.size()];
		int piv = 0;
		for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
			res[piv++] = i.next().getAge(t);
		}
		return res;
	}
	
	public double getMeanHerbivoreAge(double t) {
		double total = 0, totalN = 0;
		for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
			Individual ind = i.next(); 
			total += ind.getAge(t)*ind.getN();
			totalN += ind.getN();
		}
		return total/totalN;
	}
	
	/**
	 * Returns the number of herbivores represented in this population, taking into account each Herbivore object's
	 * internal population count.
	 * @return
	 */
	public double getHerbivoreCount() {
		return herbivoreCount;
	}

 	private void updateHerbivoreCount() {
		double total = 0;
		for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
			total += i.next().getN();
		}
		herbivoreCount = total;
	}

    /**
	 * Returns the number of Herbivore objects contained in this population, without regard to their internal population counts.
	 * @return
	 */
	public int getHerbivoreObjectCount() {
		return herbivores.size();
	}
	
	/**
	 * Returns the number of Predator objects contained in this population, without regard to their internal population counts.
	 * @return
	 */
	public int getPredatorObjectCount() {
		return predators.size();
	}
	
	/**
	 * Returns a specific Herbivore. Note that this method is not efficient for large populations, because 
	 * it needs to iterate the linked list to find the given individual.
	 * @param no The index of the herbivore to find.
	 * @return A Herbivore, or null if the index is out of bounds.
	 */
	public Herbivore getHerbivore(int no) {
		if (no < herbivores.size())
			return herbivores.get(no);
		else
			return null;
	}

	/**
	 * Returns the entire List of herbivores.
	 * @return
	 */
	public List getHerbivores() {
		return herbivores;
	}
	
	/**
	 * Returns a List of all herbivores with type.equals(Individual.getType()). This method can be used to
	 * extract one species/strain from a mixed population.
	 * @return
	 */
	public List getMatchingHerbivores(Strain type) {
		FastTable al = new FastTable();
                for (int i=0; i<herbivores.size(); i++) {
		//for (Iterator i=herbivores.iterator(); i.hasNext();) {

                    //Herbivore herb = (Herbivore)i.next();
                    Herbivore herb = herbivores.get(i);
                    if (type == herb.getType()) {
                        al.add(herb);
                    }
		}
		return al;
	}
	
	/**
	 * Returns a specific Herbivore. Note that this method is not efficient for large populations, because 
	 * it needs to iterate the linked list to find the given individual.
	 * @param no The index of the herbivore to find.
	 * @return A Herbivore, or null if the index is out of bounds.
	 */
	public Predator getPredator(int no) {
		if (no < predators.size())
			return (Predator)predators.get(no);
		else
			return null;
	}
	
	
	/**
	 * Returns a state set representing the mean of the states in the herbivore population. Beware that
	 * this can be a time-consuming operation for large populations. This operation can only be used if
	 * all herbivores are comparable, e.g. have the same state space.
	 * TODO: Enable weighting if herbivores' internal population counts differ.
	 * @return
	 */
	public double[] getMeanHerbivoreStates() {
		int count = getHerbivoreObjectCount();
		if (count == 0)
			return null;
		double totalN = 0;
		Iterator<Herbivore> i = herbivores.iterator();
		Herbivore herb = i.next();
		totalN += herb.getN();
		// Get the states of the first one, and clone it to avoid messing it up.
		double[] sum = (double[])(herb.getBioStates().clone());
		for (int index=0; index<sum.length; index++)
			sum[index] *= totalN;
		for (; i.hasNext();) {
			herb = (Herbivore)i.next();
			totalN += herb.getN();
			StepSolver.multAndAdd(sum, herb.getN(), herb.getBioStates());
		}
		//if (totalN < 1)
		//System.out.println("totalN="+totalN);
		for (int j=0; j<sum.length; j++)
			sum[j] /= totalN;
		
		return sum;
	}
	
	/**
	 * Returns a state set representing the mean of the states in the herbivore population. Beware that
	 * this can be a time-consuming operation for large populations. This operation can only be used if
	 * all herbivores are comparable, e.g. have the same state space.
	 * TODO: Enable weighting if herbivores' internal population counts differ.
	 * @return
	 */
	public double[] getMeanPredatorStates() {
		int count = getPredatorObjectCount();
		if (count == 0)
			return null;
		double totalN = 0;
		Iterator i = predators.iterator();
		Predator herb = (Predator)i.next();
		totalN += herb.getN();
		// Get the states of the first one, and clone it to avoid messing it up.
		double[] sum = (double[])(herb.getBioStates().clone());
		for (int index=0; index<sum.length; index++)
			sum[index] *= totalN;
		for (; i.hasNext();) {
			herb = (Predator)i.next();
			totalN += herb.getN();
			StepSolver.multAndAdd(sum, herb.getN(), herb.getBioStates());
		}
		//if (totalN < 1)
		//System.out.println("totalN="+totalN);
		for (int j=0; j<sum.length; j++)
			sum[j] /= totalN;
		
		return sum;
	}
	

	/**
	 * Returns a state set representing the mean of the states in the herbivore population. Beware that
	 * this can be a time-consuming operation for large populations. This operation can only be used if
	 * all herbivores are comparable, e.g. have the same state space.
	 * @return
	 */
	public double[][] getMeanAndStdevHerbivoreStates() {
		double[] means = getMeanHerbivoreStates();
		if (means == null)
			return null;
		double[] sqSum = new double[means.length];
		double[][] states = getHerbivoreBioStates();
		if (states.length == 0)
			return null;
		for (int ind=0; ind<states.length; ind++)
			for (int st=0; st<states[ind].length; st++) {
				double diff = states[ind][st] - means[st];
				sqSum[st] += diff*diff;
			}
		double[][] res = new double[2][means.length];
		for (int i=0; i<res[0].length; i++) {
			res[0][i] = means[i];
			if (states.length >= 2)
				res[1][i] = Math.sqrt(sqSum[i]/(states.length-1));
			else
				res[1][i] = 0;
			//System.out.println(sqSum[i]+"\t"+states.length);
		}
		return res;
	}
	
	
	public double[][] getHerbivoreDerivedOutputs(double t) {
		double[][] res = new double[herbivores.size()][];
		int piv = 0;
		for (Iterator<Herbivore> i=herbivores.iterator(); i.hasNext();) {
			res[piv++] = i.next().getDerivedOutputs(t);
		}
		return res;
	}
	
	public String[] getHerbivoreDerivedOutputNames(int no) {
		if (no >= getHerbivoreCount())
			return null;
		return getHerbivore(no).getDerivedOutputNames();
	}
	
	/**
	 * Returns a compact collection of all the bio states of all herbivores. 
	 * @return A two-dimensional array containing all states of each individual.
	 */
	public double[][] getPredatorBioStates() {
		double[][] res = new double[predators.size()][];
		int piv = 0;
		for (Iterator i=predators.iterator(); i.hasNext();) {
			res[piv++] = ((Individual)i.next()).getBioStates();
		}	
		return res;
	}
	
	public double getBaseFeedConcentration() {
		return baseFeedDynamic.getLevel();
	}
	
	public double getWasteLevel() {
		return wasteDynamic.getLevel();
	}
	
	/**
	 * @param simulatePredation The simulatePredation to set.
	 */
	public void setSimulatePredation(boolean simulatePredation) {
		this.simulatePredation = simulatePredation;
	}

	/**
	 * @return
	 */
	public List getPredators() {
		return predators;
	}
	
	/**
	 * Adds an Operation to this population. At each time step the Operation will be given the chance to
	 * be performed on the population.
	 * When several operations are added to the same population, they are performed in a predictable order,
	 * by default in the same order as they are added.
	 * @param op The Operation to add.
	 */
	public void addOperation(Operation op) {
		operations.add(op);
	}
	
	public void addOperation(int index, Operation op) {
		operations.add(index, op);
	}
	
	public void removeOperation(Operation op) {
		operations.remove(op);
	}
	
	public void removeOperation(int index) {
		operations.remove(index);
	}
	
	public int getOperationCount() {
		return operations.size();
	}

	/**
	 * @return
	 */
	public double[][] getPredatorDerivedOutputs(double t) {
		double[][] res = new double[predators.size()][];
		int piv = 0;
		for (Iterator i=predators.iterator(); i.hasNext();) {
			res[piv++] = ((Individual)i.next()).getDerivedOutputs(t);
		}
		return res;
	}
        
        //public Population clone() {
            
        //}
}
