/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.util.ArrayList;
import java.util.Iterator;

import javolution.util.FastTable;
import deb.control.Controller;

/**
 * @author alver
 *         <p/>
 *         To change the template for this generated type comment go to
 *         Window>Preferences>Java>Code Generation>Code and Comments
 */
public class Simulator extends Thread {

    Population pop;
    Parameters envPar = null;
    ArrayList listeners = new ArrayList();
    boolean startedIteration;
    FastTable<Population> iterPops;
    FastTable<FastTable<SimulationMonitor>> monitorLists;

    int monitorInterval = 1, recombineInterval = 0;
    private double dt = 1, t_0 = 0, t_end = 1, t, iterationStartTime = Double.MAX_VALUE;
    boolean running = false, stopped = false, reportSimulationFinished = false,
            paused = false, recombine = false;
    private SimulationTrigger trigger;
    private boolean externalTriggingEnabled;
    private int ticksLeft = 0;
    private int printTimePiv = 0, printTimeInterval = 50, iterations = -1;
    private Controller controller = null;    
    private double lastPredictiveControllerRun = 0;
//  I ADDED THIS...
//    private Controller ref_controller = null;
//    private double lastReferenceControllerRun = 0;

    /**
     *
     */
    public Simulator(Population pop) {
        super();
        this.pop = pop;
    }

    /**
     * Adds a simulation monitor. This class will be notified on regular intervals while the simulation is going on.
     *
     * @param mon The simulation monitor.
     */
    public void addSimulationMonitor(SimulationMonitor mon) {
        listeners.add(mon);
    }

    public void setSimulationMonitorInterval(int steps) {
        monitorInterval = steps;
    }

    public void setSimulationBounds(double t_0, double t_end, double dt) {
        this.dt = dt;
        this.t_0 = t_0;
        this.t_end = t_end;
    }

    public void setIteration(double startTime, int iterations) {
        iterationStartTime = startTime;
        this.iterations = iterations;
    }

    /**
     * Set a predictive controller for this simulation. The controller is allowed to run test simulations
     * on copies of the population for purposes of determining input values.
     * @param controller
     */
    public void setPredictiveController(Controller controller) {

        this.controller = controller;
    }

    /**
     * Set a reference controller for this simulation. The controller is allowed to run test simulations
     * on copies of the population for purposes of determining input values.
     * @param controller
     */
/*
    public void setReferenceController(Controller ref_controller) {

        this.ref_controller = ref_controller;
    }
*/
    /**
     * @return Returns the dt.
     */
    public double getDt() {
        return dt;
    }

    /**
     * @return Returns the t_0.
     */
    public double getT_0() {
        return t_0;
    }

    /**
     * @return Returns the t_end.
     */
    public double getT_end() {
        return t_end;
    }

    /**
     * Run the simulation.
     *
     * @param envPar The environmental parameter set.
     */
    public void runSimulation(Parameters envPar) {
        running = true;
        stopped = false;
        this.envPar = envPar;
        start();
    }

    public void stopSimulation(boolean reportSimulationFinished) {
        stopped = true;
        running = false;
        this.reportSimulationFinished = reportSimulationFinished;
    }

    public void reportSimulationInterval() {
        for (Iterator itt = listeners.iterator(); itt.hasNext();) {
            ((SimulationMonitor) itt.next()).simulationInterval(t, pop);
        }
    }

    public void run() {
        pop.resetAll();

        boolean shouldIterate = iterations > 0;
        startedIteration = false;
        iterPops = new FastTable<Population>();
        monitorLists = new FastTable<FastTable<SimulationMonitor>>();
        lastPredictiveControllerRun = t_0;
//        
//        lastReferenceControllerRun = t_0;
//        
        final int steps = Math.round((float) ((t_end - t_0) / dt));
        int piv = 0;
        t = t_0;

        // Notify all simulation monitors that the simulation starts.
        for (Iterator itt = listeners.iterator(); itt.hasNext();) {
            ((SimulationMonitor) itt.next()).simulationStarted(t, pop);
        }

        mainLoop:
        for (int i = 0; i < steps; i++) {

            // If we are running on external triggering, check if we should be paused:
            if (externalTriggingEnabled) {
                if (ticksLeft == 0) {
                    if (trigger != null)
                        trigger.ticksCompleted();
                    setPaused(true);
                } else
                    adjustTicksLeft(-1);
            }

            if (!running)
                break mainLoop;
            while (paused) {
                try {
                    sleep(200);
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                }
            }

            // If we have a predictive controller, check if it is time for it to run. If so, make the
            // run and update the time stamp:
            if ((controller != null) &&
                    (t - lastPredictiveControllerRun >= controller.getCallInterval())) {
                // We refuse to run the predictive controller if an iteration has been started. This could
                // be allowed in the future:
                if (startedIteration)
                    System.err.println("Cannot call predictive controller after iteration has started!");
                else {
                    controller.makeControlRun(t, this, pop, envPar);
                    lastPredictiveControllerRun = t;
                }
            }
/*            
//          If we have a reference controller, check if it is time for it to run. If so, make the
            // run and update the time stamp:
            if ((ref_controller != null) &&
                    (t - lastReferenceControllerRun >= ref_controller.getCallInterval())) {
                // We refuse to run the predictive controller if an iteration has been started. This could
                // be allowed in the future:
                if (startedIteration)
                    System.err.println("Cannot call reference controller after iteration has started!");
                else {
                    ref_controller.makeControlRun(t, this, pop, envPar);
                    lastReferenceControllerRun = t;
                }
            }
*/

            if (!startedIteration) {
                pop.stepAll(t, dt, envPar);
            } else {
                for (int num = 0; num < iterPops.size(); num++) {
                    Population p = iterPops.get(num);
                    envPar.setIterationMode(num);
                    p.setIterationMode(num);
                    p.stepAll(t, dt, envPar);
                    //System.out.println("t="+t+"\t iter="+num);
                }
            }
            //t += dt;

            //System.out.println(t+"\t"+(t_0 + (t_end - t_0)*((double)(i+1)/(double)steps)));
            dt = t;
            t = t_0 + (t_end - t_0) * ((double) (i + 1) / (double) steps);
            dt = t - dt;

            if (shouldIterate && !startedIteration && (t >= iterationStartTime)) {
                startedIteration = true;
                System.out.println(iterations + " iterations started.");
                for (int num = 0; num < iterations; num++) {
                    Population newPop = pop.copy(false);
                    iterPops.add(newPop);
                    FastTable<SimulationMonitor> monitors = new FastTable<SimulationMonitor>();
                    for (Iterator itt = listeners.iterator(); itt.hasNext();) {
                        monitors.add(((SimulationMonitor) itt.next()).copy(newPop.par, String.valueOf(num)));
                    }
                    monitorLists.add(monitors);
                }

            }

            piv++;

            if (piv == monitorInterval) {
                if (!startedIteration) {

                    // Notify all normal simulation monitors.
                    if (i + 1 < steps) for (Iterator itt = listeners.iterator(); itt.hasNext();) {
                        ((SimulationMonitor) itt.next()).simulationUpdate(t, pop);
                    }
                    printTimePiv++;
                    if (printTimePiv == printTimeInterval) {
                        double nEach = pop.getHerbivoreCount() / ((double) pop.getHerbivoreObjectCount());
                        System.out.println("t = " + t + "\t\tN each = " + nEach);
                        printTimePiv = 0;
                    }
                } else {
                    // We are iterating, so we need to instruct all sets of monitors to record values.
                    int num = 0;
                    for (FastTable<SimulationMonitor> monitorSet : monitorLists) {
                        for (SimulationMonitor monitor : monitorSet) {
                            monitor.simulationUpdate(t, iterPops.get(num));
                            //System.out.println("Ping, num="+num);
                        }
                        num++;
                    }
                }
                piv = 0;

            }

        }

        // Call the simulation monitors if the simulation ended as planned, or if it was stopped
        // externally but we were told to report simulation finished.
        if (!stopped || reportSimulationFinished)
            reportSimulationFinished(t);
    }

    protected void reportSimulationFinished(double t) {
        // Notify all simulation monitors that the simulation is finished.
        if (!startedIteration) {
            for (Iterator itt = listeners.iterator(); itt.hasNext();) {
                ((SimulationMonitor) itt.next()).simulationFinished(t, pop);
            }
        } else {
            // We have been iterating, so we need to instruct all sets of monitors to record values.
            int num = 0;
            for (FastTable<SimulationMonitor> monitorSet : monitorLists) {
                for (SimulationMonitor monitor : monitorSet) {
                    monitor.simulationFinished(t, iterPops.get(num));
                    //System.out.println("Finishing, num="+num);
                }

                num++;
            }
        }
    }

    public void setExternalTriggingEnabled(boolean enabled) {
        externalTriggingEnabled = enabled;
    }

    public void trig(SimulationTrigger trigger, int steps) {
        this.trigger = trigger;
        adjustTicksLeft(steps);
        setPaused(false);
    }

    private synchronized void adjustTicksLeft(int adj) {
        ticksLeft += adj;
    }

    /**
     * @param paused The paused to set.
     */
    public void setPaused(boolean paused) {
        this.paused = paused;
    }

    /**
     * @return Returns the paused.
     */
    public boolean isPaused() {
        return paused;
    }
}
