/*
 * Created on 28.apr.2005
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class SingleStrainSimulationMonitor implements SimulationMonitor {

	protected ArrayList count, meanStates;
	Strain strain;
	boolean herbivore;
	String directory; 
	String filePrefix;
	
	/**
	 * 
	 */
	public SingleStrainSimulationMonitor(String directory, String filePrefix, Strain strain, boolean herbivore) {
		this.directory = directory;
		this.filePrefix = filePrefix;
		this.strain = strain;
		this.herbivore = herbivore;
		count = new ArrayList();
		meanStates = new ArrayList();
	}


        
	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationStarted(double, deb.Population)
	 */
	public void simulationStarted(double t, Population pop) {
		record(t, pop);
	}

	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationUpdate(double, deb.Population)
	 */
	public void simulationUpdate(double t, Population pop) {
		record(t, pop);
	}

	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationInterval(double, deb.Population)
	 */
	public void simulationInterval(double t, Population pop) {
		simulationFinished(t, pop);
	}

	/* (non-Javadoc)
	 * @see deb.SimulationMonitor#simulationFinished(double, deb.Population)
	 */
	public void simulationFinished(double t, Population pop) {
		//record(t, pop);
		
		try {
			// Store count:
			File countFile = new File(directory+File.separator+filePrefix+"Count.dlm");
			Functions.saveToFile(count, countFile);
			
			File meanFile = new File(directory+File.separator+filePrefix+"Mean.dlm");
			Functions.saveToFile(meanStates, meanFile);
			
		} catch (IOException ex) {
			ex.printStackTrace();
		}
	}
	
	protected void record(double t, Population pop) {
		
		// First gather a List of all individuals that should be stored:
		List toStore = new ArrayList();
		if (herbivore) {
			List herbivores = pop.getHerbivores();
			for (Iterator i=herbivores.iterator(); i.hasNext();) {
				Herbivore herb = (Herbivore)i.next();
				if (herb.getType() == strain) {
					toStore.add(herb);
				}
			}
		} else {
			List predators = pop.getPredators();
			for (Iterator i=predators.iterator(); i.hasNext();) {
				Predator pred = (Predator)i.next();
				if (pred.getType() == strain) {
					toStore.add(pred);
				}
			}
		}
		
		double n = Functions.countIndividuals(toStore);
		count.add(new Double(n));
		
		double[][] states = Functions.getBioStates(toStore);
		double[] mean = Functions.mean(states);
		meanStates.add(mean);
		
		//	herbivoreStates.add(pop.getHerbivoreBioStates());
	}
	
	public SimulationMonitor copy(Parameters par, String directory) {
            SingleStrainSimulationMonitor newMon = new SingleStrainSimulationMonitor(directory, filePrefix, strain, herbivore);
            newMon.count = (ArrayList)count.clone();
            newMon.meanStates = (ArrayList)meanStates.clone();
            return newMon;
        }
}
