/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public abstract class StepSolver {

	final static double f_16 = 1d/6d,
		f_13 = 1d/3d;

	/**
	 * Calculates the next time step for a herbivore individual using Runge-Kutta 4 calculation.
	 * @param hrb
	 * @param dt
	 * @param states
	 * @param envPar
	 * @return The state value after the step.
	 */
	public static double stepBaseFeedDynamic(BaseFeedDynamic hrb, double t, double dt, double concentration, double predation, Parameters envPar) {
		//System.out.println("> "+concentration);
		//double[] states = new double[] {concentration};
		//double[] res = new double[states.length];
		//double[] k = new double[states.length];
		//double[] tmp = new double[states.length];
		double k, res, tmp;
		//System.out.println("** "+predation);
		res = concentration;
		k = hrb.f(t, concentration, predation, envPar); // k*dt = k1
		res += f_16*dt*k; // add k1/6 to result
		tmp = concentration + 0.5d*dt*k; // tmp = x+k1/2 
		//k = hrb.f(t, concentration, tmp, envPar); // k*dt = k2
		k = hrb.f(t, tmp, predation, envPar); // k*dt = k2
		res += f_13*dt*k; // add k2/3 to result
		tmp = concentration + 0.5d*dt*k; // tmp = x+k2/2 
		k = hrb.f(t, tmp, predation, envPar); // k*dt = k3
		res += f_13*dt*k; // add k3/3 to result
		tmp = concentration + dt*k; // tmp = x+k3 
		k = hrb.f(t, tmp, predation, envPar); // k*dt = k4
		res += f_16*dt*k; // add k4/6 to result
		//System.out.println(">>> "+res);
		return res;
	}
	
	/**
	 * Calculates the next time step for a herbivore individual using Runge-Kutta 4 calculation.
	 * @param hrb
	 * @param dt
	 * @param states
	 * @param envPar
	 * @return The state value after the step.
	 */
	public static double[] stepIndividual(Individual hrb, double t, double dt, double[] states, Population population, Parameters envPar) {
		double[] res = new double[states.length];
		double[] k = new double[states.length];
		double[] tmp = new double[states.length];
       
                setEqual(res, states);
                hrb.f(k, t, states, population, envPar); // k*dt = k1
		multAndAdd(res, f_16*dt, k); // add k1/6 to result
		setEqual(tmp, states);
		multAndAdd(tmp, 0.5d*dt, k); // tmp = x+k1/2 
		hrb.f(k, t+dt/2d, tmp, population, envPar); // k*dt = k2
		multAndAdd(res, f_13*dt, k); // add k2/3 to result
		setEqual(tmp, states);
		multAndAdd(tmp, 0.5d*dt, k); // tmp = x+k2/2 
		hrb.f(k, t+dt/2d, tmp, population, envPar); // k*dt = k3
		multAndAdd(res, f_13*dt, k); // add k3/3 to result
		setEqual(tmp, states);
		multAndAdd(tmp, dt, k); // tmp = x+k3 
		hrb.f(k, t+dt, tmp, population, envPar); // k*dt = k4
		multAndAdd(res, f_16*dt, k); // add k4/6 to result
		
                
		return res;
	}
	
        public static double stepScalar(ScalarEquation fun, double t, double dt, double state, Parameters par) {
            double res = 0, k = 0, tmp = 0;
            
            k = fun.f(t, state, par); // k*dt = k1
            res = state + f_16*dt*k; // add k1/6 to result
            tmp = state + 0.5d*dt*k; // tmp = x+k1/2 
            k = fun.f(t+dt/2d, tmp, par); // k*dt = k2
            res += f_13*dt*k; // add k2/3 to result
            tmp = state + 0.5d*dt*k; // tmp = x+k2/2 
            k = fun.f(t+dt/2d, tmp, par); // k*dt = k3
            res += f_13*dt*k; // add k3/3 to result
            tmp = state + dt*k; // tmp = x+k3 
            k = fun.f(t+dt, tmp, par); // k*dt = k4
            res += f_16*dt*k; // add k4/6 to result
            //System.out.println("k="+k+"\t dt="+dt+"\t diff="+(res-state));
            
            return res;
        }
	
	/**
	 * Add the contents of an array to another.
	 * @param res Array to add to.
	 * @param arg Array to add.
	 */
	public static void add(double[] res, double[] arg) {
		for (int i=0; i<res.length; i++)
			res[i] += arg[i];
	}

	/**
	 * Copy the contents of an array into another.
	 * @param res Array to copy into.
	 * @param arg Array to copy from.
	 */
	public static void setEqual(double[] res, double[] arg) {
		for (int i=0; i<res.length; i++)
			res[i] = arg[i];
	}
	
	/**
	 * Multiply an array with a scalar and add to another array.
	 * @param res The vector to add to.
	 * @param mult The scalar to multiply by.
	 * @param arg The vector to multiply and add.
	 */
	public static void multAndAdd(double[] res, double mult, double[] arg) {
		for (int i=0; i<res.length; i++)
			res[i] += mult*arg[i];
	}
	
	/**
	 * Multiply an array with a scalar and put the result into another array.
	 * @param res The vector to add to.
	 * @param mult The scalar to multiply by.
	 * @param arg The vector to multiply and add.
	 */
	public static void multAndSet(double[] res, double mult, double[] arg) {
		for (int i=0; i<res.length; i++)
			res[i] = mult*arg[i];
	}
}
