package deb.control;

import deb.Population;
import deb.Parameters;
import deb.Simulator;

/**
 * Abstract partial implementation of the PredictiveController interface. Provides the mechanics for
 * easily making a single predictive run.
 */
public abstract class AbstractPredictiveController implements PredictiveController {


    public double getPredictionInterval() {
        return 1;
    }

    public Population makeRun(double t, double t_end, Simulator sim, Population origPop, Parameters envPar) {
        Population pop = origPop.copy(false);

        double dt = sim.getDt(), t_0 = t;

        final int steps = Math.round((float) ((t_end - t_0) / dt));
        t = t_0;

        for (int i = 0; i < steps; i++) {

            dt = t;
            t = t_0 + (t_end - t_0) * ((double) (i + 1) / (double) steps);
            dt = t - dt;
            pop.stepAll(t, dt, envPar);
        }

        return pop;
    }

    public int getStepCount(double t, double t_end, Simulator sim) {
       return Math.round((float) ((t_end - t) / sim.getDt()));
    }


    public double[] makeRunWithMetric(double t, double t_end, Simulator sim, Population origPop, Parameters envPar,
                                          PopulationMetric metric, String parameterName, double parameterValue) {

        return makeRunWithMetric(t, t_end, sim, origPop, envPar, metric, parameterName, new double[] {parameterValue});
    }
    

    public double[] makeRunWithMetric(double t, double t_end, Simulator sim, Population origPop, Parameters envPar,
                                      PopulationMetric metric, String parameterName, double[] parameterValues) {

        Population pop = origPop.copy(false);

        double dt, t_0 = t;

        final int steps = getStepCount(t, t_end, sim);
        double[] values = new double[steps];
        t = t_0;

        for (int i = 0; i < steps; i++) {

            envPar.put(parameterName, parameterValues[Math.min(parameterValues.length-1, i)]);

            dt = t;
            t = t_0 + (t_end - t_0) * ((double) (i + 1) / (double) steps);
            dt = t - dt;
            pop.stepAll(t, dt, envPar);

            // Store metric:
            values[i] = metric.getValue(pop);
        }

        return values;
    }
}
