package deb.control;

import deb.Parameters;
import deb.Population;
import deb.math.Matrix;

/**
 * Created by IntelliJ IDEA.
 * User: alver
 * Date: Nov 29, 2006
 * Time: 4:31:16 PM
 * To change this template use File | Settings | File Templates.
 */
public class ArmaxMinimumVarianceController implements InternalController {

    double u_value = 0;
    private double[] bf, g;
    private double bf0;
    private double inputLowerBound;
    private double inputUpperBound;
    private PopulationMetric metric;
    private double[] y_h, u_h;

    int n_g, n_bf, piv_y = 0, piv_u = 0;
    private double[] y_hist, u_hist;
    private Parameters par;
    private String referencePar;

    public ArmaxMinimumVarianceController(Parameters par, String referencePar, String bf, String g, PopulationMetric metric,
                                          double inputLowerBound, double inputUpperBound) {
        this.par = par;
        this.referencePar = referencePar;

        double[] bf_coef = parseMatrix(bf);
        double[] g_coef = parseMatrix(g);

        /*for (int i = 0; i < a_coef.length; i++) {
            double v = a_coef[i];
            System.out.println("a["+i+"] = "+v);
        }
        for (int i = 0; i < b_coef.length; i++) {
            double v = b_coef[i];
            System.out.println("b["+i+"] = "+v);
        }*/

        init(bf_coef, g_coef, metric, inputLowerBound, inputUpperBound);
    }

    public ArmaxMinimumVarianceController(String referencePar, double[] bf, double[] g, PopulationMetric metric,
                                          double inputLowerBound, double inputUpperBound) {
        this.referencePar = referencePar;
        init(bf, g, metric, inputLowerBound, inputUpperBound);
    }

    private void init(double[] bf_, double[] g, PopulationMetric metric,
                                          double inputLowerBound, double inputUpperBound) {

        this.g = g;
        this.inputLowerBound = inputLowerBound;
        this.inputUpperBound = inputUpperBound;
        // We split b into the first:
        bf0 = bf_[0];
        // ... and the rest:
        this.bf = new double[bf_.length-1];
        for (int i = 0; i < this.bf.length; i++) {
            this.bf[i] = bf_[i+1];
        }

        this.metric = metric;
        n_g = this.g.length;
        n_bf = this.bf.length;
        y_h = new double[n_g];
        y_hist = new double[n_g];
        u_h = new double[n_bf];
        u_hist = new double[n_bf];

        updateHistY(3.3e5);
        updateHistY(3.3e5);
        updateHistY(3.3e5);
        updateHistY(3.3e5);
        updateHistY(3.3e5);
        updateHistY(3.3e5);
        updateHistY(3.3e5);
        updateHistU(0.5);
        updateHistU(0.5);
        updateHistU(0.5);
        updateHistU(0.5);
        updateHistU(0.5);
        updateHistU(0.5);
        updateHistU(0.5);
    }


    public double getValue() {
        return u_value;
    }

    public void update(Object system, double t, double dt, Parameters envPar) {
        // Obtain the latest output value:
        double y = metric.getValue((Population)system);// - par.get(referencePar, t);  /// !!! Subtracting reference, so y should approach 0 !!!
        System.out.println("Current pop: "+y);
        // Add it to our rolling array:
        updateHistY(y);
        // Set our y-history and u-history vectors:
        buildYUVectors();

        System.out.println("y history: ");
        Matrix.printMatrix(y_h);
        System.out.println("u history: ");
        Matrix.printMatrix(u_h);

        u_value = calculateInput(par.get(referencePar, t));
        System.out.println("Input (unbounded): "+u_value);
        // Bound input:
        u_value = Math.min(inputUpperBound, Math.max(inputLowerBound, u_value));
        System.out.println("Calculated input: "+u_value);
        updateHistU(u_value);
    }

    private double calculateInput(double reference) {
        System.out.println("scalarprod y_h x G: "+Matrix.scalarProd(y_h, g));
        System.out.println("scalarprod u_h x BF: "+Matrix.scalarProd(u_h, bf));


        //return (reference + Matrix.scalarProd(y_h, g) - Matrix.scalarProd(u_h, bf));
        return (1/bf0)*(reference*Matrix.sumVector(g) - Matrix.scalarProd(y_h, g) - Matrix.scalarProd(u_h, bf));
    }

    private void buildYUVectors() {
        int piv = 1;
        for (int i=0; i<n_g; i++) {

            y_h[n_g - piv] = y_hist[(i+piv_y) % n_g];
            piv++;
        }
        piv = 1;
        for (int i=0; i<n_bf; i++) {

            u_h[n_bf - piv] = u_hist[(i+piv_u) % n_bf];
            piv++;
        }
    }

    /**
     * Update the array of the latest y.
     * @param y The output of time step t
     */
    private void updateHistY(double y) {
        y_hist[piv_y] = y;
        piv_y++;
        if (piv_y == n_g)
            piv_y = 0;

    }

    /**
     * Update the array of the latest u.
     * @param u The input of last time step.
     */
    private void updateHistU(double u) {
        u_hist[piv_u] = u;
        piv_u++;
        if (piv_u == n_bf)
            piv_u = 0;
    }

    public InternalController copy(Parameters par) {
        throw new UnsupportedOperationException("ArmaxMinimumVarianceController");
    }

    public double getCallInterval() {
        return 1;

    }

    /**
     * Parse a Matlab-style polynomial for an ARX model to acquire the coefficients.
     * @param a_str The polynomial description as output by Matlab
     * @param offset The number of initial empty terms in the polynomial (if the lowest time delay is more than one time step).
     * @return An array containing the coefficients.
     */
    private double[] parsePolynomial(String a_str, int offset) {

        a_str = a_str.replaceAll("[\\s]+", "");
        //System.out.println("remove whitespace: "+a_str);
        String[] parts = a_str.split("\\(\\+-\\d\\.?[e\\d]*\\)q\\^-\\d+");
        double[] a = new double[offset+parts.length];
        for (int i = 0; i < parts.length; i++) {
            String part = parts[i];
            a[offset+i] = Double.parseDouble(part);
        }

        return a;
    }

    private double[] parseMatrix(String a_str) {
        double mult = 1;
        if (a_str.indexOf('*') > 0) {
            int index = a_str.indexOf('*');
            String multip = a_str.substring(0, index);
            a_str = a_str.substring(index+1).trim();
            System.out.println("Multip: "+multip);
            mult = Double.parseDouble(multip.trim());
        }
        System.out.println("Mult: "+mult);
        System.out.println("A_str: "+a_str);
        String[] parts = a_str.split("[\\s]+");
        double[] a = new double[parts.length];
        for (int i = 0; i < parts.length; i++) {
            String part = parts[i];
            a[i] = mult*Double.parseDouble(part);
        }
        return a;
    }

}
