package deb.control;

import deb.Simulator;
import deb.Population;
import deb.Parameters;
import deb.math.Matrix;

import java.io.Writer;
import java.io.FileWriter;
import java.io.File;
import java.io.IOException;

/**
 * Created by IntelliJ IDEA.
 * User: alver
 * Date: Nov 28, 2006
 * Time: 12:55:00 PM
 * To change this template use File | Settings | File Templates.
 */
public class ArmaxRecursiveLeastSquares implements Controller {
    private int n_a;
    private int n_b;
    private double[] theta, phi;
    private double[] y_hist, u_hist;
    private int piv_y = 0, piv_u = 0;
    private double lambda; // Forgetting factor
    private PopulationMetric metric;
    private double u, y, t;

    private double T_w = 0;

    private double[][] P_t, K_t, I_n_th;

    class MyMetric implements PopulationMetric {
            double[] a = new double[] {-0.6, -0.1};
            double[] b = new double[] {1, 0};
            double lastY=0, lastLastY=0, nowY=0;
            double lastU=0, lastLastU=0;
            Writer out=null;
            public MyMetric() {
                try {
                    out = new FileWriter(new File("/home/alver/hjemmek/dr-ing/full_stoic_rotifer/in_out.dlm"));
                } catch (IOException e) {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }
            }
            public double getValue(Population p) {
                lastLastY = lastY;
                lastY = nowY;
                lastLastU = lastU;
                lastU = u;
                //nowY = -a[0]*lastY - a[1]*lastLastY + b[0]*lastU + b[1]*lastLastU;// + 0.1*(Math.random()-0.5);
                nowY = p.getHerbivoreCount();
                try {
                    out.write(u+"\t"+y+"\n");
                    if (t >= 19) {
                        System.out.println("closing");
                        out.close();
                    }

                } catch (IOException e) {
                    e.printStackTrace();
                }

                //System.out.println("lastU="+lastU+"\t lastLastU="+lastLastU);
                //System.out.println("lastY="+lastY+"\t lastLastY="+lastLastY);
                //System.out.println("nowY="+nowY);
                return nowY;
            }
        };

    class TempMetric implements PopulationMetric {
        public double getValue(Population p) {
            return T_w;
        }
    }

    public ArmaxRecursiveLeastSquares() {
        this(14, 14, 0.95, null);
    }

    public ArmaxRecursiveLeastSquares(int n_a, int n_b, double lambda, PopulationMetric metric) {

        this.n_a = n_a;
        this.n_b = n_b;
        this.lambda = lambda;
        if (metric == null)
            this.metric = new MyMetric();
        else
            this.metric = metric;

        y_hist = new double[n_a];
        u_hist = new double[n_b];

        // Define helper matrixes for recursive solution:
        P_t = Matrix.matrMultScalar(100, Matrix.eye(n_a+n_b));
        K_t = new double[n_a+n_b][n_a+n_b];

        phi = new double[n_a + n_b];
        theta = new double[n_a + n_b];
        for (int i = 0; i < theta.length; i++) {
            theta[i] = 0;
        }


        I_n_th = Matrix.eye(n_a+n_b);
    }

    public double getCallInterval() {
        return 1;
    }

    public void makeControlRun(double t, Simulator sim, Population pop, Parameters envPar) {
        this.t = t;
        u = envPar.get("baseFeedReference", t);
        y = metric.getValue(pop);
        System.out.println("u = "+u+"\t y = "+y);
        nextStep(y, u);
        double k_filter = 0.5;
        double newU = k_filter*u + (1-k_filter)*(1*Math.random());
        if (t - Math.floor(t) < 0.25)
            envPar.put("baseFeedReference", newU);
    }

    /**
     * Compute the estimates for next time step, based on the last output and input values.
     * @param y The output of time step t
     * @param u The input of time step t-1
     */
    private void nextStep(double y, double u) {
        updateHistU(u);
        buildPhiVector();
        updateHistY(y);

        for (int i = 0; i < phi.length; i++) {
            System.out.println("phi["+i+"] = "+phi[i]);
        }


        // Calculate K_t from the previous P_t:
        double[][] P_t_phi = Matrix.matrMult(P_t, phi);
        K_t = Matrix.matrMultScalar(1/(lambda + Matrix.scalarProd(phi, Matrix.matr2Vector(P_t_phi))),
                P_t_phi);

        P_t = Matrix.matrMultScalar(1/lambda, Matrix.matrMult(Matrix.matrSub(I_n_th, Matrix.matrMult(K_t, Matrix.transpVector(phi))), P_t));
        
        double epsilon = y - Matrix.scalarProd(phi, theta);
        theta = Matrix.vectorAdd(theta, Matrix.matr2Vector(Matrix.matrMultScalar(epsilon, K_t)));
        System.out.print("theta: ");
        Matrix.printMatrix(theta);

    }

    private void buildPhiVector() {
        int piv_phi = 1;
        for (int i=0; i<n_a; i++) {

            phi[n_a - piv_phi] = -y_hist[(i+piv_y) % n_a];
            piv_phi++;
        }
        piv_phi = 1;
        for (int i=0; i<n_b; i++) {

            phi[n_a + n_b - piv_phi] = u_hist[(i+piv_u) % n_b];
            piv_phi++;
        }
    }

    /**
     * Update the array of the latest y.
     * @param y The output of time step t
     */
    private void updateHistY(double y) {
        y_hist[piv_y] = y;
        piv_y++;
        if (piv_y == n_a)
            piv_y = 0;

    }

    /**
     * Update the array of the latest u.
     * @param u The input of last time step.
     */
    private void updateHistU(double u) {
        u_hist[piv_u] = u;
        piv_u++;
        if (piv_u == n_b)
            piv_u = 0;
    }
}
