/*
 * Created on 14.des.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.control;

import deb.Parameters;

/**
 * This interface defines the role of a regulator that can be used in various contexts - both to represent real-life
 * regulators used to control the modelled system, and to represent regulator structures imitating regulative functions
 * internal to the modelled system.
 */
public interface InternalController {

	/**
	 * Return the current regulator value. This method should return a precomputed value
	 * already determined by the update() method.
	 * @return
	 */
	public double getValue();
	
	/**
	 * A regulator can have state, e.g. an integral regulator.
	 * @param t
	 * @param dt
	 * @param envPar
	 */
	public void update(Object system, double t, double dt, Parameters envPar);
	
        /**
         *Return a copy of this controller.
         */
        public InternalController copy(Parameters par);

    /**
     * Get how often this controller should be called, in simulation time.
     * @return The call interval, or -1 if it should be called each time step.
     */
    public double getCallInterval();
}
