/*
 * ParameterSeeker.java
 *
 * Created on October 26, 2005, 11:28 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package deb.control;

import deb.Parameters;

/**
 *
 * @author alver
 */
public class ParameterSeeker implements InternalController {
    
    private double currentValue, lastObjectValue=0, lastAdjustment, minValue, maxValue;
    private int timeStepsPerUpdate, timeStepsSinceUpdate;
    private double tendency = 0, k_sign=0, k_abs=0;
    
    /** Creates a new instance of ParameterSeeker */
    public ParameterSeeker(double initValue, int timeStepsPerUpdate, double k, double minValue, double maxValue) {
        currentValue = initValue;
        lastAdjustment = 0;
        this.minValue = minValue;
        this.maxValue = maxValue;
        this.timeStepsPerUpdate = timeStepsPerUpdate;
        this.timeStepsSinceUpdate = 0;
        //lastObjectValue = getDeviation(t0);
        k_sign = Math.signum(k);
        k_abs = Math.abs(k);
    }

    public double getValue() {
        return currentValue;
    }

    /**
     * If the required number of time steps has passed since the last update, perform adjustment 
     * of the parameter based on the object function. To override the adjustment procedure without
     * changing timing, override makeParameterAdjustment() instead of this method.
     */
    public void update(Object system, double t, double dt, deb.Parameters envPar) {
        timeStepsSinceUpdate++;
        if (timeStepsSinceUpdate == timeStepsPerUpdate) {
            timeStepsSinceUpdate = 0;
            makeParameterAdjustment(t, dt, envPar);
        }
    }
    
    /**
     * Perform adjustment of the parameter based on the object function. Override this method to change
     * the algorithm used to adjust the parameter.
     */
    public void makeParameterAdjustment(double t, double dt, deb.Parameters envPar) {
        double currentDev = getDeviation(t);
        //System.out.println("Object value: "+currentObject);
        double dDev = currentDev-lastObjectValue;
        double newValue;
        //if (lastAdjustment != 0) {
            tendency += Math.signum(lastAdjustment)*Math.signum(dDev);
            double multiplier = Math.signum(tendency);
            if (multiplier == 0)
                multiplier = 1;
            double relEffect = dDev/lastAdjustment;
            //System.out.println("Rel effect: "+relEffect+"\t dObject: "+dObject+"\t lastAdj: "+lastAdjustment);
            newValue = currentValue +k_sign*k_abs*currentDev; // 1% adjustment in right direction
        //}
        //else {
        //    newValue = 1.05*currentValue;
        //}
        
        // Make sure parameter bounds are honoured:
        newValue = Math.max(minValue, Math.min(maxValue, newValue));
        lastAdjustment = newValue-currentValue;
        currentValue = newValue;
        lastObjectValue = currentDev;
        System.out.println("Current value: "+currentValue+"\t Deviation: "+currentDev+"\t k="+(k_abs*k_sign));
    }
    
    /**
     * Get the current object value which this parameter should seek to tune to 0. Subclasses must override this method.
     */
    public double getDeviation(double t) {
	return 0;
	}

    public InternalController copy(Parameters par) {
        ParameterSeeker copy = new ParameterSeeker(currentValue, timeStepsPerUpdate, k_sign*k_abs, minValue, maxValue);
        copy.timeStepsSinceUpdate = timeStepsSinceUpdate;
        copy.lastAdjustment = lastAdjustment;
        copy.tendency = tendency;
        return copy;
    }

    public double getCallInterval() {
        return -1;
    }


}
