/*
 * Created on 17.des.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import deb.Herbivore;
import deb.HerbivoreFactory;
import deb.Parameters;
import deb.Population;
import deb.SimulationMonitor;
import deb.operations.Operation;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.*;

/**
 * @author alver
 *
 * Operation for adding rotifers continuously according to some function. PRBS.
 */
public class AddContinuous implements Operation, SimulationMonitor {

	ArrayList values = new ArrayList();
	
	final static int PRBS = 1;
	
	private String directory = null;
	private boolean isOn = false;
	private double nextChangeT, lambda, dailyAmount, n_each, lastAdded = 0;
	private int mode;
	private HerbivoreFactory factory = null;
	
	// Block default constructor, so only factory methods can be used:
	private AddContinuous() {
	}
	
	public static AddContinuous getPrbsAdder(HerbivoreFactory factory, Parameters par, double lambda, String directory) {
		AddContinuous ac = new AddContinuous();
	
		ac.factory = factory;
		ac.lambda = lambda;
		ac.n_each = par.get("n_each", 0);
		ac.mode = PRBS;
		ac.dailyAmount = 500000;
		ac.directory = directory;
		return ac;
	}
	
	/* (non-Javadoc)
	 * @see deb.operations.Operation#performNow(double, double)
	 */
	public boolean performNow(double t, double dt) {
		if (mode == PRBS) {
			if (t>=nextChangeT) {
				isOn = !isOn;
				// Time until next switch:
				nextChangeT = t + 2*lambda*Math.random();
			}
			return true;
		}
		else
			return false;
	}

	/* (non-Javadoc)
	 * @see deb.operations.Operation#performOperation(double, double, deb.Population, deb.Parameters)
	 */
	public void performOperation(double t, double dt, Population pop,
			Parameters envPar) {
		
		// PRBS:
		double toAdd = 0;
		if (mode == PRBS) {
			if (isOn)
				toAdd = dt*dailyAmount*envPar.get("V_w", t);
		}
			

		// Remember value:
		lastAdded = toAdd;
		
		if (toAdd > 0) {
			
			// Add the given amount of rotifers. Make sure to add a sensible amount of super individuals:
			double n_each_now = n_each/2;
			int n_to_add = 0, MIN=2, MAX=20;
			while ((n_to_add < MIN) || (n_to_add > MAX)) {
				if (n_to_add < MIN)
					n_each_now /= 2;
				else
					n_each_now *= 2;
				
				n_to_add = (int)(Math.floor(toAdd/n_each_now));
				
			}
			// Adjust so we add the exact correct amount:
			n_each_now = toAdd/n_to_add;
			
			//System.out.println("NY: "+toAdd+"\tN: "+n_to_add+"\tN_each: "+n_each_now+"\tTot:"+(n_each_now*n_to_add));
			for (int i=0; i<n_to_add; i++) {
				Herbivore herb = factory.getHerbivore(FishtankStrains.BRACHIONUS_PLICATILIS, n_each_now, t);
				pop.addHerbivore(herb);
			}
		}
	}

	public void simulationFinished(double t, Population pop) {
		values.add(new Double(lastAdded));
		if (values.size() == 0)
			return;
		try {
			FileWriter fw = new FileWriter(new File(directory+"/addition.dlm"));
			for (Iterator i=values.iterator(); i.hasNext();) {
				double d = ((Double)i.next()).doubleValue();
				fw.write(d+"\n");
			}
			fw.close();
		} catch (IOException ex) {
			ex.printStackTrace();
		}
	}
	public void simulationInterval(double t, Population pop) {
		simulationFinished(t, pop);
	}
	public void simulationStarted(double t, Population pop) {		
	}
	public void simulationUpdate(double t, Population pop) {
		values.add(new Double(lastAdded));
	}
	
        public Operation copy(Parameters par) throws UnsupportedOperationException {
            return new AddContinuous();
        }

    public SimulationMonitor copy(Parameters par, String directory) {
        AddContinuous ac = new AddContinuous();
        ac.isOn = isOn;
        ac.dailyAmount = dailyAmount;
        ac.directory = directory;
        ac.factory = factory;
        ac.lambda = lambda;
        ac.lastAdded = lastAdded;
        ac.mode = mode;
        ac.n_each = n_each;
        ac.nextChangeT = nextChangeT;
        ac.values = (ArrayList)values.clone();
        return ac;
    }

        
}
