/*
 * Created on 11.okt.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import deb.math.OptimizationResult;
import deb.math.ScalarFunction;

import java.awt.Point;
import java.io.File;
import java.io.IOException;

import deb.*;
import deb.control.InternalController;

/**
 * @author alver
 *         <p/>
 *         To change the template for this generated type comment go to
 *         Window>Preferences>Java>Code Generation>Code and Comments
 */
public class CodLarva extends BasicPredator implements Predator {

    final int DRY_WEIGHT_OUTPUT = 0;
    static int counter = 0, prolim = 0;
    protected int totalEaten = 0;
    double feedint = 0, dt = 0, t0 = 0;
    InternalController lipidReg;
    double seenRotifers = 0, ingestionRate = 0, assimRate = 0, pam = 0;

    /**
     * @param par
     * @param initVal
     * @param n_0
     */
    public CodLarva(Parameters par, InitialValues initVal, double n_0, double t0) {
        super(par, FishtankStrains.GADUS_MORHUA, initVal, n_0);
        this.t0 = t0;
        init();
    }

    /**
     * @param par
     * @param initVal
     */
    public CodLarva(Parameters par, InitialValues initVal) {
        super(par, FishtankStrains.GADUS_MORHUA, initVal);
        init();
    }

    ScalarFunction visualRangeFunction = new ScalarFunction() {

        public double f(double r, double... par) {
            // Parameters: par[0] = left_side, par[1] = c_turb
            double val = par[0] - r * r * Math.exp(par[1] * r);
            return val * val;
        }

        public double getDerivative(double x) {
            return 0;
        }

        public boolean hasDerivative() {
            return false;
        }
    };

    /**
     * Object function for calculating the implicitly defined growth rate.
     */
    ScalarFunction growthFunction = new ScalarFunction() {
        public double getDerivative(double x) {
            return 0;
        }

        public boolean hasDerivative() {
            return false;
        }

        public double f(double x, double... par) {
            double d_V;
            double E_p_ = par[0],
                    E_l_ = par[1],
                    V23 = par[2],
                    v_dot = par[3],
                    p_M_p = par[4],
                    p_M_l = par[5],
                    n_p = par[6],
                    n_l = par[7],
                    sigma_G = par[8],
                    gamma_G = par[9],
                    d_V_mrk = x;
            double p_C_p = E_p_ * (v_dot * V23 - d_V_mrk);
            double p_C_l = E_l_ * (v_dot * V23 - d_V_mrk);
            //p_C_C = E_c_*(v_dot*V23 - d_V_mrk);
            double
                    j_P_C = p_C_p - p_M_p,
                    j_L_C = p_C_l - p_M_l;
            //j_C_C = p_C_C - j_C_M;

            if ((j_P_C > 0) && (j_L_C > 0)) { // && (j_C_C > 0)) {
                double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_L_C,
                        n_p, n_l, sigma_G);
                d_V = growth_fluxes[0] / gamma_G;

            } else
                d_V = 0;

            // Return squared difference between projected and resulting growth:
            return (d_V - d_V_mrk) * (d_V - d_V_mrk);
        }

    };

    public double getAge(double t) {
        return t - t0;
    }

    protected final void init() {
        lipidReg = new CodLipidReg(this);
        addRegulator(lipidReg);
    }

    /* (non-Javadoc)
      * @see deb.Predator#update(double, double, deb.Population, deb.Parameters)
      */
    public IndividualReturnValues update(double t, double dt, Population population,
                                         Parameters envPar) {
        super.update(t, dt, population, envPar);

        this.dt = dt;
        // Step ordinary states:
        states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);

        // Add up how many units we eat during this time step:
        totalEaten += ni.get("numbers") * (dt * getN());

        if ((states[1] <= 0) || (states[2] <= 0))
            kill();

        return new IndividualReturnValues(0, 0);
    }

    /* (non-Javadoc)
      * @see deb.Predator#f(double[], double[], deb.Population, deb.Parameters)
      */
    public IndividualReturnValues f(double[] f, double t, double[] states,
                                    Population population, Parameters envPars) {

        // Convenience variables for states:
        double
                V = states[0],
                E_p = states[1],
                E_l = states[2],

                E = E_p + E_l,
                E_ = E / V,
                E_p_ = E_p / V,
                E_l_ = E_l / V,
                yolk = states[3],
                Ms_p = states[4],
                Ms_l = states[5],
                Ms = Ms_p + Ms_l;

        /*for (int i=0; i<states.length; i++) {
              if ((i != 3) && states[i]<0)
                  System.out.println("<0 states["+i+"] = "+states[i]);
          }*/

        // Primary parameters:
        double
                E_m_ = par.get("E_m_", t),
                k_as_p = par.get("k_as", t),
                k_as_l = par.get("k_as", t),
                k_g = par.get("k_g", t),
                p_dot_M_ = par.get("p_dot_M_", t),
                //E_G_ = par.get("E_G_", t),
                sigma_G = par.get("sigma_G", t),
                gamma_G = par.get("gamma_G", t) + 1,
                //kappa = par.get("kappa", t),
                kappa_R = par.get("kappa_R", t),
                p_Am_ = par.get("p_Am_", t),
                p_Am_yolk_ = par.get("p_Am_yolk_", t),
                shape_coeff = par.get("shape_coeff", t),
                T_ref = par.get("T_ref", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                P_Y = par.get("P_Y", t),
                L_Y = par.get("L_Y", t),
                DW_to_VOL = par.get("DW_to_VOL", t),

                // Extract the feed intake levels of the various nutrients, already
                // calculated and set externally by Population:
                dw_in = ni.get("dw"),
                protein_in = ni.get("protein"),
                lipid_in = ni.get("lipid"),
                dha_in = ni.get("dha"),
                epa_in = ni.get("epa"),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                energy_in = prot_E * protein_in
                        + lip_E * lipid_in, // dha is included in lipid_in
                dha_energy_in = lip_E * dha_in,

                protein_E_in = protein_in * prot_E,
                lipid_E_in = lipid_in * lip_E;

        //V_dig = par.get("V_dig", t);

        //if (ingestionRate > 0)
        //    System.out.println("p/l per rot: "+(protein_in/ingestionRate)+" / "+(lipid_in/ingestionRate));
        //System.out.println(t+"\t"+protein_in+"\t"+lipid_in+"\t"+dha_in);
        //System.out.println(""+((protein_in+lipid_in))+"\t"+(dt*feedint)+"\t");

        double[] Tpars = par.getArray("Tpars", t);

        // Environmental parameters
        double
                T_w = envPars.get("T_w", t),
                V_w = envPars.get("V_w", t),
                X = population.getHerbivoreCount() / V_w; // Rotifer density
        // if (X>0) System.out.println(""+X);
        double
                V23 = Math.pow(V, 0.66667),
                V13 = Math.pow(V, 0.33333);

        //		 Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)

        p_Am_ *= TC;
        p_Am_yolk_ *= TC;
        p_dot_M_ *= TC;

        /*if (V < V_dig) {
              k_as_p *= V/V_dig;
              k_as_l *= V/V_dig;
          }*/
        /*
          if (E_ > 3000) {
              k_as_l *= 3000/E_;
              k_as_p *= 3000/E_;
          }*/

        double
                fromGut = k_g * (Ms_p + Ms_l),
                p_Am = p_Am_ * V23,

                A_p = Ms > 0 ? (Ms_p / Ms) * k_as_p * p_Am * fromGut / (k_as_p * fromGut + p_Am) : 0,
                A_l = Ms > 0 ? (Ms_l / Ms) * k_as_l * p_Am * fromGut / (k_as_l * fromGut + p_Am) : 0,

                //A_p = Ms>0 ? k_as_p*fromGut*Ms_p/Ms : 0,
                //A_l = Ms>0 ? k_as_l*fromGut*Ms_l/Ms : 0,

                A = A_p + A_l,


                v_dot = par.get("v_dot", t), //p_Am_/E_m_,
                v_dot_yolk = p_Am_yolk_ / E_m_;

        double assim_frac;
        if (fromGut > 0)
            assim_frac = A / fromGut;
        else
            assim_frac = 0;
        assimRate = A;
        pam = dw_in;//p_Am;
        //System.out.println("assim = "+A+"\t Assim_frac="+assim_frac);
        //SSystem.out.println("fromgut/p_Am = "+(k_as_p*fromGut/p_Am));

        double
                yolk_V = yolk / par.get("yolk_energy_density", t),
                d_yolk = -p_Am_yolk_ * Math.pow(Math.max(0, Math.min(V, yolk_V)), 2.0 / 3.0), // (3) in Lika & Nisbet
                controller = lipidReg.getValue(),
                p_M = V * p_dot_M_,
                p_M_p = (1 - controller) * p_M, //Math.max(0, E_p*p_M/E), //
                p_M_l = controller * p_M; //Math.max(0, E_l*p_M/E), //

        // Growth phase. No egg production.

        double d_V_mrk = 0;

        double p_C_p, p_C_l, d_V = 0, j_P_G, j_L_G;
        // d_V is implicit and must be solved numerically. We iterate to find the right value:
        boolean done = false;
        int count = 0;

        // Use optimization to find the implicitly defined growth rate d_V:
        OptimizationResult optRes = deb.math.Functions.nonLinearScalarMinimumSearch(growthFunction, 0, 10,
                E_p_, E_l_, V23, v_dot, p_M_p, p_M_l, P_V * prot_E / DW_to_VOL,
                L_V * lip_E / DW_to_VOL, sigma_G, gamma_G);
        d_V_mrk = optRes.getOptimum();

        p_C_p = E_p_ * (v_dot * V23 - d_V_mrk);
        p_C_l = E_l_ * (v_dot * V23 - d_V_mrk);
        //p_C_C = E_c_*(v_dot*V23 - d_V_mrk);
        double
                j_P_C = p_C_p - p_M_p,
                j_L_C = p_C_l - p_M_l;
        //j_C_C = p_C_C - j_C_M;

        if ((j_P_C > 0) && (j_L_C > 0)) { // && (j_C_C > 0)) {
            double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_L_C,
                    P_V * prot_E / DW_to_VOL,
                    L_V * lip_E / DW_to_VOL, sigma_G);
            d_V = growth_fluxes[0] / gamma_G;
            j_P_G = growth_fluxes[1];
            j_L_G = growth_fluxes[2];
            //j_C_G = growth_fluxes[3];

            //pG = j_P_G + j_L_G + j_C_G; // Energy spent on growth
            //pG_out = d_V*(P_V*prot_E + L_V*lip_E + C_V*carb_E)/DW_to_VOL; // Energy incorporated into structure
            double eff_P = j_P_G / j_P_C;
            double eff_L = j_L_G / j_L_C;

            //System.out.println(eff_P+"\t "+eff_L);
            //eff_C=j_C_G/j_C_C;
            //N_somatic = j_P_G/prot_E/6.25;

            //System.out.println("Eff: "+eff_P+"\t"+eff_L);
        } else {
            d_V = 0;
            j_P_G = 0;
            j_L_G = 0;
        }
        // Starvation
        if (d_V <= 0) {
            d_V = 0;
            ///p_M *= 0.5;
            p_C_p = p_M_p;
            p_C_l = p_M_l;
            //p_C_C = j_C_M;
            j_P_G = 0;
            j_L_G = 0;
        }

        // Stomach energy content:
        double
                d_Ms_p = protein_E_in - k_g * Ms_p,
                d_Ms_l = lipid_E_in - k_g * Ms_l;

        // Growth:
        //d_V = (kappa*p_C - p_dot_M_*V)/E_G_;

        // Returned parts of rejected fluxes:
        double
                rej_P = kappa_R * (p_C_p - p_M_p - j_P_G),
                rej_L = kappa_R * (p_C_l - p_M_l - j_L_G);

        // Energy storage:
        //double d_E = p_A - p_C;
        double d_E_p = A_p - d_yolk * P_Y - p_C_p + rej_P,
                d_E_l = A_l - d_yolk * L_Y - p_C_l + rej_L;

        //System.out.println("prot: "+(A_p/(k_g*Ms_p))+"lip: "+(A_l/(k_g*Ms_l)));
        //System.out.println("in: "+(A_p+A_l)+"\t maint: "+(p_M_p+p_M_l));
        //System.out.println("\t j_L_G/p_C_l = "+(j_L_G/p_C_l));
        //if (t > 5)
        //    System.out.println("prot+lipid inn = "+((protein_in+lipid_in)/dt));//"reell E_G = "+((j_P_G+j_L_G)/d_V));

        if (++counter == 10000) {
            System.out.println("CodLarva.java: assim_frac = " + assim_frac + "\tk_as_p = " + k_as_p + "\tk_as_l = " + k_as_l);
            //System.out.println("CodLarva.java: lip_in = "+lipid_in+"\tprot_in = "+protein_in);
            //System.out.println("CL: d_V = "+d_V+"\td_V_by_p = "+d_V_by_p+"\td_V_by_l = "+d_V_by_l);
            //System.out.println("CL: p_C_by_p = "+p_C_by_p+"\tp_C_by_l = "+p_C_by_l);//+"\td_V_by_l = "+d_V_by_l);
            //System.out.println("CL: prolim = "+prolim+"\tdiff_p = "+(p_C_p_rej-p_C_p)+"\tdiff_l = "+(p_C_l_rej-p_C_l));
            //System.out.println("CL: controller = "+controller);
            counter = 0;
        }
        /*
          if (new Double(d_E_p).isNaN()) {
              System.out.println("A_p = "+A_p+"\tP_Y = "+P_Y+"\td_yolk = "+d_yolk);//+"\tE_p = "+E_p+"\tE_l = "+E_l);
          }*/


        f[0] = d_V;
        f[1] = d_E_p;
        f[2] = d_E_l;
        f[3] = d_yolk;
        f[4] = d_Ms_p;
        f[5] = d_Ms_l;

        // Accumulation of ingested nutrients:
        f[6] = dw_in;
        f[7] = protein_in;
        f[8] = lipid_in;
        f[9] = dha_in;
        f[10] = epa_in;
        return new IndividualReturnValues(0, 0);
    }

    public Predator spawnAndGetChild(double t) {
        return null;
    }

    ScalarEquation mortalityEq = new ScalarEquation() {
        public double f(double t, double value, Parameters par) {
            return -n * par.get("mortalityRate", t);
        }
    };

    public boolean checkMortality(double t, double dt) {
        /*double number = Math.random();
        if (number < dt*par.get("mortalityRate", t)) {
                died = true;
        }*/

        setN(StepSolver.stepScalar(mortalityEq, t, dt, getN(), par));

        return n < 1;
    }

    public double[] getDerivedOutputs(double t) {

        // Convenience variables for states:
        double
                V = states[0],
                E_p = states[1],
                E_l = states[2],
                E = E_p + E_l,
                E_ = E / V,
                yolk = states[3],
                M_s = states[4] + states[5];

        double
                mu_E = par.get("mu_E", t),
                mu_S = par.get("mu_S", t),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                L_V = par.get("L_V", t),
                P_V = par.get("P_V", t),
                L_Y = par.get("L_Y", t),
                P_Y = par.get("P_Y", t);

        // Calculate dry weights for the respective states.
        double
                M_V = V / par.get("DW_to_VOL", t),
                M_E_p = E_p / prot_E,
                M_E_l = E_l / lip_E,
                M_E = M_E_p + M_E_l,
                M_S = M_s / mu_S,
                M_Y = yolk / mu_E,
                M_p = M_E_p + P_V * M_V + P_Y * M_Y,
                M_l = M_E_l + L_V * M_V + L_Y * M_Y;

        return new double[]{(M_V + M_E + M_Y + M_S), M_V, M_E_p, M_E_l, M_Y, M_S, M_p, M_l, seenRotifers, ingestionRate, assimRate, pam};

    }

    public String[] getDerivedOutputNames() {
        return new String[]{"Dry weight", "Dry weight in structure", "Dry weight in protein reserves", "Dry weight in lipid reserves",
                "Dry weight in yolk", "Dry weight in gut", "Dry weight in protein", "Dry weight in lipid"};
    }

    public Point getPosition() {
        return null;
    }

    public boolean readyToSpawn(double t) {
        return false;
    }

    public void storeCurrentState(File file) throws IOException {
        // TODO Auto-generated method stub
    }

    /* (non-Javadoc)
      * @see deb.Individual#getProteinContent(double)
      */
    public double getProteinContent(double t) {
        // TODO Auto-generated method stub
        return 0;
    }

    /* (non-Javadoc)
      * @see deb.Individual#getLipidContent(double)
      */
    public double getLipidContent(double t) {
        // TODO Auto-generated method stub
        return 0;
    }

    /* (non-Javadoc)
      * @see deb.Predator#predation(double, deb.Population, deb.Parameters)
      */
    public double predation(double t, Population population, Parameters envPars) {

        double
                V = states[0],
                V23 = Math.pow(V, 0.66667);

        // Environmental parameters
        double
                T_w = envPars.get("T_w", t),
                V_w = envPars.get("V_w", t),
                X = population.getHerbivoreCount() / V_w, // Rotifer density
                T_ref = par.get("T_ref", t),
                J_Xm_ = par.get("J_Xm_", t),
                X_K = par.get("X_K", t),
                shape_coeff = par.get("shape_coeff", t);

        ingestionRate = J_Xm_ * Math.pow(V, 0.667) * (X/(X+X_K));

        return ingestionRate;
        /*
        double length = Math.pow(V, 0.3333) / shape_coeff;

        // Calculate visual range estimate:
        double K_e = par.get("K_e", t),
                E_b = par.get("E_b", t),
                contrast = par.get("contrast", t),
                c_turb = par.get("c_turb", t),
                A_p = par.get("preyArea", t),
                f_pause = par.get("f_pause", t),
                t_h = 140;//*0.001/V23;//par.get("t_h", t);
        double E_l = Math.pow(10, 5.04 / (1 + Math.exp(-(length - 5.33) / 0.64)));
        double left_side = Math.abs(contrast) * A_p * E_l * E_b / (K_e + E_b);
        double r = 100 * deb.math.Functions.nonLinearScalarMinimumSearch(visualRangeFunction, length, 30, left_side, c_turb).getOptimum();
        seenRotifers = (2d / 3d) * 3.14 * r * r * r * X * 1e-3;
        double searchFrequency = 1 * 60 * 24;
        double attackProbability = seenRotifers / (seenRotifers + 0.75);
        double attackRate = attackProbability * searchFrequency;
        double captureProbability = length / (length + 6);
        ingestionRate = captureProbability * attackRate;

        double[] Tpars = par.getArray("Tpars", t);
        double TC = Functions.TC(T_w, T_ref, Tpars);

        // Kioerboe & MacKenzie (1995). Prey density in rot/ml.
        // "tw" = 2.
        double encounterRate = TC * X * 1e-6 * 3.14 * r * r * f_pause * (0.667 * r - 0 * 1.41 * 2);
        double ingestion = 0;
        if (encounterRate > 0) {
            ingestion = 86400 / (t_h + 1 / encounterRate);
        }
        //System.out.println("Length="+length+"\t r = "+r+"\t sqrt(left)="+(Math.sqrt(left_side))+"\t enc="+encounterRate+"\t preyInRange="+preyInRange
        //        +"\t I="+ingestion);

        feedint = ingestion;

        //return ingestion;
        return ingestionRate;
        */

        /*double[] Tpars = par.getArray("Tpars", t);
double TC = Functions.TC(T_w, T_ref, Tpars);
double
    fl = X/(X_K + X),
    J_Xm = TC*J_Xm_*V23; // Predation = fl*J_Xm

feedint = fl*J_Xm;
        //System.out.println(fl+"\t "+feedint+"");
return fl*J_Xm;
        */
    }

    public double getDryWeight(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[DRY_WEIGHT_OUTPUT];
    }

    public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
    }

    /* (non-Javadoc)
      * @see deb.Individual#compareTo(deb.Individual)
      */
    public double compareTo(Individual ind) throws ClassCastException {
        // TODO CompareTo for CodLarva
        return 0;
    }

    /* (non-Javadoc)
      * @see deb.Individual#combineWith(deb.Individual)
      */
    public Individual combineWith(Individual ind, double t) throws ClassCastException {
        // TODO CombineWith for CodLarva
        return null;
    }


    public BasicIndividual copy(Parameters par) {
        System.out.println(this.par.containsKey("J_Xm_")+" "+par.containsKey("J_Xm_"));
        CodLarva newInd = new CodLarva(par,
                this.initVal, this.n_0, this.t0);
        newInd.n = this.n;

        // Copy state values:
        newInd.states = (double[])states.clone();

        return newInd;
    }

    public int unitsEaten() {
        return totalEaten;
    }


}
