package deb.fishtank;

import deb.*;
import deb.math.ScalarFunction;
import deb.math.OptimizationResult;
import deb.math.Matrix;

import java.awt.*;
import java.io.File;
import java.io.IOException;

/**
 * @author alver
 *         <p/>
 *         To change the template for this generated type comment go to
 *         Window>Preferences>Java>Code Generation>Code and Comments
 */
public class CodLarvaStoich extends BasicPredator implements Predator {

    final int DRY_WEIGHT_OUTPUT = 0;
    static int counter = 0, prolim = 0;
    protected int totalEaten = 0;
    double feedint = 0, dt = 0, t0 = 0;

    double seenRotifers = 0, ingestionRate = 0, assimRate = 0, pam = 0;
    double[] mineralFluxes = new double[] {0,0,0,0},
        organicFluxes = new double[] {0,0,0,0};
    /**
     * @param par
     * @param initVal
     * @param n_0
     */
    public CodLarvaStoich(Parameters par, InitialValues initVal, double n_0, double t0) {
        super(par, FishtankStrains.GADUS_MORHUA, initVal, n_0);
        this.t0 = t0;
        init();
    }

    /**
     * @param par
     * @param initVal
     */
    public CodLarvaStoich(Parameters par, InitialValues initVal) {
        super(par, FishtankStrains.GADUS_MORHUA, initVal);
        init();
    }

    ScalarFunction visualRangeFunction = new ScalarFunction() {

        public double f(double r, double... par) {
            // Parameters: par[0] = left_side, par[1] = c_turb
            double val = par[0] - r * r * Math.exp(par[1] * r);
            return val * val;
        }

        public double getDerivative(double x) {
            return 0;
        }

        public boolean hasDerivative() {
            return false;
        }
    };

    /**
     * Object function for calculating the implicitly defined growth rate.
     */
    ScalarFunction growthFunction = new ScalarFunction() {
        public double getDerivative(double x) {
            return 0;
        }

        public boolean hasDerivative() {
            return false;
        }

        public double f(double x, double... par) {
            double d_V;
            double E_p_ = par[0],
                    E_l_ = par[1],
                    E_dha_ = par[2],
                    E_epa_ = par[3],
                    V23 = par[4],
                    v_dot = par[5],
                    p_M_p = par[6],
                    p_M_l = par[7],
                    p_M_dha = par[8],
                    p_M_epa = par[9],
                    n_p = par[10],
                    n_l = par[11],
                    n_dha = par[12],
                    n_epa = par[13],
                    sigma_G = par[14],
                    gamma_G = par[15],
                    d_V_mrk = x;
            double p_C_p = E_p_ * (v_dot * V23 - d_V_mrk);
            double p_C_l = E_l_ * (v_dot * V23 - d_V_mrk);
            double p_C_dha = E_dha_ * (v_dot * V23 - d_V_mrk);
            double p_C_epa = E_epa_ * (v_dot * V23 - d_V_mrk);

            //p_C_C = E_c_*(v_dot*V23 - d_V_mrk);
            double
                    j_P_C = p_C_p - p_M_p,
                    j_L_C = p_C_l - p_M_l,
                    j_DHA_C = p_C_dha - p_M_dha,
                    j_EPA_C = p_C_epa - p_M_epa;
            //j_C_C = p_C_C - j_C_M;

            if ((j_P_C > 0) && (j_L_C > 0)) { // && (j_C_C > 0)) {
                double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_L_C, j_DHA_C, j_EPA_C,
                        n_p, n_l, n_dha, n_epa, sigma_G);
                d_V = growth_fluxes[0] / gamma_G;

            } else
                d_V = 0;

            // Return squared difference between projected and resulting growth:
            return (d_V - d_V_mrk) * (d_V - d_V_mrk);
        }

    };

    public double getAge(double t) {
        return t - t0;
    }

    protected final void init() {

    }

    /* (non-Javadoc)
      * @see deb.Predator#update(double, double, deb.Population, deb.Parameters)
      */
    public IndividualReturnValues update(double t, double dt, Population population,
                                         Parameters envPar) {
        super.update(t, dt, population, envPar);

        this.dt = dt;
        // Step ordinary states:
        states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);

        // Add up how many units we eat during this time step:
        totalEaten += ni.get("numbers") * (dt * getN());

        if ((states[1] <= 0) || (states[2] <= 0))
            kill();

        return new IndividualReturnValues(0, 0);
    }

    /* (non-Javadoc)
      * @see deb.Predator#f(double[], double[], deb.Population, deb.Parameters)
      */
    public IndividualReturnValues f(double[] f, double t, double[] states,
                                    Population population, Parameters envPars) {

        // Convenience variables for states:
        double
                V = states[0],
                E_p = states[1],
                E_l = states[2],
                E_dha = states[3],
                E_epa = states[4],
                E = E_p + E_l + E_dha + E_epa,
                E_ = E / V,
                E_p_ = E_p / V,
                E_l_ = E_l / V,
                E_dha_ = E_dha / V,
                E_epa_ = E_epa / V,
                yolk = states[5],
                Ms_p = states[6],
                Ms_l = states[7],
                Ms_dha = states[8],
                Ms_epa = states[9],
                Ms_lambda = Ms_l+Ms_dha+Ms_epa,
                Ms = Ms_p + Ms_l + Ms_dha + Ms_epa;



        // Primary parameters:
        double
                E_m_ = par.get("E_m_", t),
                k_as_p = par.get("k_as_p", t),
                k_as_l = par.get("k_as_l", t),
                k_as_dha = par.get("k_as_dha", t),
                k_as_epa = par.get("k_as_epa", t),
                k_g = par.get("k_g", t),
                p_dot_M_ = par.get("p_dot_M_", t),
                //E_G_ = par.get("E_G_", t),
                sigma_G = par.get("sigma_G", t),
                gamma_G = par.get("gamma_G", t) + 1,
                //kappa = par.get("kappa", t),
                kappa_R = par.get("kappa_R", t),
                p_Am_ = par.get("p_Am_", t),
                p_Am_yolk_ = par.get("p_Am_yolk_", t),
                shape_coeff = par.get("shape_coeff", t),
                T_ref = par.get("T_ref", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                DHA_V = par.get("DHA_V", t),
                EPA_V = par.get("EPA_V", t),
                P_Y = par.get("P_Y", t),
                L_Y = par.get("L_Y", t),
                DHA_Y = par.get("DHA_Y", t),
                EPA_Y = par.get("EPA_Y", t),
                DW_to_VOL = par.get("DW_to_VOL", t),

                // Extract the feed intake levels of the various nutrients, already
                // calculated and set externally by Population:
                dw_in = ni.get("dw"),
                protein_in = ni.get("protein"),
                dha_in = ni.get("dha"),
                epa_in = ni.get("epa"),
                lipid_in = ni.get("lipid") - dha_in - epa_in,
                lip_E = par.get("lip_energy", t),
                prot_E = par.get("prot_energy", t),
                energy_in = prot_E * protein_in
                        + lip_E * lipid_in, // dha is included in lipid_in
                dha_E_in = lip_E * dha_in,
                epa_E_in = lip_E * epa_in,

                protein_E_in = protein_in * prot_E,
                lipid_E_in = lipid_in * lip_E,

                mu_Ms = (Ms_p*prot_E + Ms_lambda*lip_E)/Ms;


        //if (ingestionRate > 0)
        //    System.out.println("p/l per rot: "+(protein_in/ingestionRate)+" / "+(lipid_in/ingestionRate));
        //System.out.println(t+"\t"+protein_in+"\t"+lipid_in+"\t"+dha_in);
        //System.out.println(""+((protein_in+lipid_in))+"\t"+(dt*feedint)+"\t");

        double[] Tpars = par.getArray("Tpars", t);

        // Environmental parameters
        double
                T_w = envPars.get("T_w", t),
                V_w = envPars.get("V_w", t),
                X = population.getHerbivoreCount() / V_w; // Rotifer density
        // if (X>0) System.out.println(""+X);
        double
                V23 = Math.pow(V, 0.66667),
                V13 = Math.pow(V, 0.33333);

        //		 Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)

        p_Am_ *= TC;
        p_Am_yolk_ *= TC;
        p_dot_M_ *= TC;



        double
                fromGut = k_g * (Ms_p + Ms_lambda),
                fromGut_g = k_g * (Ms_p/prot_E + Ms_lambda/lip_E),
                p_Am = p_Am_ * V23,

                A_p = Ms > 0 ? (Ms_p / Ms) * k_as_p * p_Am * fromGut / (k_as_p * fromGut + p_Am) : 0,
                A_l = Ms > 0 ? (Ms_l / Ms) * k_as_l * p_Am * fromGut / (k_as_l * fromGut + p_Am) : 0,
                A_dha = Ms > 0 ? (Ms_dha / Ms) * k_as_dha * p_Am * fromGut / (k_as_dha * fromGut + p_Am) : 0,
                A_epa = Ms > 0 ? (Ms_epa / Ms) * k_as_epa * p_Am * fromGut / (k_as_epa * fromGut + p_Am) : 0,

                A = A_p + A_l,


                v_dot = par.get("v_dot", t), //p_Am_/E_m_,
                v_dot_yolk = p_Am_yolk_ / E_m_;

        double assim_frac;
        if (fromGut > 0)
            assim_frac = A / fromGut;
        else
            assim_frac = 0;
        assimRate = A;
        pam = dw_in;//p_Am;


        double
                yolk_V = yolk / par.get("yolk_energy_density", t),
                d_yolk = -p_Am_yolk_ * Math.pow(Math.max(0, Math.min(V, yolk_V)), 2.0 / 3.0), // (3) in Lika & Nisbet
                controller = E_l/E, // TODO: fractions of maintenance contribution
                p_M = V * p_dot_M_,
                p_M_p = (1 - controller) * p_M, //Math.max(0, E_p*p_M/E), //
                p_M_l = E_l/(E_l + E_dha + E_epa) * controller * p_M,
                p_M_dha = E_dha/(E_l + E_dha + E_epa) * controller * p_M,
                p_M_epa = E_epa/(E_l + E_dha + E_epa) * controller * p_M;

        // Growth phase. No egg production.

        double d_V_mrk = 0;

        double p_C_p, p_C_l, p_C_dha, p_C_epa, d_V = 0, j_P_G, j_L_G, j_DHA_G, j_EPA_G;
        // d_V is implicit and must be solved numerically. We iterate to find the correct value:
        boolean done = false;
        int count = 0;

        // Use optimization to find the implicitly defined growth rate d_V:
        OptimizationResult optRes = deb.math.Functions.nonLinearScalarMinimumSearch(growthFunction, 0, 10,
                E_p_, E_l_, E_dha_, E_epa_, V23, v_dot, p_M_p, p_M_l, p_M_dha, p_M_epa,
                P_V * prot_E / DW_to_VOL, L_V * lip_E / DW_to_VOL,
                DHA_V * lip_E / DW_to_VOL, EPA_V * lip_E / DW_to_VOL,
                sigma_G, gamma_G);
        d_V_mrk = optRes.getOptimum();

        p_C_p = E_p_ * (v_dot * V23 - d_V_mrk);
        p_C_l = E_l_ * (v_dot * V23 - d_V_mrk);
        p_C_dha = E_dha_ * (v_dot * V23 - d_V_mrk);
        p_C_epa = E_epa_ * (v_dot * V23 - d_V_mrk);

        //p_C_C = E_c_*(v_dot*V23 - d_V_mrk);
        double
                j_P_C = p_C_p - p_M_p,
                j_L_C = p_C_l - p_M_l,
                j_DHA_C = p_C_dha - p_M_dha,
                j_EPA_C = p_C_epa - p_M_epa;
        //j_C_C = p_C_C - j_C_M;

        if ((j_P_C > 0) && (j_L_C > 0)) { // && (j_C_C > 0)) {
            double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_L_C, j_DHA_C, j_EPA_C,
                    P_V * prot_E / DW_to_VOL, L_V * lip_E / DW_to_VOL,
                    DHA_V * lip_E / DW_to_VOL, EPA_V * lip_E / DW_to_VOL,
                    sigma_G);
            d_V = growth_fluxes[0] / gamma_G;
            j_P_G = growth_fluxes[1];
            j_L_G = growth_fluxes[2];
            j_DHA_G = growth_fluxes[3];
            j_EPA_G = growth_fluxes[4];

            //pG = j_P_G + j_L_G + j_C_G; // Energy spent on growth
            //pG_out = d_V*(P_V*prot_E + L_V*lip_E + C_V*carb_E)/DW_to_VOL; // Energy incorporated into structure
            double eff_P = j_P_G / j_P_C;
            double eff_L = j_L_G / j_L_C;

            //System.out.println(eff_P+"\t "+eff_L);
            //eff_C=j_C_G/j_C_C;
            //N_somatic = j_P_G/prot_E/6.25;

            //System.out.println("Eff: "+eff_P+"\t"+eff_L);
        } else {
            d_V = 0;
            j_P_G = 0;
            j_L_G = 0;
            j_DHA_G = 0;
            j_EPA_G = 0;
        }
        // Starvation
        if (d_V <= 0) {
            d_V = 0;
            ///p_M *= 0.5;
            p_C_p = p_M_p;
            p_C_l = p_M_l;
            p_C_dha = p_M_dha;
            p_C_epa = p_M_epa;

            j_P_G = 0;
            j_L_G = 0;
            j_DHA_G = 0;
            j_EPA_G = 0;
        }

        // Total energy flux used for growth:
        double sum_p_G = j_P_G + j_L_G + j_DHA_G + j_EPA_G;

        // Stomach energy content:
        double
                d_Ms_p = protein_E_in - k_g * Ms_p,
                d_Ms_l = lipid_E_in - k_g * Ms_l,
                d_Ms_dha = dha_E_in - k_g * Ms_dha,
                d_Ms_epa = epa_E_in - k_g * Ms_epa;

        // Growth:
        //d_V = (kappa*p_C - p_dot_M_*V)/E_G_;

        // Returned parts of rejected fluxes:
        double
                rej_P = kappa_R * (p_C_p - p_M_p - j_P_G),
                rej_L = kappa_R * (p_C_l - p_M_l - j_L_G),
                rej_DHA = kappa_R * (p_C_dha - p_M_dha - j_DHA_G),
                rej_EPA = kappa_R * (p_C_epa - p_M_epa - j_EPA_G);

        // Energy storage:
        //double d_E = p_A - p_C;
        double d_E_p = A_p - d_yolk * P_Y - p_C_p + rej_P,
                d_E_l = A_l - d_yolk * L_Y - p_C_l + rej_L,
                d_E_dha = A_dha - d_yolk * DHA_Y - p_C_dha + rej_DHA,
                d_E_epa = A_epa - d_yolk * EPA_Y - p_C_epa + rej_EPA;


        // Mineral fluxes:
        double[] feedCompClasses = new double[] {0.5*Ms_p/prot_E, 0.5*Ms_p/prot_E,
            0.4*Ms_lambda/lip_E, 0.3*Ms_lambda/lip_E, 0.3*Ms_lambda/lip_E};
        // Assume only protein and no free amino acids in structure. Should lipid classes be taken into account?
        double[] VCompClasses = new double[] {0, P_V, 0.4*(L_V+DHA_V+EPA_V), 0.3*(L_V+DHA_V+EPA_V),
            0.3*(L_V+DHA_V+EPA_V)};
        // Assume only free amino acids and no protein in reserves. Should lipid classes be taken into account?
        double[] ECompClasses = new double[] {E_p/prot_E, 0, 0.4*(E_l+E_dha+E_epa)/lip_E, 0.3*(E_l+E_dha+E_epa)/lip_E,
            0.3*(E_l+E_dha+E_epa)/lip_E};
        double[] feedComp = (Ms_p+Ms_lambda > 0) ? MolecularComponents.molecularComp(feedCompClasses)
                : new double[] {0,0,0,0,0};
        double[] VComp = MolecularComponents.molecularComp(VCompClasses);
        double[] EComp = MolecularComponents.molecularComp(ECompClasses);
        Matrix.printMatrix(feedComp);
        double feedCompCmol = MolecularComponents.gram2CMole(feedCompClasses);
        double VCompCmol = MolecularComponents.gram2CMole(VCompClasses);
        double ECompCmol = MolecularComponents.gram2CMole(ECompClasses);

        double eta_VG = d_V > 0 ? VCompCmol*d_V/DW_to_VOL/sum_p_G : 0; // has no effect if d_V == 0.
        double eta_XA = feedCompCmol*fromGut_g/A; // couples assimilation energy flux to feed mass flux (from gut).
        double mu_E_inv = ECompCmol/(E/(E_p/prot_E + (E_l+E_dha+E_epa)/lip_E));
        double p_D = p_M; // TODO: p_D (dissipating power) more than p_M?
        double[][] n_M_inv = new double[][] {{1, 0, 0, 0}, {0, 0.5, 0, -1.5}, {-1, -0.25, 0.5, 0.75}, {0, 0, 0, 1}};
        // TODO: put in correct values for n_O:
        /*double[][] n_O = new double[][] {
                {feedComp[0], VComp[0], EComp[0], 1},
                {feedComp[1], VComp[1], EComp[1], 1.8},
                {feedComp[2], VComp[2], EComp[2], 0.5},
                {feedComp[3], VComp[3], EComp[3], 0.2}
        };*/
        double[][] n_O = new double[][] {
                {1, 1, 1, 1},
                {1.8, 1.8, 1.8, 1.8},
                {0.5, 0.5, 0.5, 0.5},
                {0.2, 0.2, 0.2, 0.2}
        };
        if (d_V > 0)
            System.out.println("[E_G]="+(sum_p_G/d_V)+"eta_VG="+eta_VG);


       // Calculate mineral fluxes C, H, O, N in mol/day:
       mineralFluxes = mineralFluxes(new double[] {A, -d_yolk, p_D, sum_p_G}, n_O, eta_VG, eta_XA, mu_E_inv, (1-assim_frac)*eta_XA, 0, 0, 0, n_M_inv);

        
        if (++CodLarvaStoich.counter == 10000) {
            System.out.println("CodLarva.java: assim_frac = " + assim_frac + "\tk_as_p = " + k_as_p + "\tk_as_l = " + k_as_l);
            //System.out.println("CodLarva.java: lip_in = "+lipid_in+"\tprot_in = "+protein_in);
            //System.out.println("CL: d_V = "+d_V+"\td_V_by_p = "+d_V_by_p+"\td_V_by_l = "+d_V_by_l);
            //System.out.println("CL: p_C_by_p = "+p_C_by_p+"\tp_C_by_l = "+p_C_by_l);//+"\td_V_by_l = "+d_V_by_l);
            //System.out.println("CL: prolim = "+prolim+"\tdiff_p = "+(p_C_p_rej-p_C_p)+"\tdiff_l = "+(p_C_l_rej-p_C_l));
            //System.out.println("CL: controller = "+controller);
            CodLarvaStoich.counter = 0;
        }


        f[0] = d_V;
        f[1] = d_E_p;
        f[2] = d_E_l;
        f[3] = d_E_dha;
        f[4] = d_E_epa;
        f[5] = d_yolk;
        f[6] = d_Ms_p;
        f[7] = d_Ms_l;
        f[8] = d_Ms_dha;
        f[9] = d_Ms_epa;

        // Accumulation of ingested nutrients:
        f[10] = dw_in;
        f[11] = protein_in;
        f[12] = lipid_in;
        f[13] = dha_in;
        f[14] = epa_in;
        return new IndividualReturnValues(0, 0);
    }

    public Predator spawnAndGetChild(double t) {
        return null;
    }

    ScalarEquation mortalityEq = new ScalarEquation() {
        public double f(double t, double value, Parameters par) {
            return -n * par.get("mortalityRate", t);
        }
    };

    public boolean checkMortality(double t, double dt) {
        /*double number = Math.random();
        if (number < dt*par.get("mortalityRate", t)) {
                died = true;
        }*/

        setN(StepSolver.stepScalar(mortalityEq, t, dt, getN(), par));

        return n < 1;
    }

    public double[] getDerivedOutputs(double t) {

        // Convenience variables for states:
        double
                V = states[0],
                E_p = states[1],
                E_l = states[2],
                E_dha = states[3],
                E_epa = states[4],
                E = E_p + E_l + E_dha + E_epa,
                E_ = E / V,
                yolk = states[5],
                Ms_p = states[6],
                Ms_l = states[7],
                Ms_dha = states[8],
                Ms_epa = states[9];


        double
                mu_E = par.get("mu_E", t),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                L_V = par.get("L_V", t),
                P_V = par.get("P_V", t),
                DHA_V = par.get("DHA_V", t),
                EPA_V = par.get("EPA_V", t),
                L_Y = par.get("L_Y", t),
                P_Y = par.get("P_Y", t);

        // Calculate dry weights for the respective states.
        double
                M_V = V / par.get("DW_to_VOL", t),
                M_E_p = E_p / prot_E,
                M_E_l = E_l / lip_E,
                M_E_dha = E_dha / lip_E,
                M_E_epa = E_epa / lip_E,
                M_E = M_E_p + M_E_l,
                M_s = Ms_p/prot_E + (Ms_l + Ms_dha + Ms_epa)/lip_E,
                M_Y = yolk / mu_E,
                M_p = M_E_p + P_V * M_V + P_Y * M_Y,
                M_l = M_E_l + L_V * M_V + L_Y * M_Y;

        return new double[]{(M_V + M_E + M_Y + M_s), M_V, M_E_p, M_E_l, M_E_dha, M_E_epa,
                M_V*P_V, M_V*L_V, M_V*DHA_V, M_V*EPA_V,
                M_Y, M_s, M_p, M_l, seenRotifers, ingestionRate, assimRate, pam,
                mineralFluxes[0], mineralFluxes[1], mineralFluxes[2], mineralFluxes[3],
                organicFluxes[0], organicFluxes[1], organicFluxes[2], organicFluxes[3]};

    }

    public String[] getDerivedOutputNames() {
        return new String[]{"Dry weight", "Dry weight in structure", "Dry weight in protein reserves", "Dry weight in lipid reserves",
                "Dry weight in DHA reserves", "Dry weight in EPA reserves",
                "Dry weight in protein structure", "Dry weight in lipid structure", "Dry weight in DHA structure",
                "Dry weight in EPA structure",
                "Dry weight in yolk", "Dry weight in gut", "Dry weight in protein", "Dry weight in lipid"};
    }

    public Point getPosition() {
        return null;
    }

    public boolean readyToSpawn(double t) {
        return false;
    }

    public void storeCurrentState(File file) throws IOException {
        // TODO Auto-generated method stub
    }

    /* (non-Javadoc)
      * @see deb.Individual#getProteinContent(double)
      */
    public double getProteinContent(double t) {
        // TODO Auto-generated method stub
        return 0;
    }

    /* (non-Javadoc)
      * @see deb.Individual#getLipidContent(double)
      */
    public double getLipidContent(double t) {
        // TODO Auto-generated method stub
        return 0;
    }

    /* (non-Javadoc)
      * @see deb.Predator#predation(double, deb.Population, deb.Parameters)
      */
    public double predation(double t, Population population, Parameters envPars) {

        double
                V = states[0],
                V23 = Math.pow(V, 0.66667);

        // Environmental parameters
        double
                T_w = envPars.get("T_w", t),
                V_w = envPars.get("V_w", t),
                X = population.getHerbivoreCount() / V_w, // Rotifer density
                T_ref = par.get("T_ref", t),
                J_Xm_ = par.get("J_Xm_", t),
                X_K = par.get("X_K", t),
                shape_coeff = par.get("shape_coeff", t);

        ingestionRate = J_Xm_ * Math.pow(V, 0.667) * (X/(X+X_K));

        return ingestionRate;
        /*
        double length = Math.pow(V, 0.3333) / shape_coeff;

        // Calculate visual range estimate:
        double K_e = par.get("K_e", t),
                E_b = par.get("E_b", t),
                contrast = par.get("contrast", t),
                c_turb = par.get("c_turb", t),
                A_p = par.get("preyArea", t),
                f_pause = par.get("f_pause", t),
                t_h = 140;//*0.001/V23;//par.get("t_h", t);
        double E_l = Math.pow(10, 5.04 / (1 + Math.exp(-(length - 5.33) / 0.64)));
        double left_side = Math.abs(contrast) * A_p * E_l * E_b / (K_e + E_b);
        double r = 100 * deb.math.Functions.nonLinearScalarMinimumSearch(visualRangeFunction, length, 30, left_side, c_turb).getOptimum();
        seenRotifers = (2d / 3d) * 3.14 * r * r * r * X * 1e-3;
        double searchFrequency = 1 * 60 * 24;
        double attackProbability = seenRotifers / (seenRotifers + 0.75);
        double attackRate = attackProbability * searchFrequency;
        double captureProbability = length / (length + 6);
        ingestionRate = captureProbability * attackRate;

        double[] Tpars = par.getArray("Tpars", t);
        double TC = Functions.TC(T_w, T_ref, Tpars);

        // Kioerboe & MacKenzie (1995). Prey density in rot/ml.
        // "tw" = 2.
        double encounterRate = TC * X * 1e-6 * 3.14 * r * r * f_pause * (0.667 * r - 0 * 1.41 * 2);
        double ingestion = 0;
        if (encounterRate > 0) {
            ingestion = 86400 / (t_h + 1 / encounterRate);
        }
        //System.out.println("Length="+length+"\t r = "+r+"\t sqrt(left)="+(Math.sqrt(left_side))+"\t enc="+encounterRate+"\t preyInRange="+preyInRange
        //        +"\t I="+ingestion);

        feedint = ingestion;

        //return ingestion;
        return ingestionRate;
        */

        /*double[] Tpars = par.getArray("Tpars", t);
double TC = Functions.TC(T_w, T_ref, Tpars);
double
    fl = X/(X_K + X),
    J_Xm = TC*J_Xm_*V23; // Predation = fl*J_Xm

feedint = fl*J_Xm;
        //System.out.println(fl+"\t "+feedint+"");
return fl*J_Xm;
        */
    }

    public double getDryWeight(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[DRY_WEIGHT_OUTPUT];
    }

    public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
    }

    /* (non-Javadoc)
      * @see deb.Individual#compareTo(deb.Individual)
      */
    public double compareTo(Individual ind) throws ClassCastException {
        // TODO CompareTo for CodLarva
        return 0;
    }

    /* (non-Javadoc)
      * @see deb.Individual#combineWith(deb.Individual)
      */
    public Individual combineWith(Individual ind, double t) throws ClassCastException {
        // TODO CombineWith for CodLarva
        return null;
    }


    public BasicIndividual copy(Parameters par) {
        System.out.println(this.par.containsKey("J_Xm_")+" "+par.containsKey("J_Xm_"));
        CodLarvaStoich newInd = new CodLarvaStoich(par,
                this.initVal, this.n_0, this.t0);
        newInd.n = this.n;

        // Copy state values:
        newInd.states = (double[])states.clone();

        return newInd;
    }

    public int unitsEaten() {
        return totalEaten;
    }


    /**
     *
     * @param powers
     * @param n_O
     * @param eta_VG
     * @param eta_XA
     * @param mu_E_inv
     * @param eta_PA
     * @param eta_PY
     * @param eta_PD
     * @param eta_PG
     * @param n_M_inv
     * @return
     */
    public double[] mineralFluxes(double[] powers, double[][] n_O, double eta_VG, double eta_XA,
                                      double mu_E_inv, double eta_PA, double eta_PY, double eta_PD, double eta_PG,
                                      double[][] n_M_inv) {


        double[][] eta_o = new double[][] {
                { -eta_XA, -eta_XA, 0, 0},
                { 0, 0, 0, eta_VG},
                {mu_E_inv, mu_E_inv, -mu_E_inv, mu_E_inv},
                {eta_PA, eta_PY, eta_PD, eta_PG}
        };

        organicFluxes = Matrix.matr2Vector(Matrix.matrMult(eta_o, powers));

        return Matrix.matr2Vector(Matrix.matrMultScalar(-1, Matrix.matrMult(n_M_inv, Matrix.matrMult(n_O, organicFluxes))));
    }



}
