/*
 * Created on 29.des.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import deb.Parameters;
import deb.control.InternalController;

/**
 * @author alver
 *
 * This controller outputs the fraction of the maintenance flux that should be covered by lipids.
 */
public class LipidPReg implements InternalController {

	private Object lock = new Object();
	private double value = 0;
	private final double E_lambda_ref, k_reg, k_int, int_limit;
	private double integral = 0;
	private int teller=0;
	/**
	 * 
	 */
	public LipidPReg(double E_lambda_ref, double k_reg) {
		this(E_lambda_ref, k_reg, 0, 0);
	}

	public LipidPReg(double E_lambda_ref, double k_reg, double k_int, double int_limit) {
		this.E_lambda_ref = E_lambda_ref;
		this.k_reg = k_reg;
		this.k_int = k_int;
		this.int_limit = int_limit;
	}
	
	/* (non-Javadoc)
	 * @see deb.control.Regulator#getValue(double, double, deb.Parameters)
	 */
	public double getValue() {
		
		return value;
	}

	/* (non-Javadoc)
	 * @see deb.control.Regulator#update(double, double, deb.Parameters)
	 */
	public void update(Object system, double t, double dt, Parameters envPar) {
		RotiferModel rot = (RotiferModel)system;
		double[] states = rot.getBioStates();
		double
			V = states[0] + states[1] + states[2],			
			E_p_ = states[3]/V,
			E_lambda_ = (states[4]+states[5]+states[6])/V,
			E_ = E_p_ + E_lambda_;
		synchronized (lock) {
			double deviation = (E_lambda_-E_lambda_ref)/E_lambda_ref;
                        double skew = 0d;
			double baseValue = (1d+skew)*E_lambda_/(E_+skew*E_lambda_);
			value = baseValue;
			value += E_lambda_ > E_lambda_ref ? k_reg*deviation : 0; 
			value += Math.max(value+integral, baseValue);
                        value = Math.min(1, Math.max(0, value));
			integral += k_int*deviation*dt;
                        
                        //System.out.println("value = "+value);
			// Limit integral:
			//integral = Math.min(int_limit, Math.max(-int_limit, integral));
			/*if (t > 167)
				System.out.println("t="+t+"\tintegral="+integral);
			if ((t > 165) && (Math.abs(k_int*deviation*dt) > 0.2)) 
				System.out.println("t="+t+"\tintegral="+integral+"\t value="+value+"\tdeviation="+deviation);*/
		}

	}

	public void setIntegratorValue(double newVal) {
		synchronized (lock) {
			integral = newVal;
		}
	}
	
	public double getIntegratorValue() {
		return integral;
	}
        
        public InternalController copy(Parameters par) {
            LipidPReg newReg = new LipidPReg(E_lambda_ref, k_reg, k_int, int_limit);
            newReg.integral = this.integral;
            newReg.teller = this.teller;
            newReg.value = this.value;
            return newReg;
        }

    public double getCallInterval() {
        return -1;
    }
}
