/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import deb.*;
import deb.control.Controller;
import deb.fishtank.gui.ControlPanel;
import deb.operations.AddHerbivoresTabled;
import deb.operations.Operation;

import java.util.List;
import java.io.*;
import java.util.ArrayList;

/**
 * @author alver
 *         <p/>
 *         To change the template for this generated type comment go to
 *         Window>Preferences>Java>Code Generation>Code and Comments
 */
public class Main {

    public static final String DEFAULT_FILENAME = "params.xml";
    public static boolean runInSequence = true;
    public static long startTime = System.currentTimeMillis(); // Tic

    /**
     * This is the starting point of the application.
     * @param args The command-line arguments.
     */
    public static void main(String[] args) {

        // The user can specify a parameter file as an argument:
        if (args.length == 1) {
            File f = new File(args[0]);
            if (!f.exists()) {
                System.out.println("File not found: "+args[0]);
                return;
            }
            main(f.getName(), f.getParent(), true);
            return;
        }

        // Otherwise, use the directory file:
        String[] directory = readDirectory("D:\\Documents and Settings\\Alejandro\\My Documents\\PROYECTO - NTNU\\Copia de model\\testDeblife\\dir.txt");
//        String[] directory = readDirectory("N:\\model\\testDeblife\\ndir.txt");
        for (int i = 0; i < directory.length; i++) {
            System.out.println("Working directory: " + directory[i]);
            String filename = DEFAULT_FILENAME, dirname = directory[i];
            File f = new File(directory[i]);
            if (!f.exists()) {
                System.out.println("Illegal directory.");
                continue;
            } else if (!f.isDirectory()) {
                filename = f.getName();
                dirname = f.getParent();

            }
            main(filename, dirname, runInSequence);
        }
    }

    /**
     * This is the method where we set up and launch the simulation.
     * @param filename The name of the parameter file to use.
     * @param directory The directory in which files are read and written.
     * @param runInSequence Determines whether this method should return immediately after
     *  starting the simulation (in a separate thread), or whether it should return after
     *  simulation is finished. true signals that it should wait.
     */
    public static void main(String filename, String directory, boolean runInSequence) {
        try {

            // Read the main parameter file:
            Parameters par = new Parameters(new File(directory + "/" + filename));
            // Create a factory object for instantiating rotifer individuals:
            RotiferFactory factory = new RotiferFactory(directory, par);

            // See if cod larvae should be included in this simulation (if so, the parameter "larvaeParams"
            // should point to a parameter file for larvae):
            LarvaFactory lFactory = null;
            Parameters lPar = null;
            if (par.containsKey("larvaeParams")) {
                String lFilename = par.getString("larvaeParams", 0);
                File lFile = new File(directory + "/" + lFilename);
                if (lFile.exists()) {
                    lFactory = new LarvaFactory(directory);
                    lPar = new Parameters(lFile);
                }
            }

            // Instantiate feed dynamics for the tank:
            AlgaDynamic bfd = new AlgaDynamic(par);
            // Waste materials can also be simulated (experimental stuff):
            WasteDynamic wsd = new WasteDynamic(par);

            // The rotifer population reference:
            RotiferPopulation pop = null;

            // There are several different rotifer models available. The "modelType" parameter determines
            // which one we should use:
            String modelType = par.getString("modelType", 0);
            Strain strain = null;
            if (modelType.equals("RotiferSingleReserve"))
                strain = FishtankStrains.BRACHIONUS_PLICATILIS_SIMPLE;
            else if (modelType.equals("RotiferStoich4"))
                strain = FishtankStrains.BRACHIONUS_PLICATILIS_FULL_STOICH;
            else
                strain = FishtankStrains.BRACHIONUS_PLICATILIS_STOICH;

            // Instantiate the rotifer population, giving the feed/waste dynamics, rotifer strain, factory object
            // and parameter values. The RotiferPopulation itself sets up many operations such as feed addition,
            // based on parameter values:
            pop = new RotiferPopulation(bfd, wsd, strain, factory, par);

            // Instantiate as many rotifers as indicated in the parameter file:
            double nMult = par.get("n_each", 0);
            for (int i = 0; i < par.get("n_0", 0); i++) {
                pop.addHerbivore(factory.getHerbivore(strain, nMult, 0));
            }

            // Create a Simulator, which is responsible for running the simulation. The Simulator requires
            // a model to simulate, in this case represented by our RotiferPopulation object:
            Simulator sim = new Simulator(pop);

            // The following is for a special "iteration" feature in which the simulation can be split into
            // a series of parallel simulations with different values for certain parameters. If the parameter
            // "iteration" is absent or set to false, no iteration will take place:
            if (par.containsKey("iteration") && par.getBool("iteration", 0)) {
                sim.setIteration(par.get("iterationStart", 0), (int) par.get("iterationCount", 0));
            }

            // See if we should add a predictive controller instance:
            if (par.containsKey("predictiveController")) {
                try {
                    String className = par.getString("predictiveController", 0);
                    Class c = Class.forName(className);
                    Controller controller = (Controller)c.newInstance();
                    sim.setPredictiveController(controller);
                    System.out.println("Added predictive controller: "+className);
                } catch (Throwable e) {
                    e.printStackTrace();
                }

            }
/*
//			I ADDED THIS...
//          See if we should add a reference controller instance:
            if (par.containsKey("referenceController")) {
                try {
                    String className = par.getString("referenceController", 0);
                    Class ref_c = Class.forName(className);
                    Controller ref_controller = (Controller)ref_c.newInstance();
                    sim.setReferenceController(ref_controller);
                    System.out.println("Added reference controller: "+className);
                } catch (Throwable e) {
                    e.printStackTrace();
                }

            }
*/

            // Adding larvae, if specified:
            LarvaMonitor lMon = null;
            if ((lPar != null) && lPar.containsKey("n_each")) {
                nMult = lPar.get("n_each", 0);
                if (nMult > 0) {
                    System.out.println("Adding larvae.");
                    for (int i = 0; i < lPar.get("n_0", 0); i++) {
                        pop.addPredator(lFactory.getPredator(FishtankStrains.GADUS_MORHUA, nMult, 0));
                    }
                }
                lMon = new LarvaMonitor(lPar, directory);
            }

            // The Simulator itself does not keep any data from the simulation. To store data we need to add a
            // SimulationMonitor (look at the SimulationMonitor interface for details). The monitor is called
            // by the Simulator regularly, and at each point the monitor
            // can store whatever data is needed. After the simulation is finished, the monitor is allowed to store
            // the gathered data to disk:
            RotiferMonitor mon = new RotiferMonitor(par, directory, strain);

            sim.addSimulationMonitor(mon);
            if (lMon != null)
                sim.addSimulationMonitor(lMon);

            pop.setSimulateReproduction(par.getBool("simulateReproduction", 0));
            pop.setRemoveDeadIndividuals(par.getBool("removeDeadIndividuals", 0));
            pop.setSimulateMortality(par.getBool("simulateMortality", 0));
            sim.setSimulationBounds(0, par.get("t_end", 0), par.get("dt", 0));
            sim.setSimulationMonitorInterval((int) par.get("monitorInterval", 0));

            // We have a special kind of SimulationMonitor available, which shows a simple control panel where
            // you can pause and stop the simulation, as well as bringing up online plots during the simulation.
            // The control panel appears if the parameter "controlPanel" is set to true:
            if (par.getBool("controlPanel", 0)) {
                try {
                    ControlPanel pan = new ControlPanel(sim);
                    pan.setVisible(true);
                } catch (InternalError ex) {
                    // We assume we are running over an SSH connection, or something else that makes the GUI
                    // unavailable. So we'll run without the control panel.
                    System.err.println("Could not create GUI. Running without one.");
                }
            }

            // Now we start the simulator. The call to runSimulation() starts the simulation in a new thread,
            // and returns immediately:
            sim.runSimulation(par);


            // We may be running several simulations defined in several directories. If so, we can either
            // run them strictly in sequence, or just start them all at once. If running in sequence, we
            // call Thread.join() below to wait for the current simulation to finish.
            if (runInSequence) {
                try {
                    sim.join();
                } catch (InterruptedException iex) {
                    iex.printStackTrace();
                }
            }


        } catch (IOException ex) {
            System.out.println(ex.getMessage());
            //ex.printStackTrace();
        }

        System.out.println("Total time: " + ((System.currentTimeMillis() - startTime) / 1000) + " s");
    }

    /**
     * Reads all lines of the file pointed to, and returns them in a String array.
     *
     * @param string The path to the file which defines the working directory.
     * @return The paths to the working directories.
     */
    private static String[] readDirectory(String string) {
        String res = null;
        List dirs = new ArrayList();
        try {
            File f = new File(string);
            BufferedReader in = new BufferedReader(new FileReader(f));
            while (((res = in.readLine()) != null) && (!res.toLowerCase().trim().equals("end"))) {
                dirs.add(res);
            }
            in.close();

        } catch (IOException ex) {
            ex.printStackTrace();
            return new String[]{string};
        }
        String[] res2 = new String[dirs.size()];
        res2 = (String[]) dirs.toArray(res2);
        return res2;
	}

}
