package deb.fishtank;

import deb.math.Matrix;

/**
 * Created by IntelliJ IDEA.
 * User: alver
 * Date: Oct 31, 2006
 * Time: 3:24:32 PM
 * To change this template use File | Settings | File Templates.
 */
public class MolecularComponents {

    /*
     Molecular composition (C, H, O, N) of:
     0: Free Amino Acids
     1: Protein Amino Acids
     2: Phosphatidyl Choline
     3: Triacylglycerol
     4: Cholesterol
    */
    public final static double[] weight =
            {123.55, 126.44, 804.6, 911.31, 386.66};
    public final static double[][] molC =
            {       {4.81, 4.87, 43.92, 59.34, 26.0},
                    {10.17, 9.83, 79.52, 101.81, 44.0},
                    {2.28, 2.45, 8.0, 6.0, 1.0},
                    {1.31, 1.22, 1.0, 0, 0} };

    /**
     * Calculate the relative molecular content of the substrate composed of a combination of the ones defined
     * in weight and molC. Given as content of C, H, O and N, normalized so C=1.
     * @param comp The composition of the substrate.
     * @return The molecular composition in C, H, O, N.
     */
    public static double[] molecularComp(double[] comp) {
        double[] tmp = new double[weight.length];
        // Divide by nutrient class molecular weight:
        for (int i=0; i<weight.length; i++)
            tmp[i] = comp[i]/weight[i];
        tmp = Matrix.matr2Vector(Matrix.matrMult(molC, tmp));
        for (int i=tmp.length-1; i >= 0; i--)
            tmp[i] = tmp[i]/tmp[0];
        return tmp;

    }

    /**
     * Calculate the conversion factor from grams to C-moles for the given substate
     * @param comp The relative contents of the various nutrient classes. Does not have to be
     *  normalized values,
     * @return
     */
    public static double gram2CMole(double[] comp) {
        double res = 0, sumComps = 0;
        for (int i=0; i<weight.length; i++) {
            res += molC[0][i]*comp[i]/weight[i];
            sumComps += comp[i];
        }
        return res/sumComps;
    }

    public static void main(String[] args) {
        Matrix.printMatrix(molecularComp(new double[] {0.15, 0.15, 0.08, 0.06, 0.06}));
        Matrix.printMatrix(molecularComp(new double[] {0.124, 0.332, 0.102*0.3839, 0.102*0.1875, 0.102*0.2143}));
        System.out.println("gram2CMole: "+gram2CMole(new double[] {0.15, 0.15, 0.2, 0.2, 0.2}));
        System.out.println("gram2CMole: "+gram2CMole(new double[] {0.05, 0.05, 0.3, 0.3, 0.3}));
    }
}
