/*
 * Created on 29.nov.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import java.io.File;
import java.util.HashMap;
import java.util.ArrayList;

import deb.Herbivore;
import deb.HerbivoreFactory;
import deb.InitialValues;
import deb.Parameters;
import deb.Strain;

import java.io.IOException;

/**
 * This class provides RotiferModel instances based on the given parameters.
 */
public class RotiferFactory implements HerbivoreFactory {

    Parameters[] valset, valset_simple;
    InitialValues childInit, srChildInit;
    Parameters par, srPar;
    int piv = 0;
    HashMap typePars = new HashMap();

    public RotiferFactory(String directory, Parameters par) {

        try {
            this.par = par;

            // There are two ways to get initial values for a rotifer individual. One is to use the
            // "initValues" parameter given in par. Another is to read a previously stored set of
            // initial values. The latter method is used when we want to recreate a population from
            // a previous simulation. At the end of that simulation we can store a random selection of
            // individuals, preserving the state values of those individuals. Those state values can be
            // used as initial values for a later simulation, in effect creating a new population in the
            // same state as the previous one.


            // The file set describing where old initial rotifer state values should be found in
            // a directory specified in the parameter file:
            String stateValDir = par.getString("storedInitialValuesDir", 0);
            if (!stateValDir.endsWith(File.separator))
                stateValDir = stateValDir + File.separator;
            String stateValName = par.getString("storedInitialValuesName", 0);

            // Read all parameter sets and store them in an array:
            ArrayList<Parameters> tmp = new ArrayList<Parameters>();
            if (par.getBool("useStoredInitialValues", 0)) {
                int i=0;
                File f;
                while ((f = new File(stateValDir + "stateVal" + i + ".xml")).exists()) {
                    tmp.add(new Parameters(f));
                    i++;
                }
            }
            valset = tmp.toArray(new Parameters[0]);
            System.out.println("Stored rotifer value sets found: "+valset.length);

            // Create initial values based on the parameters:
            childInit = new InitialValues(par.getArray("initValues", 0));

        } catch (Exception ex) {
            System.out.println(ex.getMessage());
            //ex.printStackTrace();
        }
    }

    /*
      * (non-Javadoc)
      *
      * @see deb.HerbivoreFactory#getHerbivore(deb.Parameters, deb.Parameters,
      *      deb.InitialValues, double, double)
      */
    public Herbivore getHerbivore(Strain type, double n_each, double t) {
        //System.out.println("Creating rotifer... piv="+piv);
        Rotifer rot = null;
        if (type == FishtankStrains.BRACHIONUS_PLICATILIS) {
            if (par.getBool("useStoredInitialValues", t))
                rot = new RotiferModel(type, par, valset[piv], childInit, n_each, 0);
                //rot = new TestRotif(type, par, valset[piv%valset.length], childInit, n_each, 0);
            else
                rot = new RotiferModel(type, par, par, childInit, n_each, 0);
        } else if (type == FishtankStrains.BRACHIONUS_PLICATILIS_SIMPLE) {

            if (par.getBool("useStoredInitialValues", t))
                rot = new RotiferSingleReserve(type, par, valset[piv], childInit, n_each, 0);
            else
                rot = new RotiferSingleReserve(type, par, par, childInit, n_each, 0);
        } else if (type == FishtankStrains.BRACHIONUS_PLICATILIS_FULL_STOICH) {
            if (par.getBool("useStoredInitialValues", t))
                rot = new RotiferStoich4(type, par, valset[piv], childInit, n_each, 0);
            else
                rot = new RotiferStoich4(type, par, par, childInit, n_each, 0);
        } else if (type == FishtankStrains.BRACHIONUS_ROTUNDIFORMIS)
            rot = new RotiferModel(type, srPar, srPar, srChildInit, n_each, 0);
        piv++;
        if (piv == valset.length)
            piv = 0;

        if ((t == 0) && par.containsKey("initialMicticRatio")) {
            double micticProb = par.get("initialMicticRatio", 0);
            if (Math.random() < micticProb) {
                rot.setAmicticFemale(false);
            }
        }
        return rot;
    }

}