/*
 * RotiferMale.java
 *
 * Created on December 5, 2005, 6:32 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package deb.fishtank;

import deb.BasicIndividual;
import deb.Functions;
import deb.Herbivore;
import deb.IndividualReturnValues;
import deb.InitialValues;
import deb.Parameters;
import deb.StepSolver;
import deb.Strain;
import java.util.List;
import java.util.Random;

/**
 *
 * @author alver
 */
public class RotiferMale extends BasicIndividual implements Herbivore /* Not really herbivore, as it doesn't eat... */ {
    
    private static Random rand = new Random();
    private final double DW=0.1e-6;
    private int spermCount = 20;
    private double t0;
    
    /** Creates a new instance of RotiferMale */
    public RotiferMale(Parameters par, Strain type, double n_0, double t0) {
        super(par, type, new InitialValues(new double[] {0,0}), n_0);
        this.t0 = t0;
    }

    public double getDryWeight(double t) {
        return DW;
    }

    public double[] getDerivedOutputs(double t) {
        return null;
    }

    public double getAge(double t) {
        return t-t0;
    }

    public double compareTo(deb.Individual ind) throws ClassCastException {
        RotiferMale other = (RotiferMale)ind;
        double difference = 0;
        for (int i=0; i<states.length; i++) {
            difference += Math.abs((states[i] - other.states[i])/states[i]);
        }
        return difference;
    }

    public deb.Individual combineWith(deb.Individual ind, double t) throws ClassCastException {
        RotiferMale other = (RotiferMale)ind;
        double[] newStates = new double[states.length];
        double[] indStates = other.states;
        for (int i=0; i<newStates.length; i++) {
            newStates[i] = 0.5*(indStates[i] + states[i]); // Mean value
        }
        RotiferMale newInd = new RotiferMale(par, type, getN()+ind.getN(), 0.5*(t0+other.t0));
        newInd.states = newStates;
        return newInd;
    }

    public void addNutritionalContent(deb.NutrientIntake ni, double multiplier, double t) {
    }

    
    public deb.IndividualReturnValues f(double[] f, double t, double[] states, deb.Population population, deb.Parameters envPars) {
        // Convenience variables for states:
        double
                M_Q = states[0], // Damage-inducing compounds
                h = states[1];   // Hazard rate

        double[] Tpars = par.getArray("Tpars", t);

        double T_w = envPars.get("T_w", t),
                T_ref = par.get("T_ref", t);


        // Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)

        // Damage-inducing compounds and hazard rate:
        double p_C = par.get("p_dot_C_male", t)*TC, // Catabolic rate.... temperature dependent.
                V = 0.5e-6; // Size

        double
                nu_QC = par.get("eta_QC_male", t),
                d_M_Q = nu_QC*p_C,
                d_h = (M_Q/V);
        
        f[0] = d_M_Q;
        f[1] = d_h;
        return new IndividualReturnValues(0, 0);
    }

    public java.awt.Point getPosition() {
        return null;
    }

    public String[] getDerivedOutputNames() {
        return null;
    }

    public boolean checkMortality(double t, double dt) {
        // TODO: male lifespan, deterministic h=0.5
        // Hazard rate:
        double
                h = states[1],
                randNum = Math.random();
        if (h > 0.5) {
        //if (randNum < h*dt) {
            kill();
            //System.out.println("Male age: "+(t-t0));
            return true;
        }
        return !died;
    }

    public void storeCurrentState(java.io.File file) throws java.io.IOException {
    }

    public boolean readyToSpawn(double t) {
        return false;
    }

    public int spermCount() { return spermCount; }
    
    public Herbivore spawnAndGetChild(double t) {
        throw new UnsupportedOperationException();
    }

    public IndividualReturnValues update(double t, double dt, deb.Population population, Parameters envPar) {
        
        super.update(t, dt, population, envPar);
        
        RotiferPopulation rotPop = (RotiferPopulation)population;
        
        // Step ordinary states:
        states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);
        
        double encounterRate = par.get("encounterRate",t);
        
        // TODO: carrying capacity-virkning på fertilization... 
        double pressure = rotPop.densityPerLiter/rotPop.carryingCapacity,
                multiplier = 1;///(1+pressure*pressure);//1 - 0.75*pressure*pressure;
        //System.out.println(multiplier);
        //encounterRate *= multiplier;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 
        
        // Try to fertilize females, if there is sperm left
        if (spermCount > 0) {
            List females = population.getMatchingHerbivores(type);
            // TODO: messing with the density dependence of fertilization
            // double density = 100000;//
            double density = Functions.countIndividuals(females)/envPar.get("V_w", t);
            //double density = population.getHerbivoreCount()/envPar.get("V_w", t);

            // Probability of encounter is calculated a P(1; lambda*t) of a Poisson process:
            //double chanceOfEncounter = encounterRate*dt*density*Math.exp(-encounterRate*dt*density);
            //System.out.println(chanceOfEncounter+"\t "+chanceOfTwoEnc);
            double chanceOfNoEncounter = Math.exp(-encounterRate*dt*density);
            double chanceOfSuccess = 0.2; // probability of mating given encounter
            if (Math.random() < chanceOfSuccess*(1-chanceOfNoEncounter)) {
                Herbivore herb = (Herbivore)females.get(rand.nextInt(females.size()));
                if (herb instanceof Rotifer) {
                    Rotifer rot = (Rotifer)herb;
                    if (rot.eligibleForMating()) {
                        boolean success = rot.fertilizeIfPossible(rotPop);
                        //System.out.println(success);
                        spermCount -= 0*4;
                    }
                }            
            }
        }
        
        return new IndividualReturnValues(0,0);
    }

    public deb.Individual copy(Parameters par) throws UnsupportedOperationException {
        deb.Individual retValue;
        
        RotiferMale newM = new RotiferMale(par, this.type, n, this.t0);
        newM.states = (double[])states.clone();
        newM.spermCount = spermCount;
        return newM;
    }
    
    
}
