/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
//import java.util.ArrayList;
import javolution.util.FastTable;
import java.util.Iterator;
import java.util.List;
import deb.*;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class RotiferModel extends BasicIndividual implements Rotifer {

    static int teller = 0;

    FastTable eggs = new FastTable();
    InitialValues childInitialValue;
    boolean adv = false, old = false;
    double t0, senescence = 0;
    double pC=0, pM=0, pR=0, pJ=0, pA=0, pG=0, pW=0, waste_rel_p=0, waste_rel_la=0;
    LipidPReg lipidReg = null;

    final int REPR_BUFFER = 9, // State number for reproductive buffer
            DRY_WEIGHT_OUTPUT = 1,
            PROTEIN_OUTPUT = 5,
            LIPID_OUTPUT = 6,
            DHA_OUTPUT = 7,
            EPA_OUTPUT = 8;

    final static double SENESCENCE_INTAKE_MOD = 0.5;

    /**
     *
     */
    public RotiferModel(Strain type, Parameters par, InitialValues initVal, InitialValues childInitialValue, double n_0, double t0) {
        super(par, type, initVal, n_0);

        this.childInitialValue = childInitialValue;
        this.t0 = t0;
        init();
    }

    /**
     * Construct a new Rotifer, taking initial values and information on eggs from the Parameters
     * object supplied.
     */
    public RotiferModel(Strain type, Parameters par, Parameters initValues, InitialValues childInitialValue, double n_0, double t0) {
        super(par, type, new InitialValues(initValues.getArray("initValues", t0)), n_0);

        this.childInitialValue = childInitialValue;
        addEggs(initValues, t0);
        this.t0 = t0;
        init();
    }

    private void init() {
        lipidReg = new LipidPReg(500, 0.5, 0, 0.08);
        addRegulator(lipidReg);
        //throw new RuntimeException();
    }

    public double getAge(double t) {
        return t-t0;
    }

    public void setT0(double t0) {
        this.t0 = t0;
    }

    public double getV() {
        return states[0] + states[1]+ states[2];
    }

    public double getE() {
        return states[3] + states[4] + states[5] + states[6];
    }

    public double getReprBuffer() {
        return states[REPR_BUFFER];
    }

    protected void setReprBuffer(double newValue) {
        states[REPR_BUFFER] = newValue;
    }

    protected double getHazardRate() {
        return states[11];
    }



    /**
     * This method reads information about eggs from the parameter object, and adds the eggs, if any.
     */
    private void addEggs(Parameters prm, double t) {

        int eggCount = (int)prm.get("eggCount", t0);
        if (eggCount > 0) {
            double[] eggHatchDDG = prm.getArray("egg_hatch_DDG", t);
            double[] eggDDG = prm.getArray("egg_DDG", t);
            for (int i=0; i<eggCount; i++) {
                double[] initVal = prm.getArray("egg_initValues_"+i, t);
                RotiferEgg egg = new RotiferEgg(eggHatchDDG[i], new InitialValues(initVal), 1, t);
                // TODO: Survival rate for initial value eggs.
                egg.setCurrentDDG(eggDDG[i]);
                eggs.add(egg);
                //System.out.println("added egg");
            }
        }
    }

        /* (non-Javadoc)
        * @see deb.Herbivore#update(double, double)
        */
    public IndividualReturnValues update(double t, double dt, Population population, Parameters envPar) {
        super.update(t, dt, population, envPar);



        // Step eggs, if any:
        if (eggs.size() > 0)
            for (int i=0; i<eggs.size(); i++) {
                ((RotiferEgg)eggs.get(i)).update(dt, envPar.get("T_w", t));
            }

        // See if we are ready to create a new egg. Need to calculate the cost of an egg, based on energy density.
        // Assumption: child has standard initial structural volume, and energy density and relative DHA content equal to parent.
        double
                E_G_ = (par.get("P_V", t)*par.get("prot_energy", t)+par.get("L_V", t)*par.get("lip_energy", t))/par.get("DW_to_VOL", t);

        double[] values = childInitialValue.getValues();
        double
                V = getV(),
                V0 = values[0] + values[1] + values[2],
                E_ = getE()/V,
                costOfEgg = (E_*V0 + V0*E_G_)*par.get("egg_cost_overhead",t);
        //System.out.println(getReprBuffer()+" "+costOfEgg+" "+V);
        if (getReprBuffer() > costOfEgg) {
            eggs.add(createEgg(t, costOfEgg));
            //createEgg(t, costOfEgg); // TEST TEST TEST: create and throw away
            //System.out.println("Created egg");
        }

        if (old)
            senescence += dt*envPar.get("T_w", t);

        // Step ordinary states:
        states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);

        double
            E_p = states[3],
            E_l = states[4],
            E_dha = states[5],
            E_epa = states[6];

        // Death occurs if energy reserves are negative
        if ((E_p <= 0) || (E_l <= 0) || (E_dha <= 0) || (E_epa <= 0)) {
            kill();
        }


        return ingestion(t, population, envPar);
    }

    public IndividualReturnValues f(double[] f, double t, double[] states, Population population, Parameters envPars) {

        // Model with 3 energy storages (protein, lipid less DHA, and DHA),
        // and 2 structural volumes (protein+lipid-DHA and DHA).
        // The fat fraction of structure is assumed to be constant, so
        // lipid-DHA plus DHA should make up a constant fraction of V.

        // Convenience variables for states:
        double
                V_pl = states[0],
                V_dha = states[1],
                V_epa = states[2],
                E_p = states[3],
                E_l = states[4],
                E_dha = states[5],
                E_epa = states[6],
                M_s = states[7],
                Mat = states[8],
                R = states[9],   // Reproductive buffer
                M_Q = states[10], // Damage-inducing compounds
                h = states[11];   // Hazard rate

        // Death occurs if energy reserves are negative
        if ((E_p <= 0) || (E_l <= 0) || (E_dha <= 0) || (E_epa <= 0)) {
            kill();
            return null;
        }

        // Primary parameters:
        double
                V_p = par.get("V_p", t),
                v_dot = par.get("v_dot", t),
                k_as_P = par.get("k_as_P", t),
                k_as_L = par.get("k_as_L", t),
                k_as_DHA = par.get("k_as_DHA", t),
                k_as_EPA = par.get("k_as_EPA", t),
                p_dot_M_ = par.get("p_dot_M_", t),
                p_dot_J_ = par.get("p_dot_J_", t),
                E_g_p_ = par.get("E_G_p_", t),
                E_g_l_ = par.get("E_G_l_", t),
                //X_K = par.get("X_K", t),
                kappa = par.get("kappa", t),
                p_Am_ = par.get("p_Am_", t),
                shape_coeff = par.get("shape_coeff", t),
                alpha_M = par.get("alpha_M", t),
                beta_M = par.get("beta_M", t),
                //alpha_G = par.get("alpha_G", t),
                //beta_G = par.get("beta_G", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                T_ref = par.get("T_ref", t),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                DW_to_VOL = par.get("DW_to_VOL", t),
                p_dot_E_ = par.get("p_dot_E_", t);

        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;


        // Derived values:
        double
                V = (V_pl + V_dha + V_epa),
                E = Math.max(0, E_p + E_l + E_dha + E_epa),
                E_ = E/V,
                E_p_ = Math.max(0, E_p/V),
                E_l_ = Math.max(0, E_l/V),
                E_dha_ = Math.max(0, E_dha/V),
                E_epa_ = Math.max(0, E_epa/V),
                E_la_ = E_l_+E_dha_+E_epa_,
                V23 = Math.pow(V, 0.66667),
                V13 = Math.pow(V, 0.33333);

        double[] Tpars = par.getArray("Tpars", t);

        double L = V13 / shape_coeff; //L has direct physical meaning: standard length (cm)

        // Inputs:
        double
                P_X = envPars.getFromStruct("feed_comp", "P_X", t), // Protein fraction in feed
                L_X = envPars.getFromStruct("feed_comp", "L_X", t), // Lipid fraction in feed, including DHA
                C_X = envPars.getFromStruct("feed_comp", "C_X", t), // Carbohydrate fraction in feed (not assimilated)
                DHA_X = envPars.getFromStruct("feed_comp", "DHA_X", t), // DHA fraction in feed. Replaces part of L_X.
                EPA_X = envPars.getFromStruct("feed_comp", "EPA_X", t), // EPA fraction in feed. Replaces part of L_X.
                T_w = envPars.get("T_w", t);

        // Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)
        //System.out.println(""+TC);
        p_Am_ *= TC; // Temp.corrected maximum surface spec. assimilation rate.
        p_dot_M_ *= TC; // Temp.corrected vol.specific maintenance.
        p_dot_J_ *= TC; // Temp.corrected vol.specific maintenance.
        v_dot *= TC;
        p_dot_E_ *= TC;

        // Somatic and maturity maintenance.
        double
                p_M = p_dot_M_*V,
                p_J = Math.min(V, V_p)*p_dot_J_;//*(1-kappa)/kappa;

        // ----------------------------------------------------------------------
        // Modulating maintenance flux for low feed concentrations:
        // ----------------------------------------------------------------------
        //p_M *= 0.5+0.5*f_feed;
        //p_J *= 0.5+0.5*f_feed;
        // ----------------------------------------------------------------------

        /*
        * *************************************************************
        * STEP 1+2: Determine maintenance fluxes for protein and lipid:
        * *************************************************************
        */

        // beta_ltot is set by the controller to the fraction of maintenance that should be
        // covered by lipid:
        //double beta_ltot = Math.max(0.05, Math.min(0.95, E_la_/E_));
        double beta_ltot = Math.max(0.025, Math.min(0.975, lipidReg.getValue()));
        // Test: decouple catabolic fraction from lipids from maintenance fraction:
        double lip_catab_frac = beta_ltot; //Math.max(0.05, Math.min(0.95, (1d+skew)*E_la_/(skew*E_la_+E_)));
        double
                p_M_l = p_M*beta_ltot*E_l_/E_la_,
                p_M_dha = p_M*beta_ltot*E_dha_/E_la_,
                p_M_epa = p_M*beta_ltot*E_epa_/E_la_,
                p_M_la = p_M_l+p_M_dha+p_M_epa,
                p_M_p = p_M*(1 - beta_ltot),
                p_J_p = p_J*(1 - beta_ltot),
                p_J_l = p_J*beta_ltot*E_l_/E_la_,
                p_J_dha = p_J*beta_ltot*E_dha_/E_la_,
                p_J_epa = p_J*beta_ltot*E_epa_/E_la_,
                p_J_la = p_J_l+p_J_dha+p_J_epa;

        double
                mu_V = P_V*prot_E + L_V*lip_E,
                //E_G_ = E_g_overhead * mu_V / DW_to_VOL,
                E_G_p_ = E_g_p_ * mu_V / DW_to_VOL,
                E_G_l_ = E_g_l_ * mu_V / DW_to_VOL,
                pe_V = P_V*prot_E/mu_V,
                le_V = L_V*lip_E/mu_V;


        // Coerce reproduction effort into a standard prot/lip dist:
        double repr_lip_part = par.get("L_frac_R", t); //0.35;//beta_ltot;//0.35;


        // Get the feed ingestion rate in Joules per day:
        IndividualReturnValues ingestion = ingestion(t, population, envPars);
        double
                mu_X = prot_E*P_X + lip_E*L_X + carb_E*0.13, // The energy density of the feed
                I_DW = ingestion.getBaseFeedConsumption(), // Ingestion, in dry weight.
                //faeces_g = ingestion.getFaecesProduction(),
                p_A_p = I_DW*P_X*k_as_P*prot_E,
                p_A_l = I_DW*(L_X-DHA_X-EPA_X)*k_as_L*lip_E,
                p_A_dha = I_DW*DHA_X*k_as_DHA*lip_E,
                p_A_epa = I_DW*EPA_X*k_as_EPA*lip_E,
                p_A = p_A_p+p_A_l+p_A_dha+p_A_epa;


        /*
        * *************************************************************
        * STEP 3: Calculate growth under protein- and lipid-limitation:
        * *************************************************************
        */
        double
            d_V_p = (kappa*E_*v_dot*V23 - ((1-beta_ltot)/(1-lip_catab_frac))*p_M)/((pe_V/(1-lip_catab_frac))*E_G_p_ + kappa*E_),
            d_V_la = (kappa*E_*v_dot*V23 - (beta_ltot/lip_catab_frac)*p_M)/((le_V/lip_catab_frac)*E_G_l_ + kappa*E_),
            d_V = Math.min(d_V_p, d_V_la), //Math.sqrt(d_V_p*d_V_la), // Geometric mean?   //
            p_C_tot = E_*(v_dot*V23 - d_V);


        double
            p_C_p = (1-lip_catab_frac)*p_C_tot,
            p_C_la = lip_catab_frac*p_C_tot;

        /*
        * *************************************************************
        * STEP 4: Determine max reproductive investment from
        *         each nutrient. The minimum value is chosen.
        * *************************************************************
        */
        double
                p_R_p = (1-kappa)*p_C_p - p_J_p,
                p_R_la = (1-kappa)*p_C_la - p_J_la,
                d_R_p = p_R_p/(1-repr_lip_part),
                d_R_la = p_R_la/repr_lip_part,
                p_R = Math.min(d_R_p, d_R_la);

        /*
        * *************************************************************
        * STEP 4.5: Split lipid allocation between compartments.
        * *************************************************************
        */
        double p_C_l, p_C_dha, p_C_epa;
        p_C_l = p_C_la*E_l_/E_la_;
        p_C_dha = p_C_la*E_dha_/E_la_;
        p_C_epa = p_C_la*E_epa_/E_la_;

        // Split growth between the volumes. 
        double
                d_V_dha = d_V*E_dha_/E_la_,
                d_V_epa = d_V*E_epa_/E_la_,
                d_V_pl = d_V-d_V_dha-d_V_epa;

        // XXX: TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
                                /*double flatut = Math.max(0, 1-(E_-1200)/(2100-1200)),
                                        a1 = alpha_G*flatut,
                                        a2 = beta_G*flatut,
                                        adj = a1*p_C_dha + a2*p_C_epa;
                                p_C_dha *= 1+a1;
                                p_C_epa *= 1+a2;
                                p_C_l -= adj;
                                if (++teller == 40000) {
                                        System.out.println("flatut = "+flatut);
                                        teller = 0;
                                }	*/
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST


        //String status = "normal";
        /*
         * *************************************************************
         * STEP 5: Starvation conditions
         * *************************************************************
         */
        if (d_V <= 0) { // No-growth condition
            //status = "no-growth";
            //System.out.println("no-growth");
            // Set growth and reproduction to 0.            
            d_V = 0;
            d_V_pl = 0;
            d_V_dha = 0;
            d_V_epa = 0;
            p_R = 0;
            // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
            //p_M *= 0.7;
            //p_J *= 0.7;
            // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
            p_C_tot = p_M+p_J  +p_R;
            p_C_p = p_M_p+p_J_p  +(1-repr_lip_part)*p_R;
            p_C_la = p_M_la+p_J_la  +repr_lip_part*p_R;
            p_C_l = p_M_la+p_J_l  +repr_lip_part*p_R*E_l_/E_la_;
            p_C_dha = p_M_dha+p_J_dha  +repr_lip_part*p_R*E_dha_/E_la_;
            p_C_epa = p_M_epa+p_J_epa  +repr_lip_part*p_R*E_epa_/E_la_;
            // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST

        }
        /*
         * *************************************************************
         * STEP 6: Determinate growth
         * *************************************************************
         */
        else if (V >= V_p) { // Determinate growth:
            //status = "determinate";
            //double oldpR = p_R;
            d_V = 0;
            d_V_pl = 0;
            d_V_dha = 0;
            d_V_epa = 0;

            // Setting growth to zero gives a simple expr. for p_C:
            p_C_tot = E_*v_dot*V23;
            //System.out.println(t+"\tp_C_tot="+p_C_tot);
            // Split the catabolic power as before:
            p_C_p = (1-lip_catab_frac)*p_C_tot;
            p_C_la = lip_catab_frac*p_C_tot;

            p_C_l = p_C_la*E_l_/E_la_; // - delta/kappa;
            p_C_dha = p_C_la*E_dha_/E_la_; // + delta/kappa;
            p_C_epa = p_C_la*E_epa_/E_la_; // + delta/kappa;

            kappa = p_M/p_C_tot;

            //System.out.println("Pralle: "+(p_C_p + p_C_l + p_C_dha)+"\t"+p_C);

            // Recalculate reproductive investment:
            p_R_p = (1-kappa)*p_C_p - p_J_p;
            p_R_la = (1-kappa)*p_C_la - p_J_la;
            d_R_p = p_R_p/(1-repr_lip_part);
            d_R_la = p_R_la/repr_lip_part;
            p_R = Math.min(d_R_p, d_R_la);

            //if ((t>5) &&(p_R < 0))// && !status.equals("normal"))
            //System.out.println(kappa+"\tp_R = "+p_R);//+"\t d_R = "+d_R);

            //System.out.println("kappa="+kappa+"\tp_C = "+p_C_tot+"\tp_C_p = "+p_C_p+"\tp_C_la = "+p_C_la);
            //System.out.println("p_R = "+p_R+"\td_R_p = "+d_R_p+"\td_R_la = "+d_R_la);
        }


        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST        
        double delta = alpha_M*(p_M_dha+p_J_dha),
                gamma = beta_M*(p_M_epa+p_J_epa);

        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
        //delta = 0;
        //gamma = 0;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST

        p_C_l = p_C_l - delta - gamma;
        p_C_epa = p_C_epa + gamma;
        p_C_dha = p_C_dha + delta;



        /*
        * *************************************************************
        * STEP 7: Calculate state equations.
        * *************************************************************
        */
        double
                d_E_p = p_A_p - p_C_p,// - 1*rej_p,// + rej_E_p,
                d_E_l = p_A_l - p_C_l,// - 1*rej_la,// + rej_E_l,
                d_E_dha = p_A_dha - p_C_dha,// + rej_E_dha,
                d_E_epa = p_A_epa - p_C_epa,
                d_E = d_E_p + d_E_l + d_E_dha + d_E_epa;


        //System.out.println("V = "+V+"\tV_p = "+V_p+"\td_V = "+d_V);
        //System.out.println("p_A_p = "+p_A_p);

        //  Maturity investment or reproduction
        double d_Mat, d_R;
        if (!old && (V >= V_p)) { //(Mat >= Mat_max)) {
            d_Mat = 0;
            d_R = Math.max(p_R, 0);//25.0e-4);
            //System.out.println(test+" ... d_R = "+d_R);
        } else if (!old) {
            d_R = 0;
            d_Mat = Math.max(p_R, 0);
        } else {
            d_R = 0;
            d_Mat = 0;
        }


        // Calculate wasted fractions:
        waste_rel_p = (p_C_p - (p_M_p + p_J_p + d_V*E_G_p_*pe_V + p_R*(1-repr_lip_part)))/p_C_p;
        waste_rel_la = (p_C_la - (p_M_la + p_J_la + d_V*E_G_l_*le_V + p_R*repr_lip_part))/p_C_la;

        // Absolute waste values per E compartment:
        double
                waste_p = waste_rel_p*p_C_p,
                waste_l = p_C_l - (p_M_l + p_J_l + (d_V*E_G_l_*le_V + p_R*repr_lip_part)*E_l_/E_la_),
                waste_dha = p_C_dha - (p_M_dha + p_J_dha + (d_V*E_G_l_*le_V + p_R*repr_lip_part)*E_dha_/E_la_),
                waste_epa = p_C_epa - (p_M_epa + p_J_epa + (d_V*E_G_l_*le_V + p_R*repr_lip_part)*E_epa_/E_la_),
                waste_tot = waste_p + waste_l + waste_dha + waste_epa;
        //System.out.println("Wasted: p = "+waste_p+"\t l = "+waste_l);


        // Damage-inducing compounds and hazard rate:
        double
                nu_QC = par.get("eta_QC", t),
                d_M_Q = nu_QC*p_C_tot,
                d_h = (M_Q/V);//*(1+density_effect);

        if (Double.isNaN(d_V_pl) || Double.isNaN(d_V_dha) || Double.isNaN(d_V_epa)) {
            System.out.println("V="+getV()+"\tpl=="+d_V_pl+"\tdha="+d_V_dha+"\tepa="+d_V_epa);
            for (int i=0; i<states.length; i++)
                System.out.println(i+".\t"+states[i]);
            System.exit(0);
        }

        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 
        double rate = 0;
        /*if (E_dha/E > 0.25)
            rate = Math.pow((E_dha/E/0.3), 3);
         */
        double dha_to_epa = 0.5*rate*E_dha;
        double dha_to_l = rate*E_dha;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 

        double mult = 0.3;

        f[0] = d_V_pl;
        f[1] = d_V_dha;
        f[2] = d_V_epa;
        f[3] = d_E_p + waste_p*mult;
        f[4] = d_E_l + waste_l*mult + dha_to_l;
        f[5] = d_E_dha + waste_dha*mult - dha_to_epa - dha_to_l;
        f[6] = d_E_epa + waste_epa*mult + dha_to_epa;
        f[7] = 0;//p_I - p_A;
        f[8] = d_Mat;
        f[9] = d_R;
        f[10] = d_M_Q;
        f[11] = d_h;

        pA = p_A;//p_A_p+p_A_l+p_A_dha+p_A_epa;//*(P_X+L_X);
        pC = p_C_tot;//p_C_la+p_C_p;//+p_C_l+p_C_dha+p_C_epa;//
        pJ = p_J;
        pR = p_R;
        pM = p_M;
        pG = d_V*(pe_V*E_G_p_+le_V*E_G_l_);
        pW = waste_tot;

        return ingestion;//new IndividualReturnValues(p_A, faeces_g);
    }

        /*public double ingestion(double t, BaseFeedDynamic baseFeedDynamic, Parameters envPars) {
                double X = baseFeedDynamic.getConcentration();
                return ingestion(t, X, envPars);
        }*/

    /**
     * Calculates the feed ingestion rate, as given in g per day. Feed composition is discounted by this calculation.
     */
    public IndividualReturnValues ingestion(double t, Population population, Parameters envPars) {
        double X = population.getBaseFeedConcentration(),
                TC = envPars.get("T_w", t)/25, //tempcorr(T_w+273,T_ref,par_r.Tpars); % -, Temperature Correction factor (fnct. by Kooijman)
                mu_X = 25000; // "Reference" energy density of feed
        if (X<0) {
            X = 0;
        }
        double p_Am_ = par.get("p_Am_", t);
        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;

        double I_DW = (X/(X+par.get("X_K", t)))*TC*p_Am_*Math.pow(getV(), 0.66667)/mu_X;
        // Calculate faeces is grams. Simplified - not adjusting for different assimilation rate of DHA and EPA.

        double faeces_g = I_DW*par.getFromStruct("feed_comp", "P_X", t)*(1-par.get("k_as_P", t))
            + I_DW*par.getFromStruct("feed_comp", "L_X", t)*(1-par.get("k_as_L", t));
        //System.out.println(faeces_g+"\t"+I_DW+"\t"+(par.get("P_X", t)*(1-par.get("k_as_P", t)))+"\t"+(par.get("L_X", t)*(1-par.get("k_as_L", t))));

        return new IndividualReturnValues(I_DW, faeces_g);
        //return new IndividualReturnValues(0, faeces_g);
    }

        /* (non-Javadoc)
        * @see deb.Individual#isAlive()
        */
    public boolean checkMortality(double t, double dt) {
        if (!old) {
            // Hazard rate:

            double
                    h = getHazardRate(),
                    randNum = Math.random();
            old = randNum < h*dt;
        }
        if (old && (senescence > 40))//40))
            died = true;
        return !died;
    }

    public double[] getDerivedOutputs(double t) {

        double
                V_pl = states[0],
                V_dha = states[1],
                V_epa = states[2],
                E_p = states[3],
                E_l = states[4],
                E_dha = states[5],
                E_epa = states[6],
                M_s = states[7],
                Mat = states[8],
                R = states[9];

        double
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                P_V = par.get("P_V",t),
                L_V = par.get("L_V",t),
                V = (V_pl + V_dha + V_epa),
                E = (E_p+E_l+E_dha+E_epa);

        // Calculate dry weights for the respective states.
        double
                M_V = V/par.get("DW_to_VOL", t),
                M_E = (E_p/prot_E+(E_l+E_dha+E_epa)/lip_E)/0.65d, // TEST TEST TEST TEST TEST TEST TEST
                M_prot_S = 0,//M_s*par.get("P_X",t)/prot_E,
                // TODO: Composition of stomach content.
                M_lip_S = 0,//M_s*par.get("L_X",t)/lip_E,
                M_dha_S = 0,//M_s*par.get("DHA_X",t)/lip_E,
                M_epa_S = 0,
                M_S = M_prot_S + M_lip_S + M_dha_S,
                M_lip_V = V*L_V/par.get("DW_to_VOL", t), //M_V*par_r.lip_V;
                M_lip_E = (E_l+E_dha+E_epa)/par.get("lip_energy", t),
                M_lip = M_lip_E + M_lip_V + M_lip_S,
                M_prot_V = M_V*P_V,
                M_prot_E = E_p/par.get("prot_energy", t),
                M_prot = M_prot_V + M_prot_E + M_prot_S,
                M_dha_V = V_dha*L_V/par.get("DW_to_VOL", t),
                M_dha_E = E_dha/par.get("lip_energy", t),
                M_dha = M_dha_V + M_dha_E + M_dha_S,
                M_epa_V = V_epa*L_V/par.get("DW_to_VOL", t),
                M_epa_E = E_epa/par.get("lip_energy", t),
                M_epa = M_epa_V + M_epa_E + M_epa_S;

        // Eggs also contribute to derived outputs.
        double eggDW = 0;
        if (eggs.size() > 0) {
            for (int i=0; i<eggs.size(); i++) {
                Rotifer child = ((RotiferEgg)eggs.get(i)).getChild();
                double[] outp = child.getDerivedOutputs(t);
                eggDW += outp[1];
            }


        }
        //if (t > 43) System.out.println(((LipidPReg)lipidReg).getIntegratorValue());
        return new double[] {eggs.size(), (M_E+M_V+M_S+eggDW), M_V, M_E, /*M_S,*/ eggDW, M_prot, M_lip, M_dha, M_epa, M_dha_V, M_dha_E, M_epa_V,
                M_epa_E, E/V, pA, pC, pR, pM, pG, pW, ((LipidPReg)lipidReg).getValue(), waste_rel_p, waste_rel_la};
    }

    public String[] getDerivedOutputNames() {
        return new String[] {"Eggs", "Dry weight incl. eggs", "Dry weight in structure", "Dry weight in reserves", /*"Dry weight in gut",*/ "Dry weight in eggs",
                "Dry weight in protein", "Dry weight in lipids", "Dry weight in DHA", "Dry weight in EPA", "DHA in structure", "DHA in reserves",
                "EPA in structure", "EPA in reserves", "Energy density", "Lipid controller integral value"};
    }


        /* (non-Javadoc)
        * @see deb.Individual#getPosition()
        */
    public Point getPosition() {
        return null;
    }

    protected RotiferEgg createEgg(double t, double costOfEgg) {
        double[] values = getEggInitVector(t);
        // Reduce reproductive buffer accordingly:
        setReprBuffer(getReprBuffer() - costOfEgg);
        double
                V = getV(),
                E_ = (getE())/V;

        double survivalRate = par.get("egg_survival_rate", t);//0.9; //Math.min(1, 0.3 + E_/(1400+E_));
        //System.out.println(""+survivalRate);
        return new RotiferEgg(par.get("hatch_time_DDG", t), new InitialValues(values), survivalRate, t);
    }

    protected double[] getEggInitVector(double t) {
        // Assumption: child has standard initial structural volume, and energy density and relative DHA content equal to parent.
        double[] values = childInitialValue.getValues();
        double
                V0 = values[0] + values[1] + values[2],
                V = states[0] + states[1] + states[2],
                E_p_ = states[3]/V,
                E_l_ = states[4]/V,
                E_dha_ = states[5]/V,
                E_epa_ = states[6]/V;

        values[0] = V0*states[0]/V;
        values[1] = V0*states[1]/V;
        values[2] = V0*states[2]/V;
        values[3] = E_p_*V0;
        values[4] = E_l_*V0;
        values[5] = E_dha_*V0;
        values[6] = E_epa_*V0;
        //System.out.println("E = "+(V0*(E_p_+E_l_+E_dha_)+V0*1537)); // E_G_ = 897.1

        return values;
    }

        /* (non-Javadoc)
        * @see deb.Individual#readyToSpawn()
        */
    public boolean readyToSpawn(double t) {
        if (eggs.size() == 0)
            return false;
        for (int i=0; i<eggs.size(); i++) {
                if (((RotiferEgg)eggs.get(i)).readyToHatch())
                    return true;
        }
        return false;
    }

        /* (non-Javadoc)
        * @see deb.Individual#getChild()
        */
    public Herbivore spawnAndGetChild(double t) {
        for (int i=0; i<eggs.size(); i++) {
            RotiferEgg egg = (RotiferEgg)eggs.get(i);
            if (egg.readyToHatch()) {
                eggs.remove(egg);
                Rotifer rot = egg.getChild();
                rot.setT0(t);

                                /*double[] deriv = rot.getBioStates();
                   double weight = (rot.getDerivedOutputs(t))[1];
                   double energy = (deriv[0] + deriv[1])*0.1*(0.4*23600 + 0.18*39500)
                           + deriv[2] + deriv[3] + deriv[4];*/
                //System.out.println("Egg energy = "+energy+"\tDW = "+weight);

                if (!egg.isViable())
                    rot.kill();
                return rot;
            }
        }
        return null;
    }

    /**
     * Remove eggs randomly.
     * @param fraction The likelyhood of each egg being removed.
     */
    public void killEggs(double fraction) {
        if (eggs.size() > 0) {
            for (Iterator i=eggs.iterator(); i.hasNext();) {
                double value = Math.random();
                if (value <= fraction)
                    i.remove();
            }
        }
    }

    public void storeCurrentState(File file) throws IOException {
        Parameters state = new Parameters();
        state.putArray("initValues", states);//initVal.getValues());
        state.put("eggCount", eggs.size());
        if (eggs.size() > 0) {
            double[] eggDDG = new double[eggs.size()],
                    hatch_DDG = new double[eggs.size()];
            for (int i=0; i<eggDDG.length; i++) {
                RotiferEgg egg = (RotiferEgg)eggs.get(i);
                eggDDG[i] = egg.getCurrentDDG();
                hatch_DDG[i] = egg.getHatch_time_DDG();
                state.putArray("egg_initValues_"+i, egg.getInitialValues().getValues());
            }
            state.putArray("egg_hatch_DDG", hatch_DDG);
            state.putArray("egg_DDG", eggDDG);
        }
        state.storeValues(file);
    }

        /* (non-Javadoc)
        * @see deb.Individual#getProteinContent(double)
        */
    public double getProteinContent(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[PROTEIN_OUTPUT];
    }

        /* (non-Javadoc)
        * @see deb.Individual#getLipidContent(double)
        */
    public double getLipidContent(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[LIPID_OUTPUT];
    }

    /**
     * Returns the current egg count of this rotifer.
     * @return The current number of eggs.
     */
    public int getEggCount() {
        return eggs.size();
    }

    public List getEggs() {
        return eggs;
    }

    /**
     * Simple inner class to represent rotifer eggs that hatch after a specific amount of day degrees. This class holds
     * the InitialValues that are to be used in creating the hatched rotifer.
     */
    //static int ali = 0, dea = 0;
    class RotiferEgg {
        double hatch_time_DDG, currentDDG=0;
        InitialValues initVal;
        boolean viable;
        Rotifer child;

        public RotiferEgg(double hatch_time_DDG, InitialValues initVal, double survivalRate, double t) {
            this.hatch_time_DDG = hatch_time_DDG;
            this.initVal = initVal;

            child = new RotiferModel(type, par, initVal, childInitialValue, getN(), t);

            //System.out.println(survivalRate+"");
            if (Math.random() <= survivalRate) {

                viable = true;
            } else {
                viable = false;
            }
        }
        /**
         * @return
         */
        public Rotifer getChild() {
            return child;
        }

        public void update(double dt, double T_w) {
            currentDDG += dt*T_w;
        }

        public boolean isViable() {
            return viable;
        }

        public boolean readyToHatch() {
            return currentDDG >= hatch_time_DDG;
        }
        public InitialValues getInitialValues() {
            return initVal;
        }
        /**
         * @return Returns the currentDDG.
         */
        public double getCurrentDDG() {
            return currentDDG;
        }
        /**
         * @return Returns the hatch_time_DDG.
         */
        public double getHatch_time_DDG() {
            return hatch_time_DDG;
        }
        /**
         * @param currentDDG The currentDDG to set.
         */
        public void setCurrentDDG(double currentDDG) {
            this.currentDDG = currentDDG;
        }

        /**
         * Add the nutritional value of this egg. It is currently realized by constructing a Rotifer object as
         * it would hatch from this egg, and delegating the operation to it.
         * @param ni
         * @param multiplier
         * @param t
         */
        public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
            Rotifer rot = new RotiferModel(type, par, getInitialValues(), childInitialValue, getN(), t);
            rot.addNutritionalContent(ni, multiplier, t);
        }
    }


    public double getDryWeight(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[DRY_WEIGHT_OUTPUT];
    }

        /* (non-Javadoc)
        * @see deb.Individual#addNutritionalContent(deb.NutrientIntake, double)
        */
    public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
        double[] outp = getDerivedOutputs(t);

        ni.add("dw", multiplier*outp[DRY_WEIGHT_OUTPUT]);
        ni.add("protein", multiplier*outp[PROTEIN_OUTPUT]);
        ni.add("lipid", multiplier*outp[LIPID_OUTPUT]);
        ni.add("dha", multiplier*outp[DHA_OUTPUT]);
        ni.add("epa", multiplier*outp[EPA_OUTPUT]);

        for (Iterator i=eggs.iterator(); i.hasNext();) {
            RotiferEgg e = (RotiferEgg)i.next();
            e.addNutritionalContent(ni, multiplier, t);
        }
    }

    // How much relative weight should differences in each of the state variables be given when comparing rotifers?
    private static double[] COMPARE_STATE_WEIGHTS = new double[] {1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1};

    // We need a threshold for each state below which it will be considered equal to zero. This is to prevent "wind-up" in
    // a situation where a state is very close to zero for all individuals, giving very high relative differences that are
    // in fact irrelevant:
    private static double[] ZERO_THRESHOLD = new double[] {1e-9, 1e-10, 1e-10, 1e-6, 1e-6, 1e-7, 1e-7, 0, 1e-6, 1e-6, 1e-10, 1e-5};


    private static double COMPARE_EGG_WEIGHT = 1;
    //static int[] teller = new int[11];
    //static int tell2 = 0;

        /* (non-Javadoc)
        * @see deb.Individual#compareTo(deb.Individual)
        */
    public double compareTo(Individual ind) throws ClassCastException {
        Rotifer rot = (Rotifer)ind;
        if (type != rot.getType())
            return 100; // Different types cannot be combined.
        double[] otherStates = rot.getBioStates();
        double stateDiff = 0, biggestDiff = -1;
        int biggestI = -1;
        // Calculate relative, weighted sum of deviances:
        for (int i=0; i<states.length; i++) {
            if (states[i] != 0) {
                // Check if both are above the zero threshold before counting the relative difference:
                if ((states[i] > ZERO_THRESHOLD[i]) || (otherStates[i] > ZERO_THRESHOLD[i]))
                    stateDiff += COMPARE_STATE_WEIGHTS[i]*Math.abs(states[i]-otherStates[i])/states[i];
            } else {
                stateDiff += (otherStates[i] != 0) ? COMPARE_STATE_WEIGHTS[i] : 0;
            }
        }
        stateDiff += COMPARE_EGG_WEIGHT*Math.abs(getEggCount()-rot.getEggCount());

        return stateDiff;
    }

        /* (non-Javadoc)
        * @see deb.Individual#combineWith(deb.Individual)
        */
    public Individual combineWith(Individual ind, double t) throws ClassCastException {
        // We assume that the given individual has the same number of eggs as us, because
        // a difference in egg count weighs in greatly in comparison. Also, both rotifers has
        // one or more eggs, these are likely to have a similar composition, due to the similarity
        // in other states.
        RotiferModel oth = (RotiferModel)ind;
        if (type != oth.getType())
            throw new ClassCastException("Incompatible rotifer types");
        double othN = oth.getN(),
                sumN = othN + n;

        double[] othSt = oth.getBioStates(),
                othChildInit = oth.getChildInitialValue().getValues(),
                thisChildInit = childInitialValue.getValues();


        // The new state vector should be a weighted mean of the old ones:
        double[] newSt = new double[states.length],
                newChildInit = new double[states.length];

        for (int i=0; i<newSt.length; i++) {
            newSt[i] = (n*states[i] + othN*othSt[i])/sumN;
            newChildInit[i] = (n*thisChildInit[i] + othN*othChildInit[i])/sumN;
        }

        RotiferModel newRot = new RotiferModel(type, par, new InitialValues(newSt), new InitialValues(newChildInit), sumN, (n*t0 + othN*oth.getT0())/sumN);

        newRot.lipidReg.setIntegratorValue((lipidReg.getIntegratorValue()*getN()+oth.lipidReg.getIntegratorValue()*othN)/sumN);

        // If the individuals to combine have eggs
        if (eggs.size() > 0) {
            //System.out.println("Combining with eggs: "+eggs.size());
            for (Iterator i=eggs.iterator(); i.hasNext();)
                newRot.eggs.add(i.next());
        }



        return newRot;
    }



    /**
     * "Backup version" of the f method made June 10, 2005, before rewriting allocation algorithm.
     */
    public IndividualReturnValues f_old(double[] f, double t, double[] states, Population population, Parameters envPars) {

        assert lipidReg != null;

        // Model with 3 energy storages (protein, lipid less DHA, and DHA),
        // and 2 structural volumes (protein+lipid-DHA and DHA).
        // The fat fraction of structure is assumed to be constant, so
        // lipid-DHA plus DHA should make up a constant fraction of V.

        // Convenience variables for states:
        double
                V_pl = states[0],
                V_dha = states[1],
                V_epa = states[2],
                E_p = states[3],
                E_l = states[4],
                E_dha = states[5],
                E_epa = states[6],
                M_s = states[7],
                Mat = states[8],
                R = states[9],   // Reproductive buffer
                M_Q = states[10], // Damage-inducing compounds
                h = states[11];   // Hazard rate

        // Primary parameters:
        double
                V_p = par.get("V_p", t),
                v_dot = par.get("v_dot", t),
                k_as_P = par.get("k_as_P", t),
                k_as_L = par.get("k_as_L", t),
                k_as_DHA = par.get("k_as_DHA", t),
                k_as_EPA = par.get("k_as_EPA", t),
                p_dot_M_ = par.get("p_dot_M_", t),
                p_dot_J_ = par.get("p_dot_J_", t),
                E_g_overhead = par.get("E_G_", t),
                X_K = par.get("X_K", t),
                kappa = par.get("kappa", t),
                p_Am_ = par.get("p_Am_", t),
                shape_coeff = par.get("shape_coeff", t),
                alpha_M = par.get("alpha_M", t),
                beta_M = par.get("beta_M", t),
                alpha_G = par.get("alpha_G", t),
                beta_G = par.get("beta_G", t),
                //Mat_max= par.get("Mat_max", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                T_ref = par.get("T_ref", t),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                DW_to_VOL = par.get("DW_to_VOL", t),
                p_dot_E_ = par.get("p_dot_E_", t);

        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;


        // Derived values:
        double
                V = (V_pl + V_dha + V_epa),
                E = Math.max(0, E_p + E_l + E_dha + E_epa),
                E_ = E/V,
                E_p_ = Math.max(0, E_p/V),
                E_l_ = Math.max(0, E_l/V),
                E_dha_ = Math.max(0, E_dha/V),
                E_epa_ = Math.max(0, E_epa/V),
                E_la_ = E_l_+E_dha_+E_epa_,
                V23 = Math.pow(V, 0.66667),
                V13 = Math.pow(V, 0.33333);

        // Death occurs if energy reserve cannot cover maintenance.
        if ((E <= 0) || (E_l <= 0)) {
            died = true;
            return new IndividualReturnValues();
            //System.out.println("JUUUUUBBB");
        }

        //double rot_dens = population.getHerbivoreCount()/par.get("V_w", t);

        double[] Tpars = par.getArray("Tpars", t);

        double
                L = V13 / shape_coeff; //L has direct physical meaning: standard length (cm)

        // Inputs:
        double
                P_X = envPars.get("P_X", t), // Protein fraction in feed
                L_X = envPars.get("L_X", t), // Lipid fraction in feed, including DHA
                C_X = envPars.get("C_X", t), // Carbohydrate fraction in feed (not assimilated)
                DHA_X = envPars.get("DHA_X", t), // DHA fraction in feed. Replaces part of L_X.
                EPA_X = envPars.get("EPA_X", t), // EPA fraction in feed. Replaces part of L_X.
                T_w = envPars.get("T_w", t);

        // Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)
        //System.out.println(""+TC);
        p_Am_ *= TC; // Temp.corrected maximum surface spec. assimilation rate.
        p_dot_M_ *= TC; // Temp.corrected vol.specific maintenance.
        p_dot_J_ *= TC; // Temp.corrected vol.specific maintenance.
        v_dot *= TC;
        p_dot_E_ *= TC;
        //double p_Am = p_Am_*Math.pow(V, 0.66667); //pT_Am_*V^(2/3);

        // Somatic and maturity maintenance.
        double
                p_M = p_dot_M_*V,
                p_J = Math.min(V, V_p)*p_dot_J_;//*(1-kappa)/kappa;
        //p_M = p_S + p_J;
        //Math.min(Mat, Mat_max)*p_dot_Mat_*(1-kappa)/kappa;


        // ----------------------------------------------------------------------
        // Modulating maintenance flux for low feed concentrations:
        // ----------------------------------------------------------------------
        double X = population.getBaseFeedConcentration();
        if (X<0) {
            X = 0;
        }
        double f_feed = X/(X+par.get("X_K", t));
        //p_M *= 0.5+0.5*f_feed;
        //p_J *= 0.5+0.5*f_feed;
        // ----------------------------------------------------------------------

        double
                E_ltot_ = E_l_+E_dha_+E_epa_;
        //p_M_l = p_M*E_l/E - delta,
        //p_M_p = p_M*E_p/E,
        //p_M_dha = p_M*E_dha/E + delta;


        // beta_ltot is set by the controller to the fraction of maintenance that should be
        // covered by lipid:
        double beta_ltot = lipidReg.getValue(); //E_ltot_/E_;  //

        //if (E_dha > E_l)
        //	throw new RuntimeException("E_dha="+E_dha+", E_l="+E_l);
        //assert E_l_ >= E_dha_: "\tE_l_="+E_l_+"\tE_dha_="+E_dha_;
        //assert delta <= 0 || delta < p_M*beta_ltot*E_l_/E_ltot_: "delta="+delta+"\t"+p_M*beta_ltot*E_l_/E_ltot_
        //	+"\tE_l_="+E_l_+"\tE_dha_="+E_dha_+"\nbeta="+beta_ltot;

        double
                p_M_l = p_M*beta_ltot*E_l_/E_ltot_,
                p_M_dha = p_M*beta_ltot*E_dha_/E_ltot_,
                p_M_epa = p_M*beta_ltot*E_epa_/E_ltot_,
                p_M_la = p_M_l+p_M_dha+p_M_epa,
                p_M_p = p_M*(1 - beta_ltot),
                p_J_p = p_J*(1-beta_ltot),
                p_J_l = p_J*beta_ltot*E_l_/E_ltot_,
                p_J_dha = p_J*beta_ltot*E_dha_/E_ltot_,
                p_J_epa = p_J*beta_ltot*E_epa_/E_ltot_,
                p_J_la = p_J_l+p_J_dha+p_J_epa;


        //assert p_M_l >= 0: p_M_dha+"\t"+E_dha+"\t"+delta+"\t"+E_l;
        //assert p_M_dha >= 0: p_M_dha+"\t"+E_dha+"\t"+delta+"\t"+E_l+"\t"+beta_ltot;
        assert p_M_p >= 0;

        //System.out.println("beta = "+beta_ltot+"\tp_M_p="+p_M_p+"\tp_M_la="+p_M_la);
        // Get the feed ingestion rate in Joules per day:
        IndividualReturnValues ingestion = ingestion(t, population, envPars);
        double
                mu_X = prot_E*P_X + lip_E*L_X + carb_E*0.13, // The energy density of the feed
                I_DW = ingestion.getBaseFeedConsumption(), // Ingestion, in dry weight.
                faeces_g = ingestion.getFaecesProduction(),
                p_A_p = I_DW*P_X*k_as_P*prot_E,
                p_A_l = I_DW*(L_X-DHA_X-EPA_X)*k_as_L*lip_E,
                p_A_dha = I_DW*DHA_X*k_as_DHA*lip_E,
                p_A_epa = I_DW*EPA_X*k_as_EPA*lip_E,
                p_A = p_A_p+p_A_l+p_A_dha+p_A_epa;



        assert p_A_l >= p_A_dha;
        double
                p_M_ltot = p_M_l + p_M_dha + p_M_epa;

        double
                mu_V = P_V*prot_E + L_V*lip_E,
                E_G_ = E_g_overhead * mu_V / DW_to_VOL,
                pe_V = P_V*prot_E/mu_V,
                le_V = L_V*lip_E/mu_V,

                // Calculate catabolic power based on total energy:
                p_C_tot = E_*(E_G_*v_dot*V23 + p_M) / (E_G_ + kappa*E_),

                // Calculate growth flux:
                p_G = kappa*p_C_tot - p_M,  // d_G = p_G/E_G_
                d_V = p_G/E_G_;

        double p_G_p = (P_V*prot_E/DW_to_VOL)*d_V,
                p_G_la = (L_V*lip_E/DW_to_VOL)*d_V;

        // Calculate growth energy demands:
        // TEST TEST TEST: the following limits growth if either protein or lipid is being
        // used up too quickly:
        while (Math.max(
                p_G_p/E_p,
                p_G_la/(E_ltot_*V)) > 0.5) {

            d_V = 0.75*d_V;
            p_G_p = (P_V*prot_E/DW_to_VOL)*d_V;
            p_G_la = (L_V*lip_E/DW_to_VOL)*d_V;
        }

        double
                //	p_R_p = p_R*E_p/E,
                //	p_R_la = p_R*(E_l+E_dha+E_epa)/E;

                p_R_p = ((1-kappa)*p_C_tot - p_J)*E_p_/E_,
                p_R_la = ((1-kappa)*p_C_tot - p_J)*E_la_/E_;



        double
                p_R = p_R_p + p_R_la;

        double
                // Calculate catabolic power for compartments:
                p_C_p = p_M_p + p_G_p + p_J_p + p_R_p,
                p_C_la = p_M_la + p_G_la + p_J_la + p_R_la;



        //System.out.println("DEBUG: p_C_p="+p_C_p+"\tp_M_p="+p_M_p+"\tp_G_p="+p_G_p+"\tp_J_p="+p_J_p+"\tp_R_p="+p_R_p);
        //System.out.println("DEBUG: p_C_la="+p_C_la+"\tp_M_la="+p_M_la+"\tp_G_la="+p_G_la+"\tp_J_la="+p_J_la+"\tp_R_la="+p_R_la);
        //p_C_p = (p_G_p + p_M_p)/kappa,
        //p_C_la = (p_G_la + p_M_la)/kappa;
        //System.out.println(p_C_p+"\t"+((p_G_p + p_M_p)/kappa)+"\t"+p_C_la+"\t"+((p_G_la + p_M_la)/kappa));
        double p_C_l, p_C_dha, p_C_epa;


        p_C_l = p_C_la*E_l_/E_ltot_;
        p_C_dha = p_C_la*E_dha_/E_ltot_;
        p_C_epa = p_C_la*E_epa_/E_ltot_;

        // XXX: TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
                                /*double flatut = Math.max(0, 1-(E_-1200)/(2100-1200)),
                                        a1 = alpha_G*flatut,
                                        a2 = beta_G*flatut,
                                        adj = a1*p_C_dha + a2*p_C_epa;
                                p_C_dha *= 1+a1;
                                p_C_epa *= 1+a2;
                                p_C_l -= adj;
                                if (++teller == 40000) {
                                        System.out.println("flatut = "+flatut);
                                        teller = 0;
                                }	*/
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST


                                /*
                                * Split the growth between the 2 volumes. The fraction going to the
                                * DHA compartment equals the lipid part of volume, multiplied
                                * with the fraction of lipid coming from the DHA reserve.
                                */
        double p_C = p_C_p + p_C_l + p_C_dha + p_C_epa;
        double
                d_V_dha = (p_C_dha*kappa - p_M_dha)/(E_G_*L_V),
                d_V_epa = (p_C_epa*kappa - p_M_epa)/(E_G_*L_V),
                d_V_pl = (p_C_l*kappa - p_M_l)/(E_G_*L_V);
        //System.out.println("DEBUG: p_G_la/p_G_p="+(p_G_la/p_G_p));


        // Determinate growth:
        if ((d_V > 0) && (V >= V_p)) {

            // Setting growth to zero gives a simple expr. for p_C:
            p_C = E_*v_dot*V23;
            kappa = p_dot_M_*V/p_C; // Gives zero growth.
            p_R = p_C - p_M - p_J;

            //p_C_p = p_M_p/kappa;
            //p_C_l = p_M_l/kappa;
            //p_C_dha = p_M_dha/kappa;

            p_C_p = p_M_p + p_J_p + p_R*E_p_/E_;
            p_C_l = p_M_l + p_J_l + p_R*E_l_/E_;// - delta/kappa;
            p_C_dha = p_M_dha + p_J_dha + p_R*E_dha_/E_;// + delta/kappa;
            p_C_epa = p_M_epa + p_J_epa + p_R*E_epa_/E_;// + delta/kappa;

            //System.out.println("Pralle: "+(p_C_p + p_C_l + p_C_dha)+"\t"+p_C);

            d_V = 0;
            d_V_pl = 0;
            d_V_dha = 0;
            d_V_epa = 0;
            p_R = (1-kappa)*p_C - p_J;
        } else if (d_V <= 0) { // No-growth condition

            kappa = p_M/(p_J+p_M);//p_M;
            p_C = p_J+p_M;
            p_C_p = p_M_p/kappa;
            //p_C_ltot = p_C*(E_l+E_dha)/E;
            p_C_l = p_M_l/kappa;
            p_C_dha = p_M_dha/kappa;
            p_C_epa = p_M_epa/kappa;
            //System.out.println("brille: "+(p_C_p + p_C_l + p_C_dha)+"\t"+p_C);
            d_V_pl = 0;
            d_V_dha = 0;
            d_V_epa = 0;
            d_V = 0;
            p_R = (1-kappa)*p_C - p_J;

        }

        // Coerce reproduction effort into a standard prot/lip dist:
        double repr_prot_part = 0.7, rej_p=0, rej_la=0;
        if (p_R_p/repr_prot_part > p_R_la/(1-repr_prot_part)) {
            double old = p_R_p;
            p_R_p = p_R_la*repr_prot_part/(1-repr_prot_part);
            rej_p = old-p_R_p;
                                    /*if (t>40)
                                        System.out.println("rej_p="+(rej_p/old));*/
        } else {
            double old = p_R_la;
            p_R_la = p_R_p*(1-repr_prot_part)/repr_prot_part;
            rej_la = old-p_R_la;
                                    /*if (t>40)
                                        System.out.println("rej_la="+(rej_la/old));*/

        }

        // Skew (again):

        // XXX: TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
                                /*double factor = (E_-1200)/(2100-1200);
                                alpha_M *= factor;
                                beta_M *= factor;*/
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST

        double delta = alpha_M*(p_M_dha+p_J_dha),
                //divDelta = E_l/(E_l+E_epa),
                gamma = beta_M*(p_M_epa+p_J_epa);
        //divGamma = E_l/(E_l+E_dha);

        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
        //delta = 0;
        //gamma = 0;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST


        //alpha*(p_M+p_J)*beta_ltot*E_dha_/E_ltot_; // Skew of maintenance towards more use of DHA.

        //p_C_l = p_C_l - delta*divDelta - gamma*divGamma;
        //p_C_epa = p_C_epa - delta*(1-divDelta) + gamma;
        //p_C_dha = p_C_dha + delta - gamma*(1-divGamma);
        p_C_l = p_C_l - delta - gamma;
        p_C_epa = p_C_epa + gamma;
        p_C_dha = p_C_dha + delta;

        // Reproductive or maturity development flux:
        //double p_R = (1-kappa)*p_C - p_J;

        // The variable DHA_X describes the DHA fraction of the feed. This
        // fraction goes into E2, while the rest goes into E1.
        // To calculate the assimilated amounts of each nutrient class, we
        // must take their dry weight fractions *and* energy densities into account.
        double
                d_E_p = p_A_p - p_C_p - 1*rej_p,// + rej_E_p,
                d_E_l = p_A_l - p_C_l - 1*rej_la,// + rej_E_l,
                d_E_dha = p_A_dha - p_C_dha,// + rej_E_dha,
                d_E_epa = p_A_epa - p_C_epa,
                d_E = d_E_p + d_E_l + d_E_dha + d_E_epa;

        //  Maturity investment or reproduction
        double d_Mat, d_R;
        if (!old && (V >= V_p)) { //(Mat >= Mat_max)) {
            d_Mat = 0;
            d_R = Math.max(p_R, 0);

        } else if (!old) {
            d_R = 0;
            d_Mat = Math.max(p_R, 0);
        } else {
            d_R = 0;
            d_Mat = 0;
        }


        // Damage-inducing compounds and hazard rate:
        double
                nu_QC = par.get("eta_QC", t),
                d_M_Q = nu_QC*p_C,
                d_h = (M_Q/V);//*(1+density_effect);
        //d_h_int = h/V;



        f[0] = d_V_pl;
        f[1] = d_V_dha;
        f[2] = d_V_epa;
        f[3] = d_E_p;
        f[4] = d_E_l;
        f[5] = d_E_dha;
        f[6] = d_E_epa;
        f[7] = 0;//p_I - p_A;
        f[8] = d_Mat;
        f[9] = d_R;
        f[10] = d_M_Q;
        f[11] = d_h;

        pA = p_A;//*(P_X+L_X);
        pC = p_C;
        pJ = p_J;
        pR = p_R;
        pM = p_M;
        pG = (d_V_pl + d_V_dha)*E_G_;

        return new IndividualReturnValues(p_A, faeces_g);
    }

    public double getT0() {
        return t0;
    }

    public InitialValues getChildInitialValue() {
        return childInitialValue;
    }

    public void setAmicticFemale(boolean amictic) {
        throw new UnsupportedOperationException();
    }

    public boolean fertilizeIfPossible(RotiferPopulation pop) {
        throw new UnsupportedOperationException();
    }

    public boolean eligibleForMating() {
        throw new UnsupportedOperationException();
    }

    public boolean isFertilized() {
        throw new UnsupportedOperationException();
    }

    public boolean isAmicticFemale() {
        throw new UnsupportedOperationException();
    }

    public boolean isAttached() {
        return false;
    }



}