/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import deb.fishtank.gui.PlotControlPanel;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;

import deb.*;

import java.io.*;
import java.util.Iterator;
import java.util.Map;
import javax.swing.JDialog;

/**
 * @author alver
 *         <p/>
 *         To change the template for this generated type comment go to
 *         Window>Preferences>Java>Code Generation>Code and Comments
 */
public class RotiferMonitor implements SimulationMonitor {

    ArrayList al = new ArrayList();
    ArrayList age = new ArrayList();
    ArrayList ageStruct = new ArrayList();
    ArrayList n = new ArrayList();
    ArrayList alga = new ArrayList();
    ArrayList avg = new ArrayList();
    ArrayList derived = new ArrayList();
    ArrayList derivedStd = new ArrayList();
    ArrayList nBudget = new ArrayList();
    ArrayList males = new ArrayList();
    ArrayList times = new ArrayList();

    Strain lRotType; // The strain representing large rotifers.
    String[] derivedNames;
    Parameters par;
    String directory;
    boolean recordForPlots = false; // Set to true if we should record data for realtime plotting.
    OpenEndedDataset wasteDisplay = null, fluxDisplay = null, popDisplay = null, eggDisplay = null,
        feedDisplay = null;

    /**
     *
     */
    public RotiferMonitor(Parameters par, String directory, Strain lRotType) {
        super();
        this.par = par;
        this.directory = directory;
        this.lRotType = lRotType;
        if (par.getBool("controlPanel", 0)) {
            recordForPlots = true;
            double dt = par.get("dt", 0) * par.get("monitorInterval", 0);
            int window = (int) (par.get("windowSize", 0) / dt);
            wasteDisplay = new OpenEndedDataset(dt, 0, window, new String[]{"Protein waste", "Lipid waste"},
                    "Days", "Fraction");
            fluxDisplay = new OpenEndedDataset(dt, 0, window,
                    new String[]{"pA", "pC", "pR", "pM", "pG", "pW"}, "Days", "J/day");
            popDisplay = new OpenEndedDataset(dt, 0, window,
                    new String[]{"Population"}, "Days", "Rotifers/l");
            eggDisplay = new OpenEndedDataset(dt, 0, window,
                    new String[]{"Egg rate"}, "Days", "Eggs/rot");
            feedDisplay = new OpenEndedDataset(dt, 0, window,
                    new String[] {"Feed density"}, "Days", "g (DW)/l");
            Map<String, OpenEndedDataset> datasets = new LinkedHashMap<String, OpenEndedDataset>();
            datasets.put("Rotifer density", popDisplay);
            datasets.put("Feed density", feedDisplay);
            datasets.put("Wasted fluxes", wasteDisplay);
            datasets.put("Energy fluxes", fluxDisplay);
            datasets.put("Egg rate", eggDisplay);
            if (par.getBoolean("controlPanel", 0)) {
                PlotControlPanel pcp = new PlotControlPanel(datasets);
                JDialog diag = pcp.createSimpleWindow(null);
                diag.setLocation(200, 0);
                diag.setVisible(true);
            }
        }
        /*RealTimePanel panel = new RealTimePanel(wasteDisplay, "Wasted fluxes", "Time, days", "J/day");
      panel.createSimpleWindow(null, "Egg rate").setVisible(true);
      panel = new RealTimePanel(fluxDisplay, "Dry weight", "Time, days", "g");
      panel.createSimpleWindow(null, "Dry weight").setVisible(true);*/

    }

    public SimulationMonitor copy(Parameters par, String suffix) {
        RotiferMonitor newMon = new RotiferMonitor(par, directory + "/" + suffix, lRotType);
        System.out.println("RotiferMonitor.copy: " + newMon.directory);
        newMon.al = (ArrayList) al.clone();
        newMon.age = (ArrayList) age.clone();
        newMon.ageStruct = (ArrayList) ageStruct.clone();
        newMon.n = (ArrayList) n.clone();
        newMon.alga = (ArrayList) alga.clone();
        newMon.avg = (ArrayList) avg.clone();
        newMon.derived = (ArrayList) derived.clone();
        newMon.derivedStd = (ArrayList) derivedStd.clone();
        newMon.nBudget = (ArrayList) nBudget.clone();
        newMon.males = (ArrayList) males.clone();
        newMon.times = (ArrayList)times.clone();
        return newMon;

    }

    public void simulationStarted(double t, Population pop) {
        //record(t, pop);
        System.out.println("start");
    }

    /* (non-Javadoc)
      * @see deb.SimulationMonitor#simulationUpdate(double, deb.Population)
      */
    public void simulationUpdate(double t, Population pop) {
        record(t, pop);
    }

    public void simulationInterval(double t, Population pop) {
        simulationFinished(t, pop);
    }

    /* (non-Javadoc)
      * @see deb.SimulationMonitor#simulationFinished(double, deb.Population)
      */
    public void simulationFinished(double t, Population pop) {
        final int N_STORE = 200;
        if (par.getBool("storeStatesAtFinish", 0)) {
            int divv = pop.getHerbivoreObjectCount() / N_STORE;
            if (divv < 1) {
                System.out.println("Not enough individuals to store state values.");
            } else {
                System.out.println("divv=" + divv);
                try {
                    for (int i = 0; i < N_STORE; i++) {
                        Herbivore herb = null;
                        int count = 0;
                        while ((count < N_STORE) && ((herb == null) || (!(herb instanceof Rotifer)))) {
                            int toGet = (int) Math.floor(pop.getHerbivoreObjectCount() * Math.random());
                            herb = pop.getHerbivore(toGet);
                            count++;
                        }
                        if (count == N_STORE) {
                            System.out.println("Unable to find rotifers to store");
                            return;
                        }

                        File f = new File(directory + "/stateVal" + i + ".xml");
                        //System.out.println(f.getPath());
                        herb.storeCurrentState(f);
                        //	pop.getHerbivore(i).getInitialValuesFromCurrentState().storeValues(f);
                    }
                } catch (IOException ex) {
                    ex.printStackTrace();
                }
            }
        }

        RotiferPopulation rotPop = (RotiferPopulation) pop;
        derivedNames = pop.getHerbivoreDerivedOutputNames(0);
        record(t, pop);
        System.out.println("RotiferMonitor: storing simulation results in directory '" + directory + "'");
        //System.out.println(rotPop.getBaseFeedDynamic().get)
        if (rotPop.micticTotal > 0) {
            System.out.println("Mictic females: " + rotPop.micticTotal +
                    "\nFraction fertilized females: " + (rotPop.fertilizedTotal / rotPop.micticTotal) +
                    "\nResting eggs per fertilized female: " + (rotPop.getRestingEggCount() / rotPop.fertilizedTotal) +
                    "\nTotal resting eggs: " + rotPop.getRestingEggCount());
        }
        try {
            FileWriter rotStatesFw = new FileWriter(new File(directory + "/mean.dlm"));
            FileWriter rotStatesStdFw = new FileWriter(new File(directory + "/stdev.dlm"));
            FileWriter densAndFeedFw = new FileWriter(new File(directory + "/densAndFeed.dlm"));
            FileWriter rotDerivedFw = new FileWriter(new File(directory + "/derived.dlm"));
            FileWriter rotDerivedStdFw = new FileWriter(new File(directory + "/derivedStd.dlm"));
            FileWriter malesFw = new FileWriter(new File(directory + "/males.dlm"));
            FileWriter timesFw = new FileWriter(new File(directory + "/times.dlm"));
            //FileWriter nBudFile = new FileWriter(new File(directory+"/nBudget.dlm"));

            Functions.saveToFile(age, new File(directory + "/meanAge.dlm"));
            Functions.saveToFile(ageStruct, new File(directory + "/ageStruct.dlm"));
            /*for (int j=0; j<derivedNames.length; j++) {
                   rotDerivedFw.write(derivedNames[j]+"\t");
               }
               rotDerivedFw.write("\n");*/

            for (int i = 0; i < avg.size(); i++) {
                double[][] val = (double[][]) avg.get(i);
                for (int j = 0; j < val[0].length; j++) {
                    rotStatesFw.write(val[0][j] + "\t");
                    rotStatesStdFw.write(val[1][j] + "\t");
                }
                rotStatesFw.write("\n");
                rotStatesStdFw.write("\n");
            }
            for (int i = 0; i < derived.size(); i++) {
                double[] val = (double[]) derived.get(i);
                double[] dev = (double[]) derivedStd.get(i);
                double[] time = (double[])times.get(i);
                if (val != null) {
                    for (int j = 0; j < val.length; j++) {
                        rotDerivedFw.write(val[j] + "\t");
                        rotDerivedStdFw.write(dev[j] + "\t");
                        if (j < time.length)
                            timesFw.write(time[j] + "\t");
                    }
                    rotDerivedFw.write("\n");
                    rotDerivedStdFw.write("\n");
                    timesFw.write("\n");
                }
            }

            /*int toStore = 5;
               for (int i=0; i<al.size(); i++) {
                   double[][] val = (double[][])al.get(i);
                   if (val.length > toStore) {
                       for (int j=0; j<val[toStore].length; j++)
                           fw3.write(val[toStore][j]+"\t");
                       fw3.write("\n");
                   } else fw3.write("0\t0\t0\t0\t0\t0\t0\t\n");

               }*/
            for (int i = 0; i < n.size(); i++) {
                double[] n_val = (double[]) n.get(i);
                double[] bf = ((double[]) alga.get(i));
                densAndFeedFw.write(n_val[0] + "\t" + n_val[1] + "\t" + n_val[2] 
                        + "\t" + n_val[3] + "\t" + n_val[4] + "\t" + n_val[5]
                        + "\t" + n_val[6] + "\t" + n_val[7]);
                for (int j = 0; j < bf.length; j++)
                    densAndFeedFw.write("\t" + bf[j]);
                densAndFeedFw.write("\t" + n_val[9]);
                densAndFeedFw.write("\n");
            }
            //System.out.println(males.size());
            for (int i = 0; i < males.size(); i++) {
                double[] c = (double[]) males.get(i);
                for (int j = 0; j < c.length; j++) {
                    malesFw.write(String.valueOf(c[j]));
                    if (j < c.length - 1)
                        malesFw.write("\t");
                }
                malesFw.write("\n");
            }

            /*for (int i=0; i<nBudget.size(); i++) {
                double[] val = (double[])nBudget.get(i);
                for (int j=0; j<val.length; j++)
                    nBudFile.write("\t"+val[j]);
                nBudFile.write("\n");
            }*/

            rotStatesFw.close();
            rotStatesStdFw.close();
            densAndFeedFw.close();
            rotDerivedFw.close();
            rotDerivedStdFw.close();
            timesFw.close();
            malesFw.close();
            //nBudFile.close();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }

    static int count = 0;

    protected void record(double t, Population pop) {
        if (t < 0)
            return;

        RotiferPopulation rotPop = (RotiferPopulation) pop;

        double count = pop.getHerbivoreCount();

        List lRot = pop.getMatchingHerbivores(lRotType),
                sRot = pop.getMatchingHerbivores(FishtankStrains.BRACHIONUS_ROTUNDIFORMIS);
        List maleList = new ArrayList();

        double spermCount = 0;
        double maleCount = 0, micticCount = 0, fertilizedCount = 0;
        double restingEggCount = (rotPop).getRestingEggCount();

        // Remove RotiferMale instances from the lists:
        for (Iterator i = lRot.iterator(); i.hasNext();) {
            Object o = i.next();
            if (!(o instanceof Rotifer))
                i.remove();
            else if (!((Rotifer) o).isAmicticFemale()) {
                micticCount += ((Rotifer) o).getN();
                if (((Rotifer) o).isFertilized())
                    fertilizedCount += ((Rotifer) o).getN();

            }
            if (o instanceof RotiferMale) {
                double n = ((RotiferMale) o).getN();
                maleCount += n;
                spermCount += n * ((RotiferMale) o).spermCount();
                maleList.add(o);
            }
        }
        for (Iterator i = sRot.iterator(); i.hasNext();) {
            Object o = i.next();
            if (!(o instanceof Rotifer))
                i.remove();
        }

        double countL = Functions.countIndividuals(lRot);
        double countS = Functions.countIndividuals(sRot);

        double[] maleMeans1 = Functions.mean(Functions.getBioStates(maleList));
        if (maleMeans1 != null) {
            double[] maleMeans = new double[maleMeans1.length + 1];
            for (int i = 0; i < maleMeans1.length; i++)
                maleMeans[i] = maleMeans1[i];
            maleMeans[maleMeans.length - 1] = spermCount / maleCount;
            males.add(maleMeans);
        } else
            males.add(new double[]{0, 0, 0});

        // Find out which fraction carries eggs:
        double countLWithEggs = 0;
        for (Iterator i = lRot.iterator(); i.hasNext();) {
            Rotifer rot = (Rotifer) i.next();
            if (rot.getEggCount() > 0)
                countLWithEggs += rot.getN();
        }
        double divi = 1000 * par.get("V_w", t);
        double densL = (countL) / divi;
        double densLWithEggs = (countLWithEggs) / divi;
        double densMales = maleCount / divi,
                densMictic = micticCount / divi;
        double densAnalogueWall = rotPop.getAttachedCount() / divi;
        //System.out.println("MaleDens: "+densMales);

        n.add(new double[]{densL, densLWithEggs, densMales, densMictic, fertilizedCount / divi, restingEggCount / divi, densAnalogueWall,
                rotPop.getTotalHerbivoreAddition(), countS / divi, rotPop.getHerbivoreObjectCount()});
        if (recordForPlots) {
            popDisplay.record(t, new double[]{(countL) / (1000 * par.get("V_w", t))});
            feedDisplay.record(t, new double[] {pop.getBaseFeedConcentration()});
        }
        //alga.add(new Double(pop.getHerbivoreObjectCount()));//
        alga.add(new double[]{pop.getBaseFeedConcentration(), ((AlgaDynamic) pop.getBaseFeedDynamic()).totalPredation,
                pop.getWasteDynamic().getLevel(), par.get("V_w", t), rotPop.micticSignalSubstance});
        age.add(new Double(pop.getMeanHerbivoreAge(t)));
        ageStruct.add(Functions.ageStruct(pop.getHerbivores(), t, 20, 20, false));
        if (lRot.size() > 0) {
            // Record current time to make it possible to sort out what happened even if the number
            // of individuals is 0 for periods of the simulation:
            times.add(new double[] {t});

            double[][] deriv = Functions.getDerivedOutputs(lRot, t);
            double[] meanDeriv = Functions.getWeightedMeanDerivedOutputs(lRot, t);
            derived.add(meanDeriv);
            if (recordForPlots) {
                wasteDisplay.record(t, Functions.pick(meanDeriv, 21, 23));
                fluxDisplay.record(t, Functions.pick(meanDeriv, 14, 20));
                eggDisplay.record(t, Functions.pick(meanDeriv, new int[]{0}));
            }
            derivedStd.add(Functions.stdev(deriv));
            double[][] states = Functions.getBioStates(lRot);
            double[][] average = Functions.meanAndStdev(states);
            //nBudget.add(new double[] {countL*meanDeriv});

            //System.out.println(average.length+" x "+average[0].length+" : "+average[0][1]+" "+average[0][2]);
            if (average != null)
                avg.add(average);
        }
    }


}

