/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import deb.Population;
import deb.BaseFeedDynamic;
import deb.Functions;
import deb.Herbivore;
import deb.Parameters;
import deb.Strain;
import deb.control.ParameterSeeker;
import deb.control.ArmaxMinimumVarianceController;
import deb.control.PopulationMetric;
import deb.operations.AddBaseFeed;
import deb.operations.AddBaseFeedChemostat;
import deb.operations.AddBaseFeedTabled;
import deb.operations.AddHerbivoresTabled;
import deb.operations.AddHerbivoresToSetpoint;
import deb.operations.ClearBaseFeed;
import deb.operations.DilutePopulation;
import deb.operations.RecombineHerbivores;/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */public class RotiferPopulation extends Population {
    
    private WaterVolume waterVol;
    private ArmaxMinimumVarianceController feedController;

    private RotiferFactory factory;
    private Strain rotiferType;
    private double restingEggs;
    public double micticTotal=0, fertilizedTotal=0;
    double waterColCount=0, attachedCount=0;
    protected AddHerbivoresTabled addHerbivoresTabled = null;
    protected AddHerbivoresToSetpoint addHerbivoresToSetpoint = null;

    // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 
    double densityPerLiter=0, carryingCapacity=Double.MAX_VALUE;
    double micticSignalSubstance = 0;

    /**
     *This seeker searches for the dilution rate that gives the wanted rotifer density.
     */
    ParameterSeeker dilSeeker = new ParameterSeeker(0.45, 400, -4e-4, 0, 1) {
        public double getDeviation(double t) {
            double currentDensity = 0.001*getHerbivoreCount()/par.get("V_w", t);
            double wantedDensity = 112;
            return wantedDensity-currentDensity;
        }
    };
    
    ParameterSeeker ERSeeker = new ParameterSeeker(29, 70, 1.0, 0, 30) {
        public double getDeviation(double t) {
            double[] meanDeriv = Functions.getWeightedMeanDerivedOutputs(getHerbivores(), t);
            double ER = meanDeriv[0];
            double wantedER = 0.314;
            return wantedER - ER;
        }
    };
    
    ParameterSeeker baseFeedSeeker = new ParameterSeeker(0.5, 70, 1e-4, 0, 2) {
        public double getDeviation(double t) {
            double currentDensity = 0.001*getHerbivoreCount()/par.get("V_w", t);
            double wantedDensity = 277;
            return wantedDensity-currentDensity;
        }
    };
    private double lastFeedControllerUpdate = 0;

    /**
     *
     */
    public RotiferPopulation(BaseFeedDynamic bfd, BaseFeedDynamic wsd, Strain rotiferType, RotiferFactory factory, Parameters par) {

        super(bfd, wsd, par);
        this.factory = factory;
        this.rotiferType = rotiferType;

        double V_w = par.get("V_w0", 0);
        
        waterVol = new WaterVolume(par);
        par.put("V_w", waterVol.getValue()); // Set the current water volume.
        
        if (par.containsKey("armax_bf")) {
            // Define a population metric which returns the herbivore density:
            PopulationMetric metric = new PopulationMetric() {
                public double getValue(Population p) {
                    return p.getHerbivoreCount()/waterVol.getValue()    ;
                }
            };
            PopulationMetric growthMetric = new PopulationMetric() {
                double lastDens=Double.MAX_VALUE;
                public double getValue(Population p) {
                    double result, currentDens = p.getHerbivoreCount();
                    if (lastDens == Double.MAX_VALUE)
                        result = 0;
                    else
                        result = Math.log(currentDens) - Math.log(lastDens);
                    lastDens = currentDens;
                    return result;
                }
            };
/*            feedController = new ArmaxMinimumVarianceController(par, "rotiferControllerSetpoint",
             par.getString("armax_bf", 0), par.getString("armax_g", 0),
                    growthMetric, 0, 1.2);
*/
        }

        // The following concerns water dilution rate, which is handled by an operation called
        // DilutePopulation:
        DilutePopulation dil = null;
        if (par.containsKey("dilutionRate")) {
            dil = new DilutePopulation(par);
            addOperation(dil);
        }

        // For simulation of larval feeding scenarios, it is possible to set up regular addition of rotifers
        // to the tank. This can be according to a preset schedule:
        if (par.containsKey("additionSchedule")) {
            addHerbivoresTabled = new AddHerbivoresTabled(factory, rotiferType, par, dil);
            addOperation(addHerbivoresTabled);
        }
        // ... or rotifers can be added to give a predefined density:
        if (par.containsKey("herbivoreSetpoint")) {
            addHerbivoresToSetpoint = new AddHerbivoresToSetpoint(factory, rotiferType, par);
            addOperation(addHerbivoresToSetpoint);
        }

        if (par.containsKey("clearBaseFeedTimes")) {
            ClearBaseFeed cl = new ClearBaseFeed(par);
            addOperation(cl);
        }

        // It is possible to set a constant feed density using the parameter "keepBaseFeedAtReference":
        if (par.containsKey("keepBaseFeedAtReference") && par.getBool("keepBaseFeedAtReference", 0)) {
            addOperation(new KeepBaseFeedAtReference());
        }
        // ... or use a chemostat setup:
        else if (par.containsKey("chemostatOn") && par.getBool("chemostatOn", 0))
            addOperation(new AddBaseFeedChemostat(par));
        // ... or use regular addition rates, either in batch or continuously:
        else {
            addOperation(new AddBaseFeedPerRotifer(par));
            if (par.containsKey("baseFeedAdditionTable"))
                addOperation(new AddBaseFeedTabled(par));
        }

        // Water dilution can be realized by reducing the N of model individuals to represent that some
        // individuals are washed out. This gives a regular decrease in N, which in turn can lead to a
        // regular increase in the number of model individuals. To counteract this we can regularly combine
        // individuals that are similar enough, and we have functionality to do this:
        if (par.getBool("recombine", 0)) {
            RecombineHerbivores recom = new RecombineHerbivores(rotiferType);
            recom.setRecombineInterval((int)par.get("recombineInterval", 0));
            recom.setRecombineThreshold(par.get("initialRecombineThreshold", 0),
                    par.get("minRecombineThreshold", 0),
                    par.get("maxRecombineThreshold", 0));

            addOperation(recom);
        }

        if (par.containsKey("removeEggsSchedule")) {
            RemoveEggs re = new RemoveEggs(par);
            addOperation(re);
        }

    }

    /**
     * This method is called by the Simulator to advance the model state by one time step.
     * @param t The current time.
     * @param dt The time step length.
     * @param envPar The simulation parameters.
     */
    public void stepAll(double t, double dt, Parameters envPar) {
        envPar.put("V_w", waterVol.getValue()); // Set the current water volume.
        
        // Update counts in case the rotifers are interested in them:
        attachedCount = calculateAttachedCount();
        waterColCount = calculateWaterColumnCount();

        if (feedController != null) {
            double interval = feedController.getCallInterval();
            if ((interval < 0) ||
                (t - lastFeedControllerUpdate >= interval)) {
                feedController.update(this, t, dt, envPar);
                envPar.put("baseFeedReference", feedController.getValue());
                lastFeedControllerUpdate = t;
            }


        }

        super.stepAll(t, dt, envPar);
        
        
        waterVol.update(this, t, dt, envPar);
        
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 
        densityPerLiter = getRotiferCount()/waterVol.getValue();
        if (par.containsKey("carryingCapacity"))
            carryingCapacity = par.get("carryingCapacity", t);

        double[] Tpars = envPar.getArray("Tpars", t);
        double T_w = envPar.get("T_w", t),
               T_ref = par.get("T_ref", t);


        // Temperature corrections:
        double micticSignalTC = T_w > 15 ? 0 : 0.25;
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)
        micticSignalSubstance += dt*(micticSignalTC*(500*500/((500+micticSignalSubstance)*(500+micticSignalSubstance)))
                *0.001*getRotiferCount()/waterVol.getValue() - 2*micticSignalSubstance);
        //System.out.println(micticSignalSubstance);

        //dilSeeker.update(t, dt, envPar);
        //par.put("dilutionRate", dilSeeker.getValue());
        //ERSeeker.update(t, dt, envPar);
        //par.put("hatch_time_DDG", ERSeeker.getValue());
        //baseFeedSeeker.update(t, dt, envPar);
        //par.put("baseFeedReference", baseFeedSeeker.getValue());
    }
    
    public double getTotalHerbivoreAddition() {
        double res = 0;
        if (addHerbivoresTabled != null)
            res += addHerbivoresTabled.getTotalAdded();
        if (addHerbivoresToSetpoint != null)
            res += addHerbivoresToSetpoint.getTotalAdded();
        
        return res;
    }
    
    public void addRestingEggs(double eggs) {
        restingEggs += eggs;
                
    }
    
    public double getRestingEggCount() {
        return restingEggs;
    }
    
    public void setRestingEggCount(double value) {
        restingEggs = value;
    }
    
    public Population copy(boolean cloneParameters) throws UnsupportedOperationException {
        Parameters newPar = cloneParameters ? (Parameters)par.clone() : par;
        RotiferPopulation newPop = new RotiferPopulation(baseFeedDynamic.copy(par), wasteDynamic.copy(par), rotiferType, factory, par);
        //System.out.println("OP: "+operations.size()+"\t OP oth: "+newPop.operations.size());
        newPop.operations.clear(); // We are going to copy all the operations from this pop, so remove any that are
        // created at construction time.
        super.copyFaunaAndOperations(newPop, cloneParameters); // Add Individuals from this pop to the new one
        newPop.setSimulateReproduction(this.simulateReproduction);
        newPop.setSimulateMortality(this.simulateMortality);
        newPop.setSimulatePredation(this.simulatePredation);
        newPop.waterVol = this.waterVol.copy(newPar);
        newPop.restingEggs = this.restingEggs;
        newPop.micticTotal = this.micticTotal;
        newPop.fertilizedTotal = this.fertilizedTotal;
        //System.out.println("OP: "+operations.size()+"\t OP oth: "+newPop.operations.size());
        return newPop;
    }

    public void addHerbivore(deb.Herbivore herb) {
        if (herb instanceof Rotifer) {
            if (!((Rotifer)herb).isAmicticFemale())
                micticTotal += herb.getN();
        }
        super.addHerbivore(herb);
    }
    
    
    /**
     * Return the number of rotifers that are not attached to the wall (as given by isAttached()).
     */
    public double getWaterColumnCount() {
        return waterColCount;
    }
 
    /**
     * Return the number of rotifers that are attached to the wall (as given by isAttached()).
     */
    public double getAttachedCount() {
        return attachedCount;
    }

    /**
     * @return The total number of rotifers.
     */
    public double getRotiferCount() {
        return waterColCount + attachedCount;
    }
  
    public double getEggRate() {
        double[] meanDeriv = Functions.getWeightedMeanDerivedOutputs(herbivores, 0);
        return meanDeriv[0];
    }
    
    protected double calculateWaterColumnCount() {
        double n = 0;
        for (Herbivore herb : herbivores) {
            if (herb instanceof Rotifer) {
                Rotifer rot = (Rotifer)herb;
                if (!rot.isAttached())
                    n += rot.getN();
            }
        }
        return n;
    }

    protected double calculateAttachedCount() {
        double n = 0;
        for (Herbivore herb : herbivores) {
            if (herb instanceof Rotifer) {
                Rotifer rot = (Rotifer)herb;
                if (rot.isAttached())
                    n += rot.getN();
            }
        }
        return n;
    }
}
