/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;


import java.awt.Point;
import java.io.File;
import java.io.IOException;
//import java.util.ArrayList;
import javolution.util.FastTable;
import java.util.Iterator;
import java.util.List;
import deb.*;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class RotiferSingleReserve extends BasicIndividual implements Rotifer {

    static int teller = 0;

    int restingEggsProduced = 0, eggsProduced = 0;
    double micticProbability = 0;
    FastTable eggs = new FastTable();
    InitialValues childInitialValue;

    boolean allowMictic = true; // Set to false to turn off sexual reproduction altogether

    // The following boolean variables are needed to distinguish properties in a resting egg producing strain:
    boolean
            mictic=false, // Is this a mictic female?
            fertilized=false, // Is this a mictic female that has been fertilized?
            canMate=true; // Is this an unfertilized female that is young enough to be fertilized?

    boolean adv = false, old = false;

    double t0, senescence = 0;
    double pC=0, pM=0, pR=0, pJ=0, pA=0, pG=0, pW=0, waste_rel_p=0, waste_rel_la=0;
    LipidPReg lipidReg = null;

    final int REPR_BUFFER = 3, // State number for reproductive buffer
            DRY_WEIGHT_OUTPUT = 1,
            PROTEIN_OUTPUT = 5,
            LIPID_OUTPUT = 6;
            //DHA_OUTPUT = 7,
            //EPA_OUTPUT = 8;

    final static double SENESCENCE_INTAKE_MOD = 0.5;

    /**
     *
     */
    public RotiferSingleReserve(Strain type, Parameters par, InitialValues initVal, InitialValues childInitialValue, double n_0, double t0) {
        super(par, type, initVal, n_0);

        if (states.length != 7)
            throw new RuntimeException();

        this.childInitialValue = childInitialValue;
        this.t0 = t0;
        init();
    }

    /**
     * Construct a new Rotifer, taking initial values and information on eggs from the Parameters
     * object supplied.
     */
    public RotiferSingleReserve(Strain type, Parameters par, Parameters initValues, InitialValues childInitialValue, double n_0, double t0) {
        super(par, type, new InitialValues(initValues.getArray("initValues", t0)), n_0);
        if (states.length != 7)
            throw new RuntimeException();


        this.childInitialValue = childInitialValue;
        addEggs(initValues, t0);
        this.t0 = t0;
        init();
    }

    private void init() {

    }

    public double getAge(double t) {
        return t-t0;
    }

    public void setT0(double t0) {
        this.t0 = t0;
    }

    public double getV() {
        return states[0];
    }

    public double getE() {
        return states[1];
    }

    public double getReprBuffer() {
        return states[REPR_BUFFER];
    }

    protected void setReprBuffer(double newValue) {
        states[REPR_BUFFER] = newValue;
    }

    protected double getHazardRate() {
        return states[5];
    }

    /**
     * This method reads information about eggs from the parameter object, and adds the eggs, if any.
     */
    private void addEggs(Parameters prm, double t) {

        int eggCount = (int)prm.get("eggCount", t0);
        if (eggCount > 0) {
            double[] eggHatchDDG = prm.getArray("egg_hatch_DDG", t);
            double[] eggDDG = prm.getArray("egg_DDG", t);
            for (int i=0; i<eggCount; i++) {
                double[] initVal = prm.getArray("egg_initValues_"+i, t);
                RotiferEgg egg = new RotiferEgg(eggHatchDDG[i], new InitialValues(initVal), 1, t);
                // TODO: Survival rate for initial value eggs.
                egg.setCurrentDDG(eggDDG[i]);
                eggs.add(egg);
                //System.out.println("added egg");
            }
        }
    }


    /* (non-Javadoc)
    * @see deb.Herbivore#update(double, double)
    */
    public IndividualReturnValues update(double t, double dt, Population population, Parameters envPar) {
        super.update(t, dt, population, envPar);

        double density = population.getHerbivoreCount() / par.get("V_w", t); // rotifers per liter
        if (t > 5) //(density > 50)
            micticProbability = par.get("micticEggProbability", t);
            // TODO: mictic egg probability / age etc.
            //micticProbability = (par.get("T_w", t)/20)*((RotiferPopulation)population).micticSignalSubstance/100;
            /*micticProbability = 1//par.get("micticEggProbability", t)
               *(t-t0 < 5.5 ? 1 : 0)
               ;
              */
            //*(0.01+1/(1+0.2*(t - t0)));


        else
            micticProbability = 0.1;


        // TODO: TEST TEST TEST TEST TEST TEST
        //if (par.get("T_w", t) > 15)
        //        micticProbability = 0;

        // Step eggs, if any:
        if (eggs.size() > 0)
            for (int i = 0; i < eggs.size(); i++) {
                ((RotiferEgg) eggs.get(i)).update(dt, envPar.get("T_w", t));
            }

        // See if we are ready to create a new egg. Need to calculate the cost of an egg, based on energy density.
        // Assumption: child has standard initial structural volume, and energy density and relative DHA content equal to parent.
        double
                E_G_ = (par.get("P_V", t) * par.get("prot_energy", t) + par.get("L_V", t) * par.get("lip_energy", t)) / par.get("DW_to_VOL", t);

        double[] values = childInitialValue.getValues();
        double
                V = getV(),
                V0 = values[0],
                E_ = getE() / V,
                costOfEgg = (E_ + E_G_) * V0 * par.get("egg_cost_overhead", t),
                h = states[5];   // Hazard rate

        // Modify costOfEgg in the case of a mictic female... unfertilized mictic females produce cheaper
        // males, fertilized ones produce more expensive resting eggs:
        // TODO: varying egg costs
        if (mictic) {
            if (fertilized) {
                costOfEgg *= par.get("restingEggCostMultiplier", t);

            } else {
                costOfEgg *= par.get("maleEggCostMultiplier", t);
                //costOfEgg *= Math.max(0.8, h);
                /*if (t - t0 > 5.5)
                 costOfEgg *= 2;
             else costOfEgg *= 0.6;*/
            }

        }
        if (getReprBuffer() > costOfEgg) {
            Object egg = createEgg(t, costOfEgg);
            if (egg instanceof RotiferEgg) {
                eggs.add(egg);
                eggsProduced++;
            } else if (egg instanceof RotiferMale) {
                population.addHerbivore((RotiferMale) egg);
            } else if (egg instanceof RestingEgg) {
                ((RotiferPopulation) population).addRestingEggs(getN());
                restingEggsProduced++;
            }
            //createEgg(t, costOfEgg); // TEST TEST TEST: create and throw away
            //System.out.println("Created egg");
        }

        if (old)
            senescence += dt * envPar.get("T_w", t);

        // Step ordinary states:
        states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);

        double E = states[1];

        // Death occurs if energy reserves are negative
        if (E <= 0) {
            kill();
        }

        // Simple assumption: once full size is reached, males will no longer mate with this female:
        // TODO: canMate svitsj testing
        if ((t - t0) > 1.1)
          // (V >= par.get("V_p", t))
            canMate = false;

        double
                P_X = envPar.getFromStruct("feed_comp", "P_X", t), // Protein fraction in feed
                L_X = envPar.getFromStruct("feed_comp", "L_X", t), // Lipid fraction in feed, including DHA
                C_X = envPar.getFromStruct("feed_comp", "C_X", t), // Carbohydrate fraction in feed (not assimilated)
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                p_Am_ = par.get("p_Am_", t),
                mu_X = prot_E * P_X + lip_E * L_X + carb_E * C_X; // The energy density of the feed

        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
        /*double pressure = ((RotiferPopulation)population).densityPerLiter / ((RotiferPopulation)population).carryingCapacity;
      double multiplier = 1;
      if (pressure > 1)
          multiplier = 1 + (pressure - 1) * (pressure - 1);//1 + pressure*pressure;
      p_Am_ /= multiplier;*/
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST


        return ingestion(t, population, envPar, mu_X, p_Am_);
    }

    /**
     * "Backup version" of the f method made June 10, 2005, before rewriting allocation algorithm.
     */
    public IndividualReturnValues f(double[] f, double t, double[] states, Population pop, Parameters envPars) {

        RotiferPopulation population = (RotiferPopulation)pop;
        // Convenience variables for states:
        double
                V = states[0],
                E = states[1],
                Mat = states[2],
                R = states[3],   // Reproductive buffer
                M_Q = states[4], // Damage-inducing compounds
                h = states[5],   // Hazard rate
                age = states[6]; // Age...

        // Primary parameters:
        double
                V_p = par.get("V_p", t),
                v_dot = V>=V_p ? par.get("v_dot_adult", t) : par.get("v_dot", t),
                k_as = par.get("k_as", t),
                p_dot_M_ = par.get("p_dot_M_", t),
                p_dot_J_ = par.get("p_dot_J_", t),
                E_g_overhead = par.get("E_G_", t),
                X_K = par.get("X_K", t),
                kappa = par.get("kappa", t),
                p_Am_ = par.get("p_Am_", t),
                shape_coeff = par.get("shape_coeff", t),
                nu_QC = par.get("eta_QC", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                T_ref = par.get("T_ref", t),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                DW_to_VOL = par.get("DW_to_VOL", t),
                p_dot_E_ = par.get("p_dot_E_", t);

        // TODO: testing av carrying capacity
        double pressure = population.densityPerLiter/population.carryingCapacity;
        double multiplier = 1;
        if (pressure > 1)
                multiplier = 1+(pressure-1)*(pressure-1);//1 + pressure*pressure;

        //multiplier = 1+h*(pressure-1)*(pressure-1);//1 + pressure*pressure;

        //p_Am_ /= multiplier;
       // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
           //
        //nu_QC *= multiplier;
        p_dot_M_ *= multiplier;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 

        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;

        // Derived values:
        double
                E_ = E/V,
                V23 = Math.pow(V, 0.66667),
                V13 = Math.pow(V, 0.33333);

        // Death occurs if energy reserve cannot cover maintenance.
        if ((V <= 0) || (E <= 0)) {
            died = true;
            return new IndividualReturnValues();
            //System.out.println("JUUUUUBBB");
        }

        double[] Tpars = par.getArray("Tpars", t);

        double
                L = V13 / shape_coeff; //L has direct physical meaning: standard length (cm)

        // Inputs:
        double
                P_X = envPars.getFromStruct("feed_comp", "P_X", t), // Protein fraction in feed
                L_X = envPars.getFromStruct("feed_comp", "L_X", t), // Lipid fraction in feed, including DHA
                C_X = envPars.getFromStruct("feed_comp", "C_X", t), // Carbohydrate fraction in feed (not assimilated)
                T_w = envPars.get("T_w", t);

        // Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)
        //System.out.println(""+TC);
        p_Am_ *= TC; // Temp.corrected maximum surface spec. assimilation rate.
        p_dot_M_ *= TC; // Temp.corrected vol.specific maintenance.
        p_dot_J_ *= TC; // Temp.corrected vol.specific maintenance.
        v_dot *= TC;
        p_dot_E_ *= TC;

        // Somatic and maturity maintenance.
        double
                p_M = p_dot_M_*V,
                p_J = Math.min(V, V_p)*p_dot_J_;//*(1-kappa)/kappa;

        // ----------------------------------------------------------------------
        // Modulating maintenance flux for low feed concentrations:
        // ----------------------------------------------------------------------
        double X = population.getBaseFeedConcentration();
        if (X<0) {
            X = 0;
        }

        // Get the feed ingestion rate in Joules per day:
        double
                mu_X = prot_E*P_X + lip_E*L_X + carb_E*C_X; // The energy density of the feed
        IndividualReturnValues ingestion = ingestion(t, population, envPars, mu_X, p_Am_);

        double I_DW = ingestion.getBaseFeedConsumption(), // Ingestion, in dry weight.
                I_waste = ingestion.getWasteConsumption(),
            p_A = (I_DW+I_waste)*mu_X*k_as;

        // TODO: test av minimum repr.-rate (ikke aktivert)
        double
            p_Rmin = 0;
        p_M += p_Rmin;

        double
                mu_V = P_V*prot_E + L_V*lip_E,
                E_G_ = E_g_overhead * mu_V / DW_to_VOL,

                // Calculate catabolic power based on total energy:
                p_C_tot = E_*(E_G_*v_dot*V23 + p_M) / (E_G_ + kappa*E_),

                // Calculate growth flux:
                p_G = kappa*p_C_tot - p_M,  // d_G = p_G/E_G_
                d_V = p_G/E_G_;
        double
                p_R = (1-kappa)*p_C_tot - p_J;

        // Determinate growth:
        if ((d_V > 0) && (V >= V_p)) {

            // Setting growth to zero gives a simple expr. for p_C:
            p_C_tot = E_*v_dot*V23;
            kappa = p_dot_M_*V/p_C_tot; // Gives zero growth.
            //p_R = p_C_tot - p_M - p_J;
            //System.out.println(p_C_tot+"\t"+p_A);
            d_V = 0;
            p_R = (1-kappa)*p_C_tot - p_J;
        } else if (d_V <= 0) { // No-growth condition

            //if (true) 
            { //p_R < -p_G) {
                kappa = p_M/(p_J+p_M);//p_M;
                p_C_tot = p_J+p_M;
                d_V = 0;
                p_R = 0;//(1-kappa)*p_C_tot - p_J;
            }
             /*else {
                //if (t >= 40)
                //    System.out.println("... p_R = "+p_R+"\t p_G = "+p_G);
                p_R += p_G; // Reduce reproductive effort
                d_V = 0;
            }
          */
        }


        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
        p_M -= p_Rmin;
        p_R += p_Rmin;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST

        //  Maturity investment or reproduction
        double d_Mat, d_R;
        if (!old && (V >= V_p)) { //(Mat >= Mat_max)) {
            d_Mat = 0;
            d_R = Math.max(p_R, 0);

        } else if (!old) {
            d_R = 0;
            d_Mat = Math.max(p_R, 0);
        } else {
            d_R = 0;
            d_Mat = 0;
        }


        // Damage-inducing compounds and hazard rate:
        double
                d_M_Q = nu_QC*p_C_tot,
                d_h = (M_Q/V);//*(1+density_effect);      

        double d_E = p_A - p_C_tot;

        //if ((t > 40) && (d_V < 0))
        //    System.out.println(":: "+d_V);

        f[0] = d_V;
        f[1] = d_E;
        f[2] = d_Mat;
        f[3] = d_R;
        f[4] = d_M_Q;
        f[5] = d_h;
        f[6] = 1; // age

        pA = p_A;//*(P_X+L_X);
        pC = p_C_tot;
        pJ = p_J;
        pR = p_R;
        pM = p_M;
        pG = d_V*E_G_;

        return new IndividualReturnValues(p_A, 0);
    }


    /**
     * Calculates the feed ingestion rate, as given in g per day. Feed composition is discounted by this calculation.
     */
    public IndividualReturnValues ingestion(double t, Population population, Parameters envPars, double mu_X, double p_Am_) {
        double X = population.getBaseFeedConcentration(),
                TC = envPars.get("T_w", t)/25; //tempcorr(T_w+273,T_ref,par_r.Tpars); % -, Temperature Correction factor (fnct. by Kooijman)
                //mu_X = 25000; // "Reference" energy density of feed
        if (X<0) {
            X = 0;
        }
        //double p_Am_ = par.get("p_Am_", t);

        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
        RotiferPopulation rotPop = (RotiferPopulation)population;
        double pressure = rotPop.densityPerLiter/rotPop.carryingCapacity,
                multiplier = 1/(1+pressure*pressure);
        //        //multiplier = 1 + pressure*pressure;
        //p_Am_ *= multiplier;
        ////System.out.println(multiplier);
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST 



        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;

        double V23 = Math.pow(getV(), 0.66667),
                k_as = par.get("k_as", t);
        double I_DW = (X/(X+par.get("X_K", t)))*TC*p_Am_*V23/mu_X;

        double waste = I_DW*(1-k_as);

        double X_K_w = par.get("X_K_waste", t),
            p_Am_wst = par.get("p_Am_waste", t);
        p_Am_ += p_Am_wst;
        double X_w = population.getWasteLevel();// getBaseFeedConcentration();

        double f_wst = X_w/(X_w + X_K_w);
        double wasteConsumption = f_wst*p_Am_wst*TC*V23/mu_X;

        return new IndividualReturnValues(I_DW, wasteConsumption, waste);

    }

        /* (non-Javadoc)
        * @see deb.Individual#isAlive()
        */
    public boolean checkMortality(double t, double dt) {
        if (!old) {
            // Hazard rate:

            double
                    h = getHazardRate(),
                    randNum = Math.random();
            old = randNum < h*dt;
            //old = h >= 0.1;
        }
        if (old) { // && (senescence > 40)) {
            died = true;
           /*if (mictic) { 
                if (fertilized)
                    System.out.println("Total resting egg prod: "+restingEggsProduced);
            } else
                System.out.println("Age: "+getAge(t)+"\t \tTotal normal egg prod:"+eggsProduced);*/

        }
        return !died;
    }

    public double[] getDerivedOutputs(double t) {

         double
                V = states[0],
                E = states[1],
                Mat = states[2],
                R = states[3],   // Reproductive buffer
                //                M_Q = states[4], // Damage-inducing compounds
                h = states[5];   // Hazard rate

        double
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                P_V = par.get("P_V",t),
                L_V = par.get("L_V",t),
                P_E = par.get("P_E",t),
                L_E = par.get("L_E",t);


        // Calculate dry weights for the respective states.
        double
                M_V = V/par.get("DW_to_VOL", t),
                M_E = (E*P_E/prot_E+E*L_E/lip_E)/0.65d,

                M_prot_S = 0,//M_s*par.get("P_X",t)/prot_E,
                // TODO: Composition of stomach content.
                M_lip_S = 0,//M_s*par.get("L_X",t)/lip_E,
                M_S = M_prot_S + M_lip_S,
                M_lip_V = V*L_V/par.get("DW_to_VOL", t), //M_V*par_r.lip_V;
                M_lip_E = E*L_E/par.get("lip_energy", t),
                M_lip = M_lip_E + M_lip_V + M_lip_S,
                M_prot_V = M_V*P_V,
                M_prot_E = E*P_E/par.get("prot_energy", t),
                M_prot = M_prot_V + M_prot_E + M_prot_S;

        // Eggs also contribute to derived outputs.
        double eggDW = 0;
        if (eggs.size() > 0) {
            for (int i=0; i<eggs.size(); i++) {
                RotiferSingleReserve child = ((RotiferEgg)eggs.get(i)).getChild();
                double[] outp = child.getDerivedOutputs(t);
                eggDW += outp[1];
            }


        }
        //if (t > 43) System.out.println(((LipidPReg)lipidReg).getIntegratorValue());
        return new double[] {eggs.size(), (M_E+M_V+M_S+eggDW), M_V, M_E, /*M_S,*/ eggDW, M_prot, M_lip, E/V, pA, pC, pR, pM, pG, pW};
    }

    public String[] getDerivedOutputNames() {
        return new String[] {"Eggs", "Dry weight incl. eggs", "Dry weight in structure", "Dry weight in reserves", /*"Dry weight in gut",*/ "Dry weight in eggs",
                "Dry weight in protein", "Dry weight in lipids", "Dry weight in DHA", "Dry weight in EPA", "DHA in structure", "DHA in reserves",
                "EPA in structure", "EPA in reserves", "Energy density", "Lipid controller integral value"};
    }


        /* (non-Javadoc)
        * @see deb.Individual#getPosition()
        */
    public Point getPosition() {
        return null;
    }

    protected Object createEgg(double t, double costOfEgg) {

        // In all cases, subtract the cost:
        setReprBuffer(getReprBuffer() - costOfEgg);

        if (!mictic) {

            double[] values = getEggInitVector(t);

            // Reduce reproductive buffer accordingly:
            double
                V = getV(),
                E_ = (getE())/V;

            double survivalRate = par.get("egg_survival_rate", t);
            return new RotiferEgg(par.get("hatch_time_DDG", t), new InitialValues(values), survivalRate, t);
        } else {
            if (fertilized) {
                // Create resting egg:
                return new RestingEgg();
            } else {
                // Create new male:
                RotiferMale newMale = new RotiferMale(par, this.type, this.n, t);
                return newMale;
            }
        }
    }

    protected double[] getEggInitVector(double t) {
        // Assumption: child has standard initial structural volume, and energy density and relative DHA content equal to parent.
        double[] values = childInitialValue.getValues();
        double
                V0 = values[0],
                V = states[0],
                E_ = states[1]/V;

        values[1] = E_*V0;

        return values;
    }

        /* (non-Javadoc)
        * @see deb.Individual#readyToSpawn()
        */
    public boolean readyToSpawn(double t) {
        if (eggs.size() == 0)
            return false;
        for (int i=0; i<eggs.size(); i++) {
                if (((RotiferEgg)eggs.get(i)).readyToHatch())
                    return true;
        }
        return false;
    }

        /* (non-Javadoc)
        * @see deb.Individual#getChild()
        */
    public Herbivore spawnAndGetChild(double t) {
        for (int i=0; i<eggs.size(); i++) {
            RotiferEgg egg = (RotiferEgg)eggs.get(i);
            if (egg.readyToHatch()) {
                eggs.remove(egg);
                RotiferSingleReserve rot = egg.getChild();
                rot.setT0(t);

                                /*double[] deriv = rot.getBioStates();
                   double weight = (rot.getDerivedOutputs(t))[1];
                   double energy = (deriv[0] + deriv[1])*0.1*(0.4*23600 + 0.18*39500)
                           + deriv[2] + deriv[3] + deriv[4];*/
                //System.out.println("Egg energy = "+energy+"\tDW = "+weight);
                if (!egg.isViable())
                    rot.kill();

                return rot;
            }
        }
        return null;
    }

    /**
     * Remove eggs randomly.
     * @param fraction The likelyhood of each egg being removed.
     */
    public void killEggs(double fraction) {
        if (eggs.size() > 0) {
            for (Iterator i=eggs.iterator(); i.hasNext();) {
                double value = Math.random();
                if (value <= fraction)
                    i.remove();
            }
        }
    }

    public void storeCurrentState(File file) throws IOException {
        Parameters state = new Parameters();
        state.putArray("initValues", states);//initVal.getValues());
        state.put("eggCount", eggs.size());
        if (eggs.size() > 0) {
            double[] eggDDG = new double[eggs.size()],
                    hatch_DDG = new double[eggs.size()];
            for (int i=0; i<eggDDG.length; i++) {
                RotiferEgg egg = (RotiferEgg)eggs.get(i);
                eggDDG[i] = egg.getCurrentDDG();
                hatch_DDG[i] = egg.getHatch_time_DDG();
                state.putArray("egg_initValues_"+i, egg.getInitialValues().getValues());
            }
            state.putArray("egg_hatch_DDG", hatch_DDG);
            state.putArray("egg_DDG", eggDDG);
        }
        state.storeValues(file);
    }

        /* (non-Javadoc)
        * @see deb.Individual#getProteinContent(double)
        */
    public double getProteinContent(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[PROTEIN_OUTPUT];
    }

        /* (non-Javadoc)
        * @see deb.Individual#getLipidContent(double)
        */
    public double getLipidContent(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[LIPID_OUTPUT];
    }

    /**
     * Returns the current egg count of this rotifer.
     * @return The current number of eggs.
     */
    public int getEggCount() {
        return eggs.size();
    }

    public List getEggs() {
        return eggs;
    }

    /**
     * Simple inner class to represent rotifer eggs that hatch after a specific amount of day degrees. This class holds
     * the InitialValues that are to be used in creating the hatched rotifer.
     */
    //static int ali = 0, dea = 0;
    class RotiferEgg {
        double hatch_time_DDG, currentDDG=0;
        InitialValues initVal;
        boolean viable;
        RotiferSingleReserve child;

        public RotiferEgg(double hatch_time_DDG, InitialValues initVal, double survivalRate, double t) {
            this.hatch_time_DDG = hatch_time_DDG;
            this.initVal = initVal;

            child = new RotiferSingleReserve(type, par, initVal, childInitialValue, getN(), t);
            if (allowMictic && (Math.random() < micticProbability)) { //par.get("micticEggProbability", t))) {
                child.mictic = true;
                child.canMate = true;
            }

            //System.out.println(survivalRate+"");
            if (Math.random() <= survivalRate) {

                viable = true;
            } else {
                viable = false;
            }
        }
        /**
         * @return child The child rotifer.
         */
        public RotiferSingleReserve getChild() {
            return child;
        }

        public void update(double dt, double T_w) {
            currentDDG += dt*T_w;
            }

        public boolean isViable() {
            return viable;
        }

        public boolean readyToHatch() {
            return currentDDG >= hatch_time_DDG;
        }
        public InitialValues getInitialValues() {
            return initVal;
        }
        /**
         * @return Returns the currentDDG.
         */
        public double getCurrentDDG() {
            return currentDDG;
        }
        /**
         * @return Returns the hatch_time_DDG.
         */
        public double getHatch_time_DDG() {
            return hatch_time_DDG;
        }
        /**
         * @param currentDDG The currentDDG to set.
         */
        public void setCurrentDDG(double currentDDG) {
            this.currentDDG = currentDDG;
        }

        public RotiferEgg copy() {
            RotiferEgg newEgg = new RotiferEgg(this.hatch_time_DDG, this.initVal, 1, child.getT0());
            newEgg.viable = this.viable;
            newEgg.currentDDG = this.currentDDG;
            return newEgg;
        }

        /**
         * Add the nutritional value of this egg. It is currently realized by constructing a Rotifer object as
         * it would hatch from this egg, and delegating the operation to it.
         * @param ni
         * @param multiplier
         * @param t
         */
        public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
            RotiferSingleReserve rot = new RotiferSingleReserve(type, par, getInitialValues(), childInitialValue, getN(), t);
            rot.addNutritionalContent(ni, multiplier, t);
        }
    }


    public double getDryWeight(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[DRY_WEIGHT_OUTPUT];
    }

        /* (non-Javadoc)
        * @see deb.Individual#addNutritionalContent(deb.NutrientIntake, double)
        */
    public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
        double[] outp = getDerivedOutputs(t);

        ni.add("dw", multiplier*outp[DRY_WEIGHT_OUTPUT]);
        ni.add("protein", multiplier*outp[PROTEIN_OUTPUT]);
        ni.add("lipid", multiplier*outp[LIPID_OUTPUT]);

        for (Iterator i=eggs.iterator(); i.hasNext();) {
            RotiferEgg e = (RotiferEgg)i.next();
            e.addNutritionalContent(ni, multiplier, t);
        }
    }

    // How much relative weight should differences in each of the state variables be given when comparing rotifers?
    private static double[] COMPARE_STATE_WEIGHTS = new double[] {1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1};

    // We need a threshold for each state below which it will be considered equal to zero. This is to prevent "wind-up" in
    // a situation where a state is very close to zero for all individuals, giving very high relative differences that are
    // in fact irrelevant:
    private static double[] ZERO_THRESHOLD = new double[] {1e-9, 1e-10, 1e-10, 1e-6, 1e-6, 1e-7, 1e-7, 0, 1e-6, 1e-6, 1e-10, 1e-5};


    private static double COMPARE_EGG_WEIGHT = 1;
    //static int[] teller = new int[11];
    //static int tell2 = 0;

        /* (non-Javadoc)
        * @see deb.Individual#compareTo(deb.Individual)
        */
    public double compareTo(Individual ind) throws ClassCastException {
        RotiferSingleReserve rot = (RotiferSingleReserve)ind;
        if (type != rot.getType())
            return 100; // Different types cannot be combined.
        if ((mictic != rot.mictic) || (fertilized != rot.fertilized))
            return 100;
        double[] otherStates = rot.getBioStates();
        double stateDiff = 0, biggestDiff = -1;
        int biggestI = -1;
        // Calculate relative, weighted sum of deviances:
        for (int i=0; i<states.length; i++) {
            if (states[i] != 0) {
                // Check if both are above the zero threshold before counting the relative difference:
                if ((states[i] > ZERO_THRESHOLD[i]) || (otherStates[i] > ZERO_THRESHOLD[i]))
                    stateDiff += COMPARE_STATE_WEIGHTS[i]*Math.abs(states[i]-otherStates[i])/states[i];
            } else {
                stateDiff += (otherStates[i] != 0) ? COMPARE_STATE_WEIGHTS[i] : 0;
            }
        }
        stateDiff += COMPARE_EGG_WEIGHT*Math.abs(getEggCount()-rot.getEggCount());
        if ((this.isAmicticFemale() != rot.isAmicticFemale())
            || (this.isFertilized() != rot.isFertilized())
            || (this.isAttached() != rot.isAttached())
            || (this.eligibleForMating() != rot.eligibleForMating()))
            stateDiff += 1;

        return stateDiff;
    }

        /* (non-Javadoc)
        * @see deb.Individual#combineWith(deb.Individual)
        */
    public Individual combineWith(Individual ind, double t) throws ClassCastException {
        // We assume that the given individual has the same number of eggs as us, because
        // a difference in egg count weighs in greatly in comparison. Also, both rotifers has
        // one or more eggs, these are likely to have a similar composition, due to the similarity
        // in other states.
        RotiferSingleReserve oth = (RotiferSingleReserve)ind;
        if (type != oth.getType())
            throw new ClassCastException("Incompatible rotifer types");
        double othN = oth.getN(),
                sumN = othN + n;

        double[] othSt = oth.getBioStates(),
                othChildInit = oth.childInitialValue.getValues(),
                thisChildInit = childInitialValue.getValues();


        // The new state vector should be a weighted mean of the old ones:
        double[] newSt = new double[states.length],
                newChildInit = new double[states.length];
        //System.out.println(newChildInit.length+"\t"+othChildInit.length+"\t"+states.length+"\t"+getType().getName());
        for (int i=0; i<newSt.length; i++) {
            newSt[i] = (n*states[i] + othN*othSt[i])/sumN;
            newChildInit[i] = (n*thisChildInit[i] + othN*othChildInit[i])/sumN;
        }

        RotiferSingleReserve newRot = new RotiferSingleReserve(type, par, new InitialValues(newSt), new InitialValues(newChildInit), sumN, (n*t0 + othN*oth.t0)/sumN);

        newRot.eggsProduced = (eggsProduced + oth.eggsProduced)/2;
        newRot.restingEggsProduced = (restingEggsProduced + oth.restingEggsProduced)/2;
        // The following two properties are the same in both individuals, by requirement:
        newRot.mictic = mictic;
        newRot.fertilized = fertilized;

        // If the individuals to combine have eggs
        if (eggs.size() > 0) {
            //System.out.println("Combining with eggs: "+eggs.size());
            for (Iterator i=eggs.iterator(); i.hasNext();)
                newRot.eggs.add(i.next());
        }



        return newRot;
    }

    public InitialValues getChildInitialValue() {
        return childInitialValue;
    }

    public double getT0() {
        return t0;
    }

    public boolean fertilizeIfPossible(RotiferPopulation pop) {
        if (canMate) {
            if (mictic && !fertilized)
                pop.fertilizedTotal += getN();
            fertilized = true;
        }
        return fertilized;
    }

    public boolean eligibleForMating() {
        return canMate;
    }

    public boolean isAmicticFemale() {
        return !mictic;
    }

    public boolean isFertilized() {
        return fertilized;
    }

    public Individual copy(Parameters par) throws UnsupportedOperationException {
        RotiferSingleReserve newInd = new RotiferSingleReserve(this.type, par,
                this.initVal, this.childInitialValue, this.n_0, this.t0);
        newInd.n = this.n;
        FastTable newEggs = new FastTable();
        for (Iterator i = this.eggs.iterator(); i.hasNext();) {
            RotiferEgg egg = (RotiferEgg)i.next();
            newEggs.add(egg.copy());
        }
        // Copy state values:
        newInd.states = (double[])states.clone();

        newInd.eggs = newEggs;
        newInd.mictic = mictic;
        newInd.canMate = canMate;
        newInd.fertilized = fertilized;
        newInd.eggsProduced = eggsProduced;
        newInd.restingEggsProduced = restingEggsProduced;

        return newInd;
    }

    public void setAmicticFemale(boolean amictic) {
        mictic = !amictic;
    }

    public boolean isAttached() {
        return false;
    }

    class RestingEgg {

    }

}