/*
 * Created on 30.aug.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.fishtank;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
//import java.util.ArrayList;
import javolution.util.FastTable;

import java.util.Iterator;
import java.util.List;

import deb.*;

/**
 * Rotifer model complete with protein, lipid, carbohydrate, DHA and EPA content.
 */
public class RotiferStoich4 extends BasicIndividual implements Rotifer {

    static int teller = 0;

    boolean isAttached = false;
    FastTable eggs = new FastTable();
    InitialValues childInitialValue;
    boolean adv = false, old = false;
    double t0, senescence = 0;
    double pI = 0, pC = 0, pM_P = 0, pM_L = 0, pM_C = 0, pR = 0, pR_out = 0, pJ = 0, pA = 0, pG = 0, pG_out = 0, pW = 0, waste_rel_p = 0, waste_rel_la = 0;
    double eff_P = 0, eff_L = 0, eff_C = 0, condition_index = 0;
    // N budget:
    double N_assim = 0, N_ingested = 0, N_somatic = 0, N_reprod = 0;
    LipidPReg lipidReg = null;

    final int REPR_BUFFER = 8, // State number for reproductive buffer
            DRY_WEIGHT_OUTPUT = 1,
            PROTEIN_OUTPUT = 5,
            LIPID_OUTPUT = 6,
            DHA_OUTPUT = 8,
            EPA_OUTPUT = 9;

    final static double SENESCENCE_INTAKE_MOD = 0.5;
    private boolean starving = false;

    /**
     *
     */
    public RotiferStoich4(Strain type, Parameters par, InitialValues initVal, InitialValues childInitialValue, double n_0, double t0) {
        super(par, type, initVal, n_0);
        if (states.length != 11)
            throw new RuntimeException();

        this.childInitialValue = childInitialValue;
        this.t0 = t0;
        init();
    }

    /**
     * Construct a new Rotifer, taking initial values and information on eggs from the Parameters
     * object supplied.
     */
    public RotiferStoich4(Strain type, Parameters par, Parameters initValues, InitialValues childInitialValue, double n_0, double t0) {
        super(par, type, new InitialValues(initValues.getArray("initValues", t0)), n_0);

        if (states.length != 11)
            throw new RuntimeException();


        this.childInitialValue = childInitialValue;
        addEggs(initValues, t0);
        this.t0 = t0;
        init();
    }

    private void init() {

    }

    public InitialValues getChildInitialValue() {
        return childInitialValue;
    }

    public double getAge(double t) {
        return t - t0;
    }

    public void setT0(double t0) {
        this.t0 = t0;
    }

    public double getT0() {
        return t0;
    }

    public double getV() {
        return states[0] + states[1] + states[2];
    }

    public double getE() {
        return states[3] + states[4] + states[5] + states[6] + states[7];
    }

    public double getReprBuffer() {
        return states[REPR_BUFFER];
    }

    protected void setReprBuffer(double newValue) {
        states[REPR_BUFFER] = newValue;
    }

    protected double getHazardRate() {
        return states[10];
    }

    /**
     * This method reads information about eggs from the parameter object, and adds the eggs, if any.
     */
    private void addEggs(Parameters prm, double t) {
        if (!prm.containsKey("eggCount"))
            return;
        int eggCount = (int) prm.get("eggCount", t0);
        if (eggCount > 0) {
            double[] eggHatchDDG = prm.getArray("egg_hatch_DDG", t);
            double[] eggDDG = prm.getArray("egg_DDG", t);
            for (int i = 0; i < eggCount; i++) {
                double[] initVal = prm.getArray("egg_initValues_" + i, t);
                RotiferEgg egg = new RotiferEgg(eggHatchDDG[i], new InitialValues(initVal), 1, t);
                // TODO: Survival rate for initial value eggs.
                egg.setCurrentDDG(eggDDG[i]);
                eggs.add(egg);
                //System.out.println("added egg");
            }
        }
    }

    /* (non-Javadoc)
    * @see deb.Herbivore#update(double, double)
    */
    public IndividualReturnValues update(double t, double dt, Population population, Parameters envPar) {
        super.update(t, dt, population, envPar);

        // Step eggs, if any:
        if (eggs.size() > 0)
            for (int i = 0; i < eggs.size(); i++) {
                ((RotiferEgg) eggs.get(i)).update(dt, envPar.get("T_w", t));
            }

        // See if we are ready to create a new egg. Need to calculate the cost of an egg, based on energy density.
        // Assumption: child has standard initial structural volume, and energy density and relative DHA content equal to parent.
        double
                E_G_ = par.get("mu_V", t) / par.get("DW_to_VOL", t);

        double[] values = childInitialValue.getValues();
        double
                V = getV(),
                V0 = values[0],
                E_ = getE() / V,
                costOfEgg = (E_ + E_G_) * V0;
        if (getReprBuffer() > costOfEgg) {
            eggs.add(createEgg(t, costOfEgg));
            //createEgg(t, costOfEgg); // TEST TEST TEST: create and throw away
            //System.out.println("Created egg");
        }

        if (old)
            senescence += dt * envPar.get("T_w", t);

        // Step ordinary states:
        states = StepSolver.stepIndividual(this, t, dt, states, population, envPar);

        double E = getE();

        // Check if this individual should attach or de-attach:
        if (isAttached) {
            RotiferPopulation rotPop = (RotiferPopulation) population;
            double likelyhood = dt * Math.pow(rotPop.getAttachedCount() / par.get("A_w", t) / par.get("rho_cap", t), 4);
            //System.out.println("Attached: "+likelyhood);
            if (Math.random() < likelyhood)
                isAttached = false;
        } else {
            if (par.containsKey("k_m")) {
                double likelihood = dt * par.get("k_m", t) * par.get("A_w", t) / 10000 / par.get("V_w", t);
                //System.out.println("Not attached: "+likelihood);
                if (Math.random() < likelihood)
                    isAttached = true;
            }
        }

        // Death occurs if energy reserves are negative
        if (E <= 0) {
            kill();
        }

        double[] Tpars = par.getArray("Tpars", t);
        // Inputs:
        double
                P_X = envPar.getFromStruct("feed_comp", "P_X", t), // Protein fraction in feed
                L_X = envPar.getFromStruct("feed_comp", "L_X", t), // Lipid fraction in feed, including DHA
                C_X = envPar.getFromStruct("feed_comp", "C_X", t), // Carbohydrate fraction in feed (not assimilated)
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                mu_X = prot_E * P_X + lip_E * L_X + carb_E * C_X, // The energy density of the feed
                T_ref = par.get("T_ref", t),
                T_w = envPar.get("T_w", t);

        // Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)
        double montagnes_feed_TC = 1.1817 / (1 + 219000.0 * Math.pow(T_w, -4.35)); // Normalized so T=25 C gives 1.0

        return ingestion(t, population, envPar, montagnes_feed_TC, mu_X);
    }

    /**
     * "Backup version" of the f method made June 10, 2005, before rewriting allocation algorithm.
     */
    public IndividualReturnValues f(double[] f, double t, double[] states, Population population, Parameters envPars) {

        // Convenience variables for states:
        double
                V_plc = states[0],
                V_dha = states[1],
                V_epa = states[2],
                V = V_plc + V_dha + V_epa,
                E_p = states[3],
                E_l = states[4],
                E_c = states[5],
                E_dha = states[6],
                E_epa = states[7],
                E_Lambda = E_l + E_dha + E_epa,
                E = E_p + E_l + E_c + E_dha + E_epa,
                R = states[8],   // Reproductive buffer
                M_Q = states[9], // Damage-inducing compounds
                h = states[10];   // Hazard rate

        for (int i = 0; i < states.length; i++) {
            if (Double.isNaN(states[i])) {
                System.out.println("State "+i+" is NaN");
            }
        }

        // Primary parameters:
        double
                V_p = par.get("V_p", t),
                v_dot = V >= V_p ? par.get("v_dot_adult", t) : par.get("v_dot", t),
                p_dot_M_ = par.get("p_dot_M_", t),
                X_K = par.get("X_K", t),
                p_Am_ = par.get("p_Am_", t),
                //shape_coeff = par.get("shape_coeff", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                C_V = par.get("C_V", t),
                mu_V = par.get("mu_V", t),
                T_ref = par.get("T_ref", t),
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                DW_to_VOL = par.get("DW_to_VOL", t),
                //p_dot_E_ = par.get("p_dot_E_", t);
                k_as_P = par.get("k_as_P", t),
                k_as_L = par.get("k_as_L", t),
                k_as_C = par.get("k_as_C", t),
                k_as_DHA = par.get("k_as_DHA", t),
                k_as_EPA = par.get("k_as_EPA", t),
                kappa_R = par.get("kappa_R", t),
                rho_P = par.get("rho_P", t),
                rho_L = par.get("rho_L", t),
                rho_C = par.get("rho_C", t),
                n_P_M = par.get("n_P_M", t),
                n_L_M = par.get("n_L_M", t),
                n_C_M = par.get("n_C_M", t),
                sigma_G = par.get("sigma_G", t), // Growth overhead without stoichiometric restriction
                gamma_G = 1 + par.get("gamma_G", t); // 1 + growth overhead with same stoichiometric restriction as tissue.



        // Equal v_dot values for all compartments:
        /*double
            v_dot_P = v_dot,
            v_dot_L = v_dot,
            v_dot_C = v_dot;
        */

        // TODO: try compartment differences in v_dot instead of in affinity:
        double avg = (rho_P+rho_L+rho_C)/3.0;
        double
            v_dot_P = v_dot * rho_P/avg,
            v_dot_L = v_dot * rho_L/avg,
            v_dot_C = v_dot * rho_C/avg;
        rho_P = 1;
        rho_L = 1;
        rho_C = 1;


        // TODO: try different nutrient affinities for juvenile vs. grown rotifers
        /*if (V < V_p) {
            // juvenile
            rho_P = 0.05;
            rho_L = 0.9;
            rho_C = 0.5;
        }
        else {
            // grown
            rho_P = 0.9;
            rho_L = 0.05;
            rho_C = 0.5;
        }*/

        // TODO: try reduced assimilation ratio of suboptimally utilized nutrients:
        /*if (eff_P != 0)
            k_as_P *= eff_P;
        if (eff_L != 0)
            k_as_L *= eff_L;
        if (eff_C != 0)
            k_as_C *= eff_C;*/
        // TODO: 100% assimilation rate of limiting nutrient:
        /*if (eff_P > 0.98)
            k_as_P = 1;
        if (eff_L > 0.98)
            k_as_L = 1;
        if (eff_C > 0.98)
            k_as_C = 1;
        */

        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;

        // Derived values:
        double
                E_p_ = E_p / V,
                E_Lambda_ = E_Lambda / V,
                E_c_ = E_c / V,
                V23 = Math.pow(V, 0.66667);

        // Death occurs if energy reserve cannot cover maintenance.
        if ((V <= 0) || (E <= 0) || (E_p < 0) || (E_l < 0) || (E_c < 0) || (E_dha < 0) || (E_epa < 0)) {
            died = true;
            return new IndividualReturnValues();

        }

        double[] Tpars = par.getArray("Tpars", t);

        // Inputs:
        double
                P_X = envPars.getFromStruct("feed_comp", "P_X", t), // Protein fraction in feed
                DHA_X = envPars.getFromStruct("feed_comp", "DHA_X", t), // DHA fraction in feed
                EPA_X = envPars.getFromStruct("feed_comp", "EPA_X", t), // EPA fraction in feed
                L_X = envPars.getFromStruct("feed_comp", "L_X", t) - DHA_X - EPA_X, // Lipid fraction in feed except DHA/EPA
                C_X = envPars.getFromStruct("feed_comp", "C_X", t), // Carbohydrate fraction in feed (not assimilated)
                mu_X = prot_E * P_X + lip_E * (L_X + DHA_X + EPA_X) + carb_E * C_X, // The energy density of the feed
                T_w = envPars.get("T_w", t);

        // Temperature corrections:
        double TC = Functions.TC(T_w, T_ref, Tpars); // -, Temperature Correction factor (fnct. by Kooijman)
        double montagnes_feed_TC = 1.1817 / (1 + 219000.0 * Math.pow(T_w, -4.35)); // Normalized so T=25 C gives 1.0
        //TC = montagnes_feed_TC;
        //System.out.println(""+TC);
        p_Am_ *= TC; // Temp.corrected maximum surface spec. assimilation rate.
        p_dot_M_ *= TC; // Temp.corrected vol.specific maintenance.
        v_dot_P *= TC;
        v_dot_L *= TC;
        v_dot_C *= TC;
        //p_dot_E_ *= TC;

        // Somatic and maturity maintenance.
        double p_M = p_dot_M_ * V;

        // Get the feed ingestion rate in Joules per day:
        IndividualReturnValues ingestion = ingestion(t, population, envPars, montagnes_feed_TC, mu_X);
        double
                I_DW = ingestion.getBaseFeedConsumption(), // Ingestion, in dry weight.
                p_I = I_DW * (prot_E * P_X + lip_E * (L_X + DHA_X + EPA_X) + carb_E * C_X), // Gross energy ingestion
                j_P_A = I_DW * prot_E * k_as_P * P_X,
                j_L_A = I_DW * (lip_E * k_as_L * L_X),
                j_C_A = I_DW * (carb_E * k_as_C * C_X),
                j_DHA_A = I_DW * (lip_E * k_as_DHA * DHA_X),
                j_EPA_A = I_DW * (lip_E * k_as_EPA * EPA_X),

                p_A = j_P_A + j_L_A + j_C_A + j_DHA_A + j_EPA_A;

        N_ingested = P_X * I_DW / 6.25;
        N_assim = k_as_P * N_ingested;

        // TODO: Let rotifers ingest an amount of waste material from faeces and dead rotifers, when available.
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST
        boolean usesWaste = true;
        double wasteConsumption = 0;
        double X_K_w=1, p_Am_wst=0;
        if (par.containsKey("X_K_waste")) {
            X_K_w = par.get("X_K_waste", t);
            p_Am_wst = par.get("p_Am_waste", t);
            p_Am_ += p_Am_wst;
        }
        double X_w = population.getWasteLevel();// getBaseFeedConcentration();
        double f_wst = X_w / (X_w + X_K_w);
        wasteConsumption = f_wst * p_Am_wst * TC * V23 / mu_X;
        //energyIntake += wasteConsumption;
        //if (t > 20)
        //System.out.println("f_wst = "+f_wst);
        double P_W = 0.075, L_W = 0.2, C_W = 0.25;
        j_P_A += 0.75 * P_W * prot_E * wasteConsumption;
        j_L_A += 0.75 * L_W * lip_E * wasteConsumption;
        j_C_A += 0.75 * C_W * carb_E * wasteConsumption;
        p_A = j_P_A + j_L_A + j_C_A;
        // TEST TEST TEST TEST TEST TEST TEST TEST TEST


        double
                E_G_ = mu_V / DW_to_VOL;

        double c2l_rate = par.get("k_carb2lipid", t);//*5*(1-condition_index);
        double p2h_rate = par.get("k_prot2h", t);
        // TODO: starvation on/off parameter.

        //starving = par.getBool("starving", t);
        //starving = false;//((pG+pR)/p_M) < 2;

        double starvInd = Math.max(0.6, Math.min(6, 1+1.5*((pG+pR)/p_M - 1)));
        c2l_rate = 0.3/starvInd;
        p2h_rate = 0.3/starvInd;
        rho_L = Math.min(0.95, 0.3*starvInd);
        /*if (starving) {
            c2l_rate = 0.3;
            p2h_rate = 0.3;
            rho_L = 0.2;
        }
        else {
            c2l_rate = 0;
            p2h_rate = 0;
            rho_L = 0.8;
        } */

        // ----------------------------------------------------------------------
        // Maintenance        
        // ----------------------------------------------------------------------

        // TODO: condition factor affect lipid metabolism:
        // Low condition -> lower rho_L -> more accumulation of lipid (?)
        // unfortunately, condition index is calculated below... not good.
        //rho_L *= Math.max(0.01, condition_index);
        // high condition -> better lipid efficiency (counteract reduction in body DW)
        //n_P_M /= Math.max(0.01, condition_index);

        // Steady state SU values:
        double
                rel_p_C_P = v_dot_P * E_p / E, // dummy value; p_C_P is not yet known. We only need relative values.
                rel_p_C_Lambda = v_dot_L * E_Lambda / E, // dummy value; p_C_P is not yet known. We only need relative values.
                rel_p_C_C = v_dot_C * E_c / E, // dummy value; p_C_P is not yet known. We only need relative values.
                denominator = ((rho_P / n_P_M) * rel_p_C_P + ((rho_L / n_L_M) * rel_p_C_Lambda) + ((rho_C / n_C_M) * rel_p_C_C)),
                omega_P = (rho_P / n_P_M) * rel_p_C_P / denominator,
                omega_Lambda = (rho_L / n_L_M) * rel_p_C_Lambda / denominator,
                omega_C = 1 - omega_P - omega_Lambda;

        // Quantities of P, L and C consumed for maintenance:
        double
                j_P_M = p_M * n_P_M * omega_P,
                j_Lambda_M = p_M * n_L_M * omega_Lambda,
                j_C_M = p_M * n_C_M * omega_C;

        double
                d_V = 0,
                j_X_G = 0,
                p_C_P = 0,
                p_C_Lambda = 0,
                p_C_C = 0,
                j_P_G = 0,
                j_Lambda_G = 0,
                j_C_G = 0;

        pG = 0;
        pG_out = 0;
        pR = 0;
        pR_out = 0;
        N_somatic = 0;
        N_reprod = 0;

        double
                carb2lipid = 0,
                prot2h = 0;
        eff_P = 0;
        eff_L = 0;
        eff_C = 0;

        if (V < V_p) {
            // Growth phase. No egg production.

            double d_V_mrk = 0;

            // d_V is implicit and must be solved numerically. We iterate to find the right value:
            boolean done = false;
            int count = 0;
            while (!done) {
                p_C_P = E_p_ * (v_dot_P * V23 - d_V_mrk);
                p_C_Lambda = E_Lambda_ * (v_dot_L * V23 - d_V_mrk);
                p_C_C = E_c_ * (v_dot_C * V23 - d_V_mrk);
                double
                        j_P_C = p_C_P - j_P_M,
                        j_L_C = p_C_Lambda - j_Lambda_M,
                        j_C_C = p_C_C - j_C_M;

                if ((j_P_C > 0) && (j_L_C > 0) && (j_C_C > 0)) {
                    double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_L_C, j_C_C,
                            P_V * prot_E / DW_to_VOL,
                            L_V * lip_E / DW_to_VOL,
                            C_V * carb_E / DW_to_VOL, sigma_G);
                    d_V = growth_fluxes[0] / gamma_G;

                    //d_V = Functions.non_su(1e9, j_P_C, j_Lambda_C, j_C_C, P_V*prot_E/psi_P_G/DW_to_VOL, L_V*lip_E/psi_L_G/DW_to_VOL,
                    //        C_V*carb_E/psi_C_G/DW_to_VOL);
                    //System.out.println("i="+i+"\tdiff: "+(2*(d_V-d_V_mrk)));
                } else
                    d_V = 0;
                //System.out.println(i+": "+(100*(d_V_mrk-d_V)/d_V)+"");
                double rel_dev;
                if (d_V > 0)
                    rel_dev = (d_V_mrk - d_V) / d_V;
                else {
                    //System.out.println("dd");
                    if (d_V_mrk > 0)
                        rel_dev = 1;
                    else
                        rel_dev = 0;
                }

                if ((count > 5) && (rel_dev < 1e-4))
                    done = true;

                d_V_mrk = (d_V_mrk + d_V) / 2;
                count++;
            }

            //System.exit(0);
            p_C_P = E_p_ * (v_dot_P * V23 - d_V_mrk);
            p_C_Lambda = E_Lambda_ * (v_dot_L * V23 - d_V_mrk);
            p_C_C = E_c_ * (v_dot_C * V23 - d_V_mrk);
            double
                    j_P_C = p_C_P - j_P_M,
                    j_Lambda_C = p_C_Lambda - j_Lambda_M,
                    j_C_C = p_C_C - j_C_M;

            if ((j_P_C > 0) && (j_Lambda_C > 0) && (j_C_C > 0)) {
                double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_Lambda_C, j_C_C,
                        P_V * prot_E / DW_to_VOL,
                        L_V * lip_E / DW_to_VOL,
                        C_V * carb_E / DW_to_VOL, sigma_G);
                d_V = growth_fluxes[0] / gamma_G;
                j_P_G = growth_fluxes[1];
                j_Lambda_G = growth_fluxes[2];
                j_C_G = growth_fluxes[3];
                //System.out.println("Growth: "+growth_fluxes[0]+"\t "+(growth_fluxes[2]/growth_fluxes[1])+"\t "+(growth_fluxes[3]/growth_fluxes[1]));
                //d_V = Functions.non_su(1e9, j_P_C, j_Lambda_C, j_C_C, P_V*prot_E/psi_P_G/DW_to_VOL, L_V*lip_E/psi_L_G/DW_to_VOL,
                //        C_V*carb_E/psi_C_G/DW_to_VOL);

                // Quantities of P and H consumed:
                //j_P_G = d_V*P_V*prot_E/psi_P_G/DW_to_VOL;
                //j_Lambda_G = d_V*L_V*lip_E/psi_L_G/DW_to_VOL;
                //j_C_G = d_V*C_V*carb_E/psi_C_G/DW_to_VOL;
                pG = j_P_G + j_Lambda_G + j_C_G; // Energy spent on growth
                pG_out = d_V * (P_V * prot_E + L_V * lip_E + C_V * carb_E) / DW_to_VOL; // Energy incorporated into structure
                eff_P = j_P_G / j_P_C;
                eff_L = j_Lambda_G / j_Lambda_C;
                eff_C = j_C_G / j_C_C;
                N_somatic = j_P_G / prot_E / 6.25;

                //System.out.println("Eff: "+(j_P_G/j_P_C)+"\t"+(j_Lambda_G/j_Lambda_C)+"\t"+(j_C_G/j_C_C));
            } else
                d_V = 0;

            // Starvation
            if (d_V <= 0) {
                d_V = 0;
                // TODO: reduction of maintenance under starvation conditions:
                //p_M *= 0.5;
                double fr = (j_P_M+j_C_M)/(j_P_M+j_C_M+j_Lambda_M);
                p_C_P = j_P_M;// /fr;
                p_C_Lambda = j_Lambda_M;
                p_C_C = j_C_M;// /fr;
                j_P_G = 0;
                j_Lambda_G = 0;
                j_C_G = 0;
            }
        } else {
            // Egg production phase:

            // // TODO: lower nitrogen level needed for egg production:
            // psi_P_G *= 1.25;


            p_C_P = E_p_ * v_dot_P * V23;
            p_C_Lambda = E_Lambda_ * v_dot_L * V23;
            p_C_C = E_c_ * v_dot_C * V23;

            double
                    j_P_C = p_C_P - j_P_M,
                    j_Lambda_C = p_C_Lambda - j_Lambda_M,
                    j_C_C = p_C_C - j_C_M;

            if ((j_P_C > 0) && (j_Lambda_C > 0) && (j_C_C > 0)) {
                double egg_E_ = (P_V * prot_E + L_V * lip_E + C_V * carb_E);


                double[] growth_fluxes = Functions.hybrid_su(1e9, j_P_C, j_Lambda_C, j_C_C,
                        P_V * prot_E,
                        L_V * lip_E,
                        C_V * carb_E, sigma_G);
                j_X_G = egg_E_ * growth_fluxes[0] / gamma_G;
                j_P_G = growth_fluxes[1];
                j_Lambda_G = growth_fluxes[2];
                j_C_G = growth_fluxes[3];
                //System.out.println("Eggprod: "+growth_fluxes[0]+"\t "+(growth_fluxes[2]/growth_fluxes[1])+"\t "+(growth_fluxes[3]/growth_fluxes[1]));


                eff_P = j_P_G / j_P_C;
                eff_L = j_Lambda_G / j_Lambda_C;
                eff_C = j_C_G / j_C_C;
                //System.out.println("Eff: "+(j_P_G/j_P_C)+"\t"+(j_Lambda_G/j_Lambda_C)+"\t"+(j_C_G/j_C_C));
                pR = j_P_G + j_Lambda_G + j_C_G;
                pR_out = j_X_G;
                N_reprod = j_P_G / prot_E / 6.25;
            } else
                j_X_G = 0;

            // Starvation
            if (j_X_G <= 0) {
                j_X_G = 0;
                //p_M *= 0.5;
                double fr = (j_P_M+j_C_M)/(j_P_M+j_C_M+j_Lambda_M);
                p_C_P = j_P_M;// /fr;
                p_C_Lambda = j_Lambda_M;
                p_C_C = j_C_M;// /fr;
                j_P_G = 0;
                j_Lambda_G = 0;
                j_C_G = 0;
            }
            //    System.out.println(".");

        }

        double mult;
        //if ((eff_L > 0.98) && (eff_L > eff_C+0.4)) {
        if (eff_C < 1) {
            carb2lipid = c2l_rate * E_c;
        }

        if (eff_P < 0.95) {
            prot2h = p2h_rate * E_p;
        }



        double p_C_tot = p_C_P + p_C_Lambda + p_C_C;

        double
                k_dha = E_dha / E_Lambda,
                k_epa = E_epa / E_Lambda,
                k_l = 1 - k_dha - k_epa;

        // Returned parts of rejected fluxes:
        double
                rej_P = kappa_R * (p_C_P - j_P_M - j_P_G),
                rej_Lambda = kappa_R * (p_C_Lambda - j_Lambda_M - j_Lambda_G),
                rej_C = kappa_R * (p_C_C - j_C_M - j_C_G),

                rej_tot = rej_P + rej_Lambda + rej_C;

        // Damage-inducing compounds and hazard rate:
        double
                nu_QC = par.get("eta_QC", t),
                d_M_Q = nu_QC * p_C_tot,
                d_h = (M_Q / V);//*(1+density_effect);

        // TODO: accounting for excessive spending of DHA compared to other fatty acids:
        double mod_DHA = 0 * 1 * k_dha * j_Lambda_M;

        f[0] = d_V * k_l;
        f[1] = d_V * k_dha;
        f[2] = d_V * k_epa;
        f[3] = j_P_A + rej_P - p_C_P - prot2h;
        f[4] = j_L_A + k_l * (rej_Lambda - p_C_Lambda) + 0.83 * carb2lipid + 0.6 * prot2h + mod_DHA * k_l / (k_l + k_epa);
        f[5] = j_C_A + rej_C - p_C_C - carb2lipid + 0.3 * prot2h;
        f[6] = j_DHA_A + k_dha * (rej_Lambda - p_C_Lambda) - mod_DHA;
        f[7] = j_EPA_A + k_epa * (rej_Lambda - p_C_Lambda) + mod_DHA * k_epa / (k_l + k_epa);
        f[8] = j_X_G;//old ? 0 : j_X_G;
        f[9] = d_M_Q;
        f[10] = d_h;

        pI = p_I;
        pA = p_A;//*(P_X+L_X);
        pC = p_C_tot - rej_tot;
        //pR = j_X_G;
        pM_P = j_P_M;
        pM_L = j_Lambda_M;
        pM_C = j_C_M;
        pW = rej_tot * (1 - kappa_R) / kappa_R; // The wasted part, not the recycled part.

        /*if (t>=12) {
            System.out.println(".");
        }*/

        return new IndividualReturnValues(p_A, 0);
    }


    /**
     * Calculates the feed ingestion rate, as given in g per day. Feed composition is discounted by this calculation.
     */
    public IndividualReturnValues ingestion(double t, Population population, Parameters envPars, double TC, double mu_X) {
        double X = population.getBaseFeedConcentration()/par.get("WW_DW_X", t);
        //TC = envPars.get("T_w", t)/25; //tempcorr(T_w+273,T_ref,par_r.Tpars); % -, Temperature Correction factor (fnct. by Kooijman)

        //mu_X = 25000; // "Reference" energy density of feed

        if (X < 0) {
            X = 0;
        }
        double p_Am_ = par.get("p_Am_", t);
        if (old)
            p_Am_ *= SENESCENCE_INTAKE_MOD;
        double V23 = Math.pow(getV(), 0.66667);
        double I_DW = (X / (X + par.get("X_K", t))) * TC * p_Am_ * V23 / mu_X;

        double
                k_as_P = par.get("k_as_P", t),
                k_as_L = par.get("k_as_L", t),
                k_as_C = par.get("k_as_C", t),
                P_X = envPars.getFromStruct("feed_comp", "P_X", t), // Protein fraction in feed
                L_X = envPars.getFromStruct("feed_comp", "L_X", t), // Lipid fraction in feed, including DHA
                C_X = envPars.getFromStruct("feed_comp", "C_X", t); // Carbohydrate fraction in feed (not assimilated)
        double waste = I_DW * (P_X * (1 - k_as_P) + L_X * (1 - k_as_L) + C_X * (1 - k_as_C));


        double X_K_w=1, p_Am_wst=0;
        if (par.containsKey("X_K_waste")) {
            X_K_w = par.get("X_K_waste", t);
            p_Am_wst = par.get("p_Am_waste", t);
            p_Am_ += p_Am_wst;
        }

        double X_w = population.getWasteLevel();// getBaseFeedConcentration();
        double f_wst = X_w / (X_w + X_K_w);
        double wasteConsumption = f_wst * p_Am_wst * TC * V23 / mu_X;

        return new IndividualReturnValues(I_DW, wasteConsumption, waste);

        //return new IndividualReturnValues(I_DW, waste);

        //return new IndividualReturnValues(0, faeces_g);
    }

    /* (non-Javadoc)
    * @see deb.Individual#isAlive()
    */
    public boolean checkMortality(double t, double dt) {
        if (!old) {
            // Hazard rate:

            double
                    h = getHazardRate(),
                    randNum = Math.random();
            old = randNum < h * dt;
        }
        if (old && (senescence > par.get("senescent_period", t)))//40))
            died = true;
        return !died;
    }

    public double[] getDerivedOutputs(double t) {

        // Convenience variables for states:
        double
                V_plc = states[0],
                V_dha = states[1],
                V_epa = states[2],
                V = V_plc + V_dha + V_epa,
                E_p = states[3],
                E_l = states[4],
                E_c = states[5],
                E_dha = states[6],
                E_epa = states[7],
                E_lambda = E_l + E_dha + E_epa,
                E = E_p + E_lambda + E_c,
                R = states[8],   // Reproductive buffer
                M_Q = states[9], // Damage-inducing compounds
                h = states[10];   // Hazard rate

        double
                prot_E = par.get("prot_energy", t),
                lip_E = par.get("lip_energy", t),
                carb_E = par.get("carb_energy", t),
                P_V = par.get("P_V", t),
                L_V = par.get("L_V", t),
                C_V = par.get("C_V", t);

        // Calculate dry weights for the respective states.
        double
                M_V = V / par.get("DW_to_VOL", t),
                M_E = (E_p / prot_E + E_lambda / lip_E + E_c / carb_E) / 0.9d,

                M_prot_S = 0,//M_s*par.get("P_X",t)/prot_E,
                // TODO: Composition of stomach content.
                M_lip_S = 0,//M_s*par.get("L_X",t)/lip_E,
                M_carb_S = 0,
                M_S = M_prot_S + M_lip_S + M_carb_S,
                M_lip_V = M_V * L_V,  //V*L_V/par.get("DW_to_VOL", t), //M_V*par_r.lip_V;
                M_lip_E = E_lambda / lip_E,
                M_lip = M_lip_E + M_lip_V + M_lip_S,
                M_carb_V = M_V * C_V,  //V*L_V/par.get("DW_to_VOL", t), //M_V*par_r.lip_V;
                M_carb_E = E_c / carb_E,
                M_carb = M_carb_E + M_carb_V + M_carb_S,
                M_prot_V = M_V * P_V,
                M_prot_E = E_p / prot_E,
                M_prot = M_prot_V + M_prot_E + M_prot_S,
                M_dha_V = M_lip_V * V_dha / V,
                M_dha_E = E_dha / lip_E,
                M_dha = M_dha_V + M_dha_E,
                M_epa_V = M_lip_V * V_epa / V,
                M_epa_E = E_epa / lip_E,
                M_epa = M_epa_V + M_epa_E;

        // Eggs also contribute to derived outputs.
        double eggDW = 0;
        if (eggs.size() > 0) {
            for (int i = 0; i < eggs.size(); i++) {
                Rotifer child = ((RotiferEgg) eggs.get(i)).getChild();
                double[] outp = child.getDerivedOutputs(t);
                eggDW += outp[1];
            }


        }
        //if (t > 43) System.out.println(((LipidPReg)lipidReg).getIntegratorValue());
        return new double[]{eggs.size(), (M_E + M_V + M_S + eggDW), M_V, M_E, /*M_S,*/ eggDW, M_prot, M_lip, M_carb,
                M_dha, M_epa, M_dha_V, M_epa_V, (E_p + E_lambda + E_c) / V,
                pI, pA, pC, pR, pR_out, pM_P, pM_L, pM_C, pG, pW,
                eff_P, eff_L, eff_C, N_ingested, N_assim, N_somatic, N_reprod, (starving ? 1 : 0)
        };
    }

    public String[] getDerivedOutputNames() {
        return new String[]{"Eggs", "Dry weight incl. eggs", "Dry weight in structure", "Dry weight in reserves", /*"Dry weight in gut",*/ "Dry weight in eggs",
                "Dry weight in protein", "Dry weight in lipids", "Dry weight in DHA", "Dry weight in EPA", "DHA in structure", "DHA in reserves",
                "EPA in structure", "EPA in reserves", "Energy density", "Lipid controller integral value"};
    }


    /* (non-Javadoc)
    * @see deb.Individual#getPosition()
    */
    public Point getPosition() {
        return null;
    }

    protected RotiferEgg createEgg(double t, double costOfEgg) {
        double[] values = getEggInitVector(t);
        // Reduce reproductive buffer accordingly:
        setReprBuffer(getReprBuffer() - costOfEgg);
        double
                V = getV(),
                E_ = (getE()) / V;

        double survivalRate = par.get("egg_survival_rate", t);
        return new RotiferEgg(par.get("hatch_time_DDG", t), new InitialValues(values), survivalRate, t);
    }

    protected double[] getEggInitVector(double t) {
        // Assumption: child has standard initial structural volume, and energy density and relative DHA content equal to parent.
        double[] values = childInitialValue.getValues();
        double
                V0 = values[0] + values[1] + values[2],
                V = getV();
        //             E_ = getE()/V;

        for (int i = 0; i < 8; i++)
            values[i] = states[i] * V0 / V;
        //values[2] = states[2]*V0/V;
        //values[3] = states[3]*V0/V;

        return values;
    }

    /* (non-Javadoc)
    * @see deb.Individual#readyToSpawn()
    */
    public boolean readyToSpawn(double t) {
        if (eggs.size() == 0)
            return false;
        for (int i = 0; i < eggs.size(); i++) {
            if (((RotiferEgg) eggs.get(i)).readyToHatch())
                return true;
        }
        return false;
    }

    /* (non-Javadoc)
    * @see deb.Individual#getChild()
    */
    public Herbivore spawnAndGetChild(double t) {
        for (int i = 0; i < eggs.size(); i++) {
            RotiferEgg egg = (RotiferEgg) eggs.get(i);
            if (egg.readyToHatch()) {
                eggs.remove(egg);
                Rotifer rot = egg.getChild();
                rot.setN(this.getN());
                rot.setT0(t);

                if (!egg.isViable())
                    rot.kill();
                return rot;
            }
        }
        return null;
    }

    /**
     * Remove eggs randomly.
     *
     * @param fraction The likelyhood of each egg being removed.
     */
    public void killEggs(double fraction) {
        if (eggs.size() > 0) {
            for (Iterator i = eggs.iterator(); i.hasNext();) {
                double value = Math.random();
                if (value <= fraction)
                    i.remove();
            }
        }
    }

    public void storeCurrentState(File file) throws IOException {
        Parameters state = new Parameters();
        state.putArray("initValues", states);//initVal.getValues());
        state.put("eggCount", eggs.size());
        if (eggs.size() > 0) {
            double[] eggDDG = new double[eggs.size()],
                    hatch_DDG = new double[eggs.size()];
            for (int i = 0; i < eggDDG.length; i++) {
                RotiferEgg egg = (RotiferEgg) eggs.get(i);
                eggDDG[i] = egg.getCurrentDDG();
                hatch_DDG[i] = egg.getHatch_time_DDG();
                state.putArray("egg_initValues_" + i, egg.getInitialValues().getValues());
            }
            state.putArray("egg_hatch_DDG", hatch_DDG);
            state.putArray("egg_DDG", eggDDG);
        }
        state.storeValues(file);
    }

    /* (non-Javadoc)
    * @see deb.Individual#getProteinContent(double)
    */
    public double getProteinContent(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[PROTEIN_OUTPUT];
    }

    /* (non-Javadoc)
    * @see deb.Individual#getLipidContent(double)
    */
    public double getLipidContent(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[LIPID_OUTPUT];
    }

    /**
     * Returns the current egg count of this rotifer.
     *
     * @return The current number of eggs.
     */
    public int getEggCount() {
        return eggs.size();
    }

    public List getEggs() {
        return eggs;
    }

    /**
     * Simple inner class to represent rotifer eggs that hatch after a specific amount of day degrees. This class holds
     * the InitialValues that are to be used in creating the hatched rotifer.
     */
    //static int ali = 0, dea = 0;
    class RotiferEgg {
        double hatch_time_DDG, currentDDG = 0, survivalRate;
        InitialValues initVal;
        boolean viable;
        Rotifer child;

        public RotiferEgg(double hatch_time_DDG, InitialValues initVal, double survivalRate, double t) {
            this.hatch_time_DDG = hatch_time_DDG;
            this.initVal = initVal;
            this.survivalRate = survivalRate;
            child = new RotiferStoich4(type, par, initVal, childInitialValue, getN(), t);

            //System.out.println(survivalRate+"");
            if (Math.random() <= survivalRate) {

                viable = true;
            } else {
                viable = false;
            }
        }

        /**
         * @return
         */
        public Rotifer getChild() {
            return child;
        }

        public RotiferEgg copy() {
            RotiferEgg newEgg = new RotiferEgg(this.hatch_time_DDG, this.initVal, this.survivalRate, child.getT0());
            newEgg.currentDDG = this.currentDDG;
            return newEgg;
        }

        public void update(double dt, double T_w) {
            currentDDG += dt * T_w;
        }

        public boolean isViable() {
            return viable;
        }

        public boolean readyToHatch() {
            return currentDDG >= hatch_time_DDG;
        }

        public InitialValues getInitialValues() {
            return initVal;
        }

        /**
         * @return Returns the currentDDG.
         */
        public double getCurrentDDG() {
            return currentDDG;
        }

        /**
         * @return Returns the hatch_time_DDG.
         */
        public double getHatch_time_DDG() {
            return hatch_time_DDG;
        }

        /**
         * @param currentDDG The currentDDG to set.
         */
        public void setCurrentDDG(double currentDDG) {
            this.currentDDG = currentDDG;
        }

        /**
         * Add the nutritional value of this egg. It is currently realized by constructing a Rotifer object as
         * it would hatch from this egg, and delegating the operation to it.
         *
         * @param ni
         * @param multiplier
         * @param t
         */
        public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
            Rotifer rot = new RotiferStoich4(type, par, getInitialValues(), childInitialValue, getN(), t);
            rot.addNutritionalContent(ni, multiplier, t);
        }
    }


    public double getDryWeight(double t) {
        double[] outp = getDerivedOutputs(t);
        return outp[DRY_WEIGHT_OUTPUT];
    }

    /* (non-Javadoc)
    * @see deb.Individual#addNutritionalContent(deb.NutrientIntake, double)
    */
    public void addNutritionalContent(NutrientIntake ni, double multiplier, double t) {
        double[] outp = getDerivedOutputs(t);

        double prot = outp[PROTEIN_OUTPUT],
                lipid = outp[LIPID_OUTPUT],
                dha = outp[DHA_OUTPUT],
                epa = outp[EPA_OUTPUT];
        //System.out.println("Nutr: "+prot+"\t "+lipid);

        // TEST TEST TEST TEST TEST TEST
        // TODO: overstyrer kroppssammensetning av rottiser:
        double dw = 4e-7;
        double prt = 0.3;
        double lpd = 0.2;

        ni.add("dw", multiplier * dw);
        ni.add("protein", multiplier * prt * dw);
        ni.add("lipid", multiplier * lpd * dw);
        ni.add("dha", multiplier * 0.14 * lpd * dw);
        ni.add("epa", multiplier * 0.01 * lpd * dw);


        /*
        ni.add("dw", multiplier * outp[DRY_WEIGHT_OUTPUT]);
        ni.add("protein", multiplier * prot);
        ni.add("lipid", multiplier * lipid);
        ni.add("dha", multiplier * dha);
        ni.add("epa", multiplier * epa);
        */
        for (Iterator i = eggs.iterator(); i.hasNext();) {
            RotiferEgg e = (RotiferEgg) i.next();
            e.addNutritionalContent(ni, multiplier, t);
        }
    }

    // How much relative weight should differences in each of the state variables be given when comparing rotifers?
    private static double[] COMPARE_STATE_WEIGHTS = new double[]{1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1};

    // We need a threshold for each state below which it will be considered equal to zero. This is to prevent "wind-up" in
    // a situation where a state is very close to zero for all individuals, giving very high relative differences that are
    // in fact irrelevant:
    private static double[] ZERO_THRESHOLD = new double[]{1e-9, 1e-9, 1e-9, 1e-6, 1e-6, 1e-6, 1e-6, 1e-6,
            0, 0, 0, 0}; //1e-6, 1e-6, 1e-10, 1e-5};


    private static double COMPARE_EGG_WEIGHT = 1;
    //static int[] teller = new int[11];
    //static int tell2 = 0;

    /* (non-Javadoc)
    * @see deb.Individual#compareTo(deb.Individual)
    */
    public double compareTo(Individual ind) throws ClassCastException {
        Rotifer rot = (Rotifer) ind;
        if (type != rot.getType())
            return 100; // Different types cannot be combined.
        double[] otherStates = rot.getBioStates();
        double stateDiff = 0, biggestDiff = -1;
        int biggestI = -1;
        // Calculate relative, weighted sum of deviances:
        for (int i = 0; i < states.length; i++) {
            if (states[i] != 0) {
                // Check if both are above the zero threshold before counting the relative difference:
                if ((states[i] > ZERO_THRESHOLD[i]) || (otherStates[i] > ZERO_THRESHOLD[i]))
                    stateDiff += COMPARE_STATE_WEIGHTS[i] * Math.abs(states[i] - otherStates[i]) / states[i];
            } else {
                stateDiff += (otherStates[i] != 0) ? COMPARE_STATE_WEIGHTS[i] : 0;
            }
        }
        stateDiff += COMPARE_EGG_WEIGHT * Math.abs(getEggCount() - rot.getEggCount());
        //System.out.println(stateDiff);
        return stateDiff;
    }

    /* (non-Javadoc)
    * @see deb.Individual#combineWith(deb.Individual)
    */
    public Individual combineWith(Individual ind, double t) throws ClassCastException {
        // We assume that the given individual has the same number of eggs as us, because
        // a difference in egg count weighs in greatly in comparison. Also, both rotifers has
        // one or more eggs, these are likely to have a similar composition, due to the similarity
        // in other states.
        Rotifer oth = (Rotifer) ind;
        if (type != oth.getType())
            throw new ClassCastException("Incompatible rotifer types");
        double othN = oth.getN(),
                sumN = othN + n;
        if (sumN <= 0)
            return this; // both individuals have n=0 ????? No point.
        double[] othSt = oth.getBioStates(),
                othChildInit = oth.getChildInitialValue().getValues(),
                thisChildInit = childInitialValue.getValues();

        // The new state vector should be a weighted mean of the old ones:
        double[] newSt = new double[states.length],
                newChildInit = new double[states.length];
        //System.out.println(newChildInit.length+"\t"+othChildInit.length+"\t"+states.length+"\t"+getType().getName());
        for (int i = 0; i < newSt.length; i++) {
            newSt[i] = (n * states[i] + othN * othSt[i]) / sumN;
            newChildInit[i] = (n * thisChildInit[i] + othN * othChildInit[i]) / sumN;
        }

        Rotifer newRot = new RotiferStoich4(type, par, new InitialValues(newSt), new InitialValues(newChildInit), sumN, (n * t0 + othN * oth.getT0()) / sumN);
        //System.out.println(sumN+"\t"+newRot.getN());
        // If the individuals to combine have eggs
        if (eggs.size() > 0) {
            //System.out.println("Combining with eggs: "+eggs.size());
            for (Iterator i = eggs.iterator(); i.hasNext();)
                newRot.getEggs().add(i.next());
        }

        double[] f = newSt;
        for (int i = 0; i < f.length; i++) {
            if (Double.isNaN(f[i])) {
                System.out.println(i);
            }
        }


        return newRot;
    }

    public BasicIndividual copy(Parameters par) {
        RotiferStoich4 newInd = new RotiferStoich4(this.type, par,
                this.initVal, this.childInitialValue, this.n_0, this.t0);
        newInd.n = this.n;
        FastTable newEggs = new FastTable();
        for (Iterator i = this.eggs.iterator(); i.hasNext();) {
            RotiferEgg egg = (RotiferEgg) i.next();
            newEggs.add(egg.copy());
        }
        newInd.eggs = newEggs;

        // Copy state values:
        newInd.states = states.clone();

        return newInd;
    }




    public boolean fertilizeIfPossible(RotiferPopulation pop) {
        return false;
    }

    public void setAmicticFemale(boolean amictic) {
        throw new UnsupportedOperationException("RotiferStoich4");
    }

    public boolean isFertilized() {
        return false;
    }

    public boolean isAmicticFemale() {
        return true;
    }

    public boolean eligibleForMating() {
        return false;
    }

    public boolean isAttached() {
        return isAttached;
    }

    /**
     * Override vulnerableToDilution() to signal that this rotifer cannot be subject to dilution as long
     * as it is attached to a substrate.
     */
    public boolean vulnerableToDilution() {
        return !isAttached;
    }

}