package deb.fishtank;

import deb.control.AbstractController;
import deb.control.PopulationMetric;
import deb.Simulator;
import deb.Population;
import deb.Parameters;

/**
 * Created by IntelliJ IDEA.
 * User: alver
 * Date: Nov 8, 2006
 * Time: 11:00:05 AM
 * To change this template use File | Settings | File Templates.
 */
public class TestController extends AbstractController {

    private double conversion = 1000; // rot per ml per baseFeedReference density.
    private double targetPop = 500;
    private double lastRef = 0.5;
    private double maxRef = 1;

    public double getCallInterval() {
        return 0.25;
    }

    public void makeControlRun(final double t, Simulator sim, Population pop, final Parameters envPar) {
        double rela = pop.getHerbivoreCount()/envPar.get("V_w", t)/1000/envPar.get("baseFeedReference", t);
        // Filter in the last value:
        double a = 0.005; // filter constant
        conversion = a*rela + (1-a)*conversion;

        // Estimate needed density:
        double density = targetPop/conversion;
        envPar.put("baseFeedReference", density);

        System.out.println("conversion="+conversion+"\t reference="+density);
    }

    public void _makePredictiveRun(final double t, Simulator sim, Population pop, final Parameters envPar) {
        long time = System.currentTimeMillis();
        //System.out.println("Pop count before: "+pop.getHerbivoreCount());

        PopulationMetric density = new PopulationMetric() {
            public double getValue(Population p) {
                return p.getHerbivoreCount()/envPar.get("V_w", t)/1000;
            }
        };

        double t_end = t+6;
        int steps = getStepCount(t, t_end, sim);
        double[] times = new double[steps];
        for (int i=0; i<steps; i++)
                times[i] = /*t+*/(i+1)*sim.getDt();
        double S_xx = sumSquares(times, 0);
        double S_x = sum(times);

        //System.out.println("steps: "+steps);
        double[] densities;// = makeRunWithMetric(t, t_end, sim, pop, envPar, density, "baseFeedReference",
                //envPar.get("baseFeedReference", t));
        double deviation;// = sumSquares(densities, targetPop);

        double[] values = new double[steps];
        for (int j = 0; j < values.length; j++) {
            values[j] = lastRef;
        }
        densities = makeRunWithMetric(t, t_end, sim, pop, envPar, density, "baseFeedReference", values);
        double oldRes = mean(densities);//0.5*(densities[densities.length-1] + mean(densities)); //densities[densities.length-1];
        System.out.println("OldRes: mean = "+mean(densities)+"\t end = "+densities[densities.length-1]);
        // Try with a modified test value depending on whether we are too high or too low:
        //boolean tooHigh = oldRes > targetPop;
        double testVal = lastRef;
        System.out.println("Rel: "+(oldRes/targetPop));
        for (int i=0; i<1; i++) {
            testVal = testVal*targetPop/oldRes;
            for (int j = 0; j < values.length; j++) {
                values[j] = testVal;
            }
            densities = makeRunWithMetric(t, t_end, sim, pop, envPar, density, "baseFeedReference", values);
            double newRes = mean(densities);//0.5*(densities[densities.length-1] + mean(densities)); //densities[densities.length-1];
            System.out.println("NewRes: mean = "+mean(densities)+"\t end = "+densities[densities.length-1]);
            System.out.println("Tried: "+testVal+"\t Rel: "+(newRes/targetPop));
            oldRes = newRes;
        }
        double chosenDens = testVal*targetPop/oldRes;
        /*
        double chosenDens;
        if (tooHigh) {
            double sensitivity = (1-newRes/oldRes)/0.5;
            chosenDens = lastRef*(1 - targetPop/oldRes/sensitivity);
            System.out.println("too high, oldRes="+oldRes+", newRes="+newRes+", sens="+sensitivity);
        }
        else {
            double sensitivity = ((newRes-oldRes)/oldRes)/0.5;
            chosenDens = lastRef*(1 + oldRes/targetPop/sensitivity);
            System.out.println("too low, oldRes="+oldRes+", newRes="+newRes+", sens="+sensitivity);
        }
        chosenDens = Math.max(0.01, chosenDens);
        */
        // Filter:
        chosenDens = Math.min(chosenDens, maxRef);
        lastRef = chosenDens;
        System.out.println("Choosing reference: "+chosenDens);
        envPar.put("baseFeedReference", chosenDens);


        //System.out.println("Pop count after: "+after.getHerbivoreCount());
        //System.out.println("t = "+t+", 1 day prediction, time used: "+(System.currentTimeMillis()-time));

    }

    public static double sum(double[] values) {
        double sum = 0;
        for (int i = 0; i < values.length; i++) {
            sum += values[i];
        }
        return sum;
    }

    public static double sumSquares(double[] values, double offset) {
        double sum = 0;
        for (int i = 0; i < values.length; i++) {
            sum += (values[i]-offset)*(values[i]-offset);
        }
        return sum;
    }

    public static double mean(double[] values) {
        double sum = 0;
        for (int i = 0; i < values.length; i++) {
            sum += values[i];
        }
        return sum/(double)(values.length);
    }

    public static double variance(double[] values) {
        double sum=0, sumSq = 0;
        for (int i = 0; i < values.length; i++) {
            sum += values[i];
            sumSq += values[i]*values[i];
        }
        double mean = sum/values.length;
        return (sumSq - sum*mean)/(values.length-1);
    }


}
