/*
 * WaterVolume.java
 *
 * Created on December 1, 2005, 11:55 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package deb.fishtank;

import deb.BaseFeedDynamic;
import deb.Parameters;
import deb.Population;
import deb.control.InternalController;
import deb.operations.DilutePopulation;

/**
 *
 * @author alver
 */
public class WaterVolume implements InternalController {
    
    private double V_w, minVol, maxVol;
    
    /** Creates a new instance of WaterVolume */
    public WaterVolume(Parameters par) {
        V_w = par.get("V_w0", 0);
        if (par.containsKey("V_w_min"))
            this.minVol = par.get("V_w_min", 0);
        else
            this.minVol = 0;
        if (par.containsKey("V_w_max"))
            this.maxVol = par.get("V_w_max", 0);
        else
            this.maxVol = Double.MAX_VALUE;
        
    }

    public void update(Object system, double t, double dt, deb.Parameters envPar) {
        //V_w += dt*1000;

        if (envPar.containsKey("addWaterContinuously") && envPar.getBool("addWaterContinuously", t)) {
            addWater(system, dt*envPar.get("waterAdditionRate", t), true);
        } else {
            if (envPar.containsKey("waterAdditionTable")) {
                double timeSince = envPar.getTimeSinceScheduled("waterAdditionTable", t);
                if ((timeSince > 0) && (timeSince < dt)) {
                    addWater(system, envPar.getLastScheduledValue("waterAdditionTable", t), true);
                }
            }
        }
        if (envPar.containsKey("removeWaterContinuously") && envPar.getBool("removeWaterContinuously", t)) {
            removeWater(system, dt*envPar.get("waterRemovalRate", t));
        } else {
            if (envPar.containsKey("waterRemovalTable")) {
                double timeSince = envPar.getTimeSinceScheduled("waterRemovalTable", t);
                if ((timeSince > 0) && (timeSince < dt)) {
                    removeWater(system, envPar.getLastScheduledValue("waterRemovalTable", t));
                }
            }
         }
        
        
    }
    
    /**
     * Adds water - this shouldn't affect the inhabitants, but concentrations like base feed will be reduced.
     * If the addition overflows the maximum volume, either the addition is reduced, or the overflow is assumed
     * to be resolved by water removal (which carries out individuals).
     */
    public void addWater(Object system, double amount, boolean diluteOverflow) {
        Population pop = (Population)system;
        double oldVol = V_w;
        // Increase volume, within bounds:
        V_w += amount;
        double dilutionRate = 1-oldVol/V_w; // The dilution of concentrations.
        if (V_w > maxVol) {
            if (diluteOverflow) {
                double overflow = V_w - maxVol;
                removeWater(system, overflow);
                V_w = maxVol; // Unnecessary, should already match
            }
            else {
                V_w = maxVol;
                dilutionRate = 1-oldVol/V_w; // Less dilution because addition was reduced.
        
            }
        }
        
        //System.out.println(dilutionRate);
        // Dilute base feed and waste:
        BaseFeedDynamic bfd = pop.getBaseFeedDynamic(),
                wsd = pop.getWasteDynamic();
        bfd.dilute(dilutionRate);
        wsd.dilute(dilutionRate);
        
    }

    /**
     * Remove water - this shouldn't affect concentrations, but remove inhabitants.
     */
    public void removeWater(Object system, double amount) {
        Population pop = (Population)system;
        double oldVol = V_w;
        // Reduce volume, within bounds:
        V_w -= amount;
        if (V_w < minVol)
            V_w = minVol;
        
        // Dilute individuals:        
        double dilutionRate = 1-V_w/oldVol; // The fraction removed.
        DilutePopulation.diluteIndividuals(pop.getHerbivores(), dilutionRate, true);
        DilutePopulation.diluteIndividuals(pop.getPredators(), dilutionRate, true);
        
    }
    
    public WaterVolume copy(Parameters par) {
        WaterVolume newWv = new WaterVolume(par);
        return newWv;
    }

    public double getCallInterval() {
        return -1;
    }

    public double getValue() {
        return V_w;
    }
    
}
