package deb.fishtank.control;

import deb.control.Controller;
import deb.Simulator;
import deb.Population;
import deb.Parameters;

/**
 * Created by IntelliJ IDEA.
 * User: alver
 * Date: Feb 9, 2007
 * Time: 3:29:36 PM
 * To change this template use File | Settings | File Templates.
 */
public class SteadyStateRotiferController implements Controller {

    double
        a_plane = -433,
        b_plane = 747.27,
        c_plane = 117.68;
    double
        a_mult = 32.4088,
        b_mult = -23.8513,
        c_mult = -50.0218,
        d_mult = -0.8551,
        e_mult = -14.1750;

    double lastDens = -1;
    double integral = 0;
    double k_i = 0.0002;
    double k_p = 0.5/b_plane;
    double k_d = 0*0.05/50;

    public double getCallInterval() {
        return 1;
    }

    public void makeControlRun(double t, Simulator sim, Population pop, Parameters envPar) {
        double setpoint = envPar.get("rotiferControllerSetpoint", t);
        double dilRate = envPar.get("dilutionRate", t);
        double density = pop.getHerbivoreCount()/envPar.get("V_w", 0)/1000;
        double deviation = setpoint - density;

        // Calculate derivative of density:
        double d_density = 0;
        if (lastDens >= 0)
            d_density = (density - lastDens)/getCallInterval();
        lastDens = density;
        // Update integrator:
        integral += k_i*deviation;


        System.out.println("SteadyStateRotiferController: dens = "+density+" \tdilution = "+dilRate+" \tsetpoint = "+setpoint);

        // Linear model: a*dilRate + b*feedRate + c = density:
        //double u = (setpoint - a_plane *dilRate - c_plane)/b_plane;

        // Nonlinear model (more accurate): (a*dilRate+b)*(c*feedRate+d)+e=density:
        double u = (1/c_mult)*((setpoint-e_mult)/(a_mult*dilRate+b_mult) - d_mult);

        System.out.println("SteadyStateRotiferController: P="+(deviation*k_p)+" \tI="+integral+" \tD="+(k_d*d_density));

        // Add PID functionality:
        u += deviation*k_p + integral - k_d*d_density;

        u = Math.max(0, u);

        System.out.println("SteadyStateRotiferController: u = "+u);
        envPar.put("baseFeedReference", u);

    }
}
