package deb.math;

/**
 * Created by IntelliJ IDEA.
 * User: alver
 * Date: Oct 31, 2006
 * Time: 1:27:06 PM
 * To change this template use File | Settings | File Templates.
 */
public class Matrix {

    public static double[][] matrMultScalar(double a, double[][] b) {

        double[][] res = new double[b.length][b[0].length];
        for (int i=0; i<res.length; i++)
            for (int j=0; j<res[0].length; j++)
                res[i][j] = a*b[i][j];
        return res;
    }


    public static double[][] matrMult(double[][] a, double[][] b) {
        if (a[0].length != b.length)
            throw new IllegalArgumentException("Matrix dimensions do not agree: "+a.length+"x"+a[0].length+" * "+b.length
             +"x"+b[0].length);

        double[][] res = new double[a.length][b[0].length];
        for (int i = 0; i < res.length; i++) {
            double[] re = res[i];
            for (int j = 0; j < re.length; j++) {
                double x = 0;
                for (int k = 0; k < b.length; k++) {
                    x += a[i][k]*b[k][j];
                }
                res[i][j] = x;
            }
        }
        return res;
    }


    public static double[][] matrMult(double[][] a, double[] b) {
        if (a[0].length != b.length)
            throw new IllegalArgumentException("Matrix dimensions do not agree: "+a.length+"x"+a[0].length+" * "+b.length
             +"x1");

        double[][] res = new double[a.length][1];
        for (int i = 0; i < res.length; i++) {
            double x = 0;
            for (int j = 0; j < b.length; j++) {
                x += a[i][j]*b[j];
            }
            res[i][0] = x;

        }
        return res;
    }

    /**
     * Create a 1xn matrix from a vector with n elements
     * @param a The vector
     * @return The matrix with the transposed vector
     */
    public static double[][] transpVector(double[] a) {
        double[][] res = new double[1][a.length];
        for (int i = 0; i < res[0].length; i++) {
            res[0][i] = a[i];
        }
        return res;
    }

    /**
     * Calculate the scalar product of two vectors.
     * @param v1 Vector 1
     * @param v2 Vector 2
     * @return The scalar product.
     */
    public static double scalarProd(double[] v1, double[] v2) {
        double sum = 0;
        for (int i = 0; i < v1.length; i++) {
            sum += v1[i]*v2[i];

        }
        return sum;
    }

    public static double[][] matrAdd(double[][] a, double[][] b) {
        if ((a.length != b.length) || (a[0].length != b[0].length))
                    throw new IllegalArgumentException("Matrix dimensions do not agree: "+a.length+"x"+a[0].length+" * "+b.length
                     +"x"+b[0].length);

        double[][] res = new double[a.length][a[0].length];
        for (int i = 0; i < res.length; i++) {
            for (int j=0; j<res[0].length; j++)
                res[i][j] = a[i][j] + b[i][j];
        }
        return res;
    }

    public static double[][] matrSub(double[][] a, double[][] b) {
        if ((a.length != b.length) || (a[0].length != b[0].length))
                    throw new IllegalArgumentException("Matrix dimensions do not agree: "+a.length+"x"+a[0].length+" * "+b.length
                     +"x"+b[0].length);

        double[][] res = new double[a.length][a[0].length];
        for (int i = 0; i < res.length; i++) {
            for (int j=0; j<res[0].length; j++)
                res[i][j] = a[i][j] - b[i][j];
        }
        return res;
    }

    public static double[] vectorAdd(double[] a, double[] b) {
        if (a.length != b.length)
                    throw new IllegalArgumentException("Matrix dimensions do not agree: "+a.length+"x1 * "+b.length
                     +"x1");

        double[] res = new double[a.length];
        for (int i = 0; i < res.length; i++) {
            res[i] = a[i]+ b[i];
        }
        return res;
    }

    /**
     * Convert a two-dimensional array into a one-dimensional one. If the first dimension has length
     * one, it is removed. Otherwise the second dimension is removed.
     * @param a The matrix to reduce.
     * @return A one-dimensional array.
     */
    public static double[] matr2Vector(double[][] a) {
        if (a.length == 1) {
            double[] res = new double[a[0].length];
            for (int i=0; i<res.length; i++) {
                res[i] = a[0][i];
            }
            return res;
        } else {
            double[] res = new double[a.length];
            for (int i=0; i<res.length; i++) {
                res[i] = a[i][0];
            }
            return res;
        }
    }

    public static double sumVector(double[] a) {
        double res = 0;
        for (int i = 0; i < a.length; i++) {
            res += a[i];
        }
        return res;
    }

    /**
     * Create a unit matrix.
     * @param n Dimension of the matrix.
     * @return The unit matrix.
     */
    public static double[][] eye(int n) {
        double[][] res = new double[n][n];
        for (int i = 0; i < res.length; i++) {
            for (int j=0; j<res[0].length; j++)
                res[i][j] = i==j ? 1 : 0;
        }
        return res;
    }

    /**
     * Create a ones matrix.
     * @param n Dimension of the matrix.
     * @return The ones matrix.
     */
    public static double[][] ones(int n) {
        double[][] res = new double[n][n];
        for (int i = 0; i < res.length; i++) {
            for (int j=0; j<res[0].length; j++)
                res[i][j] = 1;
        }
        return res;
    }

    public static void printMatrix(double[][] a) {
        for (int i = 0; i < a.length; i++) {
            for (int j=0; j<a[0].length; j++)
                System.out.print(a[i][j]+"\t ");
            System.out.println();
        }
        System.out.println();
    }

    public static void printMatrix(double[] a) {
        for (int i = 0; i < a.length; i++)
            System.out.print(a[i]+"\t ");
        System.out.println();
        System.out.println();
    }

    public static void main(String[] args) {
        double[][] a = new double[][] {{1, 2, 3}, {2, 4, 6}};
        double[][] b = new double[][] {{1, 2}, {2, 4}};
        double[][] c = new double[][] {{1, 2}, {2, 4}, {3, 6}};
        double[] d = new double[] {4, 2, 1};

        printMatrix(a);
        printMatrix(b);
        printMatrix(c);
        //printMatrix(matrMult(a, b));
        printMatrix(matrMult(a, c));
        printMatrix(matrMult(c, a));
        printMatrix(matrMult(a, d));

    }

}
