/*
 * Created on 13.des.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import deb.BaseFeedDynamic;
import deb.Parameters;
import deb.Population;

/**
 * This Operation adds base feed either continuously, or at one or more specified times each day.
 *
 * Parameters:
 *  <ul>
 *  <li> baseFeedAdditionOn (boolean): decides if addition is active
 *  <li> continuousBaseFeedAddition (boolean): decides whether addition should be continuous instead of at discrete times.
 *  <li> baseFeedAdditionT0 (double): the time when addition should start. Additions will not be done before this time has been passed.
 *  <li> baseFeedAdditionTimes (double array): the list of times each day (0-1) when addition should be done.
 *  <li> baseFeedReference (double): if adding continuously, determines the daily addition per volume unit. If adding at specified
 *                                      times, determines the concentration to add to, or the concentration to increase by each time,
 *                                      depending on the value of "setBaseFeedToSetPoint".
 *  <li> setBaseFeedToSetPoint (boolean): when adding at specific times, determines whether the concentration should be adjusted to
 *                                      the reference value, or just increased by the reference value.
 *  </ul>
 */
public class AddBaseFeed implements Operation {

	//private BaseFeedDynamic baseFeedDynamic;

	private Parameters par;
	double[] additionTimes;
	double amountToAdd = 0, totalAdded = 0;
	private double bf_t0;
	boolean addToSetPoint, additionPerHerbivore;
	
	/**
	 * @param par
	 */
	public AddBaseFeed(Parameters par) {
		this.par = par;
		
		bf_t0 = par.get("baseFeedAdditionT0", 0);
		additionTimes = par.getArray("baseFeedAdditionTimes", 0);
		additionPerHerbivore = par.containsKey("additionPerHerbivore") && par.getBool("additionPerHerbivore", 0);
	}
	
        
	/* (non-Javadoc)
	 * @see deb.operations.Operation#performNow(double, double)
	 */
	public boolean performNow(double t, double dt) {
            //System.out.println(""+par.getBool("continuousBaseFeedAddition", t));
            if (par.getBool("continuousBaseFeedAddition", t)) {
                //System.out.println("her");
                double inDensity = par.get("baseFeedReference", t);
                amountToAdd =dt*inDensity;
                return true;
            }
            if (par.getBool("baseFeedAdditionOn", t) && (t > bf_t0)) {
                double dayTime = t - Math.floor(t);
                for (double additionTime : additionTimes) {
                    double delta = dayTime - additionTime;
                    if ((delta >= 0) && (delta < dt)) {
                        amountToAdd = par.get("baseFeedReference", t);
                        return true;
                    }
                }

            }
            return false;
	}

	/* (non-Javadoc)
	 * @see deb.operations.Operation#performOperation(double, double, deb.Population, deb.Parameters)
	 */
    public void performOperation(double t, double dt, Population pop,
                                 Parameters envPar) {
        BaseFeedDynamic baseFeedDynamic = pop.getBaseFeedDynamic();

        if (additionPerHerbivore) {
            double individuals = pop.getHerbivoreCount();
            amountToAdd *= individuals*1e-6/envPar.get("V_w", t);
            //System.out.println(amountToAdd+"\t "+(individuals*1e-6/envPar.get("V_w", t)));
        }

        if (par.getBool("continuousBaseFeedAddition", t)) {
            baseFeedDynamic.increaseLevel(amountToAdd);
        } else {
            // Now, amountToAdd will have been set by performNow before this method is called.

            addToSetPoint = par.getBool("setBaseFeedToSetPoint", t);
            if (addToSetPoint) {
                totalAdded += baseFeedDynamic.getLevel() - amountToAdd;
                baseFeedDynamic.setLevel(amountToAdd);
            } else {
                // Add the given amount.
                totalAdded += amountToAdd;
                baseFeedDynamic.increaseLevel(amountToAdd);
            }
        }

    }

    public Operation copy(Parameters par) throws UnsupportedOperationException {
        return new AddBaseFeed(par);
    }

}
