/*
 * Created on 26.nov.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import deb.Herbivore;
import deb.HerbivoreFactory;
import deb.MissingParameterException;
import deb.Parameters;
import deb.Population;
import deb.Strain;

/**
 * This class is used to represent a regular addition of herbivore individuals to the tank. The exact nature of the 
 * individuals added is determined by the HerbivoreFactory instance supplied with the constructor, which is
 * responsible for instantiating individuals for addition.
 */
public class AddHerbivoresRegular implements Operation {

	HerbivoreFactory factory;
	double additionT0, n_each, amountToAdd = 0;
	double[] additionTimes, additionAmounts;
	boolean addToSetPoint;
	Parameters par;
	Strain type;
	
	/**
	 * 
	 */
	public AddHerbivoresRegular(HerbivoreFactory factory, Strain type, Parameters par) {

		this.factory = factory;
		this.par = par;
		this.type = type;
		n_each = par.get("n_each", 0);
		additionT0 = par.get("additionT0", 0);
		additionTimes = par.getArray("additionTimes", 0);
		addToSetPoint = par.getBool("additionToSetPoint", 0);
		additionAmounts = par.getArray("additionAmounts", 0);

	}
	
	/* (non-Javadoc)
	 * @see deb.Operation#performNow(double, double)
	 */
	public boolean performNow(double t, double dt) {
		if (par.getBool("additionOn", t) && (t > additionT0)) {
			double dayTime = t - Math.floor(t);
			loop: for (int i=0; i<additionTimes.length; i++) {
				double delta = dayTime - additionTimes[i]; 
				if ((delta >= 0) && (delta < dt) && (additionAmounts[i] > 0)) {
					amountToAdd = additionAmounts[i];
					return true;					
				}
			}
			
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see deb.Operation#performOperation(double, double, deb.Population, deb.Parameters)
	 */
	public void performOperation(double t, double dt, Population pop, Parameters envPar) {
		// Now, amountToAdd should be set by performNow before this method is called.

		if (addToSetPoint) {
			// TODO: Adjust up to the given rotifer density.
			
		} else {
			// Add the given amount of rotifers.
			int n_to_add = (int)(Math.floor(amountToAdd*envPar.get("V_w", t)/n_each));
			for (int i=0; i<n_to_add; i++) {
				Herbivore herb = factory.getHerbivore(type, n_each, t);
				pop.addHerbivore(herb);//new Rotifer(addedRotiferParams, addedRotiferInitVals[parPivot], childInitialValue, n_each, t));
			}
		}
		
	}

    public Operation copy(Parameters par) throws UnsupportedOperationException {
        return new AddHerbivoresRegular(factory, type, par);
    }

}
