/*
 * Created on 26.nov.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import deb.Herbivore;
import deb.HerbivoreFactory;
import deb.MissingParameterException;
import deb.NutrientIntake;
import deb.Parameters;
import deb.Population;
import deb.Strain;

/**
 * This class is used to represent a regular addition of herbivore individuals to the tank. The exact nature of the 
 * individuals added is determined by the HerbivoreFactory instance supplied with the constructor, which is
 * responsible for instantiating individuals for addition.
 *
 * The class also supports negative addition values, which means that herbivores will be removed at random from the tank.
 * The number removed equals the minimum of the removal value and the current population.
 */
public class AddHerbivoresTabled implements Operation {

	HerbivoreFactory factory;
	double additionT0, n_each, amountToAdd = 0;
	long totalAdded = 0;
	//double[] additionTimes, additionAmounts;
	boolean addToSetPoint;
	Parameters par;
	Strain type;
	final static String SCHEDULE = "additionSchedule";
	//DilutePopulation dil = null;
        
	/**
	 * 
	 */
	public AddHerbivoresTabled(HerbivoreFactory factory, Strain type, Parameters par, DilutePopulation dil) {
		this.factory = factory;
		this.par = par;
		this.type = type;
        //        this.dil = dil;
		n_each = par.get("n_each", 0);
	}
	
	/* (non-Javadoc)
	 * @see deb.Operation#performNow(double, double)
	 */
	public boolean performNow(double t, double dt) {
		double value = par.getTimeSinceScheduled(SCHEDULE, t);		
		boolean res = (value >= 0) && (value < dt);
		if (res) {
			
			amountToAdd = par.getLastScheduledValue(SCHEDULE, t);
			System.out.println("add t="+t+"\tvalue="+amountToAdd);
		}
		return res;
	}

	/* (non-Javadoc)
	 * @see deb.Operation#performOperation(double, double, deb.Population, deb.Parameters)
	 */
	public void performOperation(double t, double dt, Population pop, Parameters envPar) {
            
		// Now, amountToAdd should be set by performNow before this method is called.
                //System.out.println("Adding now!");
		if (addToSetPoint) {
			// TODO: Adjust up to the given rotifer density.
			
		} else {
                    if (amountToAdd > 0)
                        addHerbivores(amountToAdd, t, pop, envPar);
                    else if (amountToAdd < 0)
                        removeHerbivores(-amountToAdd, t, pop, envPar);
		}
		
                
                // TEST TEST TEST TEST
                /*if (dil != null)
                    dil.hold(t, 1d/12d);*/
	}
        
        private void addHerbivores(double amount, double t, Population pop, Parameters envPar) {
            // Add the given amount of rotifers.
            double V_w = envPar.get("V_w", t);
            int n_to_add = (int)(Math.floor(amountToAdd*V_w/n_each));
            totalAdded += amount*V_w;
            System.out.println("To add: "+n_to_add);
            for (int i=0; i<n_to_add; i++) {
                Herbivore herb = factory.getHerbivore(type, n_each, t);
                pop.addHerbivore(herb);//new Rotifer(addedRotiferParams, addedRotiferInitVals[parPivot], childInitialValue, n_each, t));
            }
        }
        
        private void removeHerbivores(double amount, double t, Population pop, Parameters envPar) {
            amount *= envPar.get("V_w", t);
            System.out.println("t="+t+"\t remove="+amount);
            NutrientIntake result = pop.removeHerbivoresAtRandom(amount, t);
                
        }

        public Operation copy(Parameters par) throws UnsupportedOperationException {
            //System.out.println("AddHerbivoresTabled: cannot produce good copy if dil != null");
            return new AddHerbivoresTabled(factory, type, par, null);
        }

    public double getTotalAdded() {
        return totalAdded;
    }

}
