/*
 * Created on 26.nov.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import deb.Functions;
import deb.Herbivore;
import deb.HerbivoreFactory;
import deb.MissingParameterException;
import deb.NutrientIntake;
import deb.Parameters;
import deb.Population;
import deb.Strain;

/**
 * This class is used to represent a controller adding herbivores regularly to reach
 * a setpoint density. The controller adds at given intervals, and cannot remove
 * herbivores from the system. The exact nature of the individuals added is determined 
 * by the HerbivoreFactory instance supplied with the constructor, which is
 * responsible for instantiating individuals for addition.
 *
 */
public class AddHerbivoresToSetpoint implements Operation {

	HerbivoreFactory factory;
	double additionT0, n_each, amountToAdd = 0, interval, counter=0;
	long totalAdded = 0;
	Parameters par;
	Strain type;
	final static String SCHEDULE = "additionSchedule";
       
	/**
	 * 
	 */
	public AddHerbivoresToSetpoint(HerbivoreFactory factory, Strain type, Parameters par) {
		this.factory = factory;
		this.par = par;
		this.type = type;
                n_each = par.get("n_each", 0);
                interval = par.get("herbivoreAdditionInterval", 0);
	}
	
	/* (non-Javadoc)
	 * @see deb.Operation#performNow(double, double)
	 */
	public boolean performNow(double t, double dt) {
            counter += dt;
            if (counter >= interval) {
                counter = 0;
                return true;
            }
            else
                return false;
	}

	/* (non-Javadoc)
	 * @see deb.Operation#performOperation(double, double, deb.Population, deb.Parameters)
	 */
	public void performOperation(double t, double dt, Population pop, Parameters envPar) {
             // Add the given amount of rotifers.
            double V_w = envPar.get("V_w", t),
                    setpoint = envPar.get("herbivoreSetpoint", t);
            double deviation = setpoint*V_w - Functions.countIndividuals(pop.getMatchingHerbivores(type));
            System.out.println("t="+t+"\t deviation="+deviation);
            if (deviation <= 0)
                return; // High enough already, nothing to do.
            int n_to_add = (int)(Math.floor(deviation/n_each));
            totalAdded += deviation;
            System.out.println("To add: "+n_to_add);
            for (int i=0; i<n_to_add; i++) {
                Herbivore herb = factory.getHerbivore(type, n_each, t);
                pop.addHerbivore(herb);
 
        
            }
        }
        
        
        public Operation copy(Parameters par) throws UnsupportedOperationException {
            return new AddHerbivoresToSetpoint(factory, type, par);
        }

    public double getTotalAdded() {
        return totalAdded;
    }

}
