/*
 * Created on 14.sep.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import deb.fishtank.RotiferPopulation;
import java.util.Iterator;

import deb.BaseFeedDynamic;
import deb.Individual;
import deb.Parameters;
import deb.Population;
import java.util.List;

/**
 * @author alver
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class DilutePopulation implements Operation {

    protected double holdT = -1, holdDur = 0;
    
    protected boolean diluteSmoothly;
    
    protected Parameters par;
    
    /**
     * @param bfd
     */
    public DilutePopulation(Parameters par) {
        this.par = par;
        
                /*
                 * Dilute smoothly means that the super individuals are diluted by
                 * reducing their n, instead of removing entire super individuals. This
                 * gives a much better representation of the effect of dilution in a
                 * situation with relatively few super individuals. The drawback is that
                 * the mean n of super individuals will decrease over time, and in a
                 * density-limited population this will cause the number of super
                 * individuals to rise, slowing the simulation down.
                 */
        if (par.containsKey("diluteSmoothly"))
            diluteSmoothly = par.getBool("diluteSmoothly", 0);
        else
            diluteSmoothly = false;
        
    }
    
    /**
     * Sets the schedule for dilution of the population, and activates dilution
     * in the simulation.
     *
     * @param t0
     *            Time of first dilution.
     * @param interval
     *            Dilution interval.
     * @param fraction
     *            Fraction of population removed at each dilution. The removal
     *            of Individuals is determined at random, so the actual fraction
     *            may deviate, especially for small populations.
     *
     * public void setDilutionSchedule(boolean diluteBaseFeed, boolean
     * setBaseFeedToSetPoint, boolean diluteHerbivores, boolean dilutePredators) {
     * this.diluteBaseFeed = diluteBaseFeed; this.diluteHerbivores =
     * diluteHerbivores; this.dilutePredators = dilutePredators;
     * this.setBaseFeedToSetPoint = setBaseFeedToSetPoint; }
     */
    
    /**
     * Sets the enabled state of the dilution. If no schedule is set, there will
     * be no effect on the population.
     *
     * @param enabled
     *
     * public void setDilutionEnabled(boolean diluteBaseFeed, boolean
     * setBaseFeedToSetPoint, boolean diluteHerbivores, boolean dilutePredators) {
     * this.diluteBaseFeed = diluteBaseFeed; this.diluteHerbivores =
     * diluteHerbivores; this.dilutePredators = dilutePredators; }
     */
    
        /*
         * (non-Javadoc)
         *
         * @see deb.Operation#performOperation(double, double, deb.Population,
         *      deb.Parameters)
         */
    public void performOperation(double t, double dt, Population pop,
            Parameters envPar) {
        //System.out.println("t="+t);
        boolean diluteBaseFeed = par.getBool("diluteBaseFeed", t),
                diluteHerbivores = par.getBool("diluteHerbivores", t),
                dilutePredators = par.getBool("dilutePredators", t),
                continuousDilution = par.getBool("continuousDilution", t),
                diluteWaste = par.getBool("diluteWaste", t);
        
        BaseFeedDynamic baseFeedDynamic = pop.getBaseFeedDynamic(),
                wasteDynamic = pop.getWasteDynamic();

        double dil_t0 = par.get("dilutionT0", 0);
        
        if (t > dil_t0) {
            
            double dil_frac = par.get("dilutionRate", t);

            // If we are diluting contiuously, we should multiply the rate with
            // the time step.
            if (continuousDilution)
                dil_frac *= dt;

            //System.out.println("dil: "+dil_frac);

            //System.out.println("Diluting at t="+t+"\tRate="+dil_frac);
            
            // TEST TEST TEST TEST
            /*if (pop instanceof RotiferPopulation) {
                RotiferPopulation rPop = (RotiferPopulation)pop;
                double restingEggs = rPop.getRestingEggCount();
                rPop.setRestingEggCount(restingEggs*dil_frac);
            }*/ 
            
            // Dilute base feed:
            if (diluteBaseFeed)
                baseFeedDynamic.dilute(dil_frac);
            
            // Dilute waste:
            if (diluteWaste)
                wasteDynamic.dilute(dil_frac);
            
            
            // Dilute herbivores:
            if (diluteHerbivores) {
                diluteIndividuals(pop.getHerbivores(), dil_frac, diluteSmoothly);
                /*for (Iterator i = pop.getHerbivores().iterator(); i.hasNext();) {
                    Object o = i.next();
                    if (diluteSmoothly) {
                        ((Individual) o).killFraction(dil_frac);
                    } else {
                        double randomNumber = Math.random();
                        if (randomNumber < dil_frac) {// OOps, you were diluted
                            // away!
                            ((Individual) o).kill();
                        }
                    }
                }*/
            }
            
            // Dilute predators:
            if (dilutePredators) {
                diluteIndividuals(pop.getPredators(), dil_frac, diluteSmoothly);
                /*for (Iterator i = pop.getPredators().iterator(); i.hasNext();) {
                    Object o = i.next();
                    double randomNumber = Math.random();
                    if (randomNumber < dil_frac) // OOps, you were diluted away!
                        ((Individual) o).kill();
                }*/
            }
            
        }
    }
    
    public static void diluteIndividuals(List individuals, double fraction, boolean diluteSmoothly) {
        for (Iterator<Individual> i = individuals.iterator(); i.hasNext();) {
            Individual o = i.next();
            // The individual may for some reason be unaffected by dilution. Check this:
            if (!o.vulnerableToDilution())
                continue;
            if (diluteSmoothly) {
                o.killFraction(fraction);
            } else {
                double randomNumber = Math.random();
                if (randomNumber < fraction) {// OOps, you were diluted
                    // away!
                    o.kill();
                }
            }
        }
    }
    
    /*
     * (non-Javadoc)
     *
     * @see deb.Operation#performNow(double, double)
     */
    public boolean performNow(double t, double dt) {
        
        if ((holdT >= 0) && ((t-holdDur) < holdT))
            return false;
        
        boolean diluteBaseFeed = par.getBool("diluteBaseFeed", t),
                diluteHerbivores = par.getBool("diluteHerbivores", t),
                dilutePredators = par.getBool("dilutePredators", t),
                continuousDilution = par.getBool("continuousDilution", t);
        
        if (continuousDilution)
            return true;
        
        double dilutionTime = par.get("dilutionTime", t);
        double dil_t0 = par.get("dilutionT0", 0);
    
        // Return if nothing is to be diluted:
        if (!diluteBaseFeed && !diluteHerbivores && !dilutePredators)
            return false;
        if (t < dil_t0)
            return false;
        
        double dil_delta = par.get("dilutionInterval", 0);
        
        double artT = (t - dil_t0) % dil_delta;
        //while (artT < dil_t0)
        //	artT += dil_delta;
        
        // Check if it is time for dilution, or if we are diluting continuously:
        return (artT >= dilutionTime) && ((artT - dilutionTime) < dt);
    }
    
    public void hold(double t, double duration) {
        holdT = t;
        holdDur = duration;
    }
    
    public Operation copy(Parameters par) throws UnsupportedOperationException {
        return new DilutePopulation(par);
    }
    
}