/*
 * Created on 26.nov.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import deb.Parameters;
import deb.Population;

/**
 * This interface defines the role of an object taking care of an operation to be performed routinely or sporadically on a Population.
 * An Operation can be added to a Population before simulating.
 */
public interface Operation {

	/**
	 * This method will be called once per iteration, and requests whether the operation should be performed now.
	 * @param t The current time.
	 * @param dt The time step.
	 * @return true if it is time to perform the operation.
	 */
	public boolean performNow(double t, double dt);
	
	/**
	 * This method will be called once, if and only if performNow() returned true.
	 * @param t The current time.
	 * @param dt The time step.
	 * @param pop The population to perform the operation on.
	 * @param envPar Environmental parameters.
	 */
	public void performOperation(double t, double dt, Population pop, Parameters envPar);
	
        public Operation copy(Parameters par) throws UnsupportedOperationException;
}
