/*
 * Created on 29.nov.2004
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package deb.operations;

import java.util.Iterator;
import java.util.List;

import deb.CombinationTools;
import deb.CombinedPair;
import deb.Herbivore;
import deb.Individual;
import deb.Parameters;
import deb.Population;
import deb.Strain;

/**
 * @author alver
 * 
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class RecombineHerbivores implements Operation {

	private int recombPiv = 0, recombineInterval = 1;

	private double recombineThreshold = 0, minRecombineThreshold = 0,
			maxRecombineThreshold = 0;
        private Strain type;
        
	public RecombineHerbivores(Strain type) {
            this.type = type;
	}

	public void setRecombineInterval(int interval) {
		recombineInterval = interval;
	}

	public void setRecombineThreshold(double initRecombineThreshold,
			double minThreshold, double maxThreshold) {
		this.recombineThreshold = initRecombineThreshold;
		this.minRecombineThreshold = minThreshold;
		this.maxRecombineThreshold = maxThreshold;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Operation#performNow(double, double)
	 */
	public boolean performNow(double t, double dt) {
		recombPiv++;
		return (recombineInterval > 0) && (recombPiv == recombineInterval);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see deb.Operation#performOperation(double, double, deb.Population,
	 *      deb.Parameters)
	 */
	public void performOperation(double t, double dt, Population pop,
			Parameters envPar) {

		// Try to reduce individual count by combining them.
		long timeStart = System.currentTimeMillis();
		List res = CombinationTools.findClosestComparisons(pop.getMatchingHerbivores(type),
				recombineThreshold, 0.5, 50000);
		//System.out.println("Time used for recombination: "
		//		+ (System.currentTimeMillis() - timeStart));
		//System.out.println("Pairs found: "+res.size()+"\tThreshold: "+recombineThreshold);
		int pivot = 0;
		// We assume that we have adjusted the arguments well when calling
		// findClosestComparisons,
		// so we remove all non-overlapping pairs in the List res:
		for (Iterator k = res.iterator(); k.hasNext();) {
			CombinedPair pair = (CombinedPair) k.next();
			Individual newInd = pair.getOne().combineWith(pair.getTwo(), t);
                        //System.out.println("Removing: "+pair.getDistance());
			pop.getHerbivores().remove(pair.getOne());
			pop.getHerbivores().remove(pair.getTwo());
			pop.getHerbivores().add((Herbivore) newInd);
                        checkInd(pair.getOne());
                        checkInd(pair.getTwo());
                        checkInd(newInd);
			pivot++;
		}

		double loadFactor = ((double) pop.getHerbivoreObjectCount())
			/ envPar.get("desiredIndividualCount", t);
		recombineThreshold *= loadFactor;
		// Make sure to keep the threshold inside of its limits:
		recombineThreshold = Math.max(minRecombineThreshold, Math.min(
		maxRecombineThreshold, recombineThreshold));
		
		System.out.println("RecombineHerbivores: Count: " + pop.getHerbivoreObjectCount()
				+ "\tRemoved: " + pivot + "\tThreshold: " + recombineThreshold);

		recombPiv = 0;
	}

        public void checkInd(Individual ind) {
            double[] f = ind.getBioStates();
            for (int i=0; i<f.length; i++) {
                if (Double.isNaN(f[i]))  {
                    System.out.println(i);
                }
            }

        }
        
        public Operation copy(Parameters par) throws UnsupportedOperationException {
            RecombineHerbivores newRec = new RecombineHerbivores(this.type);
            newRec.recombPiv = recombPiv;
            newRec.recombineInterval = recombineInterval;
            newRec.recombineThreshold = recombineThreshold;
            newRec.minRecombineThreshold = minRecombineThreshold;
            newRec.maxRecombineThreshold = maxRecombineThreshold;
            return newRec;
        }

}

